/* @(#)out-tachyon.c	1.1 7/2/91 */

/* Subroutines for insn-output.c for TaChyon.
   Copyright (C) 1989 Free Software Foundation, Inc.
   Contributed by Y.Nakashima

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#ifndef FILE
#include <stdio.h>
#endif

/* This is where the condition code register lives.  */
rtx cc0_reg_rtx;

static rtx find_addr_reg ();

/* Return truth value of wheterh OP is EQ or NE.  */
int
eq_or_neq (op, mode)
     rtx op;
     enum machine_mode mode;
{
  return (GET_CODE (op) == EQ || GET_CODE (op) == NE);
}

/* Nonzero if OP is a valid second operand for an arithmetic insn.  */

int
arith_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  return (register_operand (op, mode) || FITS_25_BITS (op));
}

int
arith32_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  return (register_operand (op, mode) || GET_CODE (op) == CONST_INT);
}

int
int5_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  return (GET_CODE (op) == CONST_INT && (unsigned) INTVAL (op) < 0x20);
}

static char *
singlemove_string (operands)
     rtx *operands;
{
  if (GET_CODE (operands[0]) == MEM)
    {
      if (GET_CODE (operands[1]) != MEM)
	if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
	  {
	    output_asm_insn ("seti\t%m0,gr31", operands);
            if (FP_REG_P (operands[1]))
	      return "st.f\t%1,gr31,gr0";
            else
	      return "st\t%1,gr31,gr0";
	  }
	else
          if (FP_REG_P (operands[1]))
	    return "st.f\t%1,%0";
          else
	    return "st\t%1,%0";
      else
	{
	  rtx xoperands[2];

	  xoperands[0] = gen_rtx (REG, SFmode, 33);
	  xoperands[1] = operands[1];
	  output_asm_insn (singlemove_string (xoperands), xoperands);
	  xoperands[1] = xoperands[0];
	  xoperands[0] = operands[0];
	  output_asm_insn (singlemove_string (xoperands), xoperands);
	  return "";
	}
    }
  else if (GET_CODE (operands[1]) == MEM)
    {
      if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
	{
	  output_asm_insn ("seti\t%m1,gr31", operands);
          if (FP_REG_P (operands[0]))
	    return "ld.f\tgr31,gr0,%0";
          else
	    return "ld\tgr31,gr0,%0";
	}
      else
        if (FP_REG_P (operands[0]))
          return "ld.f\t%1,%0";
        else
          return "ld\t%1,%0";
    }
  else
    if (FP_REG_P (operands[1]))
      return "fadd\t%1,fr0,%0";
    else
      return "or\t%1,gr0,%0";
}

/* Output assembler code to perform a doubleword move insn
   with operands OPERANDS.  */

char *
output_move_double (operands)
     rtx *operands;
{
  enum { REGOP, OFFSOP, MEMOP, PUSHOP, POPOP, CNSTOP, RNDOP } optype0, optype1;
  rtx latehalf[2];
  rtx addreg0 = 0, addreg1 = 0;

  /* First classify both operands.  */

  if (REG_P (operands[0]))
    optype0 = REGOP;
  else if (offsettable_memref_p (operands[0]))
    optype0 = OFFSOP;
  else if (GET_CODE (operands[0]) == MEM)
    optype0 = MEMOP;
  else
    optype0 = RNDOP;

  if (REG_P (operands[1]))
    optype1 = REGOP;
  else if (CONSTANT_P (operands[1])
	   || GET_CODE (operands[1]) == CONST_DOUBLE)
    optype1 = CNSTOP;
  else if (offsettable_memref_p (operands[1]))
    optype1 = OFFSOP;
  else if (GET_CODE (operands[1]) == MEM)
    optype1 = MEMOP;
  else
    optype1 = RNDOP;

  /* Check for the cases that the operand constraints are not
     supposed to allow to happen.  Abort if we get one,
     because generating code for these cases is painful.  */

  if (optype0 == RNDOP || optype1 == RNDOP)
    abort ();

  /* If an operand is an unoffsettable memory ref, find a register
     we can increment temporarily to make it refer to the second word.  */

  if (optype0 == MEMOP)
    addreg0 = find_addr_reg (XEXP (operands[0], 0));

  if (optype1 == MEMOP)
    addreg1 = find_addr_reg (XEXP (operands[1], 0));

  /* Ok, we can do one word at a time.
     Normally we do the low-numbered word first,
     but if either operand is autodecrementing then we
     do the high-numbered word first.

     In either case, set up in LATEHALF the operands to use
     for the high-numbered word and in some cases alter the
     operands in OPERANDS to be suitable for the low-numbered word.  */

  if (optype0 == REGOP)
    latehalf[0] = gen_rtx (REG, SImode, REGNO (operands[0]) + 1);
  else if (optype0 == OFFSOP)
    latehalf[0] = adj_offsettable_operand (operands[0], 4);
  else
    latehalf[0] = operands[0];

  if (optype1 == REGOP)
    latehalf[1] = gen_rtx (REG, SImode, REGNO (operands[1]) + 1);
  else if (optype1 == OFFSOP)
    latehalf[1] = adj_offsettable_operand (operands[1], 4);
  else if (optype1 == CNSTOP)
    {
      if (CONSTANT_P (operands[1]))
	latehalf[1] = const0_rtx;
      else if (GET_CODE (operands[1]) == CONST_DOUBLE)
	{
	  latehalf[1] = gen_rtx (CONST_INT, VOIDmode,
				 CONST_DOUBLE_HIGH (operands[1]));
	  operands[1] = gen_rtx (CONST_INT, VOIDmode,
				 CONST_DOUBLE_LOW (operands[1]));
	}
    }
  else
    latehalf[1] = operands[1];

  /* If the first move would clobber the source of the second one,
     do them in the other order.

     RMS says "This happens only for registers;
     such overlap can't happen in memory unless the user explicitly
     sets it up, and that is an undefined circumstance."

     but it happens on the sparc when loading parameter registers,
     so I am going to define that circumstance, and make it work
     as expected.  */

  /* Easy case: try moving both words at once.  */
  /* First check for moving between an even/odd register pair
     and a memory location.  */
  if ((optype0 == REGOP && optype1 != REGOP && optype1 != CNSTOP
       && (REGNO (operands[0]) & 1) == 0)
      || (optype0 != REGOP && optype1 != CNSTOP && optype1 == REGOP
	  && (REGNO (operands[1]) & 1) == 0))
    {
      rtx op1, op2;
      rtx base = 0, offset = const0_rtx;

      /* OP1 gets the register pair, and OP2 gets the memory address.  */
      if (optype0 == REGOP)
	op1 = operands[0], op2 = XEXP (operands[1], 0);
      else
	op1 = operands[1], op2 = XEXP (operands[0], 0);

      /* Now see if we can trust the address to be 8-byte aligned.  */
      /* Trust global variables.  */
      if (CONSTANT_ADDRESS_P (op2))
	{
	  if (op1 == operands[0]) {
	    output_asm_insn ("seti\t%1,gr31", operands);
            if (FP_REG_P (operands[0]))
	      return "ld.d\tgr31,gr0,%0";
	    else
	      return "ld.l\tgr31,gr0,%0";
	  }
          else {
	    output_asm_insn ("seti\t%0,gr31", operands);
            if (FP_REG_P (operands[1]))
	      return "st.d\t%1,gr31,gr0";
	    else {
	      output_asm_insn ("st\t%1,gr31,gr0", operands);
              operands[1] = gen_rtx (REG, VOIDmode, REGNO (operands[1]) + 1);
	      return "st\t%1,gr31,4";
	    }
          }
	}

      if (GET_CODE (op2) == PLUS)
	{
	  if (GET_CODE (XEXP (op2, 0)) == REG)
	    base = XEXP (op2, 0), offset = XEXP (op2, 1);
	  else if (GET_CODE (XEXP (op2, 1)) == REG)
	    base = XEXP (op2, 1), offset = XEXP (op2, 0);
	}

      /* Trust round enough offsets from the stack or frame pointer.  */
      if (base
	  && (REGNO (base) == FRAME_POINTER_REGNUM
	      || REGNO (base) == STACK_POINTER_REGNUM))
	{
	  if (GET_CODE (offset) == CONST_INT
	      && (INTVAL (offset) & 0x7) == 0)
	    {
	      if (op1 == operands[0]) {
                if (FP_REG_P (operands[0])) 
		  return "ld.d\t%1,%0";
	        else
		  return "ld.l\t%1,%0";
              }
              else {
                if (FP_REG_P (operands[1])) 
		  return "st.d\t%1,%0";
	        else {
                  rtx xoperands[2];
                  xoperands[1] = gen_rtx (REG, GET_MODE (operands[1]), REGNO (operands[1]) + 1);
                  xoperands[0] = gen_rtx (MEM, GET_MODE (operands[0]), plus_constant (XEXP (operands[0], 0), 4));
	          output_asm_insn ("st\t%1,%0", operands);
	          output_asm_insn ("st\t%1,%0", xoperands);
                  return "";
                }
              }
	    }
	}
      else
	{
	  /* We know structs not on the stack are properly aligned.
	     Since a double asks for 8-byte alignment,
	     we know it must have got that if it is in a struct.
	     But a DImode need not be 8-byte aligned, because it could be a
	     struct containing two ints or pointers.  */

	  if (GET_CODE (operands[1]) == MEM && GET_MODE (operands[1]) == DFmode
	      && MEM_IN_STRUCT_P (operands[1])) {
            if (FP_REG_P (operands[0])) 
	      return "ld.d\t%1,%0";
	    else
	      return "ld.l\t%1,%0";
          }
	  else if (GET_CODE (operands[0]) == MEM
		   && GET_MODE (operands[0]) == DFmode
		   && MEM_IN_STRUCT_P (operands[0])) {
            if (FP_REG_P (operands[1]))
	      return "st.d\t%1,%0";
            else {
              rtx xoperands[2];
      	      if (GET_CODE (op2)==PLUS && GET_CODE(XEXP(op2,0))==REG && GET_CODE(XEXP(op2,1))==REG)
                {
                  xoperands[1] = gen_rtx (REG, GET_MODE (operands[1]), REGNO (operands[1]) + 1);
	          output_asm_insn ("st\t%1,%0", operands);
	          output_asm_insn ("addl\t%0,gr31", operands);
	          output_asm_insn ("st\t%1,gr31,4", xoperands);
                  return "";
                }
              else
                {
                  xoperands[1] = gen_rtx (REG, GET_MODE (operands[1]), REGNO (operands[1]) + 1);
                  xoperands[0] = gen_rtx (MEM, GET_MODE (operands[0]), plus_constant (XEXP (operands[0], 0), 4));
	          output_asm_insn ("st\t%1,%0", operands);
	          output_asm_insn ("st\t%1,%0", xoperands);
                  return "";
                 }
            }
          }
	}
    }

  if (optype0 == REGOP && optype1 == REGOP
      && REGNO (operands[0]) == REGNO (latehalf[1]))
    {
      /* Make any unoffsettable addresses point at high-numbered word.  */
      if (addreg0)
	output_asm_insn ("addli5\t%0,4,%0", &addreg0);
      if (addreg1)
	output_asm_insn ("addli5\t%0,4,%0", &addreg1);

      /* Do that word.  */
      output_asm_insn (singlemove_string (latehalf), latehalf);

      /* Undo the adds we just did.  */
      if (addreg0)
	output_asm_insn ("addli5\t%0,-4,%0", &addreg0);
      if (addreg1)
	output_asm_insn ("addli5\t%0,-4,%0", &addreg0);

      /* Do low-numbered word.  */
      return singlemove_string (operands);
    }
  else if (optype0 == REGOP && optype1 != REGOP
	   && reg_overlap_mentioned_p (operands[0], operands[1]))
    {
      /* Do the late half first.  */
      output_asm_insn (singlemove_string (latehalf), latehalf);
      /* Then clobber.  */
      return singlemove_string (operands);
    }

  /* Normal case: do the two words, low-numbered first.  */

  output_asm_insn (singlemove_string (operands), operands);

  /* Make any unoffsettable addresses point at high-numbered word.  */
  if (addreg0)
    output_asm_insn ("addli5\t%0,4,%0", &addreg0);
  if (addreg1)
    output_asm_insn ("addli5\t%0,4,%0", &addreg1);

  /* Do that word.  */
  output_asm_insn (singlemove_string (latehalf), latehalf);

  /* Undo the adds we just did.  */
  if (addreg0)
    output_asm_insn ("addli5\t%0,-4,%0", &addreg0);
  if (addreg1)
    output_asm_insn ("addli5\t%0,-4,%0", &addreg1);

  return "";
}

static char *
output_fp_move_double (operands)
     rtx *operands;
{
  if (FP_REG_P (operands[0]))
    {
      if (FP_REG_P (operands[1]))
	return "fadd.d\t%1,fr0,%0";
      if (GET_CODE (operands[1]) == REG)
	{
	  if ((REGNO (operands[1]) & 1) == 0)
            {
	    /* output_asm_insn ("st\t%1,gr2,-16\n\tst\t%2,gr2,-12\n\tld.d\tgr2,-16,%0", xoperands); */
	      return "mov.ld\t%1,%0";
            }
	  else
            {
	      rtx xoperands[3];
	      xoperands[0] = operands[0];
	      xoperands[1] = operands[1];
	      xoperands[2] = gen_rtx (REG, SImode, REGNO (operands[1]) + 1);
	      output_asm_insn ("addl\t%1,gr0,gr30\n\taddl\t%2,gr0,gr31\n\tmov.ld\tgr30,%0", xoperands);
	      return "";
	    }
	}
      if (GET_CODE (XEXP (operands[1], 0)) == PLUS
	  && (XEXP (XEXP (operands[1], 0), 0) == frame_pointer_rtx
	      || XEXP (XEXP (operands[1], 0), 0) == stack_pointer_rtx)
	  && GET_CODE (XEXP (XEXP (operands[1], 0), 1)) == CONST_INT
	  && (INTVAL (XEXP (XEXP (operands[1], 0), 1)) & 0x7) != 0)
	{
	  rtx xoperands[1];
	  xoperands[0] = gen_rtx (MEM, GET_MODE (operands[1]),
				  plus_constant (XEXP (operands[1], 0), 4));
	  output_asm_insn ("ld\t%1,gr30", operands);
	  output_asm_insn ("ld\t%0,gr31", xoperands);
	  return "mov.ld\tgr30,%0";
	}
      else if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
	{
	  output_asm_insn ("seti\t%m1,gr31", operands);
	  return "ld.d\tgr31,gr0,%0";
	}
      return "ld.d\t%1,%0";
    }
  else if (FP_REG_P (operands[1]))
    {
      if (GET_CODE (operands[0]) == REG)
	{
	  if ((REGNO (operands[0]) & 1) == 0)
	    /* return "st.d\t%1,gr2,-16\n\tld.l\tgr2,-16,%0"; */
	    return "mov.dl\t%1,%0";
	  else
	    {
	      rtx xoperands[3];
	      xoperands[2] = operands[1];
	      xoperands[1] = gen_rtx (REG, SImode, REGNO (operands[0]) + 1);
	      xoperands[0] = operands[0];
	      output_asm_insn ("mov.dl\t%2,gr30\n\taddl\tgr30,gr0,%0\n\taddl\tgr31,gr0,%1", xoperands);
	      return "";
	    }
	}
      if (GET_CODE (XEXP (operands[0], 0)) == PLUS
	  && (XEXP (XEXP (operands[0], 0), 0) == frame_pointer_rtx
	      || XEXP (XEXP (operands[0], 0), 0) == stack_pointer_rtx)
	  && GET_CODE (XEXP (XEXP (operands[0], 0), 1)) == CONST_INT
	  && (INTVAL (XEXP (XEXP (operands[0], 0), 1)) & 0x7) != 0)
	{
	  rtx xoperands[1];
	  output_asm_insn ("mov.dl\t%1,gr30", operands);
	  xoperands[0] = gen_rtx (MEM, GET_MODE (operands[0]),
				  plus_constant (XEXP (operands[0], 0), 4));
	  output_asm_insn ("st\tgr30,%0", operands);
	  output_asm_insn ("st\tgr31,%0", xoperands);
	  return "";
	}
      if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
	{
	  output_asm_insn ("seti\t%m0,gr31", operands);
	  return "st.d\t%1,gr31,gr0";
	}
      return "st.d\t%1,%0";
    }
  else abort ();
}

/* Return a REG that occurs in ADDR with coefficient 1.
   ADDR can be effectively incremented by incrementing REG.  */

static rtx
find_addr_reg (addr)
     rtx addr;
{
  while (GET_CODE (addr) == PLUS)
    {
      if (GET_CODE (XEXP (addr, 0)) == REG)
	addr = XEXP (addr, 0);
      else if (GET_CODE (XEXP (addr, 1)) == REG)
	addr = XEXP (addr, 1);
      else if (CONSTANT_P (XEXP (addr, 0)))
	addr = XEXP (addr, 1);
      else if (CONSTANT_P (XEXP (addr, 1)))
	addr = XEXP (addr, 0);
      else
        abort();
    }
  if (GET_CODE (addr) == REG)
    return addr;
  abort();
}

/* Output an ascii string.  */
output_ascii (file, p, size)
     FILE *file;
     char *p;
     int size;
{
  int i;

  fprintf (file, "\t.ascii\t\"");

  for (i = 0; i < size; i++)
    {
      register int c = p[i];
      if (c == '\"' || c == '\\')
	putc ('\\', file);
      if (c >= ' ' && c < 0177)
	putc (c, file);
      else
	{
	  fprintf (file, "\\%03o", c);
	  /* After an octal-escape, if a digit follows,
	     terminate one string constant and start another.
	     The Vax assembler fails to stop reading the escape
	     after three digits, so this is the only way we
	     can get it to parse the data properly.  */
	  if (i < size - 1 && p[i + 1] >= '0' && p[i + 1] <= '9')
	    fprintf (file, "\"\n\t.ascii\t\"");
	}
    }
  fprintf (file, "\"\n");
}

/* Since condition codes don't have logical links, we need to keep
   their setting and use together for set-cc insns.  */
void
gen_scc_insn (code, mode, operands)
     enum rtx_code code;
     enum machine_mode mode;
     rtx *operands;
{
  extern rtx sequence_stack;
  rtx last_insn = XEXP (XEXP (sequence_stack, 1), 0);
  rtx last_pat = PATTERN (last_insn);
  if (GET_CODE (last_pat) != SET
      || GET_CODE (SET_DEST (last_pat)) != CC0)
    abort ();
  SET_DEST (last_pat) = operands[0];
  SET_SRC (last_pat) = gen_rtx (code, mode, SET_SRC (last_pat), const0_rtx);
}

/* Output reasonable peephole for set-on-condition-code insns.
   Note that these insns assume a particular way of defining
   labels.  Therefore, *both* tm-tachyon.h and this function must
   be changed if a new syntax is needed.  */

char *
output_scc_insn (code, operand)
     enum rtx_code code;
     rtx operand;
{
  rtx xoperands[2];
  rtx label = gen_label_rtx ();
  int cc_in_fccr = cc_status.flags & CC_IN_FCCR;

  xoperands[0] = operand;
  xoperands[1] = label;

  output_asm_insn ("seti\t1,%0", xoperands);
  switch (code)
    {
    case NE:
      if (cc_in_fccr)
	output_asm_insn ("brncs\tfcc.e,%l0", &label);
      else
	output_asm_insn ("brncs\ticcp.z,%l0", &label);
      break;
    case EQ:
      if (cc_in_fccr)
	output_asm_insn ("brcs\tfcc.e,%l0", &label);
      else
	output_asm_insn ("brcs\ticcp.z,%l0", &label);
      break;
    case GE:
      if (cc_in_fccr)
	output_asm_insn ("brncs\tfcc.l,%l0", &label);
      else
	output_asm_insn ("brncs\ticcp.l,%l0", &label);
      break;
    case GT:
      if (cc_in_fccr)
	output_asm_insn ("brncs\tfcc.le,%l0", &label);
      else
	output_asm_insn ("brncs\ticcp.le,%l0", &label);
      break;
    case LE:
      if (cc_in_fccr)
	output_asm_insn ("brcs\tfcc.le,%l0", &label);
      else
	output_asm_insn ("brcs\ticcp.le,%l0", &label);
      break;
    case LT:
      if (cc_in_fccr)
	output_asm_insn ("brcs\tfcc.l,%l0", &label);
      else
	output_asm_insn ("brcs\ticcp.l,%l0", &label);
      break;
    case GEU:
      if (cc_in_fccr)
	abort ();
      else
	output_asm_insn ("brncs\ticcp.ul,%l0", &label);
      break;
    case GTU:
      if (cc_in_fccr)
	abort ();
      else
	output_asm_insn ("brncs\ticcp.ule,%l0", &label);
      break;
    case LEU:
      if (cc_in_fccr)
	abort ();
      else
	output_asm_insn ("brcs\ticcp.ule,%l0", &label);
      break;
    case LTU:
      if (cc_in_fccr)
	abort ();
      else
	output_asm_insn ("brcs\ticcp.ul,%l0", &label);
      break;
    default:
      abort ();
    }
  output_asm_insn ("addl\tgr0,gr0,%0\n%l1:", xoperands);
  return "";
}

/* Output a store-in-memory whose operands are OPERANDS[0,1].
   OPERANDS[0] is a MEM, and OPERANDS[1] is a reg or zero.  */

char *
output_store (operands)
     rtx *operands;
{
  enum machine_mode mode = GET_MODE (operands[1]);
  rtx address = XEXP (operands[0], 0);

  /* Store zero in two parts when appropriate.  */
  if (mode == DFmode && operands[1] == dconst0_rtx)
    {
      if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
        {
          output_asm_insn ("seti\t%m0,gr31", operands);
          return "st.d\t%1,gr31,gr0";
        }
      return "st.d\t%1,%0";
    }

  if (mode == DFmode || mode == DImode)
    {
      if (FP_REG_P (operands[1]))
        {
          if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
            {
              output_asm_insn ("seti\t%m0,gr31", operands);
              return "st.d\t%1,gr31,gr0";
            }
          return "st.d\t%1,%0";
        }
      else /* case REG_P */
        return output_move_double (operands);
    }
  else /* case SFmode */
    {
      if (FP_REG_P (operands[1]))
        {
          if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
            {
              output_asm_insn ("seti\t%m0,gr31", operands);
              return "st.f\t%1,gr31,gr0";
            } 
          return "st.f\t%1,%0";
        }
      else /* case REG_P */
        {
          if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
            {
              output_asm_insn ("seti\t%m0,gr31", operands);
              return "st\t%1,gr31,gr0";
            }
          return "st\t%1,%0";
        }
    }
}

/* Output a load-from-memory whose operands are OPERANDS[0,1].
   OPERANDS[0] is a reg, and OPERANDS[1] is a mem.  */

char *
output_load (operands)
     rtx *operands;
{
  enum machine_mode mode = GET_MODE (operands[0]);
  rtx address = XEXP (operands[1], 0);

  if (mode == DFmode || mode == DImode)
    {
      if (FP_REG_P (operands[0]))
        {
          if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
            {
              output_asm_insn ("seti\t%m1,gr31", operands);
              return "ld.d\tgr31,gr0,%0";
            } 
          return "ld.d\t%1,%0";
        }
      else /* case REG_P */
        {
          if (REGNO (operands[0]) & 1)
            return output_move_double (operands);
          else
            {
              if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
                {
                  output_asm_insn ("seti\t%m1,gr31", operands);
                  return "ld.l\tgr31,gr0,%0";
                }
              return "ld.l\t%1,%0";
            }
        }
    }
  else /* case SFmode */
    {
      if (FP_REG_P (operands[0]))
        {
          if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
            {
              output_asm_insn ("seti\t%m1,gr31", operands);
              return "ld.f\tgr31,gr0,%0";
            }
          return "ld.f\t%1,%0";
        }
      else /* case REG_P */
        {
          if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
            {
              output_asm_insn ("seti\t%m1,gr31", operands);
              return "ld\tgr31,gr0,%0";
            }
          return "ld\t%1,%0";
        }
    }
}
