/* @(#)tachyon.md	1.4 3/19/92 */

;;;- Machine description for TaChyon for GNU C compiler
;;   Copyright (C) 1988 Free Software Foundation, Inc.
;;   Contributed by Y.Nakashima

;; This file is part of GNU CC.

;; GNU CC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 1, or (at your option)
;; any later version.

;; GNU CC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU CC; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


;;- See file "rtl.def" for documentation on define_insn, match_*, et. al.

;;- cpp macro #define NOTICE_UPDATE_CC in file tm.h handles condition code
;;- updates for most instructions.

;;- Operand classes for the register allocator:

;; Compare instructions.
;; This pattern is used for generating an "insn"
;; which does just a compare and sets a (fictitious) condition code.

;; The actual TaChyon insns are compare-and-conditional-jump.
;; The define_peephole's below recognize the combinations of
;; compares and jumps, and output each pair as a single assembler insn.

;; This controls RTL generation and register allocation.
(define_insn "cmpsi"
  [(set (cc0)
	(compare (match_operand:SI 0 "arith32_operand" "r,rn")
		 (match_operand:SI 1 "arith32_operand" "n,r")))]
  ""
  "*
{
  if (FITS_5_BITS (operands[0]))
    {
      cc_status.flags |= CC_REVERSED;
      return \"subli5\\\t%1,%0,gr0\";
    }
  else if (FITS_25_BITS (operands[0]))
    {
      cc_status.flags |= CC_REVERSED;
      return \"subli25\\\t%1,%0,gr0\";
    }
  else if (GET_CODE (operands[0]) == CONST_INT)
      return \"seti\\\t%0,gr31\;subl\\\t%1,gr31,gr0\";
  else if (FITS_5_BITS (operands[1]))
      return \"subli5\\\t%0,%1,gr0\";
  else if (FITS_25_BITS (operands[1]))
      return \"subli25\\\t%0,%1,gr0\";
  else if (GET_CODE (operands[1]) == CONST_INT)
      return \"seti\\\t%1,gr31\;subl\\\t%0,gr31,gr0\";
  else
      return \"subl\\\t%0,%1,gr0\";

}")

(define_insn "cmpdf"
  [(set (cc0)
	(compare (match_operand:DF 0 "nonmemory_operand" "f,fG")
		 (match_operand:DF 1 "nonmemory_operand" "G,f")))]
  ""
  "*
{
  cc_status.flags |= CC_IN_FCCR;
  if (GET_CODE (operands[0]) == CONST_DOUBLE)
    return \"fcmp.d\\\tfr0,%1\";
  else if (GET_CODE (operands[1]) == CONST_DOUBLE)
    return \"fcmp.d\\\t%0,fr0\";
  else
    return \"fcmp.d\\\t%0,%1\";
}")

(define_insn "cmpsf"
  [(set (cc0)
	(compare (match_operand:SF 0 "nonmemory_operand" "f,fG")
		 (match_operand:SF 1 "nonmemory_operand" "G,f")))]
  ""
  "*
{
  cc_status.flags |= CC_IN_FCCR;
  if (GET_CODE (operands[0]) == CONST_DOUBLE)
    return \"fcmp\\\tfr0,%1\";
  else if (GET_CODE (operands[1]) == CONST_DOUBLE)
    return \"fcmp\\\t%0,fr0\";
  else
    return \"fcmp\\\t%0,%1\";
}")
;; We have to have this because cse can optimize the previous pattern
;; into this one.

(define_insn "tstsi"
  [(set (cc0)
	(match_operand:SI 0 "register_operand" "r"))]
  ""
  "*
{
  return \"subl\\\t%0,gr0,gr0\";
}")

(define_insn "tstdf"
  [(set (cc0)
	(match_operand:DF 0 "register_operand" "f"))]
  ""
  "*
{
  cc_status.flags |= CC_IN_FCCR;
  return \"fcmp.d\\\t%0,fr0\";
}")

(define_insn "tstsf"
  [(set (cc0)
	(match_operand:SF 0 "register_operand" "f"))]
  ""
  "*
{
  cc_status.flags |= CC_IN_FCCR;
  return \"fcmp\\\t%0,fr0\";
}")

;; Store condition code values into registers
;; There are no logical links for the condition codes.  This
;; would not normally be a problem, but on TaChyon (and possibly
;; other RISC machines), when argument passing, the insn which sets
;; the condition code and the insn which uses the set condition code
;; may not be performed adjacently (due to optimizations performed
;; in combine.c).  To make up for this, we emit insn patterns which
;; cannot possibly be rearranged on us.
(define_expand "seq"
  [(set (match_operand:SI 0 "general_operand" "=r")
        (eq (cc0) (const_int 0)))]
  ""
  "gen_scc_insn (EQ, VOIDmode, operands); DONE;")

(define_expand "sne"
  [(set (match_operand:SI 0 "general_operand" "=r")
        (ne (cc0) (const_int 0)))]
  ""
  "gen_scc_insn (NE, VOIDmode, operands); DONE;")

(define_insn ""
  [(set (match_operand:SI 0 "general_operand" "=r,r")
	(match_operator 1 "eq_or_neq"
			[(compare (match_operand:SI 2 "general_operand" "r,rI")
				  (match_operand:SI 3 "general_operand" "I,r"))
			 (const_int 0)]))]
  ""
  "*
{
  if (! REG_P (operands[2]))
    {
      if (FITS_5_BITS (operands[2])) {
          output_asm_insn (\"subli5\\\t%3,%2,gr0\", operands);
          cc_status.flags |= CC_REVERSED;
      }
      else if (FITS_25_BITS (operands[2])) {
          output_asm_insn (\"subli25\\\t%3,%2,gr0\", operands);
          cc_status.flags |= CC_REVERSED;
      }
      else
          output_asm_insn (\"seti\\\t%2,gr31\;subl\\\tgr31,%3,gr0\", operands);
    }
  else
    {
      if (FITS_5_BITS (operands[3]))
          output_asm_insn (\"subli5\\\t%2,%3,gr0\", operands);
      else if (FITS_25_BITS (operands[3]))
          output_asm_insn (\"subli25\\\t%2,%3,gr0\", operands);
      else if (GET_CODE (operands[3]) == CONST_INT)
          output_asm_insn (\"seti\\\t%3,gr31\;subl\\\t%2,gr31,gr0\", operands);
      else
          output_asm_insn (\"subl\\\t%2,%3,gr0\", operands);
    }
  return output_scc_insn (GET_CODE (operands[1]), operands[0]);
}")

(define_insn ""
  [(set (match_operand:SI 0 "general_operand" "=r")
	(match_operator 1 "eq_or_neq"
			[(match_operand:SI 2 "general_operand" "r")
			 (const_int 0)]))]
  ""
  "* output_asm_insn (\"subl\\\t%2,gr0,gr0\", operands);
     return output_scc_insn (GET_CODE (operands[1]), operands[0]);")

(define_insn ""
  [(set (match_operand:SI 0 "general_operand" "=r,r")
	(match_operator 1 "eq_or_neq"
			[(compare (match_operand:DF 2 "general_operand" "f,fG")
				  (match_operand:DF 3 "general_operand" "G,f"))
			 (const_int 0)]))]
  ""
  "*
{
  cc_status.flags |= CC_IN_FCCR;
  if (GET_CODE (operands[2]) == CONST_DOUBLE)
    output_asm_insn (\"fcmp.d\\\tfr0,%3\", operands);
  else if (GET_CODE (operands[3]) == CONST_DOUBLE)
    output_asm_insn (\"fcmp.d\\\t%2,fr0\", operands);
  else output_asm_insn (\"fcmp.d\\\t%2,%3\", operands);
  return output_scc_insn (GET_CODE (operands[1]), operands[0]);
}")

(define_insn ""
  [(set (match_operand:SI 0 "general_operand" "=r")
	(match_operator 1 "eq_or_neq"
			[(match_operand:DF 2 "general_operand" "f")
			 (const_int 0)]))]
  ""
  "*
{
  cc_status.flags |= CC_IN_FCCR;
  output_asm_insn (\"fcmp.d\\\t%2,fr0\", operands);
  return output_scc_insn (GET_CODE (operands[1]), operands[0]);
}")

(define_insn ""
  [(set (match_operand:SI 0 "general_operand" "=r,r")
	(match_operator 1 "eq_or_neq"
			[(compare (match_operand:SF 2 "general_operand" "f,fG")
				  (match_operand:SF 3 "general_operand" "G,f"))
			 (const_int 0)]))]
  ""
  "*
{
  cc_status.flags |= CC_IN_FCCR;
  if (GET_CODE (operands[2]) == CONST_DOUBLE)
    output_asm_insn (\"fcmp\\\tfr0,%3\", operands);
  else if (GET_CODE (operands[3]) == CONST_DOUBLE)
    output_asm_insn (\"fcmp\\\t%2,fr0\", operands);
  else output_asm_insn (\"fcmp\\\t%2,%3\", operands);
  return output_scc_insn (GET_CODE (operands[1]), operands[0]);
}")

(define_insn ""
  [(set (match_operand:SI 0 "general_operand" "=r")
	(match_operator 1 "eq_or_neq"
			[(match_operand:SF 2 "general_operand" "f")
			 (const_int 0)]))]
  ""
  "*
{
  cc_status.flags |= CC_IN_FCCR;
  output_asm_insn (\"fcmp\\\t%2,fr0\", operands);
  return output_scc_insn (GET_CODE (operands[1]), operands[0]);
}")

;; These control RTL generation for conditional jump insns
;; and match them for register allocation.

(define_insn "beq"
  [(set (pc)
	(if_then_else (eq (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brc\\\tfcc.e,%l0\";
  else
    return \"brc\\\ticcp.z,%l0\";
}")

(define_insn "bne"
  [(set (pc)
	(if_then_else (ne (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brnc\\\tfcc.e,%l0\";
  else
    return \"brnc\\\ticcp.z,%l0\";
}")

(define_insn "bgt"
  [(set (pc)
	(if_then_else (gt (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brnc\\\tfcc.le,%l0\";
  else
    return \"brnc\\\ticcp.le,%l0\";
}")

(define_insn "bgtu"
  [(set (pc)
	(if_then_else (gtu (cc0)
			   (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    abort ();
  else
    return \"brnc\\\ticcp.ule,%l0\";
}")

(define_insn "blt"
  [(set (pc)
	(if_then_else (lt (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brc\\\tfcc.l,%l0\";
  else
    return \"brc\\\ticcp.l,%l0\";
}")

(define_insn "bltu"
  [(set (pc)
	(if_then_else (ltu (cc0)
			   (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    abort ();
  else
    return \"brc\\\ticcp.ul,%l0\";
}")

(define_insn "bge"
  [(set (pc)
	(if_then_else (ge (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brnc\\\tfcc.l,%l0\";
  else
    return \"brnc\\\ticcp.l,%l0\";
}")

(define_insn "bgeu"
  [(set (pc)
	(if_then_else (geu (cc0)
			   (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    abort ();
  else
    return \"brnc\\\ticcp.ul,%l0\";
}")

(define_insn "ble"
  [(set (pc)
	(if_then_else (le (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brc\\\tfcc.le,%l0\";
  else
    return \"brc\\\ticcp.le,%l0\";
}")

(define_insn "bleu"
  [(set (pc)
	(if_then_else (leu (cc0)
			   (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    abort ();
  else
    return \"brc\\\ticcp.ule,%l0\";
}")

;; These match inverted jump insns for register allocation.

(define_insn ""
  [(set (pc)
	(if_then_else (eq (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brnc\\\tfcc.e,%l0\";
  else
    return \"brnc\\\ticcp.z,%l0\";
}")

(define_insn ""
  [(set (pc)
	(if_then_else (ne (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brc\\\tfcc.e,%l0\";
  else
    return \"brc\\\ticcp.z,%l0\";
}")

(define_insn ""
  [(set (pc)
	(if_then_else (gt (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brc\\\tfcc.le,%l0\";
  else
    return \"brc\\\ticcp.le,%l0\";
}")

(define_insn ""
  [(set (pc)
	(if_then_else (gtu (cc0)
			   (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    abort ();
  else
    return \"brc\\\ticcp.ule,%l0\";
}")

(define_insn ""
  [(set (pc)
	(if_then_else (lt (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brnc\\\tfcc.l,%l0\";
  else
    return \"brnc\\\ticcp.l,%l0\";
}")

(define_insn ""
  [(set (pc)
	(if_then_else (ltu (cc0)
			   (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    abort ();
  else
    return \"brnc\\\ticcp.ul,%l0\";
}")

(define_insn ""
  [(set (pc)
	(if_then_else (ge (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brc\\\tfcc.l,%l0\";
  else
    return \"brc\\\ticcp.l,%l0\";
}")

(define_insn ""
  [(set (pc)
	(if_then_else (geu (cc0)
			   (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    abort ();
  else
    return \"brc\\\ticcp.ul,%l0\";
}")

(define_insn ""
  [(set (pc)
	(if_then_else (le (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    return \"brnc\\\tfcc.le,%l0\";
  else
    return \"brnc\\\ticcp.le,%l0\";
}")

(define_insn ""
  [(set (pc)
	(if_then_else (leu (cc0)
			   (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  if (cc_prev_status.flags & CC_IN_FCCR)
    abort ();
  else
    return \"brnc\\\ticcp.ule,%l0\";
}")

;; Move instructions

(define_insn "swapsi"
  [(set (match_operand:SI 0 "general_operand" "r")
	(match_operand:SI 1 "general_operand" "r"))
   (set (match_dup 1) (match_dup 0))]
  ""
  "*
{
  CC_STATUS_INIT;
  return \"or\\\t%0,gr0,%1 & or %1,gr0,%0\";
}")

(define_insn "movsi"
  [(set (match_operand:SI 0 "general_operand" "=r,f,m")
	(match_operand:SI 1 "general_operand" "rfmi,rfm,rfJ"))]
  ""
  "*
{
  if (GET_CODE (operands[0]) == MEM)
    if (FP_REG_P (operands[1])) {
      if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
        return \"seti\\\t%0,gr31\;st.f\\\t%1,gr31,gr0\";
      else
        return \"st.f\\\t%1,%0\";
    }
    else if (REG_P (operands[1])) {
      if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
        return \"seti\\\t%0,gr31\;st\\\t%1,gr31,gr0\";
      else
        return \"st\\\t%1,%0\";
    }
    else {
      if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
        return \"seti\\\t%0,gr31\;st\\\tgr0,gr31,gr0\";
      else
        return \"st\\\tgr0,%0\";
    }
  else if (FP_REG_P (operands[0]))
    {
      if (GET_CODE (operands[1]) == MEM)
        {
          if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
            return \"seti\\\t%1,gr31\;ld.f\\\tgr31,gr0,%0\";
          else
            return \"ld.f\\\t%1,%0\";
        }
      else if (FP_REG_P (operands[1]))
        return \"fadd\\\t%1,fr0,%0\";
      else
        return \"mov.if\\\t%1,%0\";
    }
  else if (GET_CODE (operands[1]) == MEM)
    {
      if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
        return \"seti\\\t%1,gr31\;ld\\\tgr31,gr0,%0\";
      else
        return \"ld\\\t%1,%0\";
    }
  else if (FP_REG_P (operands[1]))
    return \"mov.fi\\\t%1,%0\";
  else if (REG_P (operands[1]))
    return \"or\\\t%1,gr0,%0\";
  else if (FITS_5_BITS (operands[1]))
    return \"addli5\\\tgr0,%1,%0\";
  else if (FITS_25_BITS (operands[1]))
    return \"addli25\\\tgr0,%1,%0\";
  else
    return \"seti\\\t%1,%0\";
}")

(define_insn "movhi"
  [(set (match_operand:HI 0 "general_operand" "=r,m")
	(match_operand:HI 1 "general_operand" "rmi,rJ"))]
  ""
  "*
{
  if (GET_CODE (operands[0]) == MEM) {
    if (REG_P (operands[1])) {
      if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
        return \"seti\\\t%0,gr31\;st.s\\\t%1,gr31,gr0\";
      else
        return \"st.s\\\t%1,%0\";
    }
    else {
      if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
        return \"seti\\\t%0,gr31\;st.s\\\tgr0,gr31,gr0\";
      else
        return \"st.s\\\tgr0,%0\";
    }
  }
  else if (GET_CODE (operands[1]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.s\\\tgr31,gr0,%0\";
    else
      return \"ld.s\\\t%1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"or\\\t%1,gr0,%0\";
  else if (FITS_5_BITS (operands[1]))
    return \"addli5\\\tgr0,%1,%0\";
  else if (FITS_25_BITS (operands[1]))
    return \"addli25\\\tgr0,%1,%0\";
  else
    return \"seti\\\t%1,%0\";
}")

(define_insn "movqi"
  [(set (match_operand:QI 0 "general_operand" "=r,m")
	(match_operand:QI 1 "general_operand" "rmi,rJ"))]
  ""
  "*
{
  if (GET_CODE (operands[0]) == MEM) {
    if (REG_P (operands[1])) {
      if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
        return \"seti\\\t%0,gr31\;st.c\\\t%1,gr31,gr0\";
      else
        return \"st.c\\\t%1,%0\";
    }
    else {
      if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
        return \"seti\\\t%0,gr31\;st.c\\\tgr0,gr31,gr0\";
      else
        return \"st.c\\\tgr0,%0\";
    }
  }
  else if (GET_CODE (operands[1]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.c\\\tgr31,gr0,%0\";
    else
      return \"ld.c\\\t%1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"or\\\t%1,gr0,%0\";
  else if (FITS_5_BITS (operands[1]))
    return \"addli5\\\tgr0,%1,%0\";
  else if (FITS_25_BITS (operands[1]))
    return \"addli25\\\tgr0,%1,%0\";
  else
    return \"seti\\\t%1,%0\";
}")


;; Floating point move insns

;; This pattern forces (set (reg:DF ...) (const_double ...))
;; to be reloaded by putting the constant into memory.
;; It must come before the more general movdf pattern.
(define_insn ""
  [(set (match_operand:DF 0 "general_operand" "=r,f,o")
        (match_operand:DF 1 "" "mG,m,G"))]
  "GET_CODE (operands[1]) == CONST_DOUBLE"
  "*
{
  if (FP_REG_P (operands[0]))
    return output_fp_move_double (operands);
  else if (operands[1] == dconst0_rtx && REG_P (operands[0]))
    {
      operands[1] = gen_rtx (REG, SImode, REGNO (operands[0]) + 1);
      return \"or\\\tgr0,gr0,%0\;or\\\tgr0,gr0,%1\";
    }
  else if (operands[1] == dconst0_rtx && GET_CODE (operands[0]) == MEM)
    {
      if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
	  return \"seti\\\t%0,gr31\;st.d\\\tfr0,gr31,gr0\";
      else
          return \"st.d\\\tfr0,%0\";
    }
  else
    return output_move_double (operands);
}")

(define_insn "movdf"
  [(set (match_operand:DF 0 "general_operand" "=rm,&r,?f,?rm")
	(match_operand:DF 1 "general_operand" "r,m,rfm,f"))]
  ""
  "*
{
  if (GET_CODE (operands[0]) == MEM && CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
    return output_store (operands);
  else if (GET_CODE (operands[1]) == MEM && CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
    return output_load (operands);

  else if (FP_REG_P (operands[0]) || FP_REG_P (operands[1]))
    return output_fp_move_double (operands);
  else
    return output_move_double (operands);
}")

(define_insn "movdi"
  [(set (match_operand:DI 0 "general_operand" "=rm,&r,?f,?rm")
	(match_operand:DI 1 "general_operand" "r,mi,rfm,f"))]
  ""
  "*
{
  if (GET_CODE (operands[0]) == MEM && CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
    return output_store (operands);
  else if (GET_CODE (operands[1]) == MEM && CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
    return output_load (operands);

  else if (FP_REG_P (operands[0]) || FP_REG_P (operands[1]))
    return output_fp_move_double (operands);
  else
    return output_move_double (operands);
}")

(define_insn "movsf"
  [(set (match_operand:SF 0 "general_operand" "=rf,m")
	(match_operand:SF 1 "general_operand" "rfm,rf"))]
  ""
  "*
{
  if (GET_CODE (operands[0]) == MEM && CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
    return output_store (operands);
  else if (GET_CODE (operands[1]) == MEM && CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
    return output_load (operands);
  else if (FP_REG_P (operands[0]))
    {
      if (FP_REG_P (operands[1]))
	return \"fadd\\\t%1,fr0,%0\";
      else if (REG_P (operands[1]))
        return \"mov.if\\\t%1,%0\";
      else if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
	return \"seti\\\t%1,gr31\;ld.f\\\tgr31,gr0,%0\";
      else
        return \"ld.f\\\t%1,%0\";
    }
  else if (FP_REG_P (operands[1]))
    {
      if (REG_P (operands[0]))
	return \"mov.fi\\\t%1,%0\";
      else if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
	return \"seti\\\t%0,gr31\;st.f\\\t%1,gr31,gr0\";
      else
	return \"st.f\\\t%1,%0\";
    }
  else if (GET_CODE (operands[0]) == MEM)
    return \"st\\\t%1,%0\";
  else if (GET_CODE (operands[1]) == MEM)
    return \"ld\\\t%1,%0\";
  else
    return \"or\\\t%1,gr0,%0\";
}")

;; load effective address
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
        (match_operand:QI 1 "address_operand" "p"))]
  ""
  "*
{
  rtx xoperands[3];

  xoperands[0] = operands[0];
  xoperands[1] = XEXP(operands[1],0);
  xoperands[2] = XEXP(operands[1],1);

  if (GET_CODE (operands[1]) == PLUS) {
    if (FITS_5_BITS (xoperands[1])) {
      output_asm_insn (\"addli5\\\t%2,%1,%0\", xoperands);
      return \"\";
    }
    else if (FITS_25_BITS (xoperands[1])) {
      output_asm_insn (\"addli25\\\t%2,%1,%0\", xoperands);
      return \"\";
    }
    else if (FITS_5_BITS (xoperands[2]))
      return \"addli5\\\t%a1,%0\";
    else if (FITS_25_BITS (xoperands[2]))
      return \"addli25\\\t%a1,%0\";
    else
      return \"addl\\\t%a1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"or\\\t%a1,gr0,%0\";
  else if (FITS_5_BITS (operands[1]))
      return \"addli5\\\tgr0,%a1,%0\";
  else if (FITS_25_BITS (operands[1]))
      return \"addli25\\\tgr0,%a1,%0\";
  else
    return \"seti\\\t%a1,%0\";
}")

(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
        (match_operand:HI 1 "address_operand" "p"))]
  ""
  "*
{
  rtx xoperands[3];

  xoperands[0] = operands[0];
  xoperands[1] = XEXP(operands[1],0);
  xoperands[2] = XEXP(operands[1],1);

  if (GET_CODE (operands[1]) == PLUS) {
    if (FITS_5_BITS (xoperands[1])) {
      output_asm_insn (\"addli5\\\t%2,%1,%0\", xoperands);
      return \"\";
    }
    else if (FITS_25_BITS (xoperands[1])) {
      output_asm_insn (\"addli25\\\t%2,%1,%0\", xoperands);
      return \"\";
    }
    else if (FITS_5_BITS (xoperands[2]))
      return \"addli5\\\t%a1,%0\";
    else if (FITS_25_BITS (xoperands[2]))
      return \"addli25\\\t%a1,%0\";
    else
      return \"addl\\\t%a1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"or\\\t%a1,gr0,%0\";
  else if (FITS_5_BITS (operands[1]))
      return \"addli5\\\tgr0,%a1,%0\";
  else if (FITS_25_BITS (operands[1]))
      return \"addli25\\\tgr0,%a1,%0\";
  else
    return \"seti\\\t%a1,%0\";
}")

(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
        (match_operand:SI 1 "address_operand" "p"))]
  ""
  "*
{
  rtx xoperands[3];

  xoperands[0] = operands[0];
  xoperands[1] = XEXP(operands[1],0);
  xoperands[2] = XEXP(operands[1],1);

  if (GET_CODE (operands[1]) == PLUS) {
    if (FITS_5_BITS (xoperands[1])) {
      output_asm_insn (\"addli5\\\t%2,%1,%0\", xoperands);
      return \"\";
    }
    else if (FITS_25_BITS (xoperands[1])) {
      output_asm_insn (\"addli25\\\t%2,%1,%0\", xoperands);
      return \"\";
    }
    else if (FITS_5_BITS (xoperands[2]))
      return \"addli5\\\t%a1,%0\";
    else if (FITS_25_BITS (xoperands[2]))
      return \"addli25\\\t%a1,%0\";
    else
      return \"addl\\\t%a1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"or\\\t%a1,gr0,%0\";
  else if (FITS_5_BITS (operands[1]))
      return \"addli5\\\tgr0,%a1,%0\";
  else if (FITS_25_BITS (operands[1]))
      return \"addli25\\\tgr0,%a1,%0\";
  else
    return \"seti\\\t%a1,%0\";
}")

(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
        (match_operand:SF 1 "address_operand" "p"))]
  ""
  "*
{
  rtx xoperands[3];

  xoperands[0] = operands[0];
  xoperands[1] = XEXP(operands[1],0);
  xoperands[2] = XEXP(operands[1],1);

  if (GET_CODE (operands[1]) == PLUS) {
    if (FITS_5_BITS (xoperands[1])) {
      output_asm_insn (\"addli5\\\t%2,%1,%0\", xoperands);
      return \"\";
    }
    else if (FITS_25_BITS (xoperands[1])) {
      output_asm_insn (\"addli25\\\t%2,%1,%0\", xoperands);
      return \"\";
    }
    else if (FITS_5_BITS (xoperands[2]))
      return \"addli5\\\t%a1,%0\";
    else if (FITS_25_BITS (xoperands[2]))
      return \"addli25\\\t%a1,%0\";
    else
      return \"addl\\\t%a1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"or\\\t%a1,gr0,%0\";
  else if (FITS_5_BITS (operands[1]))
      return \"addli5\\\tgr0,%a1,%0\";
  else if (FITS_25_BITS (operands[1]))
      return \"addli25\\\tgr0,%a1,%0\";
  else
    return \"seti\\\t%a1,%0\";
}")

(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
        (match_operand:DF 1 "address_operand" "p"))]
  ""
  "*
{
  rtx xoperands[3];

  xoperands[0] = operands[0];
  xoperands[1] = XEXP(operands[1],0);
  xoperands[2] = XEXP(operands[1],1);

  if (GET_CODE (operands[1]) == PLUS) {
    if (FITS_5_BITS (xoperands[1])) {
      output_asm_insn (\"addli5\\\t%2,%1,%0\", xoperands);
      return \"\";
    }
    else if (FITS_25_BITS (xoperands[1])) {
      output_asm_insn (\"addli25\\\t%2,%1,%0\", xoperands);
      return \"\";
    }
    else if (FITS_5_BITS (xoperands[2]))
      return \"addli5\\\t%a1,%0\";
    else if (FITS_25_BITS (xoperands[2]))
      return \"addli25\\\t%a1,%0\";
    else
      return \"addl\\\t%a1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"or\\\t%a1,gr0,%0\";
  else if (FITS_5_BITS (operands[1]))
      return \"addli5\\\tgr0,%a1,%0\";
  else if (FITS_25_BITS (operands[1]))
      return \"addli25\\\tgr0,%a1,%0\";
  else
    return \"seti\\\t%a1,%0\";
}")

;;- truncation instructions

(define_insn "truncsiqi2"
  [(set (match_operand:QI 0 "general_operand" "=g")
	(truncate:QI
	 (match_operand:SI 1 "register_operand" "r")))]
  ""
  "*
{
  if (GET_CODE (operands[0]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
      return \"seti\\\t%0,gr31\;st.c\\\t%1,gr31,gr0\";
    else
      return \"st.c\\\t%1,%0\";
  }
  else
    return \"or\\\t%1,gr0,%0\";
}")

(define_insn "trunchiqi2"
  [(set (match_operand:QI 0 "general_operand" "=g")
	(truncate:QI
	 (match_operand:HI 1 "register_operand" "r")))]
  ""
  "*
{
  if (GET_CODE (operands[0]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
      return \"seti\\\t%0,gr31\;st.c\\\t%1,gr31,gr0\";
    else
      return \"st.c\\\t%1,%0\";
  }
  else
    return \"or\\\t%1,gr0,%0\";
}")

(define_insn "truncsihi2"
  [(set (match_operand:HI 0 "general_operand" "=g")
	(truncate:HI
	 (match_operand:SI 1 "register_operand" "r")))]
  ""
  "*
{
  if (GET_CODE (operands[0]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
      return \"seti\\\t%0,gr31\;st.s\\\t%1,gr31,gr0\";
    else
      return \"st.s\\\t%1,%0\";
  }
  else
    return \"or\\\t%1,gr0,%0\";
}")

;;- zero extension instructions

;; Note that the one starting from HImode comes before those for QImode
;; so that a constant operand will match HImode, not QImode.

(define_insn "zero_extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(zero_extend:SI
	 (match_operand:HI 1 "general_operand" "g")))]
  ""
  "*
{
  if (GET_CODE (operands[1]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.su\\\tgr31,gr0,%0\";
    else
      return \"ld.su\\\t%1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"and\\\t%1,0xffff,%0\";
  else if (FITS_5_BITS (operands[1]))
    return \"addli5\\\tgr0,%1,%0\;and\\\t%0,0xffff,%0\";
  else if (FITS_16_BITS (operands[1]))
    return \"addli25\\\tgr0,%1,%0\;and\\\t%0,0xffff,%0\";
  else if (FITS_25_BITS (operands[1])) {
    warning (\"truncating constant `%d' to fit in half-word\", INTVAL (operands[1]));
    return \"addli25\\\tgr0,%1,%0\;and\\\t%0,0xffff,%0\";
  }
  else {
    warning (\"truncating constant `%d' to fit in half-word\", INTVAL (operands[1]));
    return \"seti\\\t%1,%0\;and\\\t%0,0xffff,%0\";
  }
}")

(define_insn "zero_extendqihi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(zero_extend:HI
	 (match_operand:QI 1 "general_operand" "g")))]
  ""
  "*
{
  if (GET_CODE (operands[1]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.cu\\\tgr31,gr0,%0\";
    else
      return \"ld.cu\\\t%1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"and\\\t%1,0xff,%0\";
  else if (FITS_5_BITS (operands[1]))
    return \"addli5\\\tgr0,%1,%0\;and\\\t%0,0xff,%0\";
  else if (FITS_25_BITS (operands[1])) {
    warning (\"truncating constant `%d' to fit in byte\", INTVAL (operands[1]));
    return \"addli25\\\tgr0,%1,%0\;and\\\t%0,0xff,%0\";
  }
  else {
    warning (\"truncating constant `%d' to fit in byte\", INTVAL (operands[1]));
    return \"seti\\\t%1,%0\;and\\\t%0,0xff,%0\";
  }
}")

(define_insn "zero_extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(zero_extend:SI
	 (match_operand:QI 1 "general_operand" "g")))]
  ""
  "*
{
  if (GET_CODE (operands[1]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.cu\\\tgr31,gr0,%0\";
    else
      return \"ld.cu\\\t%1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"and\\\t%1,0xff,%0\";
  else if (FITS_5_BITS (operands[1]))
    return \"addli5\\\tgr0,%1,%0\;and\\\t%0,0xff,%0\";
  else if (FITS_25_BITS (operands[1])) {
    warning (\"truncating constant `%d' to fit in byte\", INTVAL (operands[1]));
    return \"addli25\\\tgr0,%1,%0\;and\\\t%0,0xff,%0\";
  }
  else {
    warning (\"truncating constant `%d' to fit in byte\", INTVAL (operands[1]));
    return \"seti\\\t%1,%0\;and\\\t%0,0xff,%0\";
  }
}")

;;- sign extension instructions
;; Note that the one starting from HImode comes before those for QImode
;; so that a constant operand will match HImode, not QImode.

(define_insn "extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extend:SI
	 (match_operand:HI 1 "general_operand" "g")))]
  ""
  "*
{
  if (GET_CODE (operands[1]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.s\\\tgr31,gr0,%0\";
    else
      return \"ld.s\\\t%1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"sll\\\t%1,0x10,%0\;sra\\\t%0,0x10,%0\";
  else if (FITS_5_BITS (operands[1]))
    return \"addli5\\\tgr0,%1,%0\";
  else if (FITS_25_BITS (operands[1]))
    return \"addli25\\\tgr0,%1,%0\";
  else
    return \"seti\\\t%1,%0\";
}")

(define_insn "extendqihi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(sign_extend:HI
	 (match_operand:QI 1 "general_operand" "g")))]
  ""
  "*
{
  if (GET_CODE (operands[1]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.c\\\tgr31,gr0,%0\";
    else
      return \"ld.c\\\t%1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"sll\\\t%1,0x18,%0\;sra\\\t%0,0x18,%0\";
  else if (FITS_5_BITS (operands[1]))
    return \"addli5\\\tgr0,%1,%0\";
  else if (FITS_25_BITS (operands[1]))
    return \"addli25\\\tgr0,%1,%0\";
  else
    return \"seti\\\t%1,%0\";
}")

(define_insn "extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extend:SI
	 (match_operand:QI 1 "general_operand" "g")))]
  ""
  "*
{
  if (GET_CODE (operands[1]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.c\\\tgr31,gr0,%0\";
    else
      return \"ld.c\\\t%1,%0\";
  }
  else if (REG_P (operands[1]))
    return \"sll\\\t%1,0x18,%0\;sra\\\t%0,0x18,%0\";
  else if (FITS_5_BITS (operands[1]))
    return \"addli5\\\tgr0,%1,%0\";
  else if (FITS_25_BITS (operands[1]))
    return \"addli25\\\tgr0,%1,%0\";
  else
    return \"seti\\\t%1,%0\";
}")

;; Conversions between float and double.
(define_insn "extendsfdf2"
  [(set (match_operand:DF 0 "general_operand" "=f")
	(float_extend:DF
	 (match_operand:SF 1 "general_operand" "f")))]
  ""
  "fcnv.fd\\\t%1,%0")

(define_insn "truncdfsf2"
  [(set (match_operand:SF 0 "general_operand" "=f")
	(float_truncate:SF
	 (match_operand:DF 1 "general_operand" "f")))]
  ""
  "fcnv.df\\\t%1,%0")

;; Conversions between floating point and integer
(define_insn ""
  [(set (match_operand:SF 0 "general_operand" "=f")
        (float:SF (match_operand 1 "" "m")))]
  "GET_CODE (operands[1]) == CONST_INT"
  "*
{
  if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.f\\\tgr31,gr0,%0\;fcnv.if\\\t%0,%0\";
  else
      return \"ld.f\\\t%1,%0\;fcnv.if\\\t%0,%0\";
}")

(define_insn "floatsisf2"
  [(set (match_operand:SF 0 "general_operand" "=f")
	(float:SF (match_operand:SI 1 "general_operand" "rfm")))]
  ""
  "*
{
  if (GET_CODE (operands[1]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
        return \"seti\\\t%1,gr31\;ld.f\\\tgr31,gr0,%0\;fcnv.if\\\t%0,%0\";
    else
      return \"ld.f\\t%1,%0\;fcnv.if\\\t%0,%0\";
  }
  else if (FP_REG_P (operands[1]))
    return \"fcnv.if\\\t%1,%0\";
  else
    return \"mov.if\\\t%1,%0\;fcnv.if\\\t%0,%0\";
}")

(define_insn ""
  [(set (match_operand:DF 0 "general_operand" "=f")
        (float:DF (match_operand 1 "" "m")))]
  "GET_CODE (operands[1]) == CONST_INT"
  "*
{
  if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.f\\\tgr31,gr0,%0\;fcnv.id\\\t%0,%0\";
  else
      return \"ld.f\\\t%1,%0\;fcnv.id\\\t%0,%0\";
}")

(define_insn "floatsidf2"
  [(set (match_operand:DF 0 "general_operand" "=f")
	(float:DF (match_operand:SI 1 "general_operand" "rfm")))]
  ""
  "*
{
  if (GET_CODE (operands[1]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.f\\\tgr31,gr0,%0\;fcnv.id\\\t%0,%0\";
    else
      return \"ld.f\\t%1,%0\;fcnv.id\\\t%0,%0\";
  }
  else if (FP_REG_P (operands[1]))
    return \"fcnv.id\\\t%1,%0\";
  else
    return \"mov.if\\\t%1,%0\;fcnv.id\\\t%0,%0\";
}")

(define_insn "fix_truncdfsi2"
  [(set (match_operand:SI 0 "general_operand" "=rm")
	(fix:SI (fix:DF (match_operand:DF 1 "general_operand" "fm"))))]
  ""
  "*
{
  if (FP_REG_P (operands[1]))
    output_asm_insn (\"fcnz.di\\\t%1,fr1\", operands);
  else if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
    output_asm_insn (\"seti\\\t%1,gr31\;ld.d\\\tgr31,gr0,fr1\;fcnz.di\\\tfr1,fr1\", operands);
  else
    output_asm_insn (\"ld.d\\\t%1,fr1\;fcnz.di\\\tfr1,fr1\", operands);
  if (GET_CODE (operands[0]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
      return \"seti\\\t%0,gr31\;st.f\\\tfr1,gr31,gr0\";
    else
      return \"st.f\\\tfr1,%0\";
  }
  else
    return \"mov.fi\\\tfr1,%0\";
}")

(define_insn "fix_truncsfsi2"
  [(set (match_operand:SI 0 "general_operand" "=rm")
	(fix:SI (fix:SF (match_operand:SF 1 "general_operand" "fm"))))]
  ""
  "*
{
  if (FP_REG_P (operands[1]))
    output_asm_insn (\"fcnz.fi\\\t%1,fr1\", operands);
  else if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
    output_asm_insn (\"seti\\\t%1,gr31\;ld.f\\\tgr31,gr0,fr1\;fcnz.fi\\\tfr1,fr1\", operands);
  else
    output_asm_insn (\"ld.f\\\t%1,fr1\;fcnz.fi\\\tfr1,fr1\", operands);
  if (GET_CODE (operands[0]) == MEM) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
      return \"seti\\\t%0,gr31\;st.f\\\tfr1,gr31,gr0\";
    else
      return \"st.f\\\tfr1,%0\";
  }
  else
    return \"mov.fi\\\tfr1,%0\";
}")

;;- arithmetic instructions
;;- add instructions
(define_insn "addsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(plus:SI (match_operand:SI 1 "register_operand" "%r")
		 (match_operand:SI 2 "arith32_operand" "rn")))]
  ""
  "*
{
  int i;

  if (REG_P (operands[2]))
    return \"addl\\\t%1,%2,%0\";
  i = INTVAL (operands[2]);
  if (INT_FITS_5_BITS (i))
    return \"addli5\\\t%1,%2,%0\";
  else if (INT_FITS_25_BITS (i))
    return \"addli25\\\t%1,%2,%0\";
  else
    return \"seti\\\t%2,gr31\;addl\\\t%1,gr31,%0\";
}")

(define_insn "adddf3"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(plus:DF (match_operand:DF 1 "register_operand" "%f")
		 (match_operand:DF 2 "register_operand" "f")))]
  ""
  "fadd.d\\\t%1,%2,%0")

;; a bunch more can go in here!

(define_insn "addsf3"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(plus:SF (match_operand:SF 1 "register_operand" "%f")
		 (match_operand:SF 2 "register_operand" "f")))]
  ""
  "fadd\\\t%1,%2,%0")

;;- subtract instructions
(define_insn "subsi3"
  [(set (match_operand:SI 0 "register_operand" "=a,r")
	(minus:SI (match_operand:SI 1 "register_operand" "r,r")
		  (match_operand:SI 2 "arith32_operand" "an,rn")))]
  ""
  "*
{
  int i;

  if (REG_P (operands[2]))
    return \"subl\\\t%1,%2,%0\";
  i = INTVAL (operands[2]);
  if (INT_FITS_5_BITS (i))
    return \"subli5\\\t%1,%2,%0\";
  else if (INT_FITS_25_BITS (i))
    return \"subli25\\\t%1,%2,%0\";
  else
    return \"seti\\\t%2,gr31\;subl\\\t%1,gr31,%0\";
}")

(define_insn "subdf3"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(minus:DF (match_operand:DF 1 "register_operand" "f")
		  (match_operand:DF 2 "register_operand" "f")))]
  ""
  "fsub.d\\\t%1,%2,%0")

(define_insn "subsf3"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(minus:SF (match_operand:SF 1 "register_operand" "f")
		  (match_operand:SF 2 "register_operand" "f")))]
  ""
  "fsub\\\t%1,%2,%0")

;;- multiply instructions
(define_insn "mulsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(mult:SI (match_operand:SI 1 "register_operand" "%r")
		 (match_operand:SI 2 "arith32_operand" "rn")))]
  ""
  "*
{
  int i;

  if (REG_P (operands[2]))
    return \"mult\\\t%1,%2,%0\";
  else if (FITS_25_BITS (operands[2]))
    return \"mult\\\t%1,%2,%0\";
  else
    return \"seti\\\t%2,gr31\;mult\\\t%1,gr31,%0\";
}")

(define_insn "umulsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(umult:SI (match_operand:SI 1 "register_operand" "%r")
		  (match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "*
{
  int i;

  if (REG_P (operands[2]))
    return \"mult\\\t%1,%2,%0\";
  else if (FITS_25_BITS (operands[2]))
    return \"mult\\\t%1,%2,%0\";
  else
    return \"seti\\\t%2,gr31\;mult\\\t%1,gr31,%0\";
}")

(define_insn "muldf3"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(mult:DF (match_operand:DF 1 "register_operand" "%f")
		 (match_operand:DF 2 "register_operand" "f")))]
  ""
  "fmul.d\\\t%1,%2,%0")

(define_insn "mulsf3"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(mult:SF (match_operand:SF 1 "register_operand" "%f")
		 (match_operand:SF 2 "register_operand" "f")))]
  ""
  "fmul\\\t%1,%2,%0")

;;- divide instructions

(define_insn "divsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (div:SI (match_operand:SI 1 "register_operand" "r")
                (match_operand:SI 2 "arith32_operand" "rn")))]
  ""
  "*
{
  int i;

  if (REG_P (operands[2]))
    return \"mov.if\\\t%1,fr1\;mov.if\\\t%2,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,%0\";
  i = INTVAL (operands[2]);
  if (INT_FITS_5_BITS (i))
    return \"mov.if\\\t%1,fr1\;addli5\\\tgr0,%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,%0\";
  else if (INT_FITS_25_BITS (i))
    return \"mov.if\\\t%1,fr1\;addli25\\\tgr0,%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,%0\";
  else
    return \"mov.if\\\t%1,fr1\;seti\\\t%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,%0\";
}")

(define_insn "udivsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (udiv:SI (match_operand:SI 1 "register_operand" "r")
                 (match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "*
{
  int i;

  if (REG_P (operands[2]))
    return \"mov.if\\\t%1,fr1\;mov.if\\\t%2,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,%0\";
  i = INTVAL (operands[2]);
  if (INT_FITS_5_BITS (i))
    return \"mov.if\\\t%1,fr1\;addli5\\\tgr0,%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,%0\";
  else if (INT_FITS_25_BITS (i))
    return \"mov.if\\\t%1,fr1\;addli25\\\tgr0,%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,%0\";
  else
    return \"mov.if\\\t%1,fr1\;seti\\\t%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,%0\";
}")

(define_insn "divdf3"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(div:DF (match_operand:DF 1 "register_operand" "f")
		(match_operand:DF 2 "register_operand" "f")))]
  ""
  "fdiv.d\\\t%1,%2,%0")

(define_insn "divsf3"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(div:SF (match_operand:SF 1 "register_operand" "f")
		(match_operand:SF 2 "register_operand" "f")))]
  ""
  "fdiv\\\t%1,%2,%0")

;; Remainder instructions.

(define_insn "modsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (mod:SI (match_operand:SI 1 "register_operand" "r")
                (match_operand:SI 2 "arith32_operand" "rn")))]
  ""
  "*
{
  int i;

  if (REG_P (operands[2]))
    return \"mov.if\\\t%1,fr1\;mov.if\\\t%2,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,gr31\;mult\\\tgr31,%2,gr31\;subl\\\t%1,gr31,%0\";
  i = INTVAL (operands[2]);
  if (INT_FITS_5_BITS (i))
    return \"mov.if\\\t%1,fr1\;addli5\\\tgr0,%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,gr31\;mult\\\tgr31,%2,gr31\;subl\\\t%1,gr31,%0\";
  else if (INT_FITS_25_BITS (i))
    return \"mov.if\\\t%1,fr1\;addli25\\\tgr0,%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,gr31\;mult\\\tgr31,%2,gr31\;subl\\\t%1,gr31,%0\";
  else
    return \"mov.if\\\t%1,fr1\;seti\\\t%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,gr31\;mult\\\tgr31,%2,gr31\;subl\\\t%1,gr31,%0\";
}")

(define_insn "umodsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (umod:SI (match_operand:SI 1 "register_operand" "r")
                 (match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "*
{
  int i;

  if (REG_P (operands[2]))
    return \"mov.if\\\t%1,fr1\;mov.if\\\t%2,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,gr31\;mult\\\tgr31,%2,gr31\;subl\\\t%1,gr31,%0\";
  i = INTVAL (operands[2]);
  if (INT_FITS_5_BITS (i))
    return \"mov.if\\\t%1,fr1\;addli5\\\tgr0,%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,gr31\;mult\\\tgr31,%2,gr31\;subl\\\t%1,gr31,%0\";
  else if (INT_FITS_25_BITS (i))
    return \"mov.if\\\t%1,fr1\;addli25\\\tgr0,%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,gr31\;mult\\\tgr31,%2,gr31\;subl\\\t%1,gr31,%0\";
  else
    return \"mov.if\\\t%1,fr1\;seti\\\t%2,gr31\;mov.if\\\tgr31,fr31\;fcnv.id\\\tfr1,fr1\;fcnv.id\\\tfr31,fr31\;fdiv.d\\\tfr1,fr31,fr1\;fcnz.di\\\tfr1,fr1\;mov.fi\\\tfr1,gr31\;mult\\\tgr31,%2,gr31\;subl\\\t%1,gr31,%0\";
}")

;;- and instructions

(define_insn "andsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(and:SI (match_operand:SI 1 "register_operand" "%r")
		(match_operand:SI 2 "arith32_operand" "rn")))]
  ""
  "*
{
  if (REG_P (operands[2]))
    return \"and\\\t%1,%2,%0\";
  else if (FITS_25_BITS (operands[2]))
    return \"and\\\t%1,%2,%0\";
  else 
    return \"seti\\\t%2,gr31\;and\\\t%1,gr31,%0\";
}")

;;- Bit set (inclusive or) instructions
(define_insn "iorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ior:SI (match_operand:SI 1 "register_operand" "%r")
		(match_operand:SI 2 "arith32_operand" "rn")))]
  ""
  "*
{
  if (REG_P (operands[2]))
    return \"or\\\t%1,%2,%0\";
  else if (FITS_25_BITS (operands[2]))
    return \"or\\\t%1,%2,%0\";
  else
    return \"seti\\\t%2,gr31\;or\\\t%1,gr31,%0\";
}")

;;- xor instructions
(define_insn "xorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(xor:SI (match_operand:SI 1 "register_operand" "%r")
		(match_operand:SI 2 "arith32_operand" "rn")))]
  ""
  "*
{
  if (REG_P (operands[2]))
    return \"xor\\\t%1,%2,%0\";
  else if (FITS_25_BITS (operands[2]))
    return \"xor\\\t%1,%2,%0\";
  else
    return \"seti\\\t%2,gr31\;xor\\\t%1,gr31,%0\";
}")

;;- one complement instructions
(define_insn "one_cmplsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(not:SI (match_operand:SI 1 "register_operand" "r")))]
  ""
  "xor\\\t%1,-1,%0")

;; Optimized special case of shifting.
;; Must precede the general case.

(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashiftrt:SI (match_operand:SI 1 "memory_operand" "m")
		     (const_int 24)))]
  ""
  "*
{
  if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
    return \"seti\\\t%1,gr31\;ld.c\\\tgr31,gr0,%0\";
  else
    return \"ld.c\\\t%1,%0\";
}")

(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
	(lshiftrt:SI (match_operand:SI 1 "memory_operand" "m")
		     (const_int 24)))]
  ""
  "*
{
  if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
    return \"seti\\\t%1,gr31\;ld.cu\\\tgr31,gr0,%0\";
  else
    return \"ld.cu\\\t%1,%0\";
}")

;;- arithmetic shift instructions
(define_insn "ashlsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashift:SI (match_operand:SI 1 "register_operand" "r")
		   (match_operand:SI 2 "arith32_operand" "rn")))]
  ""
  "*
{
  if (GET_CODE (operands[2]) == CONST_INT
      && INTVAL (operands[2]) >= 32)
    return \"or\\\tgr0,gr0,%0\";
  else
    return \"sll\\\t%1,%2,%0\";
}")

(define_insn "ashrsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashiftrt:SI (match_operand:SI 1 "register_operand" "r")
		     (match_operand:SI 2 "arith32_operand" "rn")))]
  ""
  "*
{
  if (GET_CODE (operands[2]) == CONST_INT
      && INTVAL (operands[2]) >= 32)
    return \"sra\\\t%1,31,%0\";
  else
    return \"sra\\\t%1,%2,%0\";
}")

;;- logical shift instructions
(define_insn "lshrsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(lshiftrt:SI (match_operand:SI 1 "register_operand" "r")
		     (match_operand:SI 2 "arith32_operand" "rn")))]
  ""
  "*
{
  if (GET_CODE (operands[2]) == CONST_INT
      && INTVAL (operands[2]) >= 32)
    return \"or\\\tgr0,gr0,%0\";
  else
    return \"srl\\\t%1,%2,%0\";
}")

;; Special cases of bit-field insns which we should
;; recognize in preference to the general case.
;; These handle aligned 8-bit and 16-bit fields,
;; which can usually be done with move instructions.

(define_insn ""
  [(set (zero_extract:SI (match_operand:SI 0 "general_operand" "+o")
			 (match_operand:SI 1 "int5_operand" "n")
			 (match_operand:SI 2 "int5_operand" "n"))
	(match_operand:SI 3 "register_operand" "r"))]
  "(INTVAL (operands[1]) == 8 || INTVAL (operands[1]) == 16)
   && INTVAL (operands[2]) % INTVAL (operands[1]) == 0"
  "*
{
  operands[0] = adj_offsettable_operand (operands[0], INTVAL (operands[2]) / 8);
  if (INTVAL (operands[1]) == 8) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
      return \"seti\\\t%0,gr31\;st.c\\\t%1,gr31,gr0\";
    else
      return \"st.c\\\t%3,%0\";
  }
  else {
    if (CONSTANT_ADDRESS_P (XEXP (operands[0], 0)))
      return \"seti\\\t%0,gr31\;st.s\\\t%1,gr31,gr0\";
    else
      return \"st.s\\\t%3,%0\";
  }
}")

(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
	(zero_extract:SI (match_operand:SI 1 "general_operand" "ro")
			 (match_operand:SI 2 "int5_operand" "n")
			 (match_operand:SI 3 "int5_operand" "n")))]
  "(INTVAL (operands[2]) == 8 || INTVAL (operands[2]) == 16)
   && INTVAL (operands[3]) % INTVAL (operands[2]) == 0"
  "*
{
  if (REG_P (operands[1]))
    return \"sll\\\t%1,%3,%0\;srl\\\t%0,32-%2,%0\";
  operands[1] = adj_offsettable_operand (operands[1], INTVAL (operands[3]) / 8);
  if (INTVAL (operands[2]) == 8) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.cu\\\tgr31,gr0,%0\";
    else
      return \"ld.cu\\\t%1,%0\";
  }
  else {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.su\\\tgr31,gr0,%0\";
    else
      return \"ld.su\\\t%1,%0\";
  }
}")

(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extract:SI (match_operand:SI 1 "general_operand" "ro")
			 (match_operand:SI 2 "int5_operand" "n")
			 (match_operand:SI 3 "int5_operand" "n")))]
  "(INTVAL (operands[2]) == 8 || INTVAL (operands[2]) == 16)
   && INTVAL (operands[3]) % INTVAL (operands[2]) == 0"
  "*
{
  if (REG_P (operands[1]))
    return \"sll\\\t%1,%3,%0\;srr\\\t%0,32-%2,%0\";
  operands[1] = adj_offsettable_operand (operands[1], INTVAL (operands[3]) / 8);
  if (INTVAL (operands[2]) == 8) {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.c\\\tgr31,gr0,%0\";
    else
      return \"ld.c\\\t%1,%0\";
  }
  else {
    if (CONSTANT_ADDRESS_P (XEXP (operands[1], 0)))
      return \"seti\\\t%1,gr31\;ld.s\\\tgr31,gr0,%0\";
    else
      return \"ld.s\\\t%1,%0\";
  }
}")

;; negate insns
(define_insn "negsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(neg:SI (match_operand:SI 1 "arith32_operand" "rn")))]
  ""
  "*
{
  int i;

  if (REG_P (operands[1]))
    return \"subl\\\tgr0,%1,%0\";
  i = INTVAL (operands[1]);
  if (INT_FITS_5_BITS (i))
    return \"subli5\\\tgr0,%1,%0\";
  else if (INT_FITS_25_BITS (i))
    return \"subli25\\\tgr0,%1,%0\";
  else
    return \"seti\\\t%1,gr31\;subl\\\tgr0,gr31,%0\";
}")

(define_insn "negdf2"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(neg:DF (match_operand:DF 1 "register_operand" "f")))]
  ""
  "fsub.d\\\tfr0,%1,%0")

(define_insn "negsf2"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(neg:SF (match_operand:SF 1 "register_operand" "f")))]
  ""
  "fsub\\\tfr0,%1,%0")

;; Unconditional and other jump instructions
(define_insn "jump"
  [(set (pc)
	(label_ref (match_operand 0 "" "")))]
  ""
  "brc\\\ttrue,%l0")

(define_insn "tablejump"
  [(set (pc) (match_operand:SI 0 "register_operand" "r"))
   (use (label_ref (match_operand 1 "" "")))]
  ""
  "jmp\\\t%0")

;;- jump to subroutine
(define_insn "call"
  [(call (match_operand:SI 0 "memory_operand" "m")
	 (match_operand:SI 1 "general_operand" "g"))
   (use (reg:SI 3))]
  ;;- Don't use operand 1 for most machines.
  ""
  "*
{
  operands[0] = XEXP (operands[0], 0);
  if (REG_P (operands[0]))
    return \"call\\\t.+8,gr3\;subli5\\\tgr3,-16,gr3\;jmp\\\t%0\";
  else
    return \"call\\\t%0,gr3\";
}")

;;- jump to subroutine
(define_insn "call_value"
  [(set (match_operand 0 "" "=r")
	(call (match_operand:SI 1 "memory_operand" "m")
	      (match_operand:SI 2 "general_operand" "g")))
   (use (reg:SI 3))]
  ;;- Don't use operand 2 for most machines.
  ""
  "*
{
  operands[1] = XEXP (operands[1], 0);
  if (REG_P (operands[1]))
    return \"call\\\t.+8,gr3\;subli5\\\tgr3,-16,gr3\;jmp\\\t%1\";
  else
    return \"call\\\t%1,gr3\";
}")

;; A memory ref with constant address is not normally valid.
;; But it is valid in a call insns.  This pattern allows the
;; loading of the address to combine with the call.
(define_insn ""
  [(call (mem:SI (match_operand:SI 0 "" "i"))
	 (match_operand:SI 1 "general_operand" "g"))
   (use (reg:SI 3))]
  ;;- Don't use operand 1 for most machines.
  "GET_CODE (operands[0]) == SYMBOL_REF"
  "call\\\t%0,gr3")

(define_insn "nop"
  [(const_int 0)]
  ""
  "nop")

;;- Local variables:
;;- mode:emacs-lisp
;;- comment-start: ";;- "
;;- eval: (set-syntax-table (copy-sequence (syntax-table)))
;;- eval: (modify-syntax-entry ?[ "(]")
;;- eval: (modify-syntax-entry ?] ")[")
;;- eval: (modify-syntax-entry ?{ "(}")
;;- eval: (modify-syntax-entry ?} "){")
;;- End:
