static char SccsId[] = "@(#)bsd2elf.c	1.4(2/21/92)";

#include <ctype.h>
#include <stdio.h>
#include <elf.h>
#include <stdlib.h>
#include <errno.h>

int xtoi();
void read_aout();
void conv_data();
void write_elf();


/*  A.OUT STRUCT */	

struct aout_head {
	unsigned short  a_info;
	unsigned short  a_magic;
	int 		a_text;
	int		a_data;
	int		a_bss;
	int		a_syms;
	int		a_entry;
	int		a_trsize;
	int		a_drsize;
};
struct aout_head Ahead;

struct rel {
	int		r_off;
	unsigned int    r_sym: 24,
			r_extern: 1,
			     : 2,
			r_type: 5;
	int 		r_addend;
};

struct sym {
	int		s_name;
	unsigned int	s_type: 8,
			      : 24;
	int		s_value;
};

struct rel *Drel;
struct rel *Trel;
struct sym *Sym;


/*  ELF  STRUCT  */

Elf32_Ehdr	Ehead;

Elf32_Rel	*E_textrel;
Elf32_Rel	*E_datarel;
Elf32_Sym	*E_sym;

Elf32_Phdr	Phead[3];
Elf32_Shdr	Shead[8];

/*  BUFFER  */

char *Text_p, *Data_p, *Textrel_p;
char *Datarel_p, *Sym_p, *Str_p;
char *Shstr_p;
int  len, count;
int  trsz_num, drsz_num, sysz_num;


int baseaddr = 0;
char *Aout_file, *Elf_file;

main( argc, argv)
	
	int argc;
	char **argv;

{
	int i;

	for (i = 1; i < argc; i++){
		if (argv[i][0] != '-')
			break;

		switch (argv[i][1]){
		case 'T':
			baseaddr = atoi(argv[i+1]);
			i++;
			break;
		default:
			fprintf(stderr, "bsd2elf: bad option '%s'\n", argv[i]);
			exit(1);
		}
	}


	if (argc - i != 2){
		fprintf(stderr, "Usage: bsd2elf [-T baseaddr] bsd elf\n");
		exit(1);
	}
	Aout_file = argv[i];
	Elf_file = argv[i+1];


	read_aout();

	trsz_num = Ahead.a_trsize/12;
	drsz_num = Ahead.a_drsize/12;
	sysz_num = Ahead.a_syms/12 + 1;

	conv_data();
	write_elf();

	exit(0);

}


/* READ_AOUT() */
void read_aout()
{
	/* FILEOPEN */
	FILE *fp;

	fp = fopen(Aout_file, "r");
	if (fp == NULL){
		perror("Cann't open bsd_file");
		exit(1);
	}

	/* READ AOUT_HEADER */
	if (fread(&Ahead, 1, sizeof Ahead, fp) !=  (sizeof Ahead)){
		perror("Cann't read bsd_header");
		exit(1);
	}

	/* MALLOC */

	if ((Text_p = (char *)malloc(Ahead.a_text)) == NULL){
		perror("Cann't malloc bsd_text");
		exit(1);
	}
	if ((Data_p = (char *)malloc(Ahead.a_data)) == NULL){
		perror("Cann't malloc bsd_data");
		exit(1);
	}
	if ((Trel = (struct rel *)malloc(Ahead.a_trsize)) == NULL){
		perror("Cann't malloc bsd_textreloc");
		exit(1);
	}
	if ((Drel = (struct rel *)malloc(Ahead.a_drsize)) == NULL){
		perror("Cann't malloc bsd_datareloc");
		exit(1);
	}
	if ((Sym = (struct sym *)malloc(Ahead.a_syms)) == NULL){
		perror("Cann't malloc bsd_symboltab");
		exit(1);
	}

	/* READ (text, data) */
	if (fread(Text_p, 1, Ahead.a_text, fp) != Ahead.a_text){
		perror("Cann't read bsd_text");
		exit(1);
	}
	if (fread(Data_p, 1, Ahead.a_data, fp) != Ahead.a_data){
		perror("Cann't read bsd_data");
		exit(1);
	}

	/* READ (reloc) */
	if (fread(Trel, 1, Ahead.a_trsize, fp) != Ahead.a_trsize){
		perror("Cann't read bsd_textreloc");
		exit(1);
	}
	
	if (fread(Drel, 1, Ahead.a_drsize, fp) != Ahead.a_drsize){
		perror("Cann't read bsd_datareloc");
		exit(1);
	}

        if (Ahead.a_syms) {

		/* READ (symbol) */
		if (fread(Sym, 1, Ahead.a_syms, fp) != Ahead.a_syms){
			perror("Cann't read bsd_symboltab");
			exit(1);
		}
	
		/* STRING TABLE */
		fread(&len, 1, sizeof(unsigned long), fp);
		len -= sizeof(unsigned long);
	
		if ((Str_p = (char *)malloc(len + 1)) == NULL){
			perror("Cann't malloc bsd_stringtab");
			exit(1);
		}
	
		Str_p[0] = '\0';
	
		if (fread(Str_p+1, 1, len, fp) != len){
			perror("Cann't read string");
			exit(1);
		}
	}

	/* SHSTRING TABLE */
	Shstr_p = (char *)malloc(60);
	Shstr_p = "\0.text\0.data\0.rel.text\0.rel.data\0.symtab\0.strtab\0.shstrtab\0";

	fclose(fp);

}

/* CONV_DATA() */
void conv_data()
{

	int p = 0;
	int i;
	int sym, type;
	int k, l;
	
	/* A.OUT_HEADER --> ELF_HEADER */
	Ehead.e_ident[0] = ELFMAG0;
	Ehead.e_ident[1] = ELFMAG1;
	Ehead.e_ident[2] = ELFMAG2;
	Ehead.e_ident[3] = ELFMAG3;
	Ehead.e_ident[4] = ELFCLASS32;
	Ehead.e_ident[5] = ELFDATA2MSB;
	Ehead.e_ident[6] = EV_CURRENT;
	Ehead.e_ident[7] = 8;

	Ehead.e_type = 1;
	Ehead.e_machine = 9;
	Ehead.e_version = 1;
	Ehead.e_entry = Ahead.a_entry;
	Ehead.e_phoff = sizeof Ehead;
	Ehead.e_shoff = (sizeof Ehead) + (sizeof *Phead)*3
	+ Ahead.a_text + Ahead.a_data
	+ (sizeof *E_textrel)*trsz_num
	+ (sizeof *E_datarel)*drsz_num
	+ (sizeof *E_sym)*sysz_num
	+ (len + 1) + 60;
	Ehead.e_flags = 0;
	Ehead.e_ehsize = sizeof Ehead;
	Ehead.e_phentsize = sizeof *Phead;
	Ehead.e_phnum = 3;
	Ehead.e_shentsize = sizeof *Shead;
	Ehead.e_shnum = 8;
	Ehead.e_shstrndx = 7;

	/* CONV TEXT RELOC */
	if ((E_textrel = (Elf32_Rel *)malloc((sizeof *E_textrel) * trsz_num)) == NULL){
		perror("Cann't malloc elf_textrel");
		exit(1);
	}
	for (i = 0; i < trsz_num; i++){
		if (Trel[i].r_extern == 0){
			E_textrel[i].r_offset = 0;
			E_textrel[i].r_info = 0;
		}
		else {
			E_textrel[i].r_offset = Trel[i].r_off;
			sym = Trel[i].r_sym + 1;

			switch (Trel[i].r_type){
			case 6:
				type = 7;
				break;
			case 7:
				type = 6;
				break;
			case 8:
				type = 5;
				break;
			case 10:
				type = 11;
				break;
			case 11:
				type = 9;
				break;
			case 16:
				type = 3;
				break;
			default:
				fprintf(stderr, "Bad Trel[i].r_type\n");
				exit(1);
				break;
			}
			E_textrel[i].r_info = ELF32_R_INFO(sym, type);
		}
	}

	/* CONV DATA RELOC */
	if ((E_datarel = (Elf32_Rel *)malloc((sizeof *E_datarel) * drsz_num)) == NULL){
		perror("Cann't malloc elf_datarel");
		exit(1);
	}
	for (i = 0; i < drsz_num; i++){
		if (Drel[i].r_extern = 0){
			E_datarel[i].r_offset = 0;
			E_datarel[i].r_info = 0;
		}
		else {
			E_datarel[i].r_offset = Drel[i].r_off;
			sym = Drel[i].r_sym + 1;

			switch (Drel[i].r_type){
			case 6:
				type = 7;
				break;
			case 7:
				type = 6;
				break;
			case 8:
				type = 5;
				break;
			case 10:
				type = 11;
				break;
			case 11:
				type = 9;
				break;
			case 16:
				type = 3;
				break;
			default:
				fprintf(stderr, "Bad Drel[i].r_type\n");
				exit(1);
				break;
			}
			E_datarel[i].r_info = ELF32_R_INFO(sym, type);
		}
	}

        if (Ahead.a_syms) {

		/* CONV SYMBOL */
		if ((E_sym = (Elf32_Sym *)malloc((sizeof *E_sym) * sysz_num)) == NULL){
			perror("Cann't malloc elf_symbol");
			exit(1);
		}
		E_sym[0].st_name = 0;
		E_sym[0].st_value = 0;
		E_sym[0].st_size = 0;
		E_sym[0].st_info = 0;
		E_sym[0].st_other = 0;
		E_sym[0].st_shndx = SHN_UNDEF;
	
		count = 0;
		for (k = 0; k < sysz_num - 1; k++){
			int	ext, type;
	
			l = k + 1;
	
			Sym[k].s_name += -4 + 1;
			E_sym[l].st_name = Sym[k].s_name;
			E_sym[l].st_value = Sym[k].s_value;
			E_sym[l].st_size = 0;
			E_sym[l].st_other = 0;
	
			if (Sym[k].s_type & 1)
				ext = STB_GLOBAL;
			else {
				ext = STB_LOCAL;
				count = l;
			}
			type = Sym[k].s_type & ~1;
	
			switch (type) {
			case 0:				/* undefined symbol */
				E_sym[l].st_info = ELF32_ST_INFO(ext,STT_NOTYPE);
				E_sym[l].st_shndx = SHN_UNDEF;
				break;
			case 2:				/* absolute */
				E_sym[l].st_info = ELF32_ST_INFO(ext,STT_OBJECT);
				E_sym[l].st_shndx = SHN_ABS;
				break;
			case 4:				/* TEXT */
				E_sym[l].st_info = ELF32_ST_INFO(ext,STT_FUNC);
				E_sym[l].st_shndx = 1;
				break;
			case 6:				/* DATA */
				E_sym[l].st_info = ELF32_ST_INFO(ext,STT_OBJECT);
				E_sym[l].st_shndx = 2;
				break;
			case 8:				/* BSS */
				E_sym[l].st_info = ELF32_ST_INFO(ext,STT_OBJECT);
				E_sym[l].st_shndx = SHN_COMMON;
				break;
			default:
				fprintf(stderr, "Bad Sym[%d].s_type %d, name == "
				, k, Sym[k].s_type);
				if (Sym[k].s_name < 1 || Sym[k].s_name > len)
					fprintf(stderr, "<out-of-range.%d %d>"
					, Sym[k].s_name, len);
				else if (Str_p[Sym[k].s_name] == 0)
					fprintf(stderr, "<noname.%d %d>"
					, Sym[k].s_name, len);
				else
					fprintf(stderr, "<%s>"
					, &Str_p[Sym[k].s_name]);
				fprintf(stderr, "\n");
				exit(1);
			}
		}
	}

	/* SET PROGRAM HEADER */

	/* text */
	Phead[0].p_type = 1;
	Phead[0].p_offset = sizeof Ehead + (sizeof *Phead)*3;
	Phead[0].p_vaddr = baseaddr + 0;
	Phead[0].p_paddr = 0;
	Phead[0].p_filesz = Ahead.a_text;
	Phead[0].p_memsz = Ahead.a_text;
	Phead[0].p_flags = 5;
	Phead[0].p_align = 1;


	/* data */
	Phead[1].p_type = 1;
	Phead[1].p_offset = Phead[0].p_offset + Ahead.a_text;
	Phead[1].p_vaddr = baseaddr + Ahead.a_text;
	Phead[1].p_paddr = 0;
	Phead[1].p_filesz = Ahead.a_data;
	Phead[1].p_memsz = Ahead.a_data;
	Phead[1].p_flags = 7;
	Phead[1].p_align = 1;

	/* bss */
	Phead[2].p_type = 1;
	Phead[2].p_offset = 0;
	Phead[2].p_vaddr = baseaddr + Ahead.a_text + Ahead.a_data;
	Phead[2].p_paddr = 0;
	Phead[2].p_filesz = 0;
	Phead[2].p_memsz = Ahead.a_bss;
	Phead[2].p_flags = 7;
	Phead[2].p_align = 1;

	/* SET SECTION HEADER */

	/* index 0 */
	Shead[0].sh_name = 0;
	Shead[0].sh_type = SHT_NULL;
	Shead[0].sh_flags = 0;
	Shead[0].sh_addr = 0;
	Shead[0].sh_offset = 0;
	Shead[0].sh_size = 0;
	Shead[0].sh_link = SHN_UNDEF;
	Shead[0].sh_info = 0;
	Shead[0].sh_addralign = 0;
	Shead[0].sh_entsize = 0;

	/* text */
	Shead[1].sh_name = Shead[7].sh_offset + 1;
	Shead[1].sh_type = SHT_PROGBITS;
	Shead[1].sh_flags = 6;
	Shead[1].sh_addr = baseaddr + 0;
	Shead[1].sh_offset = sizeof Ehead + (sizeof *Phead)*3;
	Shead[1].sh_size = Ahead.a_text;
	Shead[1].sh_link = SHN_UNDEF;
	Shead[1].sh_info = 0;
	Shead[1].sh_addralign = 1;
	Shead[1].sh_entsize = 0;

	/* data */
	Shead[2].sh_name = Shead[1].sh_name + 6;
	Shead[2].sh_type = SHT_PROGBITS;
	Shead[2].sh_flags = 3;
	Shead[2].sh_addr = baseaddr + 0 + Ahead.a_text;
	Shead[2].sh_offset = Shead[1].sh_offset + Shead[1].sh_size;
	Shead[2].sh_size = Ahead.a_data;
	Shead[2].sh_link = SHN_UNDEF;
	Shead[2].sh_info = 0;
	Shead[2].sh_addralign = 1;
	Shead[2].sh_entsize = 0;

	/* text.reloc */
	Shead[3].sh_name = Shead[2].sh_name + 6;
	Shead[3].sh_type = SHT_REL;
	Shead[3].sh_flags = 0;
	Shead[3].sh_addr = 0;
	Shead[3].sh_offset = Shead[2].sh_offset + Shead[2].sh_size;
	Shead[3].sh_size = (sizeof *E_textrel) * trsz_num;
	Shead[3].sh_link = 5;
	Shead[3].sh_info = 1;
	Shead[3].sh_addralign = 1;
	Shead[3].sh_entsize = 8;

	/* data.reloc */
	Shead[4].sh_name = Shead[3].sh_name + 10;
	Shead[4].sh_type = SHT_REL;
	Shead[4].sh_flags = 0;
	Shead[4].sh_addr = 0;
	Shead[4].sh_offset = Shead[3].sh_offset + Shead[3].sh_size;
	Shead[4].sh_size = (sizeof *E_datarel) * drsz_num;
	Shead[4].sh_link = 5;
	Shead[4].sh_info = 2;
	Shead[4].sh_addralign = 1;
	Shead[4].sh_entsize = 8;

	/* symbol.table */
	Shead[5].sh_name = Shead[4].sh_name + 10;
	Shead[5].sh_type = SHT_SYMTAB;
	Shead[5].sh_flags = 0;
	Shead[5].sh_addr = 0;
	Shead[5].sh_offset = Shead[4].sh_offset + Shead[4].sh_size;
	Shead[5].sh_size = (sizeof *E_sym) * sysz_num;
	Shead[5].sh_link = 6;
	Shead[5].sh_info = count + 1;
	Shead[5].sh_addralign = 1;
	Shead[5].sh_entsize = 0x10;

	/* string */
	Shead[6].sh_name = Shead[5].sh_name + 8;
	Shead[6].sh_type = SHT_STRTAB;
	Shead[6].sh_flags = 0;
	Shead[6].sh_addr = 0;
	Shead[6].sh_offset = Shead[5].sh_offset + Shead[5].sh_size;
	Shead[6].sh_size = len + 1;
	Shead[6].sh_link = SHN_UNDEF;
	Shead[6].sh_info = 0;
	Shead[6].sh_addralign = 1;
	Shead[6].sh_entsize = 0;

	/* shstring */
	Shead[7].sh_name = Shead[6].sh_name + 8;
	Shead[7].sh_type = SHT_STRTAB;
	Shead[7].sh_flags = 0;
	Shead[7].sh_addr = 0;
	Shead[7].sh_offset = Shead[6].sh_offset + Shead[6].sh_size;
	Shead[7].sh_size = 0x3c;
	Shead[7].sh_link = SHN_UNDEF;
	Shead[7].sh_info = 0;
	Shead[7].sh_addralign = 1;
	Shead[7].sh_entsize = 0;

}

/* WRITE_ELF() */
void write_elf()
{

	FILE *fp;
	unsigned long where;

	/* FILE OPEN */
	fp = fopen(Elf_file, "wb");

	if (fp == NULL){
		perror("Cann't open Elf_file");
		exit(1);
	}

	/* WRITE ELF */
	if ((fwrite(&Ehead, 1, sizeof Ehead, fp)) != sizeof Ehead){
		perror("Cann't write Elf_header");
		exit(1);
	}

	/* WRITE PROGRAM HEADER */
	if ((fwrite(Phead, 3, sizeof *Phead, fp)) != sizeof *Phead){
		perror("Cann't write Program_header");
		exit(1);
	}

	/* WRITE TEXT & DATA */
	if ((fwrite(Text_p, 1, Ahead.a_text, fp)) != Ahead.a_text){
		perror("Cann't write Text");
		exit(1);
	}

	if ((fwrite(Data_p, 1, Ahead.a_data, fp)) != Ahead.a_data){
		perror("Cann't write Data");
		exit(1);
	}

	/* WRITE TEXT.RELOC & DATA.RELOC */
	if ((fwrite(E_textrel, sizeof *E_textrel, trsz_num, fp)) != trsz_num){
		perror("Cann't write Text.reloc");
		exit(1);
	}

	if ((fwrite(E_datarel, sizeof *E_datarel, drsz_num, fp)) != drsz_num){
		perror("Cann't write Data.reloc");
		exit(1);
	}

	if (Ahead.a_syms) {

		/* WRITE SYMBOL */
		if ((fwrite(E_sym, sizeof *E_sym, sysz_num, fp)) != sysz_num){
			perror("Cann't write Symbol.table");
			exit(1);
		}

		/* WRITE STRING */
		if ((fwrite(Str_p, 1, len + 1, fp)) != len + 1){
			perror("Cann't write String.table");
			exit(1);
		}
	}

	/* WRITE SHSTRING TABLE */
	if ((fwrite(Shstr_p, 1, 60, fp)) != 60){
		perror("Cann't write Shstring table");
		exit(1);
	}

	/* WRITE SECTION HEADER TABLE */
	if ((fwrite(Shead, 8, sizeof *Shead, fp))!= sizeof *Shead){
		perror("Cann't write Section header");
		exit(1);
	}

	fclose(fp);

}

/* XTOI () */
int xtoi(y)
char *y;
{
	
	int i;
	int n = 0;
	for (i = 0; y[i] >= '0' && y[i] <= 'f'; i++){
	if (y[i] >= '0' && y[i] <= '9')
		n = 16 * n + (y[i] - '0');
		else
		n = 16 * n + (y[i] - 87);
	}

	return n;

}
