#ident  "@(#)bar.c	1.4 1/8/93 16:53:32 - FUJITSU UXP/VPP"

/***********************************************/
/*  KAI BPU controller                         */
/*  Copyright (c) 1992 by FUJITSU LIMITED      */
/*   Mainframe division                        */
/*   Supercomputer development department      */
/*            Primary writer: H.Ueno           */
/*    ueno@ayame.mfd.cs.fujitsu.co.jp          */
/***********************************************/

#include	<util/types.h>
#include	<util/mcr.h>
#include	<util/cmn_err.h>
#include	<proc/user.h>
#include	<proc/proc.h>
#include	<proc/sched.h>
#include	<proc/pms/partition.h>
#include	<io/dtu/bar.h>
#include	<io/dtu/usr_dtu.h>
#include	<svc/todr.h>
#include	<svc/errno.h>

extern void	subl8y( struct todr*, struct todr*, struct todr* );

#define	REAL_WORD_ACCESS(a) *((unsigned*)((unsigned)(a)+0xc0000000u))
         
extern int	num_of_side;
bar_side_info_t	*bar_side_info ;

#ifdef DEBUG
extern int     dtrace();
#endif

int
px_bar_init( sidenum, barcon, bmmap, waitflag ) 
  px_sidenum_t		sidenum;  		/* side number */ 
  int			barcon; 
  uint_t		bmmap ; 
  int			waitflag ; 

{
  px_partition_t *partition;
  int		temp_bsdcmc, retc  ;
  int		i;
  uint_t        mcr_adrs ; 
  unsigned      u0,u1,u2 ;
  struct todr	mcr_set_time ; 
  struct todr	current_time ; 
  struct todr   diff_time ;
  unsigned      bst_cw_time ;   /* bst change wait time */
  unsigned      bst_rs_time ;   /* bst recogintion wait time */
  int		spl;
  unsigned      t_rbmsk[BMSK01_WORD_LEN]; 

#ifdef DEBUG
dtrace("barinit","enter",sidenum,0);
dbg_printf("[px_bar_init: %d,%d,%d,%d]",sidenum,barcon,bmmap,waitflag);
#endif

/*
 *	check input parameter error
 */
  if( sidenum >= num_of_side 
      || barcon != BPU_CONNECT && barcon != BPU_DISCONN 
      || barcon == BPU_CONNECT && bmmap > BPU_CON_MAP_1
      || barcon == BPU_DISCONN && bmmap > BPU_DEV_MAP_7  )
    return( EINVAL ) ;
  
/*
 *	set mcrs
 */
/*	set only bsdcon in bsdcmc 	*/
  spl = splhi();
  mcr_load( MCR_BSDCMC, temp_bsdcmc ) ;
  temp_bsdcmc &= 0xfffffffb;
  if( barcon == BPU_CONNECT )
    temp_bsdcmc |= 4;
  mcr_store( MCR_BSDCMC, temp_bsdcmc ) ;
  splx(spl);

/*	set bmmap	*/		 
  if( barcon == BPU_CONNECT || sidenum == 0 ) /* connect or side0 */ 
    mcr_store( MCR_BMMAP0, bmmap );
  else
    mcr_store( MCR_BMMAP1, bmmap );

/*	reset bstnl 	*/
  if( sidenum == 0 || barcon == BPU_CONNECT )
    mcr_store( MCR_BSTN0W, 0 );
  if( sidenum == 1 || barcon == BPU_CONNECT )
    mcr_store( MCR_BSTN1W, 0 );

/*	set bar_side_info in memory 	*/
  bar_side_info[ sidenum ].barcon = barcon ; 
  bar_side_info[ sidenum ].bmmap  = bmmap ; 
  bar_side_info[ sidenum ].bstnl  = 0;
  bar_side_info[ sidenum ].bstnl_other = 0;
  bar_side_info[ sidenum ].barinfo = NULL ; 
  bar_side_info[ sidenum ].rm_area = NULL ; 
  bar_side_info[ sidenum ].rm_len  = NULL ;
  
/*	set bmsk	*/
  if( retc = get_rbmsk( sidenum, B_TRUE, NULL, t_rbmsk ) ) 
    return( retc );

  if( barcon == BPU_CONNECT )  /* both bsds are connected */
    for( i=0, mcr_adrs=MCR_BMSK00; i< BMSK01_WORD_LEN; ++i, mcr_adrs += 8 )
      mcr_store( mcr_adrs, t_rbmsk[ i ]);
  else
    for( i=0, mcr_adrs=(sidenum==0)? MCR_BMSK00:MCR_BMSK10; i< BMSK0_WORD_LEN; ++i, mcr_adrs += 8 )
      mcr_store( mcr_adrs, t_rbmsk[ i ] );
/*
 *	wait if requested by caller
 */
    if( waitflag )
    {
#ifdef DEBUG
      dtrace("barinit","wait",waitflag,0);
#endif
      asm( "brcs %asam,." );
      rtodc( &mcr_set_time );
      
      u0 = REAL_WORD_ACCESS( 0x2038 );
      u1 = REAL_WORD_ACCESS( u0 + 0x68 );
      bst_cw_time = REAL_WORD_ACCESS( u0 + u1 );
      bst_rs_time = REAL_WORD_ACCESS( u0 + u1 + 0x08 );
      u2 = (bst_cw_time > bst_rs_time ) ? bst_cw_time : bst_rs_time ;
      do {
          rtodc( &current_time );
          subl8y( &current_time, &mcr_set_time, &diff_time );
      } while( diff_time.tod[0] == 0 && diff_time.tod[1] < u2 );
    }

    bar_side_info[ sidenum ].status = BSI_ST_INITIALIZED;

#ifdef DEBUG
    dtrace("barinit","return",0,0);
#endif
    return( 0 );
}


int
px_bar_clear( sidenum ) 
    int		sidenum ;
{
#ifdef DEBUG
    dbg_printf("[px_bar_clear(%d):%x]",sidenum,&bar_side_info[sidenum]);
    dtrace("barclr","enter",sidenum,0);
#endif

    if( sidenum >= num_of_side )
        return( EINVAL );
  
    bar_side_info[ sidenum ].status = BSI_ST_NULL;
    bar_side_info[ sidenum ].barinfo = NULL;
    bar_side_info[ sidenum ].rm_area = NULL;
    bar_side_info[ sidenum ].rm_len  = 0;
  
    /*  reset bstnl     */ /* add 5/26 */
    if( sidenum == 0 || bar_side_info[ sidenum ].barcon == BPU_CONNECT )
        mcr_store( MCR_BSTN0W, 0 );
    if( sidenum == 1 || bar_side_info[ sidenum ].barcon == BPU_CONNECT )
        mcr_store( MCR_BSTN1W, 0 );

#ifdef DEBUG
    dtrace("barclr","return",0,0);
#endif
    return( 0 );
}

/*
 *	px_bar_disp
 */
int
px_bar_disp( sidenum )
    px_sidenum_t   sidenum;
{
    int	temp_bsdcmc ; 
    int   spl;
  
#ifdef DEBUG
    /* dtrace("bardisp","enter",sidenum,0); */
#endif
    if( sidenum < 0 || sidenum >= num_of_side ) {
        cmn_err(CE_NOTE, 
            "<barr:px_bar_disp> input param sidenum range over.");
        return( -1 );   /* control cannot reach here */
    }

    spl = splhi();
    mcr_load( MCR_BSDCMC, temp_bsdcmc );
    temp_bsdcmc &= 0xfffffffc;   

    if( bar_side_info[ sidenum ].barcon == BPU_CONNECT ) {
        temp_bsdcmc |= 3; 
    }
    else {
        temp_bsdcmc |= (1 << (sidenum & 1) );
    }
      
    mcr_store( MCR_BSDCMC, temp_bsdcmc );
    splx(spl);

#ifdef DEBUG
    /* dtrace("bardisp","return",0,0); */
#endif
    return( 0 ) ;
}


/**** pending below
****/
int
px_bar_swap( )
{
}
/****  pending above
*****/

/*
 *	for px_barctl(2) system call
 */

typedef struct {
	int		svccmd;
	int		cmd;
	px_barinfo_t	*barinfo;
	caddr_t		rm_area;   /* rbmask area address */
	int		rm_len;    /* rbmask byte length */
} px_barctl_parm;


int				/* returns svc errno */
px_barctl_ent( uap )
  px_barctl_parm	*uap;
{
    unsigned      sidenum ;
    px_barinfo_t  copy_barinfo;
    unsigned      copy_mask[BMSK01_WORD_LEN];
    uint_t *      bmp;

    if( ! u.u_procp->p_parflag )    /* if not parallel process, error */
        return( EACCES );

#ifdef DEBUG
    dbg_printf("[px_barctl_ent(%d):%x]",sidenum,&bar_side_info[sidenum]);
    dtrace("barctl","enter",sidenum,0);
#endif
    sidenum = u.u_procp->p_dupid;      
    if( sidenum >= num_of_side ) {
        cmn_err(CE_NOTE, "<barr:px_barcntl> invalid u.u_procp->p_dupid " );
        return( -1 );	/* error */
    }

    if( uap->cmd == PX_BARSTAT ) {
        if ( copyin((caddr_t)uap->barinfo, &copy_barinfo, sizeof(px_barinfo_t)) )
            return( EFAULT );
/* 92/11/04 parameter check */
        if( (uap->barinfo == NULL) || (uap->rm_area == NULL ) )
            return( EFAULT );
        if( (uap->rm_len == 0) || (uap->rm_len < sizeof(uint_t)*BMSK01_WORD_LEN) || (((uap->rm_len)%(sizeof(uint_t)*BMSK01_WORD_LEN))!=0) )
            return( EINVAL );
        if( copyin( uap->rm_area, copy_mask, sizeof(uint_t)*BMSK01_WORD_LEN ))
            return( EFAULT );
        bmp = (uint_t*)(uap->rm_area) + (uap->rm_len/sizeof(uint_t) - BMSK01_WORD_LEN);
        if( copyin( bmp, copy_mask, sizeof(uint_t)*BMSK01_WORD_LEN ))
            return( EFAULT );
        if( bar_side_info[ sidenum ].status == BSI_ST_START )
            return( EINVAL );
/* 92/11/04 parameter check */
        bar_side_info[ sidenum ].status = BSI_ST_START;
        bar_side_info[ sidenum ].barinfo = uap->barinfo;
        bar_side_info[ sidenum ].rm_area = uap->rm_area;
        bar_side_info[ sidenum ].rm_len  = uap->rm_len; 
#ifdef DEBUG
        dtrace("barctl","started",0,0);
#endif
        return( 0 );
    }
    else if( uap->cmd == PX_BARSTOP ) {
/* 92/11/04 parameter check */
        if( bar_side_info[ sidenum ].status == BSI_ST_STOP )
            return( EINVAL );
/* 92/11/04 parameter check */
        bar_side_info[ sidenum ].status = BSI_ST_STOP;
        bar_side_info[ sidenum ].barinfo = NULL;
        bar_side_info[ sidenum ].rm_area = NULL; 
        bar_side_info[ sidenum ].rm_len  = 0;
#ifdef DEBUG
        dtrace("barctl","stoped",0,0);
#endif
        return( 0 );
    }
    else 
        return( EINVAL ) ;
}

/*
 *	for px_barmask(2) system call
 */

typedef struct {
	int	svccmd;
	uint_t	*vmask;
	uint_t	*rmask;
} px_barmask_parm;


int 
px_barmask_ent( uap )
  px_barmask_parm	*uap;
{
    unsigned  sidenum;
    int       retc;
    uint_t    copy_vmask[ BMSK01_WORD_LEN ];
    uint_t    copy_rmask[ BMSK01_WORD_LEN ];

    if( ! u.u_procp->p_parflag )    /* if not parallel process, error */
        return( EACCES );
    sidenum = u.u_procp->p_dupid;      
#ifdef DEBUG
    dbg_printf("[px_barmask_ent(%d)]",sidenum);
    dtrace("bmskent","enter",sidenum,0);
#endif
    if( sidenum >= num_of_side ) {
        cmn_err(CE_NOTE, "<barr:px_barmask> invalid u.u_procp->p_dupid " );
        return( -1 );	/* error */
    }

    if( copyin( uap->vmask, copy_vmask, sizeof(uint_t)*BMSK01_WORD_LEN ))
        return( EFAULT );
    if( copyin( uap->rmask, copy_rmask, sizeof(uint_t)*BMSK01_WORD_LEN ))
        return( EFAULT );
  
    {
    caddr_t	ras; 	/* rmask area start address */
    caddr_t	rae;	/* rmask area end   address */
    caddr_t	rms;	/* rmask start address */
    caddr_t	rme;	/* rmask end   address */

    ras = bar_side_info[sidenum].rm_area;
    rae = ras + bar_side_info[sidenum].rm_len;
    rms = (caddr_t) uap->rmask;
    rme = rms + sizeof(uint_t)*BMSK01_WORD_LEN; 
    if( rms < ras || rms > rae || rme < ras || rme > rae )
        return( EINVAL );
    }

    if( retc=get_rbmsk( sidenum, B_FALSE, copy_vmask, copy_rmask ) )
        return( retc );
    if(copyout(copy_rmask, uap->rmask, sizeof(uint_t)*BMSK01_WORD_LEN ))
        return( EFAULT );
#ifdef DEBUG
    dtrace("bmskent","return",retc,0);
#endif
    return( retc );
}

/*
 *	for px_barrcov(2) system call
 */

typedef struct {
	int	svccmd;
	uint_t	*rmask;
} px_barrcov_parm;


int
px_barrcov_ent( uap )
    px_barrcov_parm	*uap;
{
    int		spl, sidenum, i;
    unsigned	copy_rmask[BMSK01_WORD_LEN];
    unsigned	mcr_adrs, mcr;

    if( ! u.u_procp->p_parflag )  /* if not parallel process, error */
	return( EACCES );
    sidenum = u.u_procp->p_dupid;
#ifdef DEBUG
    dbg_printf("[px_barrcov_ent(%d):%x]",sidenum,&bar_side_info[sidenum]);
    dtrace("brcvent","enter",sidenum,0);
#endif
    if( sidenum >= num_of_side ) {
	    cmn_err(CE_NOTE, "<barr:px_barrcov> invalid u.u_procp->p-dupid");
        return( -1 );	/* error */
    }

    if( copyin( uap->rmask, copy_rmask, sizeof(uint_t)*BMSK01_WORD_LEN) )
	return( EFAULT );
    {
	caddr_t     ras;    /* rmask area start address */
	caddr_t     rae;    /* rmask area end   address */
	caddr_t     rms;    /* rmask start address */
	caddr_t     rme;    /* rmask end   address */

	ras = bar_side_info[sidenum].rm_area;
	rae = ras + bar_side_info[sidenum].rm_len;
	rms = (caddr_t) uap->rmask;
	rme = rms + sizeof(uint_t)*BMSK01_WORD_LEN;
	if( rms < ras || rms > rae || rme < ras || rme > rae )
	    return( EINVAL );
    }
    for( mcr_adrs=MCR_BMSKU00, i=0; i<BMSK01_WORD_LEN; ++i, mcr_adrs += 8 ) {
	mcr_store( mcr_adrs, copy_rmask[ i ] );
    }
    /*
     *	    wait bst recogintion stable time
     */
    {
	uint_t	u0,u1,bst_rs_time;
	struct todr	mcr_set_time, current_time, diff_time;

	asm( "brcs %asam,." );
	rtodc( &mcr_set_time );

	u0 = REAL_WORD_ACCESS( 0x2038 );
 	u1 = REAL_WORD_ACCESS( u0 + 0x68 );
	bst_rs_time = REAL_WORD_ACCESS( u0 + u1 + 0x08 );
	do {
	    rtodc ( &current_time );
        subl8y( &current_time, &mcr_set_time, &diff_time );
	} while( diff_time.tod[0] == 0 && diff_time.tod[1] < bst_rs_time );
    }
    /*
     *	    LBSY = BSTNL
     */

    if( bar_side_info[sidenum].barcon == BPU_CONNECT ) {
        if( bar_side_info[sidenum].bmmap == BPU_CON_MAP_0 ) {
            mcr_load(  MCR_BSTN0W, mcr );    
            mcr_store( MCR_LBSY0W, mcr );
        }
        else {
            mcr_load(  MCR_BSTN1W, mcr );
            mcr_store( MCR_LBSY0W, mcr );
        }
    }
    else {
        mcr_load(  sidenum==0 ? MCR_BSTN0W : MCR_BSTN1W, mcr );
        mcr_store( sidenum==0 ? MCR_LBSY0W : MCR_LBSY1W, mcr );
    }
    return(0);
}

/*
 *  < pattern of real_bmsk relative to bsdcon and bmmap >
 *
 * bsd  bm     bit0-63          bit64-127      bit128-191      bit192-255
 * con   map 
 * 0     0   (000,0)-(063,0) (064,0)-(127,0) (128,0)-(191,0) (192,0)-(255,0)
 *       1   (000,1)-(063,1) (064,1)-(127,1) (128,1)-(191,1) (192,1)-(255,1)
 * 1     0   (000,y)-(063,y) (064,y)-(127,y)
 *       1   (128,y)-(191,y) (064,y)-(127,y)
 *       2   (000,y)-(063,y) (192,y)-(255,y) 	
 *       3   (128,y)-(191,y) (192,y)-(255,y)
 *       4   (000,0)-(055,0) (000,1)-(055,1)
 *           (064,0)-(071,0) (064,1)-(071,1)   
 *       5   (064,0)-(127,0) (064,1)-(127,1)
 *       6   (128,0)-(191,0) (128,1)-(191,1)
 *       7   (192,0)-(255,0) (192,1)-(255,1)
 *
 * y: side_number
 *
 */
/*
 *  	call function get_rbmsk by one of following two method
 *          get_rbmsk( sidenum, B_TRUE,  NULL,  rbmsk );
 *          get_rbmsk( sidenum, B_FALSE, vbmsk, rbmsk );
 *      when allflag!=0, rbmsk is made for all vurtual processors 
 *        in the partition_definition_structure
 *      otherwise,  rbmsk is made by translation from vbmsk.
 */

int
get_rbmsk( sidenum, allflag, vbmsk, rbmsk )
  unsigned	sidenum;
  boolean_t	allflag;
  uint_t	*vbmsk;
  uint_t	*rbmsk;
{
    int             i ;
    unsigned        vpnbr, bsdcon, bmmap ;
    px_partition_t  *partition ; 
    unsigned        wordpos, bitpos, or_bmsk,r_side ;
    unsigned        *bmskwd ;
    unsigned short  r_netpa ; 
    unsigned        vbmskv ;  /* vbmsk value */
    int             error = 0;
    
#ifdef DEBUG
    dtrace("getrmsk","enter",sidenum,0);
    dbg_printf("[get_rbmsk: %d,%d,%d,%x]",sidenum,allflag,vbmsk,rbmsk);
#endif
    if( sidenum < 0 || sidenum >= num_of_side ) {
        cmn_err(CE_NOTE, "<barr:get_rbmsk> invalid input_var sidenum");
        return( -1 );	/* error */
    }

    partition = pms_partition[ sidenum ];
    if( partition == NULL ) {
#ifdef DEBUG
    dtrace("getrmsk","error",EACCES,0);
#endif
        return( EACCES );
    }

    vpnbr = (unsigned) partition -> ptn_vpnbr; 
    if( vpnbr > BMSK01_WORD_LEN*32 ) {
        cmn_err(CE_NOTE, "<barr:get_rbmsk> invalid partition -> vpnbr");
        return( -1 );	/* error */
    }

    if( !allflag )
    {
#ifdef DEBUG
        dtrace("getrmsk","!all",0,0);
#endif
        wordpos = vpnbr>>5 ;  /* vbmsk null start word pos */
        bitpos  = vpnbr&31 ;  /* vbmsk null start bit  pos */
        if( wordpos < BMSK01_WORD_LEN )
        {
            if( *(vbmsk+wordpos) & (0xffffffffu >> bitpos) )
                return( EINVAL );
        }
        for( bmskwd=vbmsk+wordpos+1; bmskwd < vbmsk+BMSK01_WORD_LEN; ++bmskwd )
            if( *bmskwd )
                return( EINVAL );
    }
 
    bsdcon = bar_side_info[ sidenum ].barcon ;
    bmmap  = bar_side_info[ sidenum ].bmmap ; 

#ifdef DEBUG
    dtrace("getrmsk","bsdcon",bsdcon,0);
    dtrace("getrmsk","bmmap",bmmap,0);
#endif

    if( !allflag ) {
        bmskwd = vbmsk ;
        vbmskv = *bmskwd ; 
    }
    else
        vbmskv = 0;
  
  /*add 5/26*/
  for( i=0; i < BMSK01_WORD_LEN; ++i ) {
      rbmsk[ i ] = 0;
  }

  for( i=0; i < vpnbr; ++i ) {
    if( allflag || vbmskv & 0x80000000 ) {
        r_netpa = RPID2NETPA( partition -> ptn_rpid[i] );
        r_side  = RPID2DUPID( partition -> ptn_rpid[i] );

        or_bmsk = 0x80000000u >> (r_netpa & 31u);

        if( bsdcon == BPU_CONNECT )   /* BSD connected mode */
          switch( bmmap ) {
            case 0 :
              if( r_netpa > 255 || r_side != 0 ) 
                  goto mapping_error;
              *( rbmsk + (r_netpa>>5) ) |= or_bmsk;
              break;
            case 1 :
              if( r_netpa > 255 || r_side == 0 )
                  goto mapping_error;
              *( rbmsk + (r_netpa>>5) ) |= or_bmsk;
              break;
            default :
              cmn_err(CE_NOTE, "<barr:get_rbmsk> invalid bar_side_info[sidenum].bmmap" );
              error = 1;
              break;
           }
         else         /* BSD disconnect mode */
           switch( bmmap ) {
             case 0 :
               if( r_netpa > 127 || r_side != sidenum )
                   goto mapping_error;
               *( rbmsk + (r_netpa>>5) ) |= or_bmsk;
               break;
             case 1 :
               if( r_netpa < 64 || r_netpa > 191 || r_side != sidenum )
                   goto mapping_error; 
               if( r_netpa >= 128 )
                 *( rbmsk + ((r_netpa-128u)>>5) ) |= or_bmsk ;
               else
                 *( rbmsk + (r_netpa>>5) ) |= or_bmsk;
               break;
             case 2 :
               if( r_netpa>63 && r_netpa<192 || r_netpa>255||r_side!=sidenum )
                   goto mapping_error;
               if( r_netpa < 64 )
                 *( rbmsk + (r_netpa>>5) ) |= or_bmsk;	/* add 5/27 */
               else
                 *( rbmsk + ((r_netpa-128u)>>5) ) |= or_bmsk ;
               break;
             case 3 :
               if( r_netpa < 128 || r_netpa > 255 || r_side != sidenum )
                   goto mapping_error;
               *( rbmsk + ((r_netpa-128u)>>5) ) |= or_bmsk ; 
               break;
             case 4 :
               if( r_netpa>55 && r_netpa<64 || r_netpa > 71 )
                   goto mapping_error;
               if( r_netpa < 64 ) 
                 *( rbmsk + (r_side<<1) + (r_netpa>>5) ) |= or_bmsk ;
               else
                 *( rbmsk + (r_side<<1) + ((r_netpa-8u)>>5) )
                     |= ( 0x80000000u >> ((r_netpa-8u) & 31u) );
               break;
             case 5 :
               if( r_netpa < 64 || r_netpa > 127 )
                   goto mapping_error;
               *( rbmsk+(r_side<<1) + ((r_netpa-64u)>>5) ) |= or_bmsk;
               break;
             case 6 :
               if( r_netpa < 128 || r_netpa > 191 )
                   goto mapping_error;
               *( rbmsk +(r_side<<1)+ ((r_netpa-128u)>>5) ) |= or_bmsk;
               break;
             case 7 :
               if( r_netpa < 192 || r_netpa > 255 )
                   goto mapping_error;
               *( rbmsk +(r_side<<1)+ ((r_netpa-192u)>>5) ) |= or_bmsk;
               break;
             default:
               cmn_err(CE_NOTE, "<barr:get_rbmsk> invalid bar_side_info[sidenum].bmmap" );
               error = 1;
               break;
           }  
    }
  
    if( !allflag ) {
        if( (i & 0x0000001f) == 31 )	/* add 5/27 */
          vbmskv = *( ++bmskwd );
        else
          vbmskv <<= 1; 
    }
  }  /* end of for loop */
  
  if( error ) return(EINVAL);
#ifdef DEBUG
  dtrace("getrmsk","return",0,0);
#endif
  return( 0 );

mapping_error:
  cmn_err(CE_NOTE, "<barr:get_rbmsk> PMS real_processor mapping error.");
  return( -1 );   /* control cannot reach here */
}  


/*
 *       barrier initialization for tod sync
 *         call px_bar_todinit( 0, BPU_CONNECT, BPU_CON_MAP_0 )
 */

int
px_bar_todinit( sidenum, barcon, bmmap )
    px_sidenum_t   sidenum;                /* side number */
    int            barcon;
    uint_t         bmmap ;

{
    int  temp_bsdcmc ;
    int  spl;

    /*
     *  check input parameter error
     */
    if( sidenum >= num_of_side
        || barcon != BPU_CONNECT && barcon != BPU_DISCONN
        || barcon == BPU_CONNECT && bmmap > BPU_CON_MAP_1
        || barcon == BPU_DISCONN && bmmap > BPU_DEV_MAP_7  )
        return( EINVAL ) ;

    /*
     * set mcrs
     */
    /*      set only bsdcon in bsdcmc       */
    spl = splhi();
    /********
    mcr_load( MCR_BSDCMC, temp_bsdcmc ) ;
    temp_bsdcmc &= 0xfffffffb;
    **********/
    temp_bsdcmc = 0x3;
    if( barcon == BPU_CONNECT )
        temp_bsdcmc |= 4;
    mcr_store( MCR_BSDCMC, temp_bsdcmc ) ;
    splx(spl);

    /*      set bmmap       */
    if( barcon == BPU_CONNECT || sidenum == 0 ) /* connect or side0 */
        mcr_store( MCR_BMMAP0, bmmap );
    else
        mcr_store( MCR_BMMAP1, bmmap );

    /*      reset bstnl     */
    if( sidenum == 0 || barcon == BPU_CONNECT )
        mcr_store( MCR_BSTN0W, 0 );
    if( sidenum == 1 || barcon == BPU_CONNECT )
        mcr_store( MCR_BSTN1W, 0 );

    /*      set bar_side_info in memory     */
    bar_side_info[ sidenum ].barcon = barcon ;
    bar_side_info[ sidenum ].bmmap  = bmmap ;
    bar_side_info[ sidenum ].bstnl  = 0;
    bar_side_info[ sidenum ].bstnl_other = 0;
    bar_side_info[ sidenum ].barinfo = NULL ;
    bar_side_info[ sidenum ].rm_area = NULL ;
    bar_side_info[ sidenum ].rm_len  = NULL ;
    bar_side_info[ sidenum ].status = BSI_ST_NULL;

    return( 0 );
}

