#include	<util/types.h>
#include	<util/mcr.h>
#include	<util/mcr_field.h>
#include	<util/cmn_err.h>
#include	<util/vp_param.h>
#include	<mem/immu.h>
#include	<mem/kmem.h>
#include	<mem/as.h>
#include	<mem/seg.h>
#include	<mem/seg_vp.h>
#include	<mem/seg_gs.h>
#include	<mem/px_gs.h>		/* for gtop */
#include	<proc/user.h>
#include	<proc/proc.h>
#include	<proc/sched.h>
#include	<proc/signal.h>
#include	<proc/siginfo.h>
#include	<proc/pms/partition.h>
#include	<svc/todr.h>
#include	<svc/errno.h>
#include	<io/dtu/usr_dtu.h>
#include	<io/dtu/bar.h>
#include	<io/dtu/dc.h>
#include	<io/dtu/oc.h>
#include	<util/debug.h>
#include	<util/dl.h>		/* for 2word sub */

#define DTU_Q_BOUNDARY	0x00007fff
#define R2VADR( adr )   ((adr)+0xc0000000)

STATIC	int	num_of_side; 
STATIC  int	num_of_ntre_per_proc; 
STATIC  int	push_bodyl;	/* pushout packet body byte length */

STATIC udtu_side_info_t	*udtu_side_info;
STATIC uint_t		*ntr_image;		
STATIC udturp_req_t	*urp_head; /* first request block  */
STATIC udturp_req_t	*urp_tail; /* last request block   */

extern struct hqa  *hqa;
extern int	   dtu_ss;		/* dtu sub set flag */
extern bar_side_info_t	*bar_side_info;	/* bar_side_information */
extern paddr_t gtop();
extern struct mb	*rb0, *rb1;	/* #rqo */

#ifdef DEBUG
extern int     dtrace();
#endif

/***********************************************/
/*  KAI user-DTU JOBSTOP/RESTART and SWAP      */
/***********************************************/

/*
 * UDTU JOB STOP OPERATION
 * [NOTE] if success returns 0, but fail returns -1
 */
int
udtu_stop_vp( jid, sidenum )
px_jobid_t	jid;
px_sidenum_t	sidenum;
{
	int	spl;
	uint_t	mcr;
	udtu_side_info_t * usp;

	if( sidenum < 0 || sidenum >= PX_PE_MULTIPLEXY ) {
		return(-1);
	}
	partition = pms_partition[ sidenum ];
	if( partition == NULL ) {
		return(-1);
	}

	/* stop user transfer-queue */
	spl = splhi();

	dtu_snd_stop( DTU_STOP_SLEEP );

	mcr_load(  sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, mcr );
	mcr &= 0xfffeffff;
	mcr_store( sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, mcr );

	dtu_snd_restart();

	splx(spl);     /* not necesarry */

	/* get barrier (^^; */


	/* stop user reply-queue */
	spl = splhi();

	dtu_snd_stop( DTU_STOP_SLEEP );

	mcr_load(  sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, mcr );
	mcr &= 0xfffeffff;
	mcr_store( sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, mcr );

	dtu_snd_restart();

	splx(spl);     /* not necesarry */

	/* get barrier (^^; */


	/* store MCR and controls */
	mcr_load(  sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, mcr );
	mcr_load(  sidenum==0 ? MCR_HWU0TW : MCR_HWU1TW, mcr );

	mcr_load(  sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, mcr );
	mcr_load(  sidenum==0 ? MCR_HWU0RW : MCR_HWU1RW, mcr );

	/* valid NTR entory (not need?) */

	/* success */
	return( 0 );
}

/*
 * UDTU JOB RESTART OPERATION
 * [NOTE] if success returns 0, but fail returns -1
 *        please call after user-virtual-space recover and PMS.
 */
int
udtu_start_vp( jid, sidenum )
px_jobid_t	jid;
px_sidenum_t	sidenum;
{
	int	spl;
	uint_t	mcr;
	udtu_side_info_t * usp;

	if( sidenum < 0 || sidenum >= PX_PE_MULTIPLEXY ) {
		return(-1);
	}
	partition = pms_partition[ sidenum ];
	if( partition == NULL ) {
		return(-1);
	}

	/* valid NTR entory */
	vpnbr = (int) partition->ptn_vpnbr; 
	mcr_addr = (caddr_t)(MCR_NTRE000+((num_of_ntre_per_proc<<3)*sidenum));
	ui_addr  = ntr_image + num_of_ntre_per_proc*sidenum ;
#ifdef DEBUG
	cmn_err(CE_CONT, "[udtu_start_vp(%d): (vpid,netpa)={",sidenum);
#endif
	for( i=0; i < num_of_ntre_per_proc; ++i, mcr_addr+=8, ++ui_addr ) {
		if( i >= vpnbr ) {
			mcr_store( mcr_addr, 0x00008000 );
			*ui_addr = 0x00008000u;
		}
		else {
			ui = RPID2NETPA( partition->ptn_rpid[i] );
			uj = RPID2DUPID( partition->ptn_rpid[i] );
			uk = (ui<<16) | uj;
			mcr_store( mcr_addr, uk );
			*ui_addr = uk;
#ifdef DEBUG
			cmn_err(CE_CONT, "(%d,%d)",i,ui);
#endif
		}
	}
#ifdef DEBUG
	cmn_err(CE_CONT, "}]\n");
#endif
	/* restore MCR and controls */
		/* reply quere's G->P transration and set */

		/* udtu_side_info[sidenum] recovery */

	mcr = saved->trs_read;
	mcr_store(  sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, mcr );
	mcr = saved->trs_wrt;
	mcr_store(  sidenum==0 ? MCR_HWU0TW : MCR_HWU1TW, mcr );

	mcr = saved->rpl_read;
	mcr_store(  sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, mcr );
	mcr = saved->rpl_wrt;
	mcr_store(  sidenum==0 ? MCR_HWU0RW : MCR_HWU1RW, mcr );

	/* start user queue */

	/* success */
	return( 0 );
}


/*
 * UDTU SWAP SIZE
 * [NOTE] returns swap size
 */
int
udtu_swapsize_vp( jid, sidenum )
px_jobid_t	jid;
px_sidenum_t	sidenum;
{
	return(sizeof( udtu_side_info_t ));
}

/*
 * UDTU SWAP OUT OPERATION
 * [NOTE] 
 */
int
udtu_swapout_vp( jid, sidenum )
int		jswap_handle;
px_jobid_t	jid;
px_sidenum_t	sidenum;
{

}

/*
 * UDTU SWAP OUT OPERATION
 * [NOTE] 
 */
int
udtu_swapin_vp( jid, sidenum )
int		jswap_handle;
px_jobid_t	jid;
px_sidenum_t	sidenum;
{

}

/***********************************************/
/*  KAI barrier JOBSTOP/RESTART and SWAP       */
/***********************************************/

/*
 * BAR JOB STOP OPERATION
 * [NOTE] if success returns 0, but fail returns -1
 */
int
bar_stop_vp( jid, sidenum )
px_jobid_t	jid;
px_sidenum_t	sidenum;
{
	/* wait BST .... time */

	/* store MCR and controls */
		/* BST, LBSY save */
		/* bar_side_info_t */

	/* success */
	return( 0 );
}

/*
 * BAR JOB RESTART OPERATION
 * [NOTE] if success returns 0, but fail returns -1
 */
int
bar_start_vp( jid, sidenum )
px_jobid_t	jid;
px_sidenum_t	sidenum;
{
	/* store MCR and controls */
		/* BST, LBSY save */

	/* wait BST .... time */

	/* success */
	return( 0 );
}


/*
 * BAR SWAP SIZE
 * [NOTE] returns swap size
 */
int
bar_swapsize_vp( jid, sidenum )
px_jobid_t	jid;
px_sidenum_t	sidenum;
{
	return(sizeof( bar_side_info_t ));
}

/*
 * BAR SWAP OUT OPERATION
 * [NOTE] 
 */
int
bar_swapout_vp( jid, sidenum )
int		jswap_handle;
px_jobid_t	jid;
px_sidenum_t	sidenum;
{

}

/*
 * BAR SWAP OUT OPERATION
 * [NOTE] 
 */
int
bar_swapin_vp( jid, sidenum )
int		jswap_handle;
px_jobid_t	jid;
px_sidenum_t	sidenum;
{

}

/*------------------------------------------------------------------*/
