
#ident "@(#)lc.c	1.3 12/14/92"

/*  KAI DTP driver                            */
/*  LC Layer [ LINK Control service entries ] */
/*          Primary rewriter: Y.Nakashima     */
/*    nakashim@ayame.mfd.cs.fujitsu.co.jp     */

#include <util/cmn_err.h>
#include <util/types.h>
#include <util/param.h>
#include <svc/errno.h>
#include <proc/user.h>
#include <io/stream.h>
#include <io/stropts.h>
#include <io/strlog.h>
#include <io/log/log.h>
#include <util/debug.h>
#include <svc/systm.h>
#include <proc/cred.h>
#include <net/dlpi.h>
#include <io/ddi.h>
#include <svc/edt.h>
#include <util/bootinfo.h>
#include <mem/kmem.h>
#include <fs/ioccom.h>
#include <net/tcpip/if.h>
#include <net/tcpip/in.h>
#include <util/inline.h>
#include <io/lc/lcuser.h>
#include <io/lc/lc.h>
#include <io/dtu/dc.h>

/*
 * Routines defined in this file.
 */
void	lc_init( void ) ;
STATIC	int	lc_open( queue_t* readq, dev_t* devp, int flag,
			   int sflag, cred_t* cred ) ;
STATIC	int	lc_close( queue_t* writeq, int flag, cred_t* credp ) ;
STATIC	int	lc_wput( queue_t* writeq, mblk_t* mp ) ;
STATIC	int	lc_wsrv( queue_t* writeq ) ;
STATIC	int	lc_nackdata( struct lc* lcp, mblk_t* mp, int nacknum ) ;
STATIC	int	lc_wunitdata( struct lc_ucb* lup, struct lc* lcp, mblk_t* mp ) ;
STATIC	int	lc_send( unsigned char* lpp, int len ) ;
STATIC	void	lc_rintr(int dummy, struct mb* mb);
STATIC	void	lc_nonfatal( queue_t* readq, mblk_t* mp, long prim,
			      long dlerr, long unixerr ) ;
STATIC	void	lc_fatal( queue_t* readq, mblk_t* mp ) ;
STATIC	void	lc_domajors( dev_t dev ) ;
STATIC	struct  lc_ucb*	get_lcb( dev_t dev ) ;
STATIC mblk_t*	lc_rproc(struct lc_ucb* lup, struct lc* lcp,
			    unsigned char* lfp, int size, int sap) ;

/*
 * Streams interface information.
 */
STATIC struct module_info lcm_info = {
	LC_MID, LC_NAME, 0, LCMAXPSZ-EHEADSIZE, LC_HIWAT, LC_LOWAT
};

STATIC struct qinit lc_rinit = {
	NULL, NULL, lc_open, lc_close, NULL, &lcm_info, NULL
};

STATIC struct qinit lc_winit = {
	lc_wput, lc_wsrv, NULL, NULL, NULL, &lcm_info, NULL
};

struct streamtab lcinfo = {
	&lc_rinit, &lc_winit, NULL, NULL
};

/*
 * These are allocated from the master file.
 */
STATIC  int     lc_maxdev=2;		/* max number of minors per board */
STATIC	struct	lc	lcs[2];
STATIC	struct	lc_ucb	lc_ucbs[1];

STATIC int lc_majinit = 0;	/* set when major info has been initialized */
int lcdevflag = 0;		/* we support a new-style (SVR4) interface */

/*
 *	inet-specific.	drivers really shouldn't have protocol-specific
 *	stuff, but unfortunately the precedent is already set in bsd drivers.
 */
#define SIOCGIFFLAGS	_IOWR('i',17, struct ifreq)	/* get ifnet flags */
extern int		inetstats;	/* from lc.cf */
extern struct ifstats	*ifstats;	/* per-interface statistics for inet */
struct ifstats		lcstats[1];	/* statistics for lc interface */
/*
 * Initialize the LC data structures.  Called from
 * main() at boot time.
 */
void
lc_init()
{
	int mynetpa;
	int index;

	mynetpa = MY_NETPA();
	lc_ucbs[0].lu_lcp = &lcs[0];
	lc_ucbs[0].lu_addr[0] = 0x7e;
	lc_ucbs[0].lu_addr[1] = 0xff;
	lc_ucbs[0].lu_addr[2] = 0xff;
	lc_ucbs[0].lu_addr[3] = 0xff;
	lc_ucbs[0].lu_addr[4] = mynetpa >> 8;
	lc_ucbs[0].lu_addr[5] = mynetpa & 255;

	for (index = 0; index < lc_maxdev; index++) {
		lcs[index].lc_rdq = NULL;
		lcs[index].lc_wrq = NULL;
		lcs[index].lc_state = DL_UNBOUND;
		lcs[index].lc_ucb = &(lc_ucbs[0]) ;
		lcs[index].lc_unit = 0 ;
		lcs[index].lc_sap = 0;
	}

	if( inetstats ) {
		lcstats[0].ifs_name = "lc";
		lcstats[0].ifs_mtu = LCMAXPSZ-EHEADSIZE;
		lcstats[0].ifs_next = ifstats;
		ifstats = &lcstats[0];
		lc_ucbs[0].lu_ifstat = &(lcstats[0]);
	}
	return;
}

/* 
 * LC open procedure.  Initialize the data structures.
 */
STATIC int
lc_open (q, devp, flag, sflag, credp)
	queue_t *q;
	dev_t *devp;
	int flag;
	int sflag;
	cred_t *credp;
{
	struct lc_ucb*	lup ;
	struct lc*		lcp;
	dev_t 			dev;

	if( !lc_majinit )	/* set up board-to-major mapping */
		lc_domajors(*devp);
	if( sflag == MODOPEN )
		return EINVAL ;
	if( !( lup = get_lcb(*devp) ) )
		return EINVAL ;
	if( q->q_ptr )		/* already open */
		return 0 ;
	if( sflag == CLONEOPEN ) {
		lcp = lup->lu_lcp;
		for( dev = 0; dev < lc_maxdev; dev++, lcp++ ) {
			if( !lcp->lc_rdq )
				break;
		}
		if( dev >= lc_maxdev )
			return ENOSPC ;
		*devp = makedevice( lup->lu_major, dev );
	} else {
		dev = getminor(*devp);
		if( ( dev < 0 ) || ( dev >= lc_maxdev ) )
			return EINVAL ;
		lcp = lup->lu_lcp + dev;
	}
	lcp->lc_rdq = q;
	lcp->lc_wrq = WR(q);
	lcp->lc_sap = 0;
	q->q_ptr = (char *)lcp;
	WR(q)->q_ptr = (caddr_t)lcp;
	noenable(WR(q));
	lcp->lc_state = DL_UNBOUND;
	if( inetstats && lcstats )
		lup->lu_ifstat->ifs_unit = lup->lu_major;
	return 0 ;
}

/*
 * LC close procedure.  
 * All messages are freed.  
 */
STATIC int
lc_close(q, flag, credp)
	queue_t *q;
	int flag;
	cred_t *credp;
{
	struct lc *lcp ;
	struct lc_ucb*	lup ;
	int		plx ;
	mblk_t *mp, *tmp ;

	plx = spllc();
	lcp = (struct lc *)q->q_ptr;
	lup = lcp->lc_ucb ;
	if (WR(q)->q_first)
		flushq(WR(q), FLUSHALL);
	lcp->lc_rdq = NULL;
	lcp->lc_wrq = NULL;
	lcp->lc_sap = 0;
	lcp->lc_state = DL_UNBOUND;
	q->q_ptr = NULL;
	WR(q)->q_ptr = NULL;
	splx(plx);
	if (inetstats && lcstats) {
		lcstats[lcp->lc_unit].ifs_unit = 0;
		lcstats[lcp->lc_unit].ifs_active = 0;
		lcstats[lcp->lc_unit].ifs_ipackets = 0;
		lcstats[lcp->lc_unit].ifs_ierrors = 0;
		lcstats[lcp->lc_unit].ifs_opackets = 0;
		lcstats[lcp->lc_unit].ifs_oerrors = 0;
	}
	return 0 ;
}

/*
 * LC write side put procedure.  This performs the action indicated
 * by the message provided.  If the message requires no action by
 * the interface hardware, it is handled immediately.  Otherwise,
 * the message is either enqueued (because the firmware queue is full)
 * or a firmware job is created and the message put on the lc_head list.
 * lcint() will take care of handling the response from the 
 * hardware and sending messages back upstream.
 */
STATIC int
lc_wput(q, mp)
	queue_t *q;
	mblk_t *mp;
{
	int		plx ;
	struct lc*	lcp;
	struct lc*	nlcp;
	struct lc_ucb* lup ;
	struct iocblk *	iocp ;
	queue_t*	rq ;
	union DL_primitives*	p ;
	dl_bind_req_t *		reqp ;
	int		i, j ;
	mblk_t*		tmp;
	ushort_t	sap ;
	
	lcp = (struct lc *)q->q_ptr;
	lup = lcp->lc_ucb ;
	rq = RD(q) ;
	ASSERT(lcp);

	switch(mp->b_datap->db_type) {
	case M_IOCTL:
		if ((mp->b_wptr - mp->b_rptr) < sizeof(struct iocblk)) {
			mp->b_datap->db_type = M_IOCNAK;
			putnext(rq, mp);
			return 1 ;
		}			
		iocp = (struct iocblk *)mp->b_rptr;		
		if (drv_priv(iocp->ioc_cr) != 0) {
			mp->b_datap->db_type = M_IOCNAK;
			putnext(rq, mp);
			return 1 ;
		}
		if (iocp->ioc_count == TRANSPARENT) {	/* you lose */
			mp->b_datap->db_type = M_IOCNAK;
			putnext(rq, mp);
			return 1 ;
		}

		/*
		 * Most of these ioctls cause an action to be taken
		 * by the board and so cannot be acknowledged until
		 * a completion request is retured for this stream.
		 * lccproc() handles the generation of the ack - the
		 * message is merely placed in the waiting list until the
		 * completion request comes back.
		 */

		switch (iocp->ioc_cmd) {
		case SIOCGIFFLAGS:
			((struct iocblk_in *)iocp)->ioc_ifflags |= IFF_RUNNING;
			mp->b_datap->db_type = M_IOCACK;
			putnext(rq, mp);
			return 1 ;

		default:
			mp->b_datap->db_type = M_IOCNAK;
			putnext(rq, mp);
			return 1 ;
		} /* end of ioctl switch */
		break ;

	case M_PROTO:
	case M_PCPROTO:
		p = (union DL_primitives *)mp->b_rptr;
		switch(p->dl_primitive) {
		case DL_BIND_REQ: {
			dl_bind_req_t *reqp;
			dl_bind_ack_t *ackp;

			if (lcp->lc_state != DL_UNBOUND) {
				lc_nonfatal(rq, mp, DL_BIND_REQ,
					     DL_OUTSTATE, 0);
				return 1 ;
			}
			reqp = (dl_bind_req_t *)mp->b_rptr;
			sap = reqp->dl_sap;
			for( j = 0, nlcp = lup->lu_lcp ; j < lc_maxdev ;
			     j++, nlcp++ ) {
				if( (nlcp->lc_state == DL_IDLE) &&
				    (sap == nlcp->lc_sap)) {
					lc_nonfatal(rq, mp, DL_BIND_REQ,
						     DL_BADSAP, 0);
					return 1 ;
				}
			}

			/* bind o.k., return dl_bind_ack_t */
			if (!REUSEABLE(mp, sizeof(*ackp)+PHYAD_SIZE)) {
				if (!(tmp = allocb(sizeof(*ackp)+PHYAD_SIZE,
						   BPRI_MED))) {
					lc_nonfatal(rq, mp, DL_BIND_REQ,
						     DL_SYSERR, ENOMEM);
					return 1 ;
				}
				freemsg(mp);
				mp = tmp;
			} else {
				mp->b_rptr = mp->b_datap->db_base;
				mp->b_wptr = mp->b_rptr;
				if (mp->b_cont) {
					freemsg(mp->b_cont);
					mp->b_cont = NULL;
				}
			}

			mp->b_datap->db_type = M_PCPROTO;
			ackp = (dl_bind_ack_t *)mp->b_wptr;
			ackp->dl_primitive = DL_BIND_ACK;
			ackp->dl_sap = sap;
			ackp->dl_addr_length = PHYAD_SIZE;
			ackp->dl_addr_offset = sizeof(*ackp);
			ackp->dl_max_conind = 0;
			ackp->dl_growth = 0;
			mp->b_wptr += sizeof(*ackp);
			liacpy(lcp->lc_ucb->lu_addr, mp->b_wptr);
			mp->b_wptr += PHYAD_SIZE;

			lcp->lc_state = DL_IDLE;
			lcp->lc_sap = sap;
			putnext(rq, mp);
			if (inetstats && lcstats)
				lcstats[lcp->lc_unit].ifs_active = 1;
			return 1 ;
		    }

		case DL_UNBIND_REQ: {
			dl_ok_ack_t *okp;

			if (lcp->lc_state != DL_IDLE)  {
				lc_nonfatal(rq, mp, DL_UNBIND_REQ,
					     DL_OUTSTATE, 0);
				return 1 ;
			}
	
			/* unbind o.k., return dl_ok_ack_t */
			if (!REUSEABLE(mp, sizeof(dl_ok_ack_t))) {
				if (!(tmp = allocb(sizeof(dl_ok_ack_t),
						   BPRI_MED))) {
					lc_nonfatal(rq, mp, DL_UNBIND_REQ,
						     DL_SYSERR, ENOMEM);
					return 1 ;
				}
				freemsg(mp);
				mp = tmp;
			} else {
				mp->b_rptr = mp->b_datap->db_base;
				mp->b_wptr = mp->b_rptr;
				if (mp->b_cont) {
					freemsg(mp->b_cont);
					mp->b_cont = NULL;
				}
			}

			okp = (dl_ok_ack_t *)mp->b_wptr;
			mp->b_datap->db_type = M_PCPROTO;
			okp->dl_primitive = DL_OK_ACK;;
			okp->dl_correct_primitive = DL_UNBIND_REQ;
			mp->b_wptr += sizeof(dl_ok_ack_t);
			lcp->lc_state = DL_UNBOUND;
			lcp->lc_sap = 0;
			putnext(rq, mp);
			if (inetstats && lcstats)
				lcstats[lcp->lc_unit].ifs_active = 0;
			return 1 ;
		}

		case DL_INFO_REQ: {
			dl_info_ack_t *ackp;

			if (!REUSEABLE(mp, sizeof(dl_info_ack_t)+PHYAD_SIZE)) {
				if (!(tmp = allocb(sizeof(dl_info_ack_t)+PHYAD_SIZE,
						   BPRI_MED))) {
					lc_nonfatal(rq, mp, DL_UNBIND_REQ,
						     DL_SYSERR, ENOMEM);
					return 1 ;
				}
				freemsg(mp);
				mp = tmp;
			} else {
				mp->b_rptr = mp->b_datap->db_base;
				mp->b_wptr = mp->b_rptr;
				if (mp->b_cont) {
					freemsg(mp->b_cont);
					mp->b_cont = NULL;
				}
			}

			mp->b_datap->db_type = M_PCPROTO;
			ackp = (dl_info_ack_t *)mp->b_wptr;
			ackp->dl_primitive = DL_INFO_ACK;
			ackp->dl_max_sdu = LCMAXPSZ-EHEADSIZE;
			ackp->dl_min_sdu = 1;
			ackp->dl_addr_length = PHYAD_SIZE;
			ackp->dl_mac_type = DL_ETHER;
			ackp->dl_reserved = 0;
			ackp->dl_current_state = lcp->lc_state;
			ackp->dl_service_mode = DL_CLDLS;
			ackp->dl_qos_length = 0;
			ackp->dl_qos_offset = 0;
			ackp->dl_qos_range_length = 0;
			ackp->dl_qos_range_offset = 0;
			ackp->dl_provider_style = DL_STYLE1;
			ackp->dl_growth = 0;
			if (lcp->lc_state == DL_IDLE) {
				ackp->dl_addr_offset = sizeof(dl_info_ack_t);
				mp->b_wptr += sizeof(dl_info_ack_t);
				liacpy(lcp->lc_ucb->lu_addr, mp->b_wptr);
				mp->b_wptr += PHYAD_SIZE;
			} else {
				ackp->dl_addr_offset = 0;
				mp->b_wptr += sizeof(dl_info_ack_t);
			}
			putnext(rq, mp);
			return 1;
		}

		default:
			lc_nonfatal(rq, mp, p->dl_primitive, DL_UNSUPPORTED, 0);
			return 1 ;

		/*
		 *  Followings are processed in wsrv()
		 */
		case DL_UNITDATA_REQ:
			break;
		}
		break;

	case M_FLUSH:
		if (*mp->b_rptr & FLUSHW) {
			flushq(q, FLUSHALL);
			*mp->b_rptr &= ~FLUSHW;
		}
		if (*mp->b_rptr & FLUSHR) 
			qreply(q, mp);
		else 
			freemsg(mp);
		return 1;

	default:
		lc_fatal(rq, mp);
		return 1 ; 

	}

	plx = spllc();		/* CRITICAL BEGIN */

	/*
	 *  enqueue mblk into waiting queue
	 */
	mp->b_prev = (mblk_t *)lcp;		/* XXXXX: temporary */
	mp->b_next = 0;
	if (lup->lu_sendh) {
		lup->lu_sendt->b_next = mp;
		lup->lu_sendt = mp;
	} else
		lup->lu_sendh = lup->lu_sendt = mp;
	lup->lu_swcnt++;

	/*
	 * process send waiting queue
	 */
	lc_sproc(lup);

	splx(plx);			/* CRITICAL END */

	return 1;
}

/*
 * Dummy write service procedure for
 * flow control only.
 */
STATIC int
lc_wsrv (q)
	queue_t *q;
{
	return;
}

/*
 *  Write side waiting queue process routine
 *  This routine is called in lc IPL level.
 */
STATIC
lc_sproc (lup)
struct lc_ucb	*lup;
{
	mblk_t*	mp;
	struct lc* lcp;
	queue_t *rq;		/* read queue */

	while (mp = lup->lu_sendh) {

		/* dequeue `mp' from waiting queue */
		lup->lu_swcnt--;
		if ((lup->lu_sendh = mp->b_next) == 0) {
			lup->lu_sendt = 0;
			ASSERT(lup->lu_swcnt == 0);
		}

		lcp = (struct lc *)mp->b_prev;

		rq = lcp->lc_rdq;
		switch(mp->b_datap->db_type) {
		case M_PROTO:
		case M_PCPROTO: {
			union DL_primitives *p;
	
			p = (union DL_primitives *)mp->b_rptr;
			switch(p->dl_primitive) {
			case DL_UNITDATA_REQ:
				if (!lc_wunitdata(lup, lcp, mp)) {
#ifdef DEBUG
					cmn_err(CE_CONT,"<lc_sproc:D.fail>");
#else
					cmn_err(CE_CONT,"<D.fail>");
#endif
					return;
				}
				break;
	
			default:
				freemsg(mp);
				break;
			}
			break; }
	
		default:
			lc_fatal(rq, mp);
			break;

		} /* end of message type switch */
	}
}

STATIC	int
lc_nackdata( lcp, mp, nacknum )
struct	lc*		lcp ;
mblk_t*			mp ;
int			nacknum ;
{
	dl_uderror_ind_t *errp;
	
	if( !REUSEABLE( mp, DL_UDERROR_IND_SIZE ) ) {
		freemsg(mp);
		if (!(mp = allocb(DL_UDERROR_IND_SIZE, BPRI_MED))) {
			cmn_err(CE_WARN,
				"LC - could not allocate DL_UDERROR_IND\n");
			return 1 ;
		}
	} else {
		freemsg(mp->b_cont);
		mp->b_cont = NULL;
		mp->b_rptr = mp->b_datap->db_base;
	}
	errp = (dl_uderror_ind_t *)mp->b_rptr;
	/* dl_src_addr_length and dl_src_addr_offset are set to 0 */
	bzero((char *)errp, DL_UDERROR_IND_SIZE);
	errp->dl_primitive = DL_UDERROR_IND;
	errp->dl_errno = nacknum;
	mp->b_wptr = mp->b_rptr + DL_UDERROR_IND_SIZE;
	mp->b_datap->db_type = M_PCPROTO;
	putnext(lcp->lc_rdq, mp);
	if (inetstats && lcstats)
		lcstats[lcp->lc_unit].ifs_oerrors++;
	return 1 ;
}

STATIC	int
lc_wunitdata( lup, lcp, mp )
struct	lc_ucb*	lup ;
struct	lc*		lcp ;
mblk_t*			mp ;
{
	uchar_t*	ap ;
	dl_unitdata_req_t *dp;
	struct ehead*	hdrp;
	int		i ;
	uchar_t*	addrp ;
	unsigned char*	p ;
	int		len ;
	mblk_t*		tmp ;
	int		plx ;
	unsigned char	lpp[LCMAXPSZ];
	
	dp = (dl_unitdata_req_t *)mp->b_rptr;
	if (lcp->lc_state != DL_IDLE) {
cmn_err(CE_CONT,"<W.!idle>");
		return lc_nackdata( lcp, mp, LC_NOTBIND ) ;
	}

	hdrp = (struct ehead*)lpp ;
	p = lpp+sizeof(struct ehead);
	ap = mp->b_rptr + dp->dl_dest_addr_offset;
	liacpy(lcp->lc_ucb->lu_addr, hdrp->src_addr);
	liacpy(ap, hdrp->dest_addr);
	hdrp->pro_type.val = lcp->lc_sap;

	len = EHEADSIZE;
	for (tmp = mp->b_cont; tmp; tmp = tmp->b_cont) {
		int	xl = tmp->b_wptr - tmp->b_rptr ;
		if( ( len + xl ) > LCMAXPSZ )
			return lc_nackdata( lcp, mp, LC_MAX_MSGLEN ) ;
		bcopy((caddr_t)tmp->b_rptr, (caddr_t)p, xl ) ;
		p += xl ;
		len += xl ;
	}
	(void) freemsg(mp);

	if (len < LCMINPSZ)
		len = LCMINPSZ;

	if (lc_send(lpp, len)!=0)
            return 0;

	if (inetstats && lcstats)
		lcstats[lcp->lc_unit].ifs_opackets++;
	return 1 ;
}

STATIC  int
lc_send( lpp, len )
	unsigned char*	lpp;
	int	len;
{
	int i;

        if (lpp[0] == 0xff && lpp[1] == 0xff
         && lpp[2] == 0xff && lpp[3] == 0xff
         && lpp[4] == 0xff && lpp[5] == 0xff) {
            for (i=0; i<256; i++)
	        dc_send(i,0,DC_MODE_NFUN,0,lpp,len,NULL,NULL,NULL);
        }
        else
	    dc_send((lpp[4]<<8)+lpp[5],0,DC_MODE_NFUN,0,lpp,len,NULL,NULL,NULL);

	return(0);
}

/*
 * LC interrupt handling procedure.  The interrupt id identifies the
 * stream for which the completed action was being performed and the 
 * message that initiated it.  The stream and message are then examined 
 * and the appropriate completion action carried out.  If another message
 * is waiting on that stream, it is then serviced.  Finally, if some
 * stream has been waiting for the request queue to empty out,
 * that stream is serviced (if possible).  
 * Note that there can be only one outstanding request per stream.
 */

STATIC
void
lc_rintr(dummy, mb)
        int dummy;
        struct mb* mb;
{
	struct lc_ucb*  lup = lc_ucbs;
	int		len, sap;
	mblk_t*		mp;
	struct lc*	lcp ;
	unsigned char*	datap;
	int		i;

	datap = (unsigned char*)(mb->msg);
	len = mb->len;

	sap = (datap[EHEAD_OFF_SAP] << 8) | datap[EHEAD_OFF_SAP+1];
	/* search matching SAP & push data */
	lcp = lup->lu_lcp;
	for (i = 0; i < lc_maxdev; i++, lcp++) {
		if (lcp->lc_sap == sap) {
			if( lcp->lc_state != DL_IDLE )
				break ;
			mp = lc_rproc(lup, lcp, datap, len, sap);
			dc_mbfree(mb);
			if (mp) {
				putnext(lcp->lc_rdq, mp) ;
			}
			break;
		}
	}
}

STATIC mblk_t*
lc_rproc (lup, lcp, lfp, size, sap)
struct	lc_ucb*	lup;
struct	lc*		lcp;
unsigned char*		lfp;
int			size, sap;
{
	mblk_t*		datamp = 0;
	mblk_t*		dlpimp;
	unsigned char*	ehp;
	dl_unitdata_ind_t*	duip;
	unchar*		p;

	ehp = lfp;
	size -= EHEADSIZE;
	lfp += EHEADSIZE;
	datamp = allocb(size, BPRI_MED);
	dlpimp = allocb(DATINDSIZE, BPRI_MED) ;

	if (datamp == NULL || dlpimp == NULL) {
		if (datamp)
			freemsg(datamp);
		if (dlpimp)
			freemsg(dlpimp);

		cmn_err(CE_WARN, "lc: No more buffer for recv, sap 0x%x"
		, sap) ;

		return NULL;
	}

	if (!canput(lcp->lc_rdq->q_next)) {
		cmn_err(CE_WARN, "lc: upstream queue full sap, == 0x%x"
		, sap);
{ queue_t *q;
  cmn_err(CE_CONT, "queue:");
  if (q = lcp->lc_rdq->q_next)
    for (q = lcp->lc_rdq->q_next; q; q = q->q_next) {
      cmn_err(CE_CONT," (%x %d %d)", q, q->q_flag & QFULL ? 1 : 0, q->q_blocked);
    }
  else
    cmn_err(CE_CONT, " no next");
  cmn_err(CE_CONT, "\n");
}
		if (inetstats && lcstats)
			lcstats[lcp->lc_unit].ifs_ierrors++;
		(void) freemsg(datamp);
		(void) freemsg(dlpimp);
		return NULL ;
	}

	bcopy((caddr_t)lfp, (caddr_t)datamp->b_wptr, size);
	datamp->b_wptr += size;

	duip = (dl_unitdata_ind_t *)dlpimp->b_wptr;
	duip->dl_primitive = DL_UNITDATA_IND;
	duip->dl_dest_addr_length = PHYAD_SIZE;
	duip->dl_dest_addr_offset = sizeof(*duip);
	duip->dl_src_addr_length = PHYAD_SIZE;
	duip->dl_src_addr_offset = sizeof(*duip) + PHYAD_SIZE;
	/* fill in remote address */
	p = dlpimp->b_wptr + duip->dl_dest_addr_offset;
	liacpy(&ehp[EHEAD_OFF_DEST], p);
	/* fill in local address */
	p = dlpimp->b_wptr + duip->dl_src_addr_offset;
	liacpy(&ehp[EHEAD_OFF_SRC], p);
	duip->dl_reserved = 0;
	dlpimp->b_wptr += DATINDSIZE;
	dlpimp->b_datap->db_type = M_PROTO;
	dlpimp->b_cont = datamp ;

	if (inetstats && lcstats)
		lcstats[lcp->lc_unit].ifs_ipackets++;
	return dlpimp ;
}


STATIC void
lc_nonfatal(rq, mp, prim, dlerr, unixerr)
queue_t *rq;		/* read q */
mblk_t *mp;
long prim, dlerr, unixerr;
{
	register dl_error_ack_t *errp;

	if (!REUSEABLE(mp, sizeof(*errp))) {
		mblk_t *bp;
		if ((bp = allocb(sizeof(*errp), BPRI_HI)) == NULL) {
			freemsg(mp);
			return;
		}
		freemsg(mp);
		mp = bp;
	} else {
		mp->b_rptr = mp->b_datap->db_base;
		mp->b_wptr = mp->b_rptr;
		if (mp->b_cont) {
			freemsg(mp->b_cont);
			mp->b_cont = NULL;
		}
	}
	mp->b_datap->db_type = M_PCPROTO;
	errp = (dl_error_ack_t *)mp->b_wptr;
	errp->dl_primitive = DL_ERROR_ACK;
	errp->dl_error_primitive = prim;
	errp->dl_errno = dlerr;
	errp->dl_unix_errno = unixerr;
	mp->b_wptr += sizeof(*errp);
	putnext(rq, mp);
}

/*
 * fatal error: should have set lc_state to error state, instead UNBND
 */
STATIC void
lc_fatal(rq, mp)
	queue_t *rq;
	mblk_t *mp;
{
	struct lc *lcp;

	lcp = (struct lc *)rq->q_ptr;
	lcp->lc_state = DL_UNBOUND;
	mp->b_datap->db_type = M_ERROR;
	mp->b_wptr = mp->b_rptr = mp->b_datap->db_base;
	*mp->b_wptr++ = EPROTO;
	putnext(rq, mp);
}

/*
 * Map external major numbers to ucbs.
 */
STATIC void
lc_domajors(dev)
	register dev_t dev;
{
	int	maj = -1;

	dev = getmajor(dev);
	maj = itoemajor(dev, maj);
	if (maj == -1) {
		cmn_err(CE_WARN, "lc: more boards than majors configured");
		return;
	}
	lc_ucbs[0].lu_major = maj;
	if(itoemajor(dev, maj) != -1)
		cmn_err(CE_NOTE, "lc: more majors than boards configured");
	lc_majinit = 1;
}

/*
 * Given a external dev, return the corresponding ucb.
 */
STATIC struct lc_ucb*
get_lcb(dev)
	dev_t dev;
{
	dev = getemajor(dev) ;
	if( dev == lc_ucbs[0].lu_major )
		return &(lc_ucbs[0]) ;
	return NULL ;
}

