#ident  "@(#)oc.c	1.3 1/6/93 16:14:33 - FUJITSU UXP/VPP"

/*  KAI DTP driver                           */
/*  OC Layer [ DTP Control service entries ] */
/*  Copyright (c) 1991 by FUJITSU LIMITED    */
/*   Mainframe division                      */
/*   Supercomputer development department    */
/*            Primary writer: Y.Nakashima    */
/*    nakashim@ayame.mfd.cs.fujitsu.co.jp    */

#include "util/types.h"
#include "util/param.h"
#include "util/sysmacros.h"
#include "util/debug.h"
#include "mem/immu.h"
#include "mem/page.h"
#include "mem/kmem.h"
#include "mem/kaidep.h"
#include "proc/proc.h"
#include "io/dtu/dc.h"
#include "io/dtu/oc.h"
#include "io/dtu/bar.h"
#include "util/bootinfo.h"
#include "util/cmn_err.h"
#include "svc/errno.h"
#ifdef SPMS /* for spms command support */
#include "proc/user.h"
#include "proc/sched.h"
#include "proc/job.h"
#include "proc/session.h"
#include "proc/pms/spms.h"
#include "util/petable.h"
#endif /* SPMS */

extern ppageinfo_t * vtop_pagelock();
extern void vtop_pageunlock();

#ifdef DEBUG
extern int     dtrace();
int oc_pages; /* oc using page counter */
#define UPRINTFSIZE     4096*16
char    *uprintf_top;
char    *uprintf_cur;
char    *uprintf_last;
#endif

/*
 * initialize: called from main() top
 */
void
dtu_init()
{
#ifdef DEBUG
dbg_printf("[call dc_init]");
#endif
    dc_init();
#ifdef DEBUG
dbg_printf("[call init_udtu_bar]");
#endif
    init_udtu_bar();
#ifdef SPMS
#ifdef DEBUG
dbg_printf("[call spms_tbl_init]");
#endif
    spms_tbl_init();
#endif
#ifdef DEBUG 
    oc_pages = 0; /* oc using page counter */
    /* user dbg_printf data area alloc */ 
    uprintf_top = uprintf_cur = kmem_zalloc(UPRINTFSIZE, KM_SLEEP);
    uprintf_last = uprintf_top + UPRINTFSIZE;
#endif
}

/* 
 * oc_ipl(K): send IPL packet to target machine.
 */
oc_ipl(target, iplb, len)
  int         target; /* not px_peid_t */
  char*       iplb;
  unsigned    len;
{
  extern proc_t     * curproc;
  /* struct  plist { unsigned addr, len; } *plist; */
  ppageinfo_t *plist;
  int     cc,spl;

#ifdef DEBUG
  dtrace("ocipl","enter",target,0);
  oclog('ipl ', target, iplb, len, 0, 0, 0);
  dbg_printf("[oc_ipl called netpa=%x,iplb=%x,len=%x]", target, iplb, len);
#endif

  if ((plist = (ppageinfo_t*)vtop_pagelock(curproc, iplb, len)) == NULL)
    return(DTU_EINVAL);
  
  spl = splhi(); /* 92/10/27 K.H */
  cc = dc_send((px_peid_t)target,
               0,
               DC_MODE_NIPL,
               0,
               plist->pi_addr,
               plist->pi_len,
               0x88000,
               plist,
               0);
  sleep(plist, PZERO);
  splx(spl); /* 92/10/27 K.H */
#ifdef DEBUG
  dtrace("ocipl","sndst",cc,0);
#endif
  vtop_pageunlock(curproc, iplb, len, plist);
  if (cc < 0) {
    switch (cc) {
    case DC_CC_RPDOWN:    return(DTU_RPDOWN);
    case DC_CC_RPSWAP:    return(DTU_RPSWAP);
    case DC_CC_EINVAL:    return(DTU_EINVAL);
    case DC_CC_ESNOMEM:   return(DTU_ENOMEM);
    default:              
        cmn_err(CE_PANIC,"oc_ipl: sending fail with %d.", cc);
        break;
    }
  }
#ifdef DEBUG
  dtrace("ocipl","return",0,0);
#endif
  return(0);
}

/*
 * oc_send(K): sending message
 */
oc_send(token, target, msqid, mtype, msg, len, prior, mode)
  int*        token;
  int         target;  /* not px_peid_t */
  int         mtype;   /* not use */
  unsigned    msqid;
  char*       msg;
  unsigned    len;
  unsigned    prior;
  unsigned    mode;
{
  int cc;
  
#ifdef DEBUG
    oclog('snd', target, msqid, msg, len, prior, 0);
    dtrace("ocsend","enter",msqid,0);
    dbg_printf("[oc_send called netpa=%x,qid=%x,mtype=%x,maddr=%x,mlen=%x,raddr=%x,rlen=%x]",
                target, msqid, mtype, msg, len);
#endif
  cc = dc_send((px_peid_t)target,
               prior == DTU_HIGH,
               mode == DTU_NOREPLY ? DC_MODE_AMSG : DC_MODE_TMSG,
               msqid,
               msg,
               len,
               NULL,
               NULL,
               0);
#ifdef DEBUG
    dtrace("ocsend","sndst",cc,0);
#endif
  if (cc < 0) {
    switch (cc) {
    case DC_CC_RPDOWN:    return(DTU_RPDOWN);
    case DC_CC_RPSWAP:    return(DTU_RPSWAP);
    case DC_CC_EINVAL:    return(DTU_EINVAL);
    case DC_CC_ERNOMEM:   return(DTU_ENOMEM);
    case DC_CC_ECOMM:     return(DTU_ETMO);
    case DC_CC_SIGNAL:    return(DTU_SIGNAL);
    case DC_CC_FORCEDOWN: return(DTU_ENETPA); /* temporary errcode */
    default:              
        cmn_err(CE_PANIC,"oc_send: sending fail with %d.", cc);
        break;
    }
  }
  *token = cc;
#ifdef DEBUG
    dtrace("ocsend","ok",0,0);
#endif
  return(0);
}

/*
 * oc_sends(K): sending message and wait reply.
 */
oc_sends(target, msqid, mtype, msg, len, res, res_len, prior)
  int         target;  /* not px_peid_t */
  unsigned    msqid;
  int         mtype;   /* not use */
  char*       msg;
  unsigned    len;
  char*       res;
  unsigned    res_len;
  unsigned    prior;
{
  struct mb* mb;
  int cc, tokid, rlen;

#ifdef DEBUG
    oclog('snds', target, msqid, msg, len, prior, 0);
    dtrace("ocsends","enter",msqid,0);
    dbg_printf("[oc_sends called netpa=%x,qid=%x,mtype=%x,maddr=%x,mlen=%x,raddr=%x,rlen=%x]",
                target, msqid, mtype, msg, len, res, res_len);
#endif

  cc = tokid = dc_send((px_peid_t)target,
                       prior == DTU_HIGH,
                       DC_MODE_TMSG,
                       msqid,
                       msg, 
                       len,
                       NULL,
                       NULL,
                       0);
#ifdef DEBUG
    dtrace("ocsends","sndst",cc,0);
#endif
  if (cc < 0) {
    switch (cc) {
    case DC_CC_RPDOWN:    return(DTU_RPDOWN);
    case DC_CC_RPSWAP:    return(DTU_RPSWAP);
    case DC_CC_EINVAL:    return(DTU_EINVAL);
    case DC_CC_ERNOMEM:   return(DTU_ENOMEM);
    case DC_CC_ECOMM:     return(DTU_ETMO);
    case DC_CC_SIGNAL:    return(DTU_SIGNAL);
    case DC_CC_FORCEDOWN: return(DTU_ENETPA); /* temporary errcode */
    default:              
        cmn_err(CE_PANIC,"oc_sends: sending fail with %d.", cc);
        break;
    }
  }
  cc = dc_wait(DC_MODE_WAIT,
               tokid,
               &mb);
  if (cc < 0) {
    switch (cc) {
    case DC_CC_EINVAL:    return(DTU_EINVAL);
    case DC_CC_SIGNAL:    dc_cancel(tokid); return(DTU_SIGNAL);
    case DC_CC_FORCEDOWN: return(DTU_ENETPA); /* temporary errcode */
    case DC_CC_CANTOK:    return(DTU_ENETPA); /* temporary errcode */
    default:              
        cmn_err(CE_PANIC,"oc_sends: wait fail with %d.", cc);
        break;
    }
  }
  bcopy(mb->msg, res, mb->len < res_len ? mb->len : res_len);

  rlen = mb->len;
  dc_mbfree(mb);

  if( rlen > res_len ) return( DTU_EMLEN );
#ifdef DEBUG
    dtrace("ocsends","ok",0,0);
#endif
  return(0);
}

oc_wait(token, res, res_len)
  int         token;
  char*       res;
  unsigned    res_len;
{
  struct mb* mb;
  int cc, rlen;

#ifdef DEBUG
    oclog('wait', token, res, res_len);
    dtrace("ocwait","enter",token,0);
    dbg_printf("[oc_wait called token=%x,res=%x,reslen=%x]",
                token, res, res_len);
#endif
  cc = dc_wait(DC_MODE_WAIT,
               token,
               &mb);
  if (cc < 0) {
    switch (cc) {
    case DC_CC_EINVAL:    return(DTU_EINVAL);
    case DC_CC_ENOTOK:    return(DTU_EINVAL);
    case DC_CC_SIGNAL:    dc_cancel(token); return(DTU_SIGNAL);
    case DC_CC_FORCEDOWN: return(DTU_ENETPA); /* temporary errcode */
    case DC_CC_CANTOK:    return(DTU_ENETPA); /* temporary errcode */
    default:              
        cmn_err(CE_PANIC,"oc_wait: wait fail with %d.", cc);
        break;
    }
  }
  bcopy(mb->msg, res, mb->len < res_len ? mb->len : res_len);

  rlen = mb->len;
  dc_mbfree(mb);

  if( rlen > res_len ) return( DTU_EMLEN );
#ifdef DEBUG
    dtrace("ocwait","ok",0,0);
#endif
  return(0);
}

/* 
 * oc_receive(K): receive message.
 */
/* oc_receive(msqid, msg, len, from_pe, token, options) */
oc_receive(msqid, mtype, msg, len, netpa, token, rtype, options)
  unsigned    msqid;
  int         mtype;    /* not use */
  char*       msg;
  unsigned    len;
  int*        netpa;    /* not px_peid_t */
  int*        token;
  int         rtype;    /* not use */
  unsigned    options;
{
  struct mb* mb;
  int cc;
  unsigned   rlen;
  px_peid_t  from_pe;

#ifdef DEBUG
    dtrace("oc_recv", "enter", msqid, 0);
    dbg_printf("[oc_receive called qid=%x,mtype=%x,msg=%x,mlen=%x,pe=%x,token=%x,rtype=%x]", msqid, mtype, msg, len, from_pe, token, rtype);
#endif

  cc = dc_recv(options == DTU_WAIT ? DC_MODE_WAIT|DC_MODE_PCATCH : DC_MODE_NOWAIT,
               msqid,
               &from_pe,
               token,
               &mb);
  if (cc < 0) {
    switch (cc) {
    case DC_CC_EINVAL:    return(DTU_EINVAL);
    case DC_CC_SIGNAL:    return(DTU_SIGNAL);
    case DC_CC_NOMSG:     return(DTU_NOMSG);
    /* default:              panic(); */
    default:              
        cmn_err(CE_PANIC,"oc_receive: receive fail with %d.", cc);
        break;
    }
  }
  bcopy(mb->msg, msg, mb->len < len ? mb->len : len);
  *netpa = (int)from_pe;
  rlen = mb->len;
  dc_mbfree(mb);

  if( rlen > len ) return( DTU_EMLEN );
  return(rlen);
}

/* 
 * oc_reply(K): reply sending.
 */
oc_reply(from, token, res, len)
  int         from;  /* not px_peid_t */
  int         token;
  char*       res;
  unsigned    len;
{
  int cc;
    
#ifdef DEBUG
    oclog('repl', from, token, res, len);
    dtrace("ocrepl","enter",token,0);
    dbg_printf("[oc_repl called target=%x, token=%x, res=%x, len=%x]",
                from, token, res, len);
#endif
  cc = dc_repl((px_peid_t)from,
               0,
               DC_MODE_AMSG,
               token,
               res,
               len,
               0);
#ifdef DEBUG
    dtrace("ocrepl","sndst",cc,0);
#endif
  if (cc < 0) {
    switch (cc) {
    case DC_CC_EINVAL:    return(DTU_EINVAL);
    case DC_CC_ESNOMEM:   return(DTU_ENOMEM);
    case DC_CC_ERNOMEM:   return(DTU_ETMO);
    case DC_CC_ECOMM:     return(DTU_ETMO);
    case DC_CC_RPDOWN:    return(DTU_RPDOWN);
    case DC_CC_RPSWAP:    return(DTU_RPSWAP);
    case DC_CC_SIGNAL:    return(DTU_SIGNAL);
    case DC_CC_FORCEDOWN: return(DTU_ENETPA); /* temporary errcode */
    /* default:              panic(); */
    default:              
        cmn_err(CE_PANIC,"oc_reply: reply send fail with %d.", cc);
        break;
    }
  }
#ifdef DEBUG
    dtrace("ocrepl","ok",0,0);
#endif
  return(0);
}

/*
 * oc_rpc(K):
 */
oc_rpc(target, func, argc, argv, res, res_len, prior)
  int         target;
  unsigned    func;
  unsigned    argc;
  char*       argv;
  char*       res;
  unsigned    res_len;
  unsigned    prior;
{
  int cc, tokid;

#ifdef DEBUG
    oclog('rpc', target, func, argc, argv);
    dtrace("ocrpc","enter",func,0);
    dbg_printf("[oc_rpc called target=%x, func=%x, argc=%x, argv=%x]",
                target, func, argc, argv);
#endif
  cc = tokid = dc_send((px_peid_t)target,
                       prior == DTU_HIGH,
                       DC_MODE_AFUN,
                       func,
                       argv, 
                       argc,
                       NULL,
                       NULL,
                       0);
#ifdef DEBUG
    dtrace("ocrpc","sendst",cc,0);
#endif
  if (cc < 0) {
    switch (cc) {
    case DC_CC_EINVAL:    return(DTU_EINVAL);
    case DC_CC_ESNOMEM:   return(DTU_ENOMEM);
    case DC_CC_ERNOMEM:   return(DTU_ENOMEM);
    case DC_CC_ENOFUNC:   return(DTU_ENOENT); 
    case DC_CC_ECOMM:     return(DTU_ETMO);
    case DC_CC_RPDOWN:    return(DTU_RPDOWN);
    case DC_CC_RPSWAP:    return(DTU_RPSWAP);
    case DC_CC_SIGNAL:    return(DTU_SIGNAL);
    case DC_CC_FORCEDOWN: return(DTU_ENETPA); /* temporary errcode */
    /* default:              panic(); */
    default:              
        cmn_err(CE_PANIC,"oc_rpc: sending fail with %d.", cc);
        break;
    }
  }
#ifdef DEBUG
    dtrace("ocrpc","ok",0,0);
#endif
  return(0);
}

/*
 * oc_write(K): sending 4-byte aligned data
 * [NOTE]
 */
oc_write(target, from, to, len, pid, prior, mode)
  int         target; /* not px_peid_t */
  char*       from;
  char*       to;
  unsigned    len;
  int         pid;    /* not use */
  unsigned    prior;
  unsigned    mode;
{
  extern proc_t     * curproc;
  /* struct  plist { unsigned addr, len; } *plist; */
  ppageinfo_t *plist,*pp;
  int cc,spl;

#ifdef DEBUG
    oclog('wrt ', target, from, to, len, 0, 0);
    dtrace("ocwrt","enter",target,0);
    dbg_printf("[oc_write called netpa=%x,form=%x,to=%x,len=%x]",
                 target, from, to, len);
#endif

  if ((plist = (ppageinfo_t*)vtop_pagelock(curproc, from, len)) == NULL)
    return(DTU_EINVAL);

#ifdef DEBUG /* for new vtop_pagelock(K) */
  {
  ppageinfo_t *wp;
  wp = plist;
  for ( wp = plist; wp->pi_len !=0; wp++ )
    cmn_err(CE_CONT,"[%x %x]",wp->pi_addr,wp->pi_len);
  }
#endif /* DEBUG */

  pp = plist;

  for (; pp->pi_addr; to+=pp->pi_len, pp++) {
    spl = splhi(); /* 92/10/26 K.H */
    cc = dc_send((px_peid_t)target,
                 prior == DTU_HIGH,
                 mode == DTU_NOREPLY ? DC_MODE_NLDT : DC_MODE_ALDT,
                 DC_SEL_REAL,
                 pp->pi_addr,
                 pp->pi_len,
                 to,
                 mode == DTU_NOREPLY ? (char *)pp : NULL,
                 0);
#ifdef DEBUG
    dtrace("ocwrt","sendst",cc,0);
#endif
    if (mode == DTU_NOREPLY) {
#ifdef DEBUG
      dtrace("ocwrt","sleep",pp,0);
#endif
      sleep(pp, PZERO);
    }
    splx(spl); /* 92/10/26 K.H */
  
    if (cc < 0) {
      switch (cc) {
      case DC_CC_EINVAL:  vtop_pageunlock(curproc,from,len,plist);return(DTU_EINVAL);
      case DC_CC_ESNOMEM: vtop_pageunlock(curproc,from,len,plist);return(DTU_ENOMEM);
      case DC_CC_ERNOMEM: vtop_pageunlock(curproc,from,len,plist);return(DTU_ENOMEM);
      case DC_CC_ECOMM:   vtop_pageunlock(curproc,from,len,plist);return(DTU_ETMO);
      case DC_CC_RPDOWN:  vtop_pageunlock(curproc,from,len,plist);return(DTU_RPDOWN);
      case DC_CC_RPSWAP:  vtop_pageunlock(curproc,from,len,plist);return(DTU_RPSWAP);
      case DC_CC_SIGNAL:  vtop_pageunlock(curproc,from,len,plist);return(DTU_SIGNAL);
      case DC_CC_FORCEDOWN: vtop_pageunlock(curproc,from,len,plist);return(DTU_ENETPA); /* temporary errcode */
      /* case DC_CC_ENOFUNC: panic(); */
      /* default:            panic(); */
      default:              
        cmn_err(CE_PANIC,"oc_write: sending fail with %d.", cc);
        break;
      }
    }
  }
  vtop_pageunlock(curproc, from, len, plist);
#ifdef DEBUG
    dtrace("ocwrt","ok",0,0);
#endif
  return(0);
}


/*
 * oc_writel(K): write data sending with list.
 * [NOTE] 
 */
oc_writel(target, npage, plist, prior)
  int             target; /* not px_peid_t */
  int             npage;
  px_page_info_t* plist;
  unsigned        prior;
{
  int cc,i;

#ifdef DEBUG
    oclog('wrtl', target, npage);
    dtrace("wrtl","enter",target,0);
    dbg_printf("[oc_wrtl called target=%x, npage=%x]",
                target, npage);
#endif
  for (i=0; i<npage; i++, plist++) {
    cc = oc_write(target, plist->from, plist->to, plist->len, 0, prior, DC_MODE_ALDT);
    if (cc < 0) return(cc);
  }
  return(0);
}

/*
 * oc_write_msg(K): send message nosleep ( MAX = 16 bytes )
 * [NOTE] no-token, no-sleep
 */
oc_write_msg( target, msqid, mtype, msg, from, to, len, msglen, prior, mode )
    int       target; /* target netpa (not px_peid_t)  */
    int       msqid;  /* target queue ID               */
    int       mtype;  /* message type        (not use) */
    int     * msg;    /* message address               */
    int     * from;   /* sourece address     (not use) */
    paddr_t   to;     /* destination address (not use) */
    int       len;    /* write data length   (not use) */
    int       msglen; /* message length                */
    int       prior;  /* sending priority              */
    int       mode;   /* sending mode        (not use) */
{
  int cc;

#ifdef DEBUG
    oclog('wrtm', target, msqid, msg, msglen );
    dtrace("wrtm","enter",target,0);
    dbg_printf("[oc_wrtm called target=%x, qid=%x, msg=%x, msglen=%x]",
                target, msqid, msg, msglen );
#endif
  cc = dc_send((px_peid_t)target,
               prior == DTU_HIGH,
               DC_MODE_NMSG,
               msqid,
               msg,
               msglen,
               NULL,
               NULL,
               0);
#ifdef DEBUG
    dtrace("wrtm","sndst",cc,0);
#endif
  if (cc < 0) {
    switch (cc) {
    case DC_CC_EINVAL:    return(DTU_EINVAL);
    case DC_CC_ESNOMEM:   return(DTU_ENOMEM);
    case DC_CC_ERNOMEM:   return(DTU_ETMO);
    case DC_CC_ECOMM:     return(DTU_ETMO);
    case DC_CC_RPDOWN:    return(DTU_RPDOWN);
    case DC_CC_RPSWAP:    return(DTU_RPSWAP);
    case DC_CC_SIGNAL:    return(DTU_SIGNAL);
    case DC_CC_FORCEDOWN: return(DTU_ENETPA); /* temporary errcode */
    /* case DC_CC_ENOFUNC:   panic(); */
    /* default:              panic(); */
    default:              
        cmn_err(CE_PANIC,"oc_write_msg: sending fail with %d.", cc);
        break;
    }
  }
#ifdef DEBUG
    dtrace("wrtm","ok",0,0);
#endif
  return(0);
}

/*
 *  DTU buffer allocation. Allocated memory will never freed.
 */
caddr_t
dtu_allocb(npage)
    u_int  npage;
{
    caddr_t  addr;
    page_t * pp, * pp1;
    int mode;

    /*
     *  round up size to multiple of PAGESIZE
     *  and allocate physically contiguous pages.
     */
    if(servicing_interrupt()) {
        mode = P_NOSLEEP|P_PHYSCONTIG;
    } else {
        mode = P_CANWAIT|P_PHYSCONTIG;
    }
    if ((pp = page_get(PAGESIZE*npage,mode)) == 0) {
#ifdef DEBUG
         dbg_printf("[dtu_allocb: no more page.]");
         dtrace("allocb", "nopage", npage, 0);
#endif
         cmn_err(CE_PANIC, "!dtu_allocb: no more page.");
         return 0;
    }
#ifdef DEBUG
    dbg_printf("dtu_allocb: alloc %dpages, pp->p_free %x \n",
         npage, pp->p_free);
#endif
    /*
     *  Scan all pages and set page type.
     */
    pp1 = pp;
    do {
        pp1->p_type = P_SYS_PAGE;     /* mark system page */
        pp1 = pp1->p_next;
    } while (pp1 != pp);
    addr = (caddr_t)pfntokv(page_pptonum(pp));

#ifdef DEBUG
    dtrace("allocb", "alloc", addr, 0);
    {
        int i;
        i = splhi();
        oc_pages += (int)npage;
        splx(i);
    }
#endif
    return addr;     /* return address of top page */ 
}

void
dtu_freeb( vaddr )
    caddr_t  vaddr;
{
    page_t * pp, * pp1;

#ifdef DEBUG
    int i=0;
    dbg_printf("dtu_freeb: entered\n");
    dtrace("freeb", "free", vaddr, 0);
#endif

    pp = page_numtopp(kvtopfn(vaddr));
#ifdef DEBUG
    if(pp->p_free) {
        cmn_err(CE_WARN,"dtu_freeb: invalid free page request.");
        return;
    }
#endif
    do {
#ifdef DEBUG
        i++;
        dbg_printf("dtu_freeb: pp=%x\n", pp);
#endif
        pp1 = pp;
        pp1->p_keepcnt = 0;
        pp1->p_type = P_UNIX_PAGE;
        page_sub(&pp, pp1);
        page_free(pp1);
    } while (pp != NULL);

#ifdef DEBUG
    dbg_printf("dtu_freeb: free %d page topaddr 0x%x \n", i, vaddr);
    dtrace("freeb", "return", 0, 0);
    dbg_printf("dtu_freeb: return\n");
    {
        int j;
        j = splhi();
        oc_pages -= i;
        splx(j);
    }
#endif
    return;
}

/*
 * server for DTU test program
 * [NOTE] sleep with 1 page buffer, and get 1 page data.
 *        and test some case of message receive.
 */
dtuserver(msqid, n)
    int     msqid;     /* server's message queue */
    int     n;     /* server's buffer page num (n page)*/
{
    int       tokID;    /* receive token value */
    paddr_t   buffaddr; /* buffer real address for message get */
    caddr_t   bufftop;  /* buffer virtual address for message get */
    int       bufflen;  /* buffer length */

    int       src_netpa;/* source NETPA of receive message(!px_peid_t)*/
    int       rtype;    /* received message type */
    int       oc_stat;  /* oc_receive(K) return code */
    int     * msgp;     /* message pointer, use for message type check */
#ifdef DEBUG
    dtrace("dtuserver", "enter", msqid, 0);
#endif

#ifdef DEBUG     /* for TEST */
    dtrace("dtuerrd", "test1", 0, 0);
    oc_free_test1(); 
    dtrace("dtuerrd", "test2", 0, 0);
    oc_free_test2();
#endif

    if( (bufftop = (caddr_t)dtu_allocb(n)) == NULL ) {
        cmn_err(CE_PANIC,"dtuserver(%d): cannot get memory for receive.",msqid);
    }
    msgp = (int *)bufftop; 
    buffaddr = kvtophys(bufftop);
    bufflen = PAGESIZE;

#ifdef DEBUG
    cmn_err(CE_NOTE, "dtuserver: start now, get memory %d page from %x, qid = %d", n,buffaddr,msqid);
    dtrace("dtuserver","buffaddr",buffaddr,0);
#endif

    for(;;) {
        bufflen = PAGESIZE; /* 92/04/01 K.H */
        if ((oc_stat = oc_receive(msqid, 0, bufftop, bufflen,
                          &src_netpa, &tokID, &rtype, DTU_WAIT)) < 0)
              cmn_err(CE_NOTE,"dtuserver(%d): any error occurred in oc_receive() with %d\n", msqid, oc_stat);
        cmn_err(CE_NOTE,"dtuserver(%d): oc_recv() success, message type = %x length = %x ", msqid, *msgp, oc_stat);

#ifdef DEBUG
        switch( *msgp ) {
            case DTU_TEST_NORMAL:
                bufflen = 128;
                break;
            case DTU_TEST_NOREPL:
		*(msgp+1) = DTU_ASYNCREPL+255; /* not send reply */
                bufflen = 0;
                break;
            case DTU_TEST_TIMEOUT:
                bufflen = 128;
                break;
            case DTU_TEST_PANIC:
                cmn_err(CE_PANIC,"dtuserver(%d): get massage type = %x",msqid, *msgp);
                bufflen = 128;
                break;
            default:
                bufflen = 128;
                break;
        } /* end of switch */
        if( (*(msgp+1) == DTU_ASYNCREPL) && (oc_stat >= 0) ) {
            cmn_err(CE_NOTE,"dtuserver(%d): call oc_reply", msqid);
            oc_stat = oc_reply( src_netpa, tokID, bufftop, bufflen );
            if( oc_stat < 0 ) {
                cmn_err(CE_NOTE, "dtuserver(%d): oc_reply fail(%d), target token is %x on netpa %d", msqid, oc_stat, tokID, src_netpa );
            } else {
                cmn_err(CE_NOTE, "dtuserver(%d): oc_reply success(%d), target token is %x on netpa %d", msqid, oc_stat, tokID, src_netpa );
            }
        } else {
            cmn_err(CE_NOTE,"dtuserver(%d): not call oc_reply", msqid);
        }
#endif /* DEBUG */
    } /* server loop */
}
/* end of dtuserver() */

#ifdef SPMS

/*
 * spms alloc table: if empty setting [netpa].uid(side) = -1.
 */
struct spms_alloc_table {
        uid_t uid0;
        uid_t uid1;
} spms_tbl[256];

/*
 * for spms command support
 */
struct uap {
	int	cmd;
	int	arg1;
px_partition_t  *arg2;
};

struct uap2 {
        int     mtype;
	uid_t	uid;
px_partition_t  p;
};

spms_request(uap)
struct uap *uap;
{
	int	i, error=0, token;
	px_partition_t p;
	proc_t *pp;
	int spms_stat=-1;
	struct uap2 uap2;

	dtrace("spms_rq", "enter", 0, 0);
	copyin(uap->arg2, &p, sizeof(px_partition_t));
	pp = u.u_procp;
	uap2.uid = pp->p_cred->cr_ruid;
	bcopy( &p, &uap2.p, sizeof(px_partition_t) );
	switch(uap->arg1) {
	case 1:         /* set request    */
		dtrace("spms_rq", "set-b", 0, 0);
#if 1 
/* SPMS allocate check: if 0 allocate, -1 not allocate : 92/08/26 K.H */
                uap2.mtype = 202; /* 92/10/05 K.H */
		if( error = oc_sends( MASTER_NETPA(), /*   netpa     */
				      DTU_SPMS_MQ,    /*   msqid     */
				      202,            /*   mtype     */
				      &uap2,          /*   msg addr  */
				      sizeof(struct uap2), /* msglen */
				      &spms_stat,     /* result area */
				      4,              /* result len  */
				      DTU_BASE ) < 0 ) {
			dtrace("spms_rq", "return1", error, 0);
			return(error);
		} 
		if( spms_stat != 0 ) {
			dtrace("spms_rq", "return2", spms_stat, 0);
			return(EUSERS);
		} 
#endif
		/* spms_init(&p); */
		spms_init(&uap2);

#if 1		/* Changed by T.T 92/08/25 */
		u.u_procp->p_sessp->s_jobp->j_limit.exec_mode = EM_SHARE;
		u.u_procp->p_sessp->s_jobp->j_limit.vp_nbr = p.ptn_vpnbr;
		u.u_procp->p_sessp->s_jobp->j_alloc.exec_mode = EM_SHARE;
		u.u_procp->p_sessp->s_jobp->j_alloc.vp_nbr = p.ptn_vpnbr;
#endif
		if (p.ptn_vpnbr >1) {
			for (i=1; i<p.ptn_vpnbr; i++) {
				uap2.p.ptn_vpid = (px_vpid_t)i;
				uap2.mtype = 200; /* 92/10/05 K.H */
				if ((error = oc_send(&token,
					p.ptn_rpid[i]>>1, /* netpa  */
				        DTU_SPMS_MQ,      /* msqid  */
					200,              /* mtype  */
					&uap2,
					sizeof(struct uap2),
					0,
					DTU_NOREPLY)) < 0) {
			cmn_err(CE_NOTE, "spms_request: oc_send faild rc = 0x%x\n", error);
					i = 999; /* go out this loop*/
				}
			}
		}
		dtrace("spms_rq", "set-e", 0, 0);
		break;
	case 2:		/* clear request  */
		dtrace("spms_rq", "clear-b", 0, 0);
		if (p.ptn_vpnbr >1) {
			for (i=1; i<p.ptn_vpnbr; i++) {
				uap2.p.ptn_vpid = (px_vpid_t)i;
				uap2.mtype = 201; /* 92/10/05 K.H */
				if ((error = oc_send(&token,
					p.ptn_rpid[i]>>1,
				        DTU_SPMS_MQ,
					201,
					&uap2,
					sizeof(struct uap2),
					0,
					DTU_NOREPLY)) < 0) {
			cmn_err(CE_NOTE, "spms_request: oc_send faild rc = 0x%x\n", error);
					i = 999; /* go out this loop*/
				}
			}
		}
		if (!error)   	/* if no error  */
			/* spms_clear(&p); */
			spms_clear(&uap2);
		dtrace("spms_rq", "clear-e", 0, 0);
#if 1 
		/* SPMS allocate table clear: 92/08/26 K.H */
                uap2.mtype = 203; /* 92/10/05 K.H */
		if( error = oc_send(  &token,
				      MASTER_NETPA(),
				      DTU_SPMS_MQ,
				      203,
				      &uap2,
				      sizeof(struct uap2),
				      0,
				      DTU_NOREPLY ) < 0 ) {
			dtrace("spms_rq", "return3", error, 0);
			return(error);
		} 
#endif
		break;
	
	}
	dtrace("spms_rq", "return", error, 0);
	return(error);
}

spms_init(pp)
struct uap2 *pp;
{
	int	error, sidenum,  wait=1;
	px_partition_t	*addr;
	px_partition_t *p;

	p = (px_partition_t *) & pp->p ;

	dtrace("spmsint", "enter", p, 0);
	/* sidenum = (int)p->ptn_rpid[0] & 0x1; */
	sidenum = (int)p->ptn_rpid[(int)p->ptn_vpid] & 0x1;
	addr = ( px_partition_t* )kmem_alloc(sizeof(px_partition_t), KM_SLEEP);
	*addr = *p;
	pms_partition[sidenum] = addr;
#ifdef DEBUG
	dbg_printf("[spms_init(%d): partp(0x%x)]",sidenum,pms_partition[sidenum]);
#endif
		
	if ((error = px_dtu_set(sidenum)) !=0)
		cmn_err(CE_NOTE, "px_dtu_set faild rc = 0x%x\n", error);
	if (p->ptn_vpid == (p->ptn_vpnbr-1)) {	/* if last vpid	*/
		/* if ((error = px_bar_init(0, 4, 0, addr, wait)) !=0) */
		if ((error = px_bar_init(sidenum, BPU_DISCONN, 4, wait)) !=0)
			cmn_err(CE_NOTE, "px_bar_init faild rc = 0x%x\n", error);
	} else {
		/* if ((error = px_bar_init(0, 4, 0, addr, !wait)) !=0) */
		if ((error = px_bar_init(sidenum, BPU_DISCONN, 4, !wait)) !=0)
			cmn_err(CE_NOTE, "px_bar_init faild rc = 0x%x\n", error);
	}
	dtrace("spmsint", "return", 0, 0);
}

spms_clear(pp)
struct uap2 *pp;
{
	int	error, sidenum;
	px_partition_t *p;

	p = (px_partition_t *) & pp->p ;

	dtrace("spmsclr", "enter", p, 0);
	/* sidenum = (int)p->ptn_rpid[0] & 0x1; */
	sidenum = (int)p->ptn_rpid[(int)p->ptn_vpid] & 0x1;
#if 0 /* 92/06/12 K.H */
	if ((error = px_dtu_clear(sidenum)) !=0)
		cmn_err(CE_NOTE, "px_purge_xb faild rc = 0x%x\n", error);
	if ((error = px_bar_clear(sidenum)) !=0) 
		cmn_err(CE_NOTE, "px_bar_clear faild rc = 0x%x\n", error);
	if (p->ptn_vpid == 0 && (error = px_purge_xb(sidenum)) !=0)
		cmn_err(CE_NOTE, "px_purge_xb faild rc = 0x%x\n", error);
#endif /* 0 */
	if (pms_partition[sidenum])
		kmem_free(pms_partition[sidenum], sizeof(px_partition_t));
	pms_partition[sidenum] = 0;
	dtrace("spmsclr", "return", 0, 0);
}

int
spms_tbl_init()
{
	int opri,netpa;

	opri = splhi();

	for (netpa=0; netpa<256; netpa++) {
			spms_tbl[netpa].uid0 = -1;
			spms_tbl[netpa].uid1 = -1;
	}
	splx(opri);
}

int
spms_tblchk(pp)
struct uap2 *pp;
{
	int opri,i,netpa,sidenum;
	px_partition_t *p;

	opri = splhi();
	p = &(pp->p);

	for (i=0; i<p->ptn_vpnbr; i++) {
		netpa = p->ptn_rpid[i] >> 1;
		sidenum = p->ptn_rpid[i] & 1;
		if( sidenum == 0) { 
			if( spms_tbl[netpa].uid0 != -1 ) {
				splx(opri);
				return(EUSERS); /* alloc fail */
			}
		} else {
			if( spms_tbl[netpa].uid1 != -1 ) {
				splx(opri);
				return(EUSERS); /* alloc fail */
			}
		}
	}
	for (i=0; i<p->ptn_vpnbr; i++) {
		netpa = p->ptn_rpid[i] >> 1;
		sidenum = p->ptn_rpid[i] & 1;
		if( sidenum == 0) { 
			spms_tbl[netpa].uid0 = pp->uid;
		} else {
			spms_tbl[netpa].uid1 = pp->uid;
		}
	}
	splx(opri);
	return(0); /* if no error */
}

int
spms_tblclr(pp)
struct uap2 *pp;
{
	int opri,i,netpa,sidenum;
	px_partition_t *p;

	opri = splhi();
	p = &(pp->p);

	for (i=0; i<p->ptn_vpnbr; i++) {
		netpa = p->ptn_rpid[i] >> 1;
		sidenum = p->ptn_rpid[i] & 1;
		if( sidenum == 0) { 
			spms_tbl[netpa].uid0 = -1;
		} else {
			spms_tbl[netpa].uid1 = -1;
		}
	}
	splx(opri);
	return(0); /* if no error */
}

/*
 * receive requests by msgqid = DTU_SPMS_MQ
 */
spms_server(msqid)
    int   msqid;     /* server's message queue */
{
    int    token, rtype;
    int    netpa;    /* not px_peid_t */
    struct uap2 rcvbuff;
    struct uap2 *mp;
    int    err1;     /* status of oc_xxx */
    int    err2;     /* status of spms functions */

#ifdef DEBUG	/* 92/03/23 K.H */
    cmn_err(CE_NOTE, "spmsd(%d): start now, qid = %d", msqid,msqid);
    dtrace("spmsd","qid",msqid,0);
#endif

    mp = &rcvbuff;
    
    for(;;) {
        if ((err1 = oc_receive(msqid, 0, &rcvbuff, sizeof(struct uap2),
                            &netpa, &token, &rtype, DTU_WAIT)) < 0) {
            cmn_err(CE_NOTE,"spmsd: any error occurred in oc_receive rc=%d",err1);
        } else {
#ifdef DEBUG
		dtrace("spmsd","from",(int)netpa,0);
		dtrace("spmsd","mtype",(int)(mp->mtype),0);
#endif
                switch( mp->mtype ) {
                case 200: /* if dtu&bar init. request */
                        spms_init(&rcvbuff);
                        break;
                case 201: /* if dtu&bar clear request */
                        spms_clear(&rcvbuff);
                        break;
                case 202: /* spms alloc table check */
                        err2 = spms_tblchk(&rcvbuff);
                        if((err1=oc_reply(netpa, token, &err2, 4))< 0)
                                cmn_err(CE_NOTE,"spmsd: reply fail(%d) spms check table (%x)",err1,err2);
                        break;
                case 203: /* spms alloc table clear */
                        spms_tblclr(&rcvbuff);
                        break;
                case 204: /* display spms alloc table */
                        if( (err1 = oc_reply(netpa, token, spms_tbl, sizeof(spms_tbl))) < 0 )
                                cmn_err(CE_NOTE,"tstrcv: reply fail(%d) spms display table",err1);
                        break;
                }
        }
    }
}
/* end of spms command support */

#endif /* SPMS */ /* end of spms command support */

/*
 * server for console and etc.
 */
typedef struct tstrcv_ent{
    int    mtype; 
    int    (*func)();
    char * arg; 
    int    reserved;
};

#define  MAX_TSTRCV      20
/* #define  MAX_TSTRCV_MSG 256 */
#define  MAX_TSTRCV_MSG 512 /* change for spms commad support */

static int tstrcv_ctr=0;
struct tstrcv_ent tstrcv_en_t[MAX_TSTRCV];

tstrcv_server(msqid)
    int   msqid;     /* server's message queue */
{
    int   token, i, rtype;
    int   netpa;     /* not px_peid_t */
    char  msgbuf[MAX_TSTRCV_MSG];
    int   error;

#ifdef DEBUG	/* 92/03/23 K.H */
    cmn_err(CE_NOTE, "tstrcv(%d): start now, qid = %d", msqid,msqid);
    dtrace("tstrcv","qid",msqid,0);
#endif
    
    for(;;) {
        if ((i = oc_receive(msqid, 0, msgbuf, MAX_TSTRCV_MSG, &netpa, &token, &rtype, DTU_WAIT)) < 0) {
            cmn_err(CE_NOTE,"tstrcv: any error occurred in oc_receive rc=0x%x",i);
        } else {
		switch( rtype ) {
		case 205: /* PEpanic command support  */
			cmn_err(CE_PANIC, "PEpanic: received panic dump messag \n");
			break;
		default:
			{
			char *PEPAMSG = "pepamsg";
			char *MPAMSG  = "mpamsg";

			if (strcmp(msgbuf, PEPAMSG) == 0)
				pepa_receive(msgbuf, netpa, token);
			if (strcmp(msgbuf, MPAMSG) == 0)
				mpa_receive(msgbuf, netpa, token);
			}
			break;
		}
        }
    }
}

tstrcv_func_call(p, len, netpa, mtype)
char  *p;
int   len, netpa, mtype;
{
    int i, spl, rc=0;

    spl = splhi();
    for(i = 0; i < tstrcv_ctr; i++){
        if (tstrcv_en_t[i].mtype == mtype) {
            rc = (*tstrcv_en_t[i].func)(tstrcv_en_t[i].arg, netpa, p, len);
            break;
        }
    }
    
    if (i == tstrcv_ctr) rc = -1;
    splx(spl);
    return(rc);
}

tstrcv_entry(mtype, func, arg) 
int    mtype;
int    (*func)();
char * arg;
{
    int spl;

    spl = splhi();
    if (tstrcv_ctr == MAX_TSTRCV) {
        splx(spl);
        return(-1);
    }
    tstrcv_en_t[tstrcv_ctr].mtype = mtype;
    tstrcv_en_t[tstrcv_ctr].func  = func; 
    tstrcv_en_t[tstrcv_ctr].arg   = arg;  
    tstrcv_ctr++;
    splx(spl);     
    return(0);
}

tstrcv_delete(mtype)
int mtype;
{
    int i, spl, rc=0;

    spl = splhi();
    for(i = 0; i < tstrcv_ctr; i++) {
        if (tstrcv_en_t[i].mtype == mtype) {
            tstrcv_en_t[i] = tstrcv_en_t[tstrcv_ctr-1];
            tstrcv_ctr--;
            break;
        }
    }
    
    if (i == tstrcv_ctr) rc = -1;
    splx(spl);
    return(rc);
}
/* end of server for console and etc. */

#ifdef DEBUG
/* ========================================================================= */
/*   DTU Test Programs                                                       */
/* ========================================================================= */

/*
 * message send client, called from syskai(1000, 0, ..... );
 */
dtusend( p )
dtuparm_t *p;
{
    int     oc_stat; /* tc_XXXX status * -1 */
    int     spl;

    /* sending parameters */
    int     tokID;   /* sending token ID */
    caddr_t addr;    /* message address (virtual, 4byte align) */
    int     *msgp;   /* message pointer, use for message type check */
    unsigned mode;   /* sending mode */

dtrace("dtusend","enter",0,0);

    if( (addr=(caddr_t)dtu_allocb(1)) == NULL ) {
         cmn_err(CE_NOTE,"oc_send_test failed, cannot get sending page.");
         return( DTU_ENOMEM );
    }
    msgp = (int *)addr; 

    *msgp = p->mtype;
    /* *(msgp+1) = p->mode; */
    switch( p->mode ) {
        case 0:
            *(msgp+1) = DTU_NOREPLY;
            mode = DTU_NOREPLY;
            break;
        case 1:
            *(msgp+1) = DTU_ASYNCREPL;
            mode = DTU_ASYNCREPL;
            break;
        default:
            *(msgp+1) = DTU_NOREPLY;
            mode = DTU_NOREPLY;
            break;
    }

    cmn_err(CE_NOTE,"dtusend: sending message length=%d qid=%d", p->msglen, p->msqid);

    oc_stat = oc_send( &tokID, p->target, p->msqid, p->mtype, addr,
                       p->msglen, p->prior, mode );

dtrace("dtusend","ocsend",oc_stat,0);

    if( oc_stat != 0 ) {
         cmn_err(CE_NOTE,"dtusend: oc_send returns %d, Failed.", oc_stat);
         return( oc_stat );
    }

dtrace("dtusend","sndmode",p->mode,0);

    if( (p->mode) != 1 ) goto exit_dtusend;

    *msgp = 0;

    cmn_err(CE_NOTE,"dtusend: wait token is %x", tokID);
    oc_stat = oc_wait(tokID, addr, PAGESIZE);

dtrace("dtusend","ocwait",oc_stat,0);

    if( oc_stat != 0 ) {
         cmn_err(CE_NOTE,"dtusend: oc_wait returns %d, Failed.", oc_stat);
    } else {
         if( *msgp == p->mtype ) {
              cmn_err(CE_NOTE,"dtusend: oc_wait(K) success in status(%d)", oc_stat);
         } else {
              cmn_err(CE_NOTE,"dtusend: oc_wait(K) get unknown reply message");
         }
    }

exit_dtusend:

    dtu_freeb( addr );
dtrace("dtusend","return",0,0);
    return(0);
}

/*
 * rpc_test(K): set by dc_regist(K), and call oc_rpc(K)
 */
rpc_test( from_pe, mb )
px_peid_t from_pe;
struct mb *mb;
{
    int rpc_param[10];
    bcopy( (char*)(mb->msg), (char*)rpc_param, sizeof(int)*10 );
    cmn_err(CE_NOTE,"rpc_test(): from PE#%x params %d, %d, %d",
        from_pe, rpc_param[0], rpc_param[1], rpc_param[1]);
    if( IS_SA() ) {
       cmn_err(CE_NOTE, "rpc_test() called from netpa %d: I'm SA, netpa is %d",
           mb->srcnpa, MY_NETPA() );
    } else {
       cmn_err(CE_NOTE, "rpc_test() called from netpa %d: I'm PE, netpa is %d",
           mb->srcnpa, MY_NETPA() );
    }
}

/*
 * DTU RPC client, called from syskai(1000, ... ); TESTCODE = 1
 * [NOTE] TOTSEL is max of function regist.
 */
dturpc( p )
dtuparm_t *p;
{
    int oc_stat,dc_stat;
    int rpc_param[10];
    int rpc_res[100];
    int i;

    for( i=0; i<10; i++ ) rpc_param[i]=i;
    for( i=0; i<100; i++ ) rpc_res[i]=0;

dtrace("dturpc","enter",0,0);

    /* case 0: not setuped */
    oc_stat = oc_rpc( p->target, TOTSEL-1, 10*sizeof(int),
                      rpc_param, rpc_res, 100*sizeof(int), p->prior );
    cmn_err(CE_NOTE,"test1: oc_rpc returns %d.",oc_stat);

    /* case 1: invalid sel */
    dc_stat = dc_regist( rpc_test, TOTSEL+1 );
    cmn_err(CE_NOTE,"test2: dc_regist returns %d.",dc_stat);

    /* case 2: normal set and call */
    dc_stat = dc_regist( rpc_test, TOTSEL-1 );
    cmn_err(CE_NOTE,"test3: dc_regist returns %d.",dc_stat);

    oc_stat = oc_rpc( p->target, TOTSEL-1, 10*sizeof(int),
                      rpc_param, rpc_res, 100*sizeof(int), p->prior );
    cmn_err(CE_NOTE,"test4: oc_rpc returns %d.",oc_stat);

    /* case 3: secondary call */
    dc_stat = dc_regist( rpc_test, TOTSEL-1 );
    cmn_err(CE_NOTE,"test5: dc_regist returns %d.",oc_stat);

dtrace("dturpc","return",0,0);
    return(0);
}

/*
 * DTU READ client, called from syskai(1000, ... ); TESTCODE = 2
 * [NOTE] oc_sends(K) test
 */
dturead( p )
dtuparm_t *p;
{
    int data[4];
    int rcvbuff[32];
    int rc;

/* data init */
    data[0] = p->mtype;
    data[1] = DTU_ASYNCREPL;
    data[2] = 0x02;
    data[3] = 0x03;
    bzero( rcvbuff, 128 );

dtrace("dturead","enter",0,0);

dtrace("ocsends","call",0,0);
    rc = oc_sends( p->target, p->msqid, p->mtype, data, 16, rcvbuff, 128, DTU_BASE);
dtrace("ocsends","return",rc,0);

cmn_err(CE_NOTE,"oc_sends: reply recieve %x %x %x %x %x",
        rcvbuff[0], rcvbuff[1], rcvbuff[2], rcvbuff[3], rcvbuff[4]);

dtrace("dturead","return",0,0);
    return(0);
}

/*
 * DTU WRITE client, called from syskai(1000, ... ); TESTCODE = 3
 */
dtuwrite( p )
dtuparm_t *p;
{
#define WRITE_DATA_MAX 1024

    int oc_stat,i;
    int data[WRITE_DATA_MAX];
    caddr_t saddr;

dtrace("dtuwrite","enter",0,0);

/* Test of oc_write */
    for( i=0; i<WRITE_DATA_MAX; i++ ) data[i] = i;

dtrace("dtuwrite","oc_write1",1,0);
    oc_stat = oc_write( p->target, data, bootinfo->b_trash, p->msglen, 
    		0, p->prior,  DTU_NOREPLY);

    cmn_err(CE_NOTE,"oc_write returns %d. from %x to %x",
	    oc_stat, data, bootinfo->b_trash );

dtrace("dtuwrite","oc_write1",oc_stat,0);

/* Test of oc_write */
dtrace("dtuwrite","oc_write2",2,0);

    saddr = kmem_zalloc( 128, KM_NOSLEEP );
    if( saddr == NULL ) {
        dtrace("dtuwrite","oc_write2",-1,0);
        return(-1);
    } 

    bcopy( data, saddr, 128 );

    oc_stat = oc_write( p->target, saddr, (unsigned)( bootinfo->b_trash + 128),
                        128, 0, p->prior,  DTU_NOREPLY);

    cmn_err(CE_NOTE,"oc_write returns %d. from %x to %x",
            oc_stat, saddr, bootinfo->b_trash + 128 );

    kmem_free( saddr, 128 );
dtrace("dtuwrite","oc_write2",oc_stat,0);

/* Test of oc_write 92/03/18 */
dtrace("dtuwrite","oc_write3",1,0);
    /* get memory for destination */
    if( (saddr=(caddr_t)dtu_allocb(2)) == NULL ) {
         cmn_err(CE_NOTE,"oc_write_test failed, cannot get sending page.");
         return;
    }

    for( i=0; i<8; i++ ) bcopy( data, saddr + i*4096, 128 );

    oc_stat = oc_write( MY_NETPA(), saddr, kvtophys(saddr+PAGESIZE),
                        PAGESIZE, 0, p->prior,  DTU_NOREPLY);

    cmn_err(CE_NOTE,"oc_write returns %d. from %x to %x",
            oc_stat, saddr, saddr+PAGESIZE );

dtrace("dtuwrite","oc_write3",oc_stat,0);

    dtu_freeb( saddr );
dtrace("dtuwrite","return",0,0);
    return(0);
}

/*
 * DTU OC Test client, called from syskai(1000, ... ); TESTCODE = -1
 */
/* test of vtop_page_lock(K),vtop_pageunlock(K) */
oc_free_test1()
{
    extern proc_t     * curproc;
    caddr_t vaddr;
    ppageinfo_t *p;

dtrace("free1","alloc",0,0);
    vaddr = (caddr_t)dtu_allocb( 1 );

dtrace("free1","lock",0,0);
    p = vtop_pagelock(curproc, vaddr, PAGESIZE);
dtrace("free1","locked",0,0);

dtrace("free1","unlock",0,0);
    vtop_pageunlock(curproc, vaddr, PAGESIZE, p);
dtrace("free1","unlocked",0,0);
}

/* test of dtu_allocb(K),dtu_freeb(K) */
oc_free_test2()
{
    caddr_t a;
    extern int freemem;

dtrace("free2","alloc",freemem,0);

    a = (caddr_t)dtu_allocb( 2 );
dbg_printf("[free2: page_get]");

dtrace("free2","alloced",freemem,0);

dtrace("free2","free",0,0);
    dtu_freeb( a );
dbg_printf("[free2: page_free]");
dtrace("free2","freed",freemem,0);
}

dtutest( p )
dtuparm_t *p;
{
#define DTU_WRITEL_PAGE 3
    int rc;
    int data[4];
    int rcvbuff[32];
    
    int data2[1024];
    caddr_t addr, saddr;
    px_page_info_t plist[3];

dtrace("dtutest","enter",0,0);

/* oc_write_msg test */
    data[0] = 0xF0;
    data[1] = 0xF1;
    data[2] = 0xF2;
    data[3] = 0xF3;
    bzero( rcvbuff, 128 );
    for( rc=0; rc<1024; rc++ ) data2[rc]=rc;
    rc=0;

dtrace("oc_w_msg","call",0,0);

    rc = oc_write_msg( p->target, p->msqid, p->mtype, data, bootinfo->b_trash, bootinfo->b_trash, 0, 16, p->prior, DTU_NOREPLY);

dtrace("oc_w_msg","return",rc,0);

/* oc_writel test */

dtrace("oc_writel","call",0,0);

    /* get memory for destination */
    if( (addr=(caddr_t)dtu_allocb(DTU_WRITEL_PAGE)) == NULL ) {
         cmn_err(CE_NOTE,"oc_writel_test failed, cannot get sending page.");
         return;
    }

    /* get memory for source */
    saddr = kmem_zalloc( 4096, KM_NOSLEEP );
    if( saddr == NULL ) {
        dtrace("dtutest","oc_writel",DTU_ENOMEM,0);
        return(DTU_ENOMEM);
    } 

    bcopy( data2, saddr, 4096 );

    /* set px_page_info_t , but dirty source (^^; */
    plist[0].from = plist[1].from = plist[2].from = (int *)saddr;
    plist[0].to   = (u_int *)kvtophys( addr );
    plist[0].len  = 1024*4;
    plist[1].to   = (u_int *)kvtophys( addr + PAGESIZE );
    plist[1].len  = 1024*4;
    plist[2].to   = (u_int *)kvtophys( addr + PAGESIZE*2 );
    plist[2].len  = 1024*4;

    rc = oc_writel( p->target, DTU_WRITEL_PAGE, plist, p->prior );

    cmn_err(CE_NOTE,"oc_writel returns %d. from %x to %x %x %x", 
            rc, plist[0].from, plist[0].to, plist[1].to, plist[2].to );

dtrace("oc_writel","return",rc,0);

    dtu_freeb( addr );
    kmem_free( saddr, 4096 );

dtrace("dtutest","return",0,0);
    return(0);
}

int
oc_test( parm )
dtuparm_t *parm;
{
	int error = EINVAL;
	int i,oc_cast_msg[10];

#ifdef DEBUG
        cmn_err(CE_CONT, "SKAIDTU testcode = %d\n",parm->testcode);
#endif
	switch( parm->testcode ) {
		case -5:
			error = dc_sync((int)parm->target,
			parm->mode == DC_MODE_NOWAIT ? DC_MODE_NOWAIT : DC_MODE_WAIT );
			break;
		case -4:
			for( i=0; i<10; i++ ) oc_cast_msg[i]=0;
			oc_cast_msg[0]=255;
			oc_cast_msg[1]=-1;
			error = dc_broadcast(0,
				DTU_ETC_MQ,
				(char *)oc_cast_msg,
				sizeof(oc_cast_msg));
			break;
		case -3:
			error = dc_cancel((int)parm->target);
			break;
		case -2:
			error = dc_forcedown((px_peid_t)parm->target);
			break;
		case -1:
			error = dtutest(parm);
			break;
		case 0:
			error = dtusend(parm);
			break;
		case 1:
			error = dturpc(parm);
			break;
		case 2:
			error = dturead(parm);
			break;
		case 3:
			error = dtuwrite(parm);
			break;
		default:
			error = EINVAL;
		}
	return( error );
}

#define NOCLOG  8192
struct oclog {
        int     type;
        int     use;
        int     prms[6];
} oclogs[NOCLOG];
int     oclogcnt;

oclog (type, p0, p1, p2, p3, p4, p5)
{
        struct oclog * op;
        int     pri;

        pri  = splhi();
        op = &oclogs[oclogcnt];
        if (++oclogcnt >= NOCLOG)
                oclogcnt = 0;
        op->type = type;
        op->use++;
        op->prms[0] = p0;
        op->prms[1] = p1;
        op->prms[2] = p2;
        op->prms[3] = p3;
        op->prms[4] = p4;
        op->prms[5] = p5;
        (void) splx(pri);
}

#endif /* DEBUG */
