#ident  "@(#)udtu.c	1.3 1/6/93 16:08:26 - FUJITSU UXP/VPP"

/***********************************************/
/*  KAI user-DTU system call                   */
/*  Copyright (c) 1992 by FUJITSU LIMITED      */
/*   Mainframe division                        */
/*   Supercomputer development department      */
/*            Primary writer: H.Ueno           */
/*    ueno@ayame.mfd.cs.fujitsu.co.jp          */
/***********************************************/

#include	<util/types.h>
#include	<util/mcr.h>
#include	<util/mcr_field.h>
#include	<util/cmn_err.h>
#include	<util/vp_param.h>
#include	<mem/immu.h>
#include	<mem/kmem.h>
#include	<mem/as.h>
#include	<mem/seg.h>
#include	<mem/seg_vp.h>
#include	<mem/seg_gs.h>
#include	<mem/px_gs.h>		/* for gtop */
#include	<proc/user.h>
#include	<proc/proc.h>
#include	<proc/sched.h>
#include	<proc/signal.h>
#include	<proc/siginfo.h>
#include	<proc/pms/partition.h>
#include	<svc/todr.h>
#include	<svc/errno.h>
#include	<io/dtu/usr_dtu.h>
#include	<io/dtu/bar.h>
#include	<io/dtu/dc.h>
#include	<io/dtu/oc.h>
#include	<util/debug.h>
#include	<util/dl.h>		/* for 2word sub */

#define DTU_Q_BOUNDARY	0x00007fff
#define R2VADR( adr )   ((adr)+0xc0000000)

STATIC	int	num_of_side; 
STATIC  int	num_of_ntre_per_proc; 
STATIC  int	push_bodyl;	/* pushout packet body byte length */

STATIC udtu_side_info_t	*udtu_side_info;
STATIC uint_t		*ntr_image;		
STATIC udtu_side_info_t	*urp_head; /* first request block  */
STATIC udtu_side_info_t	*urp_tail; /* last request block   */

extern struct hqa  *hqa;
extern int	   dtu_ss;		/* dtu sub set flag */
extern bar_side_info_t	*bar_side_info;	/* bar_side_information */
extern paddr_t gtop();
extern struct mb	*rb0, *rb1;	/* #rqo */

#ifdef DEBUG
extern int     dtrace();
#endif

int  /* returns svc errno */        
init_udtu_bar()
{
    int		i,j;
    caddr_t	addr;
    uint_t	mcr;

#ifdef DEBUG
dtrace("udtubar","enter",0,0);
#endif

    urp_head = NULL;
    urp_tail = NULL;

    udtu_side_info = (udtu_side_info_t*)NULL;
    ntr_image = (uint_t*)NULL;    

    /*
     *	access of model dependent information
     */    
    addr = (caddr_t) *(unsigned*) R2VADR( 0x2038 );
    i    = *(unsigned*) R2VADR( addr+0x58 );
    j    = *(char*) R2VADR( addr+i+2 );  /* j=max_of_DNUPRO */
    num_of_side = 1 << (j-1) ;
#ifdef DEBUG
dtrace("udtubar","numside",num_of_side,0);
#endif
    push_bodyl = *(unsigned short*) R2VADR( addr+i );

    mcr_load( MCR_MTC, mcr );
    num_of_ntre_per_proc = 16 << ((mcr&0x00f00000)>>20);

    /*
     *	allocate side informations 
     */
    addr = kmem_zalloc( sizeof(udtu_side_info_t)*num_of_side
             + sizeof(bar_side_info_t)*num_of_side        
             + 4*num_of_ntre_per_proc*num_of_side, KM_SLEEP);
    if( addr == NULL ) {
        cmn_err(CE_NOTE, "<dtu:init_udtu_bar> not enough memory, fail to get UDTU/BARR control tables" );
        return( ENOMEM );
	}
    udtu_side_info = (udtu_side_info_t*) addr;
    bar_side_info  = (bar_side_info_t*) &( udtu_side_info[num_of_side] );
    ntr_image = (uint_t*) &( bar_side_info[num_of_side] );

return( 0 );  
}


/*
 *	udtu communication functions 
 */
int
udtu_msg_cancel( token )
    int		token;
{
    dc_cancel(token);
    return( 0 );
}



int
udtu_send(t_pe, prior, mode, sel, src, len, dst, key, time, retryc )
    px_peid_t	t_pe;
    unsigned	prior;
    unsigned	mode;
    unsigned	sel;
    char	*src;
    unsigned	len;
    char	*dst;
    char	*key;
    long	time;
    int		retryc;
{
    int		rc;
    int		snd_count;

#ifdef DEBUG
    dtrace("udtusnd","target",t_pe,0);
#endif
    snd_count = 0;
    do {
	++snd_count;
	rc = dc_send( t_pe, prior, mode, sel, src, len, dst, key, NULL );
	if( rc >= 0 ) {
	    /* do nothing */
	} else if( rc==DC_CC_ESNOMEM || rc==DC_CC_ERNOMEM 
                                     || rc==DC_CC_ECOMM ) {
	    delay( time );
	} else {
	    cmn_err( CE_PANIC, "<dtu:udtu_send> abnormal return code of dc_send." );
	}
     } while( rc<0 && snd_count <= retryc );
#ifdef DEBUG
    dtrace("udtusnd","return",rc,0);
#endif
    return( rc );
}
 

/*  call udtu_wait with mode = DC_MODE_WAIT  */
int
udtu_wait( mode, tokid, mb, time )
    unsigned	mode;
    int		tokid;
    struct mb	**mb;
    long	time;
{
    int		timeid;
    int		rc;		/* return code */

#ifdef DEBUG
    dtrace("udtuwat","tokid",tokid,0);
#endif
    timeid = timeout( udtu_msg_cancel, (caddr_t)tokid, time );
    rc = dc_wait( mode, tokid, mb );
    untimeout( timeid );
#ifdef DEBUG
    dtrace("udtuwat","return",rc,0);
#endif
    return( rc );
}


int
udtu_repl( t_pe, prior, tokid, src, len, time, retryc )
    px_peid_t	t_pe;
    unsigned	prior;
    int		tokid;
    char	*src;
    unsigned	len;
    long	time;		/* delay_func tick value */
    int		retryc;		/* maxmum retry count */
{
    int		snd_count;
    int		rc;

#ifdef DEBUG
    dtrace("udturpl","tokid",tokid,0);
#endif
    snd_count = 0;
    do {
	++snd_count;
	rc = dc_repl(t_pe, prior, DC_MODE_AMSG, tokid, src, len, NULL);
	if( rc >= 0 ) {
	    /* do nothing */
	} else if(rc==DC_CC_ERNOMEM || rc==DC_CC_ECOMM) {
	    delay( time );
	} else {
#ifdef DEBUG
	    cmn_err(CE_PANIC,
#else
	    cmn_err(CE_WARN,
#endif
            "<dtu:udtu_repl fail> netpa=%d token=%d rc=%d",
                 t_pe, tokid, rc );
	}
    } while( rc<0 && snd_count <= retryc );
    
#ifdef DEBUG
    dtrace("udturpl","return",rc,0);
#endif
    return( rc );
}


/*
 *	udtu system call entry
 */

typedef struct dtu_ctl_parm {
    int			svc_cmd;
    uint_t		trans;
    uint_t		reply;
    int			len;
    uint_t		priority;
    int			dtu_cmd;
} dtu_ctl_parm_t;

int    /* returns svc errno */
px_dtu_ctl_ent( uap )
    dtu_ctl_parm_t	*uap;
{  
    int                sidenum;
    dtu_user_list_t    ul;
    px_partition_t    *partition;

    if( !u.u_procp->p_parflag ) /* if not parallel process, error */
        return( EACCES );
    sidenum = u.u_procp->p_dupid; 
#ifdef DEBUG
	dtrace("dctlent","enter",sidenum,0);
#endif
    if( sidenum >= num_of_side ) {
        cmn_err(CE_WARN, "<dtu:px_dtuctl(2)> invalid p_dupid(%d), PMS error", sidenum);
        return( EACCES );	/* PMS error */
    }
    partition = pms_partition[sidenum];
    if( partition == NULL ) {
        cmn_err(CE_WARN, "<dtu:px_dtuctl(2)> pms_partition[%d]==NULL, PMS error", sidenum );
        return( EACCES );	/* PMS error */
    }
    ul.reply = uap->reply;
    ul.len   = uap->len;
    ul.priority = uap->priority;
    ul.cmd = uap->dtu_cmd;

    return( udtu_ctl_core( sidenum, &ul, uap->trans ) );
}


typedef struct dtu_ctl2_parm {
    int		     	svc_cmd;
    dtu_user_list_t  	*list;
    int		     	llen;
} dtu_ctl2_parm_t;    		

int
px_dtu_ctl2_ent( uap )          /* entry for caller pe */
    dtu_ctl2_parm_t	*uap;
{
    int			sidenum;
    int			llen;
    px_partition_t	*partition;
    dtu_user_list_t	*cl; 
    px_vpid_t		self_vpid,t_vpid;
    px_rpid_t		t_rpid; 
    int			vpnbr;
    int			i,j,k;
    boolean_t		io_err;
    uint_t		ui,uj,uk,ul;
    udtu_msg_t		sbody;
    udtu_msg_t		*rbody;
    struct mb		*mbp;
    int			*tokenl;
    int			error;
    char		copy_to[ DTU_HEADER_SIZE ];

    /*		prolog
     *
     */
    error = NULL;
    if( !u.u_procp->p_parflag ) /* if not parallel process, error */
        return( EACCES );
    sidenum = u.u_procp->p_dupid;
#ifdef DEBUG
    dtrace("dctl2ent","enter",sidenum,0);
#endif
    if( sidenum >= num_of_side ) {
        cmn_err(CE_WARN, "<dtu:px_dtuctl2(2)> invalid p_dupid(%d), PMS error",sidenum);
        return( EACCES );	/* PMS error */
    }
    partition = pms_partition[ sidenum ];
    if( partition == NULL ) {
        cmn_err(CE_WARN, "<dtu:px_dtuctl2(2)> pms_partition[%d]==NULL, PMS error", sidenum);
        return( EACCES );	/* PMS error */
    }
    self_vpid = partition -> ptn_vpid;
    vpnbr = partition -> ptn_vpnbr;

    llen = uap->llen;
    if( (llen > partition->ptn_vpnbr) || (llen == 0) )
        return( EINVAL );
    /*
     *		allocate workarea  and  copyin dtu_user_list 
     */
    cl= (dtu_user_list_t*)kmem_alloc( sizeof(dtu_user_list_t)*llen,KM_SLEEP);
    tokenl = (int *)kmem_zalloc(sizeof(int)*llen,KM_SLEEP );
    if( cl == NULL || tokenl == NULL ) {
        error = ENOMEM;  goto error_end;
    }
    if( copyin( (caddr_t)uap->list, (caddr_t)cl, sizeof(dtu_user_list_t)*llen)) {
        error = EFAULT;  goto error_end;
    }
    for( i=0; i<llen; ++i ) {
#ifdef DEBUG
        dtrace("dctl2ent","vpid",cl[i].vpnum,0);
#endif
        cl[i].rc = 0;
    }
    /*
     *		check user list
     */
    for( i=0; i < llen; ++i ) {
        t_vpid = (px_vpid_t) cl[ i ].vpnum;
        j = cl[ i ].cmd; 
        ui = cl[ i ].reply;
        k = cl[ i ].len;        
        if( t_vpid < 0 || t_vpid >= vpnbr || t_vpid == self_vpid 
           || (j & PX_DTU_POSITIVE_CMD)&&(j & PX_DTU_NEGATIVE_CMD)
           || ( j & PX_DTU_NONCTL2_CMD )
           || (k<PX_DTU_HQE256 || k>PX_DTU_HQE2048)&&(j & PX_DTU_QSIZE)
           || ( ui&VPAGEMASK != (ui+(0x00008000u<<k)-1)&VPAGEMASK )
               && ( j & PX_DTU_SET_RQ )                )         {
            error = EINVAL;  goto error_end;
        }
    }
    /*
     *		send packet asynchronously 
     */
#define	UDTU_RT1	10*100
#define	UDTU_RT2	10*100
	/* Udtu Retry Ticks #1,#2 = 10seconds                 */
	/* pending here.: determine the value of UDTU_RT1  */
    io_err = B_FALSE;
    for( i=0; i < llen; ++i )  {
        t_vpid = (px_vpid_t) cl[ i ].vpnum;
        t_rpid = partition -> ptn_rpid[ t_vpid ];
	sbody.msg_kind = UDTU_CTL2_MSG_TYPE;
        sbody.dest_sidenum = RPID2DUPID( t_rpid );
        sbody.caller_vpid =  self_vpid;
        sbody.ulist = cl[ i ];        /* struct copy */

	j = udtu_send( (px_peid_t)RPID2NETPA(t_rpid), 0, DC_MODE_TMSG, 
              DTU_DTUSVR_MQ, &sbody, sizeof(udtu_msg_t), 0, 0,
 	      (long)UDTU_RT1, 1 );
        if( j > 0 ) { 
	    tokenl[i] = j;
	} else {
            cl[i].rc = EIO;
            io_err = B_TRUE;
        }
    }
    /*
     *		wait reply packet
     */
    for( i=0; i<llen; ++i ) {
        if( cl[i].rc==0 ) {	
            if( udtu_wait(DC_MODE_WAIT, tokenl[i], &mbp, UDTU_RT2) ) {
                io_err = B_TRUE;
                cl[i].rc = EIO;
            }
            else {
                rbody = (udtu_msg_t*) &(mbp->msg[0]);
                if( (cl[i].rc = rbody->rc) != 0 )
                    io_err = B_TRUE;
	    	dc_mbfree( mbp );
            }
        }
    }
    /*
     *		write rc (return code) in caller memory
     */   
    if( io_err ) {
        for( i=0; i<llen && !error; ++i ) {
            if( copyout( (caddr_t)&cl[i].rc, (caddr_t)&(uap->list[i].rc),sizeof(int) ) )
                error = EFAULT;
        }
    }
    if( !error && io_err )
        error = EIO;

error_end:
    if( cl != NULL )
        kmem_free( cl, sizeof(dtu_user_list_t)*llen );
    if( tokenl != NULL )
        kmem_free( tokenl, sizeof(int)*llen );

#ifdef DEBUG
    dtrace("dctl2ent","err_ret",error,0);
#endif
    return( error );    
}

/*
 *  This funcition is invoked when DTU_DTUSRV_MQ server recieves
 *  px_dtu_ctl2 packet  (caller->callee packet).
 */  
int
rcv_dtu_ctl2( sbody, rbody, snd_netpa )
    udtu_msg_t		*sbody;
    udtu_msg_t		*rbody;
    px_peid_t		snd_netpa;
{
    int			sidenum;
    px_partition_t	*partition; 
    uint_t		s_vpid, ui,uj;

    sidenum = sbody->dest_sidenum;
#ifdef DEBUG
    dtrace("rcvctl2","enter",sidenum,0);
#endif
    if( sidenum < 0 || sidenum >= PX_PE_MULTIPLEXY ) {
        cmn_err(CE_NOTE, "<dtu:rcv_dtu_ctl2> invalid sbody->dest_sidenum");
        return( -1 );	/* error */
    }
    partition = pms_partition[ sidenum ];

    if( partition == NULL ) {
        cmn_err(CE_NOTE, "<dtu:rcv_dtu_ctl2> pms_partition[%d]==NULL",
            sidenum);
        return( -1 );
    }
    s_vpid = sbody->caller_vpid;
    if( s_vpid < 0 || s_vpid >= partition->ptn_vpnbr ) {
	    cmn_err(CE_NOTE, "<dtu:rcv_dtu_ctl2> invalid caller_vpid");
        return( -1 );	/* error */
    }
    if( RPID2NETPA(partition->ptn_rpid[s_vpid]) != snd_netpa ) {
        cmn_err(CE_NOTE, "<dtu:rcv_dtu_ctl2> invalid caller_netpa" );
        return( -1 );	/* error */
    }
    if( partition->ptn_vpid != sbody->ulist.vpnum ) {
        cmn_err(CE_NOTE, "<dtu:rcv_dtu_ctl2> vpid set by caller != vpid of callee");
        return( -1 );	/* error */
    }

    if( partition->ptn_procp != NULL ) {   /* if parallel process exists */
	rbody->rc = EBUSY;
    } else {
        rbody->rc = udtu_ctl_core( sidenum, &(sbody->ulist), NULL );
    }
    rbody->callee_vpid = sbody->ulist.vpnum;
#ifdef DEBUG
    dtrace("rcvctl2","return",rbody->rc,0);
#endif
    return;
}

rcv_udtu_rsig( us )
    udtu_side_info_t	*us;
{
    px_partition_t	*partition;
    int			side, j;
    px_rpid_t		rpid;
    int			dummy_t;

#ifdef DEBUG
dtrace("udtusig","enter",us->dest_side,0);
#endif
    side = us->dest_side;
    if( (partition = pms_partition[side]) == NULL ) {
 	/* invalid situation : record the error */
 	/* no partition on target netpa and side */
    }
    else if( partition->ptn_procp != NULL ) { 
	sigaddq( partition->ptn_procp, &(us->sigi), KM_SLEEP );
    }
    else if( partition->ptn_vpid == 0 ) { 
	/* invalid situation : record the error */
	/* vpid ==0 and process does not exists */
    }
    else {  /*  process does not exists on this netpa-side.     */
	    /*  send remote signal to vpid==0 			*/ 
	rpid = partition->ptn_rpid[0];
	us->dest_netpa = RPID2NETPA( rpid );
	us->dest_side  = RPID2DUPID( rpid );
	us->msg_kind   = UDTU_RSIG_MSG_TYPE;

#ifdef DEBUG
        dtrace("udtusig","master",us->dest_netpa,0);
#endif
	j = udtu_send( (px_peid_t)us->dest_netpa, 0, DC_MODE_AMSG,
		DTU_DTUSVR_MQ, us, sizeof(udtu_side_info_t), 0, 0,
		(long)UDTU_RT1, 1 );
#ifdef DEBUG
        dtrace("udtusig","send",j,0);
#endif
    }
}


void
dtusvr_mq_server()
{
    union msg_union {
		udtu_side_info_t	us;
		udtu_msg_t		sbody;
		udtu_msg_t		sb_pxb;
    } msg;

    udtu_msg_t		ctl2_rbody;
    udtu_msg_t		rb_pxb;
    int			rc;
    px_peid_t		from_pe;
    int			token;
    struct mb		*mbp;

    for(;;) {
	rc = dc_recv(DC_MODE_WAIT, DTU_DTUSVR_MQ, &from_pe,
			&token, &mbp );  
	if( rc < 0 )
	    cmn_err(CE_PANIC, 
               "<dtu:dtusvr_mq_server> dc_receieve_return_code < 0" );

#ifdef DEBUG
	cmn_err(CE_NOTE,"<dtusvr> from PE#%x tokid%x type%d",
            from_pe, token, msg.sbody.msg_kind);
        dtrace("dtusvr","mtype",msg.sbody.msg_kind,0);
#endif
	bcopy( (caddr_t)mbp->msg, (caddr_t)&msg, mbp->len );
	dc_mbfree( mbp );

        switch( msg.sbody.msg_kind ) {
	    case UDTU_CTL2_MSG_TYPE :
		rcv_dtu_ctl2( &(msg.sbody), &ctl2_rbody, from_pe );
		udtu_repl( from_pe, 0, token, &ctl2_rbody,
			sizeof(udtu_msg_t), (long)UDTU_RT1, 1 );
                break;
	    case UDTU_RSIG_MSG_TYPE :
		rcv_udtu_rsig( &msg.us );
		break;
	    case UDTU_PXB_MSG_TYPE :
		rb_pxb.rc = purge_xb_side( msg.sb_pxb.dest_sidenum );
		udtu_repl(from_pe, 0, token, &rb_pxb, 
		     sizeof(udtu_msg_t), (long)UDTU_RT1, 1 );
		break;
	    default :
		cmn_err(CE_PANIC,"<dtu:dtusvr_mq_server> invalid recieve msg type");
		break;
	}
    }
}

/*
 * Transfer Queue Address Check.
 * [NOTE] if queue is on vp-page or global-area, return B_TRUE.
 *        it's locked on memory even if process is not running.
 */
boolean_t
udtu_not_vppage( addr )
    caddr_t	addr;
{
    struct seg	*seg;

    seg = as_segat( u.u_procp -> p_as, addr );
    if( seg == NULL )
        return( B_TRUE );
    else if( IS_SEGVP(seg) || IS_SEGGS(seg) )
        return( B_FALSE );
    else
        return( B_TRUE );
}


int   		/* returns svc errno */
udtu_ctl_core( sidenum, pp, trans )
    int			sidenum;
    dtu_user_list_t	*pp;
    uint_t		trans;  /* trans q addr when status==LOCAL  */
{
    uint_t	ui,uj,uk;
    uint_t	cmd;
    uint_t	copy_to[ 128/sizeof(uint_t) ];
    uint_t	q_byte_size, hqsz;
    uint_t        mcr, mcr2;
    caddr_t	mcr_addr;
    uint_t	*ui_addr;
    int		i, vpnbr;
    px_partition_t  *partition;
    int		spl;
    uint_t      *st;
    boolean_t   stop_tq_f, stop_rq_f;

#ifdef DEBUG
dbg_printf("[udtu_ctl_core: %x %x %x ]",sidenum,pp,trans);
dtrace("dctlcore","enter",sidenum,0);
#endif
    /*
     *		check command code 
     */
    cmd = pp->cmd;
    st  = &(udtu_side_info[sidenum].status);
    if( ((cmd&PX_DTU_POSITIVE_CMD) && (cmd&PX_DTU_NEGATIVE_CMD))
        || (!(*st & UDTU_ST_INITIALIZED)
            &&(cmd & (PX_DTU_VALID_CMD )))
        || ((cmd & (PX_DTU_START_TQ | PX_DTU_STOP_TQ))
            && !(*st & UDTU_ST_TQ_SET ) && !(cmd & PX_DTU_SET_TQ))
        || ((cmd & (PX_DTU_START_RQ | PX_DTU_STOP_RQ))
            && !(*st & UDTU_ST_RQ_SET ) && !(cmd & PX_DTU_SET_RQ))
      ) {
#ifdef DEBUG
        dtrace("dctlcore","Badcmd",cmd,0);
#endif
        return( EINVAL );
	}

    /*
     *		reset queues when initialize, clear, set or reset
     */
    if( cmd & ( PX_DTU_SET_TQ | PX_DTU_SET_RQ | PX_DTU_RESET_TQ
				               | PX_DTU_RESET_RQ ) ) {
        spl = splhi();
#ifdef DEBUG
        dtrace("dctlcore","sndstop",0,0);
#endif
        dtu_snd_stop( DTU_STOP_SLEEP );
        if( cmd & ( PX_DTU_SET_TQ | PX_DTU_RESET_TQ)) {
            mcr_store( sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, 0 );
            mcr_store( sidenum==0 ? MCR_HWU0TW : MCR_HWU1TW, 0 );
        }
        if( cmd & ( PX_DTU_SET_RQ | PX_DTU_RESET_RQ ) ) {
            mcr_store( sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, 0 );
            mcr_store( sidenum==0 ? MCR_HWU0RW : MCR_HWU1RW, 0 );
        }
#ifdef DEBUG
        dtrace("dctlcore","restart",0,0);
#endif
        dtu_snd_restart();
        splx(spl);
    }

    if( cmd & PX_DTU_QSIZE ) {
#ifdef DEBUG
        dtrace("dctlcore","q-size",0,0);
#endif
        hqsz = pp->len;
        if( *st & UDTU_ST_RQ_SET || *st & UDTU_ST_TQ_SET
             || hqsz < PX_DTU_HQE256 || hqsz > PX_DTU_HQE2048 )
            return( EINVAL );
        
        spl = splhi();
        mcr_load( MCR_HQSZ, mcr );
        mcr &= (sidenum==0) ? 0xff00ffff : 0xffff00ff ;
        mcr |= hqsz << ( sidenum==0 ? 16 : 8 ); 
        mcr_store( MCR_HQSZ, mcr );
        splx(spl);
    }
    else {
        mcr_load( MCR_HQSZ, mcr );
        hqsz = ( mcr>>( sidenum==0 ? 16:8 ) ) & 3 ;
    }
    q_byte_size = (128u*256u)<<hqsz;
         

    if( cmd & PX_DTU_NOPROTECT ) {
#ifdef DEBUG
        dtrace("dctlcore","L-L",0,0);
#endif
        spl = splhi();
        mcr_load( MCR_SPIDW, mcr );
        mcr |= 0x80000000 >> sidenum;
        mcr_store( MCR_SPIDW, mcr );
        splx(spl);
    }

    if( cmd & PX_DTU_PROTECT ) {
#ifdef DEBUG
        dtrace("dctlcore","nonL-L",0,0);
#endif
        spl = splhi();
        mcr_load( MCR_SPIDW, mcr );
        mcr &= ( sidenum==0 ? 0x7fffffff : 0xbfffffff );
        mcr_store( MCR_SPIDW, mcr );
        splx( spl );
    }

    if( cmd & PX_DTU_PRIORITY ) {
#ifdef DEBUG
        dtrace("dctlcore","q-prio",0,0);
#endif
        spl = splhi();
        mcr_load( MCR_UHQPRI, mcr );
        mcr &= ( sidenum==0 ? 0x3fffffff : 0xcfffffff );
        mcr |= pp->priority >> (sidenum<<1);
        mcr_store( MCR_UHQPRI, mcr );
        splx(spl);
    }

    if( cmd & PX_DTU_SET_TQ ) {
#ifdef DEBUG
        dtrace("dctlcore","setTQ",trans,0);
#endif
        if( trans&DTU_Q_BOUNDARY ) return( EFAULT ); /* 92/08/31 */
        if( udtu_not_vppage( trans ) || udtu_not_vppage(trans+q_byte_size-1) )
            return( EFAULT );
        /**** pending:  lock vp-page here ****/
        if( sidenum== 0 )
            hqa->u0tqt = trans;
        else
            hqa->u1tqt = trans;
        *st |= UDTU_ST_TQ_SET;
    }

    if( cmd & PX_DTU_SET_RQ ) {
#ifdef DEBUG
        dtrace("dctlcore","setRQ",pp->reply,0);
#endif
        ui = pp->reply;
        if( ui&DTU_Q_BOUNDARY ) return( EFAULT ); /* 92/08/31 */
        if( ui & VPAGEMASK != (ui+q_byte_size-1) & VPAGEMASK )
            return( EFAULT );
        if( ( uj = (uint_t)gtop( (caddr_t)ui, sidenum ) ) == 0 ) /* KAI-201 */
            return( EFAULT );
        if( sidenum == 0 ) {
            hqa->u0rqt = uj;
            asm( "brcs %asam,." );
            mcr_store( MCR_HWU0RW, 0x00010000 );   /* is this right ???? */
        }
        else {
            hqa->u1rqt = uj;
            asm( "brcs %asam,.");
            mcr_store( MCR_HWU1RW, 0x00010000 );    /* is this right ???? */
        }
        *st |= UDTU_ST_RQ_SET;
    }

    if( cmd & PX_DTU_START_TQ ) {
#ifdef DEBUG
        dtrace("dctlcore","validTQ",0,0);
#endif
        if( !(*st & UDTU_ST_TQ_SET) )
            return( EINVAL );
        mcr_load( sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, mcr );
        if( (mcr & 0x00010000) == 0 )
            mcr_store( sidenum==0 ? MCR_HRU0TW:MCR_HRU1TW, mcr | 0x00010000 );
    }

    if( cmd & PX_DTU_START_RQ ) {
#ifdef DEBUG
        dtrace("dctlcore","validRQ",0,0);
#endif
        if( !(*st & UDTU_ST_RQ_SET ) )
            return( EINVAL );
        mcr_load( sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, mcr );
        if( (mcr & 0x00010000) == 0 )
            mcr_store( sidenum==0 ? MCR_HRU0RW:MCR_HRU1RW, mcr | 0x00010000 );
    }
    /*
     *		stop queues
     */
    stop_tq_f = B_FALSE;
    stop_rq_f = B_FALSE;
    if( cmd & PX_DTU_STOP_TQ ) {
#ifdef DEBUG
        dtrace("dctlcore","stopTQ",0,0);
#endif
        if( !(*st & UDTU_ST_TQ_SET ) )
            return( EINVAL );
        mcr_load( sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, mcr );
        if( (mcr & 0x00010000) != 0 )
            stop_tq_f = B_TRUE;
    }        
    if( cmd & PX_DTU_STOP_RQ ) {
#ifdef DEBUG
        dtrace("dctlcore","stopRQ",0,0);
#endif
        if( !(*st & UDTU_ST_RQ_SET ) )
            return( EINVAL );
        mcr_load( sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, mcr );
        if( (mcr & 0x00010000) != 0 )
            stop_rq_f = B_TRUE;
    }

    if( stop_tq_f || stop_rq_f ) {
        spl = splhi();  /* not necesarry, to shorten mover stop period */
        dtu_snd_stop( DTU_STOP_SLEEP );
        if( stop_tq_f ) {
            mcr_load(  sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, mcr );
            mcr &= 0xfffeffff;
            mcr_store( sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, mcr );
        }
        if( stop_rq_f ) {
            mcr_load(  sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, mcr );
            mcr &= 0xfffeffff;
            mcr_store( sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, mcr );
        }
        dtu_snd_restart();
        splx(spl);     /* not necesarry */
    }

    if( cmd & PX_DTU_RESET_TQ ) {
#ifdef DEBUG
        dtrace("dctlcore","resetTQ",0,0);
#endif
        if( sidenum == 0 )
            hqa->u0tqt = NULL;
        else
            hqa->u1tqt = NULL;
        *st &= ~UDTU_ST_TQ_SET;
    }

    if( cmd & PX_DTU_RESET_RQ ) {
#ifdef DEBUG
        dtrace("dctlcore","resetRQ",0,0);
#endif
        if( sidenum == 0 )
            hqa->u0rqt = NULL;
        else
            hqa->u1rqt = NULL;
        *st &= ~UDTU_ST_RQ_SET;
    }

#ifdef DEBUG
dtrace("dctlcore","return",0,0);
#endif
    return( 0 );
}

/*
 *	user dtu interruption handler 
 */
void
udtu_sndmrp( sndmri, smiphi )
    union mcr_form sndmri;
    union mcr_form smiphi;
{
    boolean_t   	siggen;
    int         	sidenum;
    int			spl;
    udtu_side_info_t	*us;
    px_partition_t	*partition;
    int			i,j;
    uint_t		mcr;
    udtu_side_info_t	*urpp;
    paddr_t		hdrp;

#ifdef DEBUG
dtrace("usndmrp","enter",0,0);
#endif
    sidenum = smiphi.smiphi.smiuid;
    /*
     * 		stop user queue 
     */
    spl = splhi();
    mcr_load(  sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, mcr );
    mcr_store( sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, mcr & 0xfffeffff );
    mcr_load(  sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, mcr );
    mcr_store( sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, mcr & 0xfffeffff );
    splx(spl);
    /*
     *		get udtu_side_info
     */
    partition = pms_partition[ sidenum ];
    us = &( udtu_side_info[ sidenum ] );
    spl = splhi();
    if( us->rupt_status == UDTU_RST_NULL ) {
        us->rupt_status = UDTU_RST_HOLD;
        siggen = B_TRUE;
    } else {
        siggen = B_FALSE;
    }
    splx(spl);
    /*
     *		make signal if required
     */
#ifdef DEBUG
    dtrace("usndmrp","siggen",siggen,0);
#endif
    if( siggen ) {
        us->sigi._data._iot.rupt_vpid = partition->ptn_vpid;
        us->sigi._data._iot.rupt_code = (sndmri.ful & 0x0fffffff) | 0x80000000;
        us->sigi.si_signo = SIGIOT;
        us->sigi.si_code  = UDTU_INT;  /* user dtu interruption */
        us->sigi.si_errno = EIO;       /* is this right ????    */
        if( sndmri.sndmri.phav ) {
#ifdef DEBUG
            dtrace("usndmrp","phav",1,0);
#endif
            if( smiphi.smiphi.smiqty == 0 ) {  /* user reply queue */ 
#ifdef DEBUG
                dtrace("usndmrp","ureply",smiphi.ful,0);
#endif
                bcopy( (caddr_t)R2VADR(smiphi.ful & 0xffffff80), 
                       (caddr_t) &(us->sigi._data._iot.hdr),
                       DTU_HW_HDR_SIZE );
            } else {  /* user transfer queue */    
#ifdef DEBUG
                dtrace("usndmrp","utrans",smiphi.ful,0);
#endif
                hdrp=vtop((caddr_t)(smiphi.ful & 0xffffff80),
                            partition->ptn_procp);
                if( hdrp != NULL )
                    bcopy( (caddr_t) phystokv(hdrp),
                           (caddr_t) &(us->sigi._data._iot.hdr), 
                           DTU_HW_HDR_SIZE );
            }
            i = us->sigi._data._iot.hdr[0] & 0x0000ffff;
            if( i >= partition->ptn_vpnbr ) 
                j = partition->ptn_rpid[ 0 ];
            else
                j = partition->ptn_rpid[i];
            us->dest_netpa = RPID2NETPA( j );
            us->dest_side  = RPID2DUPID( j );
        }
        else {   /* phav==0 */
#ifdef DEBUG
            dtrace("usndmrp","phav",0,0);
#endif
            for(j=0; j<=15; ++j) us->sigi._data._iot.hdr[j] = 0; 
            j = partition->ptn_rpid[ 0 ];
            us->dest_netpa = RPID2NETPA( j );
            us->dest_side  = RPID2DUPID( j );
        }

#ifdef DEBUG /* debug of generate signal information */
            dtrace("usndmrp","Srsig-n",(int)(us->dest_netpa),0);
            dtrace("usndmrp","Srsig-s",(int)(us->dest_side),0);
#endif

        if( partition->ptn_procp != NULL ) {/* process exists. make signal */
#ifdef DEBUG
            dtrace("usndmrp","procp1",partition->ptn_procp,0);
#endif
            sigaddq( partition->ptn_procp, &(us->sigi), KM_NOSLEEP );
            us->rupt_status = UDTU_RST_NULL;   /* free udtu_side_info */
        }
        else {               /* remote signal */
#ifdef DEBUG
            dtrace("usndmrp","procp0",0,0);
#endif
            us->rupt_status |= UDTU_RST_RSIG;	  
            us->req.link = NULL;
            us->req.cmd  = URPCMD_RSIG;   	/* remote signal*/
            us->req.i1   = sidenum;

            spl = splhi();
            urpp = urp_tail;
            if( urpp == NULL ) {
                urp_tail = urp_head = us;
#ifdef DEBUG
                dtrace("usndmrp","wakeup",&urp_head,0);
#endif
                wakeup( (caddr_t)&urp_head );
            }
            else {
                urpp->link = urp_tail = us;
            }
            splx(spl);
        }
    }
#ifdef DEBUG
    dtrace("usndmrp","return",0,0);
#endif
}

/*
 * 	user dtu interruption handling deamon
 */
void
udtu_rupt_dmn()
{
    udtu_side_info_t	*req;
    udtu_side_info_t	*usi;
    int			spl;
    int			sidenum, dest_netpa;
    int                 j;
    int			dummy_t;

    urp_tail = NULL;
    urp_head = NULL;     

#ifdef DEBUG
dtrace("ruptdmn","enter",0,0);
#endif
    for(;;) {
	spl = splhi();
	if( urp_head == NULL ) /* if no remote-signal-request */
	    sleep( (caddr_t)&urp_head, PZERO );
        req = urp_head;
	if( urp_head == urp_tail )
	    urp_head = urp_tail = NULL;
	else
	    urp_head = urp_head->link ;
   	splx(spl);

	switch( req->req.cmd ) {
	    case URPCMD_RSIG :
	        usi = (udtu_side_info_t *) req;
#if 0
	        j = oc_send( &dummy_t, usi->dest_netpa, DTU_DTUSVR_MQ,
	              UDTU_RSIG_MSG_TYPE, usi, sizeof(udtu_side_info_t),
	              DTU_BASE, DTU_NOREPLY );
#else /* 92/12/12 K.H */
                usi->msg_kind = UDTU_RSIG_MSG_TYPE;
	        j = udtu_send( (px_peid_t)usi->dest_netpa, 0,
	              DC_MODE_AMSG, DTU_DTUSVR_MQ, usi,
	              sizeof(udtu_side_info_t), 0, 0, UDTU_RT1, 1 );
#endif
	       usi->rupt_status = UDTU_RST_NULL;
	       break;
	    default:
		cmn_err(CE_PANIC,"<dtu:udtu_rupt_process>  invalid cmd ");
	}
    }
}

/*
 *	user dtu receive interruption handler 
 */
void
udtu_rcvmrp( mrprmi, rb, snd_netpa, snd_qid, dest_side )
    union mcr_form	mrprmi;
    struct mb		*rb;
    int			snd_netpa;
    int			snd_qid;
    int			dest_side;
{
    uint_t		sf;   /* side flag to process */
    union mcr_form	mcr, mcr2;
    int			spl;
    int			i,j;	
    udtu_side_info_t	*us;
    px_partition_t	*partition;
    udtu_side_info_t	*urpp;

#ifdef DEBUG
dtrace("urcvmrp","enter",0,0);
#endif
    /*
     *		stop user queues
     */
    sf = 0;
    if( dest_side == 0 || dest_side == -1 ) {
	mcr_load( MCR_HRU0TW, mcr.ful  );
	mcr_load( MCR_HRU0RW, mcr2.ful );
	if( mcr.hru0tw.sdvu0t || mcr2.hru0rw.sdvu0r ) 
	    sf |= 0x80000000;
    }
    if( dest_side == 1 || dest_side == -1 ) {
	mcr_load( MCR_HRU1TW, mcr.ful );
	mcr_load( MCR_HRU1RW, mcr2.ful );
	if( mcr.hru1tw.sdvu1t || mcr2.hru1rw.sdvu1r )
	    sf |= 0x40000000;
    }

    if( sf ) {
	spl = splhi();
	dtu_snd_stop( DTU_STOP_NOSLEEP );   /* nosleep */
	do {
	    mcr_load( MCR_MVST, mcr.ful );
	    mcr_load( MCR_MRPRMI, mcr2.ful );
	/* } while( mcr.mvst.recvst != 0 && mcr2.mrprmi.rcvsrp == 0 ); */
	} while( mcr.mvst.sendst != 0 && mcr2.mrprmi.sndsrp == 0 );
        if( dest_side == 0 || dest_side == -1 ) {
	    mcr_load( MCR_HRU0TW, mcr.ful );
	    mcr_load( MCR_HRU0RW, mcr2.ful );
	    mcr.hru0tw.sdvu0t = 0;
	    mcr2.hru0rw.sdvu0r = 0;
	    mcr_store( MCR_HRU0TW, mcr.ful );
	    mcr_store( MCR_HRU0RW, mcr2.ful );
	}
	if( dest_side == 1 || dest_side == -1 ) {
	    mcr_load( MCR_HRU1TW, mcr.ful );
	    mcr_load( MCR_HRU1RW, mcr2.ful );
	    mcr.hru1tw.sdvu1t = 0;
	    mcr2.hru1rw.sdvu1r = 0;
	    mcr_store( MCR_HRU1TW, mcr.ful );
	    mcr_store( MCR_HRU1RW, mcr2.ful );
 	}
	dtu_snd_restart();
	splx(spl);
    }

    /*
     * 		generate signal if required 
     */
    for( i=0; i < PX_PE_MULTIPLEXY; ++i ) {
  	sf = B_FALSE;
	us = &(udtu_side_info[ i ]);
        partition = pms_partition[ i ];
	if( (dest_side==i || dest_side==-1) && partition != NULL ) {
	    spl = splhi();
	    if( us->rupt_status == UDTU_RST_NULL ) {
		us->rupt_status = UDTU_RST_HOLD;
		sf = B_TRUE;
	    }
	    splx(spl);
	}
	if( sf ) {
            us->sigi._data._iot.rupt_vpid = partition->ptn_vpid;
	    us->sigi._data._iot.rupt_code = (mrprmi.ful & 0x2fffffff);
            us->sigi.si_signo = SIGIOT;
            us->sigi.si_code  = UDTU_INT;  /* user dtu interruption */
            us->sigi.si_errno = EIO;       /* is this right ????    */

	    if( mrprmi.mrprmi.rbhv == 1 ) {
		bcopy( (caddr_t)rb, (caddr_t) &(us->sigi._data._iot.hdr), DTU_HW_HDR_SIZE );
	    }
	    else {
	        for(j=0; j<=15; ++j) us->sigi._data._iot.hdr[j] = 0;
	    }

	    if( snd_netpa != -1 && snd_qid != -1 && dest_side != -1 ) {
		us->dest_netpa = snd_netpa;
	  	us->dest_side  = snd_qid & 3;
	    }
  	    else {
		j = partition->ptn_rpid[ 0 ];
		us->dest_netpa = RPID2NETPA( j );
		us->dest_side  = RPID2DUPID( j );
	    }			    
#ifdef DEBUG /* debug of generate signal information */
            dtrace("usndmrp","Rrsig-n",(int)(us->dest_netpa),0);
            dtrace("usndmrp","Rrsig-s",(int)(us->dest_side),0);
#endif

  	    if( partition->ptn_procp != NULL ) {/* process exists. make signal*/
#ifdef DEBUG
                dtrace("urcvmrp","procp1",partition->ptn_procp,0);
#endif
    	        sigaddq( partition->ptn_procp, &(us->sigi), KM_NOSLEEP );
	        us->rupt_status = UDTU_RST_NULL;   /* free udtu_side_info */
	    }
	    else {               /* remote signal */
#ifdef DEBUG
                dtrace("urcvmrp","procp0",0,0);
#endif
	        us->rupt_status |= UDTU_RST_RSIG;	  
                us->req.link = NULL;
	        us->req.cmd  = URPCMD_RSIG;   	/* remote signal*/
	        us->req.i1   = i;  /* sidenum */

                spl = splhi();
	        urpp = urp_tail;
		if( urpp == NULL ) {
		    urp_tail = urp_head = us;
		    wakeup( (caddr_t)&urp_head );
	        }
	        else {
		    urpp->link = urp_tail = us;
	        }
	        splx(spl);
	    }
	}
    }
}

/*	rqo process                                     */

#define HWHDSZ	64		/* hardware header size */

void	/* returns svd errno */
udtu_rqo( rbnum, sidenum, rb, rb2, rupt_state, gen_sigiot )
    int			rbnum;	    /* rb number */
    int			sidenum;    /* destination side */
    struct mb    	*rb, *rb2;  /* rb1st and rb2nd  */
    int 		*rupt_state;
    boolean_t		*gen_sigiot;

{
    uint_t	        mcr, mcr2, wp, rp, wp_first, mrprmi;
    uint_t		hqmask;
    struct mb   	*rq_real;  /* real address of rq */
    uint_t		wp_mcra, rp_mcra;
    udtu_side_info_t	*usi;
    boolean_t		rcv_stop_issue = B_FALSE;

    *rupt_state = RMR_NOCHANGE;   /* Recieve Mover Rupt nochange */
    *gen_sigiot = B_FALSE;

    if( enq_rqoq( ((caddr_t)rb)+HEDRSIZE, sidenum )) {
	*gen_sigiot = B_TRUE;
	return;  
    }

    mcr_load( MCR_HQSZ, mcr );
    if( sidenum == 0 ) {
	hqmask = (256 >> ((mcr & 0x00030000) >> 16)) - 1;  
	rq_real = (struct mb*) ((hqa -> u0rqt) & 0xffff8000 ); 
	wp_mcra = MCR_HWU0RW;
	rp_mcra = MCR_HRU0RW;
	usi = &udtu_side_info[0];
      } 
    else {
	hqmask = (256 >> ((mcr & 0x00000300) >> 8 )) - 1;
	rq_real = (struct mb*) ( (hqa -> u1rqt) & 0xffff8000 ) ; 
	wp_mcra = MCR_HWU1RW;
	rp_mcra = MCR_HRU1RW;	
	usi = &udtu_side_info[1];
    }

    /*	move rqo_hdr from os_rqoq to hw_rq as well as possible */

    mcr_load( wp_mcra, wp );
    mcr_load( rp_mcra, rp );
    wp &= hqmask;
    wp_first = wp;
    rp &= hqmask;
    while( ((wp+1)& hqmask) != rp 
             && usi->rqoql != 0 ) {
        /* copy rqo_header */ 
	bcopy( (caddr_t) &((usi->rqoq_tail)->rqohdr), 
		(caddr_t) R2VADR( &rq_real[wp] ), 
		HWHDSZ );

        /* Enqueue to hw_rq. Write_ptr is not changed here  */
	wp = (wp+1) & hqmask;
	mcr_load( rp_mcra, rp );
	rp &= hqmask;

	/* dequeue from os_rqoq */
	--(usi->rqoql);
        kmem_free( usi->rqoq_tail, sizeof( rqoqe_t ) );
        if( usi->rqoql == 0 ) {
	    usi->rqoq_tail = usi->rqoq_head = NULL;
	}
	else {
	    usi->rqoq_tail = (usi->rqoq_tail)->link;
	}
    }
    if( wp_first != wp ) {   /* if write_ptr is changed */
	mcr_load( wp_mcra, mcr );
	mcr = (mcr & 0xffff0000) | (wp & hqmask );
	mcr_store( wp_mcra, mcr );
    }

    /* Process second recieve_rupt when os_rqoq is empty */

    if( usi->rqoql == 0 ) {
	/* stop dtu_recieve without sleep */

	dtu_rcv_stop( DTU_STOP_NOSLEEP );
	asm( "brcs %asam,." );
	rcv_stop_issue = B_TRUE;
	mcr_store( MCR_MRPRMI, 0x20000000 ); /* erase rcv_rupt */
	asm( "brcs %asam,." );
	do {	/* wait until stop */
	    mcr_load( MCR_MRPRMI, mcr );
	    mcr_load( MCR_MVST, mcr2 );
	} while( (mcr & 0x10000000 )==0       
		  && (mcr2 & 0x30000000)!=0 );
	    /* not rcv_stop_rupt_pend  &&  not_rcv_stop_state */
	    /* dtu stops recieving here */
	/* supply rb for erased rcv_rupt */
        mcr_load(  rbnum==0 ? MCR_RBAR0:MCR_RBAR1, mcr );
	mcr_store( rbnum==0 ? MCR_RBAR0:MCR_RBAR1, mcr|1);

	/* if 2nd_rcvrupt is available_rqo, erase and process it */
	mcr_load( MCR_MRPRMI, mrprmi );
	if( (mrprmi&0x207fe1ff)==0x20000003 ) {
	    /*  erase rupt and supply rb related to the rupt  */ 
    	    mcr_store( MCR_MRPRMI, 0x20000000 );
            mcr_load(  rbnum==0 ? MCR_RBAR1:MCR_RBAR0, mcr );
	    mcr_store( rbnum==0 ? MCR_RBAR1:MCR_RBAR0, mcr|1 );
	    *rupt_state = RMR_ERASED;	    

	    mcr_load( rp_mcra, rp );
	    rp &= hqmask;
	    if( ((wp+1)&hqmask) != rp ) {  /* if not queue_full */
		/* enqueue to hw_rq. */
		bcopy( (caddr_t)rb2, (caddr_t)R2VADR(&rq_real[wp]), HWHDSZ );
		wp = (wp+1) & hqmask;
		mcr_load( wp_mcra, mcr );
 		mcr_store( wp_mcra, (mcr&0xffff0000)| wp );
	    }
	    else {  /* hw_rq is full */		
		/* enqueue to os_rqoq */
		if( enq_rqoq( ((caddr_t)rb)+HEDRSIZE,
		              sidenum )) {
                     *gen_sigiot = B_TRUE;
		}	    
	    }
	} 
        else if( (mrprmi&0x20000000) == 0 ) {
	    *rupt_state = RMR_ERASED;
	}
   	else {
	    *rupt_state = RMR_CHANGED;
	}
    }

    /* if os_rqoq is empty,  stop rqo 
     * else  spin_wait according to os_rqoq length 
     */
    if( !*gen_sigiot ) {
        if( usi->rqoql == 0 ) { 
	    mcr_load( wp_mcra, mcr );
	    mcr &= 0xfffdffff;	/* erase overflow flag */
	    mcr_store( wp_mcra, mcr );
        }
        else {  /* os_rqoq is not empty */
	    struct todr	first_time, current_time, diff_time,
	  	    	wait_time;
#ifdef DEBUG
	    dl_t	diffs;
#endif

            if( usi->rqoql < 5 ) {
	        wait_time.tod[0] = 0;
#ifndef SIM
	        wait_time.tod[1] = 0x10000000;
#else
	        wait_time.tod[1] = 0x00001000;
#endif
	    }
	    else if( usi->rqoql < 10 ) {
	        wait_time.tod[0] = 0;
#ifndef SIM
	        wait_time.tod[1] = 0x40000000;
#else
	        wait_time.tod[1] = 0x00004000;
#endif
	    }
	    else {
	        wait_time.tod[0] = 0;
#ifndef SIM
	        wait_time.tod[1] = 0x80000000;
#else
	        wait_time.tod[1] = 0x00008000;
#endif
	}

        /*
         *	wait according to os_rqoq length
	 */
            rtodc( &first_time );
	    do {
	        rtodc( &current_time );
	        subl8y( &current_time, &first_time, &diff_time );
#ifdef DEBUG
	        diffs = lsub( current_time, first_time );
	        dbg_printf("[udtu_rqo():%x %x]",diffs.dl_hop,diffs.dl_lop);
#endif
	    } while( (diff_time.tod[0]<wait_time.tod[0])
		    || ( (diff_time.tod[0]==wait_time.tod[0])
		        &&(diff_time.tod[1]<wait_time.tod[1]) ) );
        }
    } /*  end of if( !*gen_sigiot )  */

    if( rcv_stop_issue ) {
	dtu_rcv_restart();
    }
    return;
}


int	/* returns svc errno */
enq_rqoq( hdr, sidenum )
    uint_t	hdr[];
    uint_t	sidenum;
{
    int		i;
    rqoqe_t	*rqoqe;
    udtu_side_info_t	*usi;

    usi = &udtu_side_info[ sidenum ];
    rqoqe = ( rqoqe_t* ) kmem_alloc( sizeof(rqoqe_t), KM_NOSLEEP );
    if( rqoqe == NULL ) {
	/* Make signal SIGIOT to the user program to abend.    */ 
        /* However, other part of os may die by no_memory.     */
	return( ENOMEM );
    }

    bcopy( (caddr_t)hdr, (caddr_t) rqoqe->rqohdr, HWHDSZ );	/* copy header */
    rqoqe->link = NULL;
    if( usi->rqoql == 0 ) {  
	usi->rqoq_head = usi->rqoq_tail = rqoqe;
    }
    else {
	(usi->rqoq_head)->link = rqoqe;  /* link to forward */
	usi->rqoq_head = rqoqe;
    }
    ++(usi->rqoql);  /* ++rqoq_length */ 
    return( 0 );
}

/*
 *	functions required by pms ( Kishimoto dai-sensei )
 */
int
px_dtu_set( sidenum )
    int		sidenum;
{
    px_partition_t	*partition;
    int			mcr;
    px_rpid_t		rpid;
    short		vpnbr;
    caddr_t		mcr_addr;
    uint_t		*ui_addr;
    uint_t		ui,uj,uk;
    int			spl;
    int			side,i;

#ifdef DEBUG
dtrace("dtuset","enter",sidenum,0);
#endif
    dtu_snd_stop( DTU_STOP_SLEEP );
    mcr_store( sidenum==0 ? MCR_HRU0TW : MCR_HRU1TW, 0 );
    mcr_store( sidenum==0 ? MCR_HWU0TW : MCR_HWU1TW, 0 );
    mcr_store( sidenum==0 ? MCR_HRU0RW : MCR_HRU1RW, 0 );
    mcr_store( sidenum==0 ? MCR_HWU0RW : MCR_HWU1RW, 0 );
    dtu_snd_restart();

    spl = splhi();
    mcr_load( MCR_HQSZ, mcr );
    mcr &=  (sidenum==0) ? 0xff00ffff : 0xffff00ff;
    mcr_store( MCR_HQSZ, mcr );
    mcr_load( MCR_SPIDW, mcr );
    mcr &= (sidenum==0) ? 0x7fffffff : 0xbfffffff ;
    mcr_store( MCR_SPIDW, mcr );
    mcr_load( MCR_UHQPRI, mcr );
    mcr &= (sidenum==0) ? 0x3fffffff : 0xcfffffff;
    mcr_store( MCR_UHQPRI, mcr );
    asm("brcs %asam,.");
    splx(spl);
    /*
    *		initialize NTREs
    */
#ifdef DEBUG
dtrace("dtuset","ntre",num_of_ntre_per_proc,0);
#endif
    if( ( partition = pms_partition[ sidenum ] ) == NULL )
	return( -1 );
    vpnbr = (int) partition->ptn_vpnbr; 
    mcr_addr = (caddr_t)(MCR_NTRE000+((num_of_ntre_per_proc<<3)*sidenum));
    ui_addr  = ntr_image + num_of_ntre_per_proc*sidenum ;
#ifdef DEBUG
    dbg_printf("[px_dtu_set: partp(%x) NTRimage(%x)]",partition, ui_addr);
    dtrace("dtuset","ntrimgp",ui_addr,0);
#endif
#ifdef DEBUG
    cmn_err(CE_CONT, "[px_dtu_set(%d): setting partition (vpid,netpa)={",sidenum);
#endif
    for( i=0; i < num_of_ntre_per_proc; ++i, mcr_addr+=8, ++ui_addr ) {
        if( i >= vpnbr ) {
            mcr_store( mcr_addr, 0x00008000 );      /* page invalid */
            *ui_addr = 0x00008000u;		    /* page invalid */
        }
        else {
            ui = RPID2NETPA( partition->ptn_rpid[i] );
            uj = RPID2DUPID( partition->ptn_rpid[i] );
            uk = (ui<<16) | uj;
            mcr_store( mcr_addr, uk );
            *ui_addr = uk;
#ifdef DEBUG
	        cmn_err(CE_CONT, "(%d,%d)",i,ui);
#endif
        }
    }
#ifdef DEBUG
    cmn_err(CE_CONT, "}]\n");
#endif
    udtu_side_info[ sidenum ].status = UDTU_ST_INITIALIZED;
    udtu_side_info[ sidenum ].rupt_status = UDTU_RST_NULL;
#ifdef DEBUG
dtrace("dtuset","return",0,0);
#endif
    return( 0 );
}

int
purge_xb_side( sidenum )
    int		sidenum;
{
    int			spl, i ;
    caddr_t		mcr_addr;
    uint_t		*ui_addr;
    dtu_user_list_t	ul;
    int			error;
    int			mcr; 
    px_partition_t	*partition;
    udtu_side_info_t	*usi;

#ifdef DEBUG
dbg_printf("[p_xb_side: sidenum(%d)]",sidenum);
dtrace("p_xb_sd","enter",sidenum,0);
#endif
    usi = &udtu_side_info[ sidenum ];

    /*
     *		clear queues 
     */
    ul.cmd = PX_DTU_RESET_TQ | PX_DTU_RESET_RQ;
    if( error = udtu_ctl_core( sidenum, &ul, NULL ) ) {
#ifdef DEBUG
        dtrace("p_xb_sd","return1",-1,0);
#endif
	    return( -1 );
    }

   /*
    * 	reset ntres 
    */
#ifdef DEBUG
dtrace("p_xb_sd","ntr_clr",num_of_ntre_per_proc,0);
#endif
    mcr_addr = (caddr_t)(MCR_NTRE000+((num_of_ntre_per_proc<<3)*sidenum));
    ui_addr  = ntr_image + num_of_ntre_per_proc*sidenum ;
    for( i=0; i<num_of_ntre_per_proc; ++i, mcr_addr+=8, ++ui_addr ) {
	mcr_store( mcr_addr, 0x00008000 );
	*ui_addr = 0x00008000u;
    }
    spl = splhi();
    mcr_load( MCR_HQSZ, mcr );
    mcr &= ( sidenum==0 ? 0xff00ffff : 0xffff00ff );
    mcr_store( MCR_HQSZ, mcr );
    mcr_load( MCR_SPIDW, mcr );
    mcr &= (sidenum==0) ? 0x7fffffff : 0xbfffffff ;
    mcr_store( MCR_SPIDW, mcr );
    mcr_load( MCR_UHQPRI, mcr );
    mcr &= (sidenum==0) ? 0x3fffffff : 0xcfffffff ;
    mcr_store( MCR_UHQPRI, mcr );
    splx(spl);


    /*
     *		issue push out packet
     */
    #define	PUSH_MAX_PE	256

    if( (partition = pms_partition[ sidenum ]) == NULL ) {
	/* os error : take log of the error */
    }  
    else {
	int		vpnbr,vp,rp;
	int		netpa;

	vpnbr = partition -> ptn_vpnbr;
	for( vp = 0; vp < vpnbr ; ++vp ) {
	    netpa = RPID2NETPA( partition -> ptn_rpid[vp] );
	    if( vp == vpnbr-1 ) {   /** if last packet **/
		spl = splhi();
		if( udtu_send((px_peid_t)netpa, 0, DC_MODE_PUSH, 0, 0, 
		       push_bodyl, 0, &udtu_side_info[sidenum],
		       UDTU_RT1, 1 ) >=0                        ) {
		    sleep( (caddr_t) &udtu_side_info[sidenum], PZERO );
		}
		splx(spl);
	    } else {    /** not last packet **/
		udtu_send( (px_peid_t)netpa, 0, DC_MODE_PUSH, 0, 0,
			     push_bodyl, 0, 0, UDTU_RT1, 1  );
	    }
	}
    }
    /*
     * free rqoq 
     */
    while( usi->rqoql > 0 ) {
	--(usi->rqoql);
	kmem_free( usi->rqoq_tail, sizeof( rqoqe_t ) );
	if( usi->rqoql == 0 ) {
	    usi->rqoq_tail = usi->rqoq_head = NULL;
	} else {
	    usi->rqoq_tail = (usi->rqoq_tail)->link;
	}
    }

    usi->status  = UDTU_ST_FREE;
    usi->rupt_status = UDTU_RST_NULL;
#ifdef DEBUG
dtrace("p_xb_sd","return",0,0);
#endif
    return( 0 );
}


int
px_purge_xb( sidenum )
    px_sidenum_t	sidenum;
{
    px_partition_t	*partition; 
    int			vpnbr,vp,rp, self_vp;
    int			rc,r;
    udtu_msg_t		sb_pxb;
    udtu_msg_t		rb_pxb;
    int			*tokenl;
    struct mb		*mbp;

#ifdef DEBUG
dbg_printf("[px_purge_xb: sidenum(%d)",sidenum);
dtrace("purgexb","enter",sidenum,0);
#endif
    rc = 0;
    if( (partition = pms_partition[ sidenum ] ) == NULL ) {
	/* os error. take error log */
	return( -1 );
    }
    vpnbr = partition -> ptn_vpnbr;
    self_vp = partition -> ptn_vpid;
    tokenl = (int *)kmem_alloc(sizeof(int)*vpnbr,KM_SLEEP );
    if( tokenl == NULL ) {
        return( -1 );   /* enomen */
    }
    /*
     *	issue rpc packets
     */
    for( vp=0; vp < vpnbr; ++vp ) {
        if( vp == self_vp ) {
	    tokenl[self_vp] = NULL;
            if( purge_xb_side( sidenum ) )
                rc = -1;
        }
        else {
            rp = partition -> ptn_rpid[ vp ];
	    sb_pxb.msg_kind = UDTU_PXB_MSG_TYPE;
            sb_pxb.dest_sidenum = RPID2DUPID( rp );
	    r = udtu_send( (px_peid_t)RPID2NETPA(rp), 0, DC_MODE_TMSG, 
			DTU_DTUSVR_MQ, &sb_pxb, sizeof(udtu_msg_t),
			0, 0, UDTU_RT1, 1 );
	    tokenl[vp] = r;
            if( r ) {
                rc = -1;
            }
        }
    }
    /*
     *        wait rpc result
     */
    for( vp=0; vp < vpnbr; ++vp ) {
        if( tokenl[vp] > 0 ) {
	    if( udtu_wait(DC_MODE_WAIT, tokenl[vp], &mbp, UDTU_RT2) ){
                /* take error log of os */
                rc = -1;
            } else {
                /* normal end */
                dc_mbfree( mbp );
            }
        }
    }
    
    kmem_free( tokenl, sizeof(int)*vpnbr );
#ifdef DEBUG
dtrace("purgexb","return",rc,0);
#endif
    return( rc );
}


int
px_dtu_disp( sidenum )
    int		sidenum;
{
    union mcr_form	mcr;
    int			spl;

#ifdef DEBUG
/* dtrace("dtudisp","enter",sidenum,0); */
#endif
    spl = splhi();
    mcr_load(  MCR_SPIDW, mcr.ful );
    mcr.spidw.spid = sidenum;
    mcr_store( MCR_SPIDW, mcr.ful );
    splx(spl);
 
#ifdef DEBUG
/* dtrace("dtudisp","return",0,0); */
#endif
    return( 0 );
}

/*
 * followings function are not reviewed, appended by K.H
 */
int
px_dtu_clear( sidenum )
int sidenum;
{
        /* reset user queue ? */
        return( 0 );
}
