#ident  "@(#)usr_dtu.h	1.3 1/6/93 16:08:54 - FUJITSU UXP/VPP"

/***********************************************/
/*  usr_dtu.h                                  */
/*  Copyright (c) 1992 by FUJITSU LIMITED      */
/*   Mainframe division                        */
/*   Supercomputer development department      */
/*            Primary writer: H.Ueno           */
/*    ueno@ayame.mfd.cs.fujitsu.co.jp          */
/***********************************************/

#ifdef	_KERNEL

#ifndef _UTIL_TYPES_H
#include <proc/siginfo.h>
#endif

#ifndef _PROC_TYPES_H
#include <util/types.h>
#endif

#ifndef _PROC_SIGINFO_H
#include <proc/siginfo.h>
#endif

#endif /* _KERNEL */

#ifndef _IO_DTU_USR_DTU_H
#define _IO_DTU_USR_DTU_H

/*
 *	null parameter 
 */
#define	PX_DTU_NOP		0

/*
 *	cmd of px_dtu_cntl
*/
#define PX_DTU_QSIZE        0x80000000  /* change q size            */
#define	PX_DTU_NOPROTECT    0x40000000  /* unprotect L-L transfer   */
#define PX_DTU_PROTECT      0x20000000  /* protect L-L transfer     */
#define PX_DTU_PRIORITY     0x10000000  /* change priority          */
#define	PX_DTU_SET_TQ       0x08000000  /* initialize transfer q    */
#define PX_DTU_SET_RQ       0x04000000  /* initialize recieve  q    */
#define	PX_DTU_START_TQ     0x02000000  /* start transfer q         */
#define PX_DTU_START_RQ     0x01000000  /* start reply q            */
#define	PX_DTU_STOP_TQ      0x00008000  /* stop transfer q          */
#define PX_DTU_STOP_RQ      0x00004000  /* stop reply    q          */
#define	PX_DTU_RESET_TQ     0x00002000  /* stop & clear tranfer  q  */
#define PX_DTU_RESET_RQ     0x00001000  /* stop & clear reply q     */

#define PX_DTU_POSITIVE_CMD	0xFF000000
#define PX_DTU_NEGATIVE_CMD	0x0000F000
#define PX_DTU_VALID_CMD        (PX_DTU_POSITIVE_CMD | PX_DTU_NEGATIVE_CMD)
#define PX_DTU_NONCTL2_CMD	( PX_DTU_SET_TQ | PX_DTU_START_TQ  \
                                  | PX_DTU_RESET_TQ | PX_DTU_STOP_TQ )

/*
 *	queue size
 */
enum dtu_queue_size {
	PX_DTU_HQE256,
	PX_DTU_HQE512,
	PX_DTU_HQE1024,
	PX_DTU_HQE2048
};
/*
 * 	priority of px_dtu_cntl
 */
#define	PX_DTU_TQ_HIGH		0x80000000
#define PX_DTU_RQ_HIGH  	0x40000000

typedef struct dtu_user_list_s {
	int	vpnum;		/* target virtual processor id */
        uint_t  reply;		/* reply queue global virtual address */
        int     len;		/* length code.  enum dtu_queue_size */
        uint_t  cmd;		/* command code */
        uint_t  priority;	/* priority as PX_DTU_xxxxQ_HIGH */
        int     rc;		/* return code only set when EIO */
} dtu_user_list_t;

/*
 *	supervisor calls
 */
#if 0
int px_dtu_cntl(caddr_t trans,caddr_t reply,int len,uint_t priority,int cmd);
int px_dtu_cntl2( dtu_user_list_t *list, int llen );	
#endif

#ifdef _KERNEL

/*
 *	definition for user dtu handler (OS)
 */
typedef uint_t	px_sidenum_t;

#define	DTU_HEADER_SIZE		128
#define DTU_HW_HDR_SIZE		64	


/*
 *	definition for udtu_side_info.status    
 */
#define UDTU_ST_FREE		0
#define UDTU_ST_INITIALIZED     0x80000000
#define UDTU_ST_RQ_SET          0x08000000
#define UDTU_ST_TQ_SET          0x04000000


/*	definition for rupt_status 	*/

#define UDTU_RST_NULL		0	/* signal info is empty		*/
#define UDTU_RST_HOLD		1	/* signal info is hold 		*/ 
#define UDTU_RST_RSIG		2	/* remote signal is required 	*/

/*	udtu rupt process request block 	*/

typedef struct udturp_req_s {		/* request to udtu rupt process */
	struct udturp_req_s	*link;  /* link to next request */
	uint_t			cmd;    /* command code */
	int			i1;	/* information 1 */
	int			i2;	/* information 2 */
} udturp_req_t;

#define	URPCMD_RSIG     1	/* remote signal for send rupt */
#define URPCMD_SIG_R    2	/* signal for recieve rupt */

/*      type of rqo_queue entry         */

typedef struct rqoqe_s          {                   /* #rqo */
        struct rqoqe_s  *link;       /* link to next entry. #rqo */
        uint_t          rqohdr[64/sizeof(uint_t)]; /* header. #rqo  */
} rqoqe_t;       


/*	udtu message kind 			*/

enum udtu_msg_kind {
	UDTU_CTL2_MSG_TYPE,
	UDTU_RSIG_MSG_TYPE,
	UDTU_PXB_MSG_TYPE
};
	
/*	type of udtu side info 		*/

typedef struct udtu_side_info_s {       /* set by px_dtu_ctl */
	enum udtu_msg_kind msg_kind;
	struct udtu_side_info_s *link;
	udturp_req_t	req;
        uint_t          status;
        uint_t		rupt_status;
        int		dest_netpa;
	int		dest_side;
        rqoqe_t         *rqoq_head;     /* #rqo */
        rqoqe_t         *rqoq_tail;     /* #rqo */
        int             rqoql;          /* rqo_q_length  #rqo */
        k_siginfo_t	sigi;       /* last is header hardware area */
} udtu_side_info_t;

/*
 *	packet bodys for rpc:  px_bar_ctl2, px_clear_xb 
 */

/*	typedef struct udtu_spkt_body_s {	*/
/* 		uint_t		dest_sidenum;	*/
/*	        uint_t		caller_vpid;	*/
/*		dtu_user_list_t ulist;		*/
/*	} udtu_spkt_body_t;			*/
/*						*/
/*	typedef struct udtu_rpkt_body_s {	*/
/*		uint_t		callee_vpid;	*/
/*		int		rc;		*/ /* return code */
/*	} udtu_rpkt_body_t;			*/
/*						*/
/*						*/
/*	typedef struct sbdoy_pxb_s {		*/
/*		uint_t		dest_sidenum;	*/
/*	} sbody_pxb_t;				*/
/*						*/
/*	typedef struct rbody_pxb_s {		*/
/*		int		rc;		*/ /* return code */
/*	} rbody_pxb_t;				*/

typedef struct udtu_msg_s {
	enum udtu_msg_kind  msg_kind;	/* message kind */
	uint_t		dest_sidenum;	/* ctl2s, pxbs */
	uint_t		caller_vpid;	/* ctl2s,      */
	dtu_user_list_t ulist;		/* ctl2s,      */
	uint_t		callee_vpid;	/* ctl2r,      */
	uint_t		rc;		/* ctl2r, pxbr */
} udtu_msg_t;
		
/*
 *	message types for usr_dtu_message_queue
 */
/*#define UDTU_CTL2_MSG_TYPE	1 */  /* px_dtu_ctl2 message type */
/*#define UDTU_RSIG_MSG_TYPE	2 */  /* remote signal message type */
/*#define UDTU_PXB_MSG_TYPE	3 */  /* purge xb message type */

/*
 *	rupt status code of udtu_rqo()
 */
#define RMR_NOCHANGE		0
#define RMR_ERASED		1
#define RMR_CHANGED		2

#ifdef __STDC__
extern  int             init_udtu_bar();
extern  int             px_dtu_set( int sidenum );
extern  int             px_purge_xb( px_sidenum_t sidenum );
extern  int             px_dtu_disp( int sidenum );
extern  int             px_dtu_clear( int sidenum ); /* nothing do */
#else
extern  int             init_udtu_bar();
extern  int             px_dtu_set();
extern  int             px_purge_xb();
extern  int             px_dtu_disp();
extern  int             px_dtu_clear(); /* nothing do */
#endif /* __STDC__ */

#else  /* not _KERNEL */

#include <sys/syskai.h>		/* for API */

#define px_dtuctl( trs, rpl, len, pri, cmd) syskai(SKAIDTUCTL, trs, rpl, len, pri, cmd)
#define px_dtuctl2( list, llen )            syskai(SKAIDTUCTL2, list, llen )	

#endif /* _KERNEL */

#endif /* _IO_DTU_USR_DTU_H */



