static char SccsId[] = "@(#)elf2bsd.c	1.4(10/6/92)";

#include <ctype.h>
#include <stdio.h>
#include <elf.h>
#include <stdlib.h>
#include <errno.h>

void read_elf();
void conv_data();
void write_aout();


/*  A.OUT STRUCT */	

struct aout_head {
	unsigned short  a_info;
	unsigned short  a_magic;
	int 		a_text;
	int		a_data;
	int		a_bss;
	int		a_syms;
	int		a_entry;
	int		a_trsize;
	int		a_drsize;
};
struct aout_head Ahead;

struct rel {
	int		r_off;
	unsigned int    r_sym: 24,
			r_extern: 1,
			     : 2,
			r_type: 5;
	int 		r_addend;
};

struct sym {
	int		s_name;
	unsigned int	s_type: 8,
			      : 24;
	int		s_value;
};

struct rel *Drel;
struct rel *Trel;
struct sym *Sym;


/*  ELF  STRUCT  */

Elf32_Ehdr	Ehead;

Elf32_Rel	*E_textrel;
Elf32_Rel	*E_datarel;
Elf32_Sym	*E_sym;

Elf32_Phdr	Phead[3];
Elf32_Shdr	Shead[16];

/*  BUFFER  */

char *Text_p, *Data_p, *Textrel_p;
char *Datarel_p, *Sym_p, *Str_p;
char *Shstr_p;
int  len, count;
int  trsz_num, drsz_num, sysz_num;
int  textndx   = 0;
int  datandx   = 0;
int  bssndx    = 0;
int  symtabndx = 0;
int  strtabndx = 0;


int baseaddr = 0;
char *Aout_file, *Elf_file;

main( argc, argv)
	
	int argc;
	char **argv;

{
	if (argc != 3){
		fprintf(stderr, "Usage: elf2bsd elf bsd\n");
		exit(1);
	}
	Elf_file = argv[1];
	Aout_file = argv[2];

	read_elf();

	sysz_num = Shead[symtabndx].sh_size/sizeof *E_sym;

	conv_data();
	write_aout();

	exit(0);
}

/* READ_ELF() */
void read_elf()
{
	/* FILEOPEN */
	FILE *fp;
	int dummy[16], i;

	fp = fopen(Elf_file, "r");
	if (fp == NULL){
		perror("Can't open elf_file");
		exit(1);
	}

	/* READ ELF_HEADER */
	if (fread(&Ehead, 1, sizeof Ehead, fp) !=  (sizeof Ehead)){
		perror("Can't read elf_header");
		exit(1);
	}

	fseek(fp, Ehead.e_shoff, 0);
	for (i=0; i<16 && i<Ehead.e_shnum; i++) {
		fread(&Shead[i], 1, Ehead.e_shentsize, fp);
        }

	Shstr_p = (char *)malloc(Shead[Ehead.e_shstrndx].sh_size);
	fseek(fp, Shead[Ehead.e_shstrndx].sh_offset, 0);
	if (fread(Shstr_p, 1, Shead[Ehead.e_shstrndx].sh_size, fp) != Shead[Ehead.e_shstrndx].sh_size) {
                perror("Can't read shstr");
                exit(1);
        }

	for (i=0; i<16 && i<Ehead.e_shnum; i++) {
		if      (!strcmp(Shstr_p+Shead[i].sh_name, ".text"))
			textndx = i;
		else if (!strcmp(Shstr_p+Shead[i].sh_name, ".data"))
			datandx = i;
		else if (!strcmp(Shstr_p+Shead[i].sh_name, ".bss"))
			bssndx = i;
		else if (!strcmp(Shstr_p+Shead[i].sh_name, ".symtab"))
			symtabndx = i;
		else if (!strcmp(Shstr_p+Shead[i].sh_name, ".strtab"))
			strtabndx = i;
	}
		
	fprintf(stderr, "ELF:text=%08.8x ", Shead[textndx].sh_size);
	fprintf(stderr, "data=%08.8x ", Shead[datandx].sh_size);
	fprintf(stderr, "bss=%08.8x ", Shead[bssndx].sh_size);
	fprintf(stderr, "symt=%08.8x ", Shead[symtabndx].sh_size);
	fprintf(stderr, "strt=%08.8x\n", Shead[strtabndx].sh_size-1);

	/* READ PROGRAM_HEADER */
        fseek(fp, Ehead.e_phoff, 0);
        if (fread(Phead, Ehead.e_phnum, sizeof *Phead, fp) !=  (sizeof *Phead)){
                perror("Can't read program_header");
                exit(1);
        }

	if (Shead[datandx].sh_size == 0) {
		/* MALLOC */
		if ((Text_p = (char *)malloc(Phead[0].p_memsz)) == NULL){
			perror("Can't malloc elf_text");
			exit(1);
		}
		bzero(Text_p, Phead[0].p_memsz-1);
	}
	else {
		/* MALLOC */
		if ((Text_p = (char *)malloc(Phead[1].p_vaddr)) == NULL){
			perror("Can't malloc elf_text");
			exit(1);
		}
		bzero(Text_p, Phead[1].p_vaddr);
		if ((Data_p = (char *)malloc((Phead[1].p_memsz-1|7)+1)) == NULL){
			perror("Can't malloc elf_data");
			exit(1);
		}
		bzero(Data_p, (Phead[1].p_memsz-1|7)+1);
	}

	if ((E_sym = (Elf32_Sym *)malloc(Shead[symtabndx].sh_size)) == NULL){
                perror("Can't malloc elf_symt");
                exit(1);
	}
	if ((Str_p = (char *)malloc(Shead[strtabndx].sh_size)) == NULL){
                perror("Can't malloc elf_strt");
                exit(1);
	}

	/* READ (text, data) */
	if (Shead[textndx].sh_size) {
		fseek(fp, Phead[0].p_offset, 0);
		if (fread(Text_p, 1, Phead[0].p_filesz, fp) != Phead[0].p_filesz){
			perror("Can't read elf_text");
			exit(1);
		}
	}
	if (Shead[datandx].sh_size) {
		fseek(fp, Phead[1].p_offset, 0);
		if (fread(Data_p, 1, Phead[1].p_filesz, fp) != Phead[1].p_filesz){
			perror("Can't read elf_data");
			exit(1);
		}
	}

	/* READ (symt, strt) */
	fseek(fp, Shead[symtabndx].sh_offset, 0);
	if (fread(E_sym, 1, Shead[symtabndx].sh_size, fp) != Shead[symtabndx].sh_size){
		perror("Can't read elf_symt");
		exit(1);
	}
	fseek(fp, Shead[strtabndx].sh_offset, 0);
	if (fread(Str_p, 1, Shead[strtabndx].sh_size, fp) != Shead[strtabndx].sh_size){
		perror("Can't read elf_strt");
		exit(1);
	}

	fclose(fp);
}

/* CONV_DATA() */
void conv_data()
{
	int i,j;

	if ((Sym = (struct sym *)malloc((sizeof *Sym) * sysz_num)) == NULL){
		perror("Can't malloc elf_symbol");
		exit(1);
	}

	for (i=0,j=0; i<sysz_num; i++) {
		switch (E_sym[i].st_shndx) {
		case SHN_UNDEF:
		default:
			continue;
			break;
		case SHN_ABS:
			Sym[j].s_type = 2;
			break;
		case 1:
			Sym[j].s_type = 4;
			break;
		case 2:
			Sym[j].s_type = 6;
			break;
		case 3:
			Sym[j].s_type = 8;
			break;
		case SHN_COMMON:
			Sym[j].s_type = 8;
			break;
		}
		Sym[j].s_name = E_sym[i].st_name+4;
		Sym[j].s_value = E_sym[i].st_value;
		j++;
	}
	sysz_num = j;

	/* ELF_HEADER --> A.OUT_HEADER */
	if (Shead[datandx].sh_size == 0) {
		Ahead.a_info   = 0x00c8;
		Ahead.a_magic  = 0x0107;
		Ahead.a_text   = Shead[textndx].sh_size ? Phead[0].p_memsz:0;
		Ahead.a_data   = 0;
		Ahead.a_bss    = 0;
		Ahead.a_syms   = (sizeof *Sym) * sysz_num;
		Ahead.a_entry  = Phead[0].p_vaddr;
		Ahead.a_trsize = 0;
		Ahead.a_drsize = 0;
	}
	else {
		Ahead.a_info   = 0x00c8;
		Ahead.a_magic  = 0x0107;
		Ahead.a_text   = Phead[1].p_vaddr-Phead[0].p_vaddr;
		Ahead.a_data   = (Phead[1].p_memsz-1|7)+1;
		Ahead.a_bss    = 0;
		Ahead.a_syms   = (sizeof *Sym) * sysz_num;
		Ahead.a_entry  = Phead[0].p_vaddr;
		Ahead.a_trsize = 0;
		Ahead.a_drsize = 0;
	}

	fprintf(stderr, "BSD:text=%08.8x ", Ahead.a_text);
	fprintf(stderr, "data=%08.8x ", Ahead.a_data);
	fprintf(stderr, "bss=%08.8x ", 0);
	fprintf(stderr, "symt=%08.8x ", Ahead.a_syms);
	fprintf(stderr, "strt=%08.8x\n", Shead[strtabndx].sh_size-1);
}

/* WRITE_AOUT() */
void write_aout()
{
	FILE *fp;
	int len;

	/* FILE OPEN */
	fp = fopen(Aout_file, "wb");

	if (fp == NULL){
		perror("Can't open Aout_file");
		exit(1);
	}

	/* WRITE AOUT */
	if ((fwrite(&Ahead, 1, sizeof Ahead, fp)) != sizeof Ahead){
		perror("Can't read Aout_header");
		exit(1);
	}

	/* WRITE TEXT & DATA */
	if ((fwrite(Text_p, 1, Ahead.a_text, fp)) != Ahead.a_text){
		perror("Can't write Text");
		exit(1);
	}

	if ((fwrite(Data_p, 1, Ahead.a_data, fp)) != Ahead.a_data){
		perror("Can't write Data");
		exit(1);
	}

	/* WRITE SYMT & STRT */
	if ((fwrite(Sym, 1, Ahead.a_syms, fp)) != Ahead.a_syms){
		perror("Can't write Symt");
		exit(1);
        }

	len = Shead[strtabndx].sh_size+4;
	if ((fwrite(&len, 1, 4, fp)) != 4){
                perror("Can't write Strt");
                exit(1);
        }

	if ((fwrite(Str_p, 1, len-4, fp)) != len-4){
                perror("Can't write Strt");
                exit(1);
        }

	fclose(fp);
}

