/* @(#)tachyon-opcode.h	1.17    1/7/93 */

/* tachyon_opcode.h -- Table of opcodes for the tachyon.
   Copyright (C) 1989 Free Software Foundation, Inc.

This file is part of GAS, the GNU Assembler.

GAS is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GAS; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Nass Ave, Cambridge, NA 02139, USA.  */

#if !defined(__STDC__) && !defined(const)
#define const
#endif

/*
   All tachyon opcodes are 64 bits.

   The match component is a mask saying which bits must match a
   particular opcode in order for an instruction to be an instance
   of that opcode.

   The args component is a string containing one character
   for each operand of the instruction.

 ----- STATUS -----
   COPROCESSOR-OPs are not supported yet.
*/

/*
   Type of opcode: first 4 bit of tachyon_opcode->match_u.
       0x00 NOP	
       0x01 SHIFT
       0x02 L/S/M
       0x03 L/S/M imm25
       0x04 arithmetic/logical imm25	
       0x05 arithmetic/logical regs/imm5
       0x06 branch imm12
       0x07 call/branch imm32	
       0x08 jump
       0x09 float (add)		
       0x0a float (mul)	
       0x0b coprocessor
       0x0c system
*/
   
#define   OPER_NOP     0x00
#define   OPER_SFT     0x01
#define   OPER_LSM     0x02
#define   OPER_LSMI25  0x03
#define   OPER_ALI25   0x04
#define   OPER_ALRI5   0x05
#define   OPER_BRCI12  0x06
#define   OPER_CBRI32  0x07
#define   OPER_JMP     0x08
#define   OPER_FLO_A   0x09
#define   OPER_FLO_M   0x0a
#define   OPER_COPRO   0x0b
#define   OPER_SYS     0x0c

#define   TYPE_0       0x00
#define   TYPE_1       0x01
#define   TYPE_2       0x02
#define   TYPE_3       0x03
#define   TYPE_4       0x04
#define   TYPE_5       0x05
#define   TYPE_6       0x06
#define   TYPE_7       0x07
#define   TYPE_8       0x08
#define   TYPE_9       0x09
#define   TYPE_A       0x0a
#define   TYPE_B       0x0b
#define   TYPE_C       0x0c
#define   TYPE_D       0x0d
#define   TYPE_E       0x0e
#define   TYPE_F       0x0f
   
#define _ (-1)
/*
 * 0:20+20+20 1:20+20+20(12bit) 2:20+40(25bit) 3:20+40(32bit) 4:60(4+5+5+10+10+25+5)
 * 5:60(4+4+16+20+20)
 */
int type_listing[16] = {
/* 0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f    */
   0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 2, 4, 3, 3, 3, 5 };

/*
 * true if ICCP & ICCS don't match with OP0 & OP1
 */
int type_exchange_loc[16] = {
/* Never exchange location (op0/op1)                 */
/* 0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f    */
/* 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }; */
/* Exchange location (op0/op1) ... Don't use         */
/* 0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f    */
   0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0 };

/*
 * 0:must be OP0, 1:must be OP1, 2:must be OP2, other:OP0/OP1/OP2
 */
int op_single_loc[16] = {
/* 0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f */
   _, 0, _, 1, _, _, 2, 1, _, _, _, 0, 0, _, _, _ };

/*
 * 0:can't be stared in OP2, 1:can be stared in OP2
 */
int op_avail_OP2[16] = {
/* 0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f */
   1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0 };

/*
 * combination of OPs
 */

int op_type_OP0_OP1[16][16] = {
/* Secondary-CC side
   0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f */ /*Primary-CC side		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 0 NOP     	 	*/
 { _, _, _, 9, 9, _, _,13, _, _, _, _, _, _, _, _ }, /* 1 SHIFT			*/
 { _, _, _, _, _, _, _,12, _, _, _, _, _, _, _, _ }, /* 2 L/S/M			*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 3 L/S/M imm25		*/
 { _, _, 8, _, _, _, _, _, 8, _, _, _, _, _, _, _ }, /* 4 arith/logic imm25	*/
 { _, _, _, 9, 9, _, _,13, _, _, _, _, _, _, _, _ }, /* 5 arith/logic reg/imm5	*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 6 branch imm12		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 7 call/branch		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 8 jump       		*/
 { _, _, _,10,10, _, _,14, _, _, _, _, _, _, _, _ }, /* 9 float	(add)		*/
 { _, _, _,10,10, _, _,14, _, _, _, _, _, _, _, _ }, /* a float	(mul)		*/
 {15, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* b coprocessor		*/
 {11, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* c system		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* d ---- */
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* e ---- */
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }  /* f ---- */
};

int op_type_OP2_bimm12[16][16] = {
/* Secondary-CC side
   0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f */ /*Primary-CC side		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 0 			*/
 { _, _, 6, _, _, 7, _, _, _, _, _, _, _, _, _, _ }, /* 1 SHIFT			*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 2 L/S/M			*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 3 L/S/M imm25		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 4 arith/logic imm25	*/
 { _, _, 6, _, _, 7, _, _, _, _, _, _, _, _, _, _ }, /* 5 arith/logic reg/imm5	*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 6 branch imm12		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 7 call/branch		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 8 jump       		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 9 float(add)		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* a float(mul)		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* b coprocessor		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* c system		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* d ----	*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* e ----	*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }  /* f ----  */
};

int op_type_OP2_fadd[16][16] = {
/* Secondary-CC side
   0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f */ /*Primary-CC side		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 0 NOP     	 	*/
 { _, _, 4, _, _, 5, _, _, 4, _, _, _, _, _, _, _ }, /* 1 SHIFT			*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 2 L/S/M			*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 3 L/S/M imm25		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 4 arith/logic imm25	*/
 { _, _, 4, _, _, 5, _, _, 4, _, _, _, _, _, _, _ }, /* 5 arith/logic reg/imm5	*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 6 branch imm12		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 7 call/branch		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 8 jump       		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 9 float	(add)		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* a float	(mul)		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* b coprocessor		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* c system		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* d ---- */
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* e ---- */
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }  /* f ---- */
};

int op_type_OP2_fmul[16][16] = {
/* Secondary-CC side
   0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f */ /*Primary-CC side		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 0 NOP     	 	*/
 { _, _, 4, _, _, 5, _, _, 4, 3, _, _, _, _, _, _ }, /* 1 SHIFT			*/
 { _, _, _, _, _, _, _, _, _, 2, _, _, _, _, _, _ }, /* 2 L/S/M			*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 3 L/S/M imm25		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 4 arith/logic imm25	*/
 { _, _, 4, _, _, 5, _, _, 4, 3, _, _, _, _, _, _ }, /* 5 arith/logic reg/imm5	*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 6 branch imm12		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 7 call/branch		*/
 { _, _, _, _, _, _, _, _, _, 2, _, _, _, _, _, _ }, /* 8 jump       		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* 9 float	(add)		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* a float	(mul)		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* b coprocessor		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* c system		*/
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* d ---- */
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }, /* e ---- */
 { _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _ }  /* f ---- */
};
#undef _

/*
 * Structure of an opcode table entry.
 */

struct tachyon_opcode
{
    const char *name;
    unsigned long mask_u;     /* used only for error checking */
    unsigned long mask_l;     /* used only for error checking */
    unsigned long match_u;
    unsigned long match_l;
    const char *args;
    unsigned int length;    /* used to classify length */
    unsigned int last;      /* used to simplify hashing */
};

/*
Kinds of operands:
      =:supported, ?:not supported yet
      
 = #    Number used by optimizer.  It is ignored.
 = 1    rs1 register.
 = 2    rs2 register.
 = 3    rs3 register(for 20 bit OP).
 = 4    rs3 register(for 40 bit OP).
 = 5    rs3/rs4 register(for 60 bit type-B OP).
 = e    frs1 floating point register.
 = f    frs2 floating point register.
 = g    frs3 floating point register(for 20 bit OP).
 = h    frs3 floating point register(for 40 bit OP).
 = i     5 bit Immediate.
 = u     5 bit Unsigned Immediate.(for shift OP).
 = j    12 bit immediate.
 = I    25 bit Immediate.
 = J    32 bit Immediate.
 = K    25 bit Immediate(32 bit -> base_GR1 + 25 bit).
 = r    12 bit PC relative immediate.
 = R    32 bit PC relative immediate.
 = s    integer/floating state register(for 20 bit OP).
 = t    integer/floating state register(for 40 bit OP).
 = c    Coprocessor state register(for 20 bit OP).
 = d    Coprocessor state register(for 40 bit OP).

 = U    SRrsr( 5 bits wide 0-31 for type-B).
 = V    SRrsr(10 bits wide 0-1023 for type-B).
 = X     4 bit Unsigned Immediate(for type-B).
 = Y     8 bit Unsigned Immediate(for type-B).
 = y    Base register(for type-B).
 = z    25 bit Immediate(for type-B).
 = w    25 bit Immediate(32 bit -> base_GR2 + 25 bit (for type-B)).

 = A    VCP vR1 register (26-33)
 = B    VCP vR2 register (36-43)
 = C    VCP vR3 register (48-55)
 = D    VCP vR4 register (56-63)
 = L    VCP vR1 mask register (26-33)
 = M    VCP vR2 mask register (36-43)
 = N    VCP vR3 mask register (48-55)
 = O    VCP vR4 mask register (56-63)
 = 6    VCP rr1 GR (14-18)
 = 7    VCP rr2 GR (29-33)
 = 8    VCP rw1 GR (39-43)
 = 9    VCP rw2 GR (59-63)
 = E    VCP rr2 FR (29-33)
 = F    VCP rw2 FR (59-63)
 = P    VCP R2 VCR (29-33)
 = Q    VCP R3 VCR (59-63)
 = S    VCP R2 VUT (29-33)
 = T    VCP R3 VUT (59-63)
 = H    VCP M (54-55)
 = G    VCP ID (26-28) (disassembler only)
*/

/*
 *   The assembler requires that all instances of the same mnemonic must be
 *   consecutive.  If they aren't, the assembler will bomb at runtime.
 */

static const struct tachyon_opcode tachyon_opcodes[] =
{

{ "nop",      0xff800000, 0x00000000, 0x00000000, 0x00000000, "",      1,1 },

{ "call",     0xfe000000, 0x00000000, 0x72000000, 0x00000000, "R,4",   2,1 },
{ "brc",      0xfe000000, 0x00000000, 0x74000000, 0x00000000, "t,R",   2,0 },
{ "brc",      0xfe000000, 0x00000000, 0x78000000, 0x00000000, "d,R",   2,1 },
{ "brcs",     0xfe000000, 0x00000000, 0x64000000, 0x00000000, "s,r",   1,0 },
{ "brcs",     0xfe000000, 0x00000000, 0x68000000, 0x00000000, "c,r",   1,1 },
{ "brnc",     0xfe000000, 0x00000000, 0x76000000, 0x00000000, "t,R",   2,0 },
{ "brnc",     0xfe000000, 0x00000000, 0x7a000000, 0x00000000, "d,R",   2,1 },
{ "brncs",    0xfe000000, 0x00000000, 0x66000000, 0x00000000, "s,r",   1,0 },
{ "brncs",    0xfe000000, 0x00000000, 0x6a000000, 0x00000000, "c,r",   1,1 },
{ "seti",     0xfe000000, 0x00000000, 0x7c000000, 0x00000000, "J,4",   2,1 },
{ "jmp",      0xff800000, 0x00000000, 0x80800000, 0x00000000, "1",     1,1 },

{ "ld.c",     0xff800000, 0x00000000, 0x21000000, 0x00000000, "1,2,3", 1,0 },
{ "ld.c",     0xff800000, 0x00000000, 0x31000000, 0x00000000, "1,I,4", 2,0 },
{ "ld.c",     0xff800000, 0x00000000, 0x31000000, 0x00000000, "K,4",   2,1 },
{ "ld.cu",    0xff800000, 0x00000000, 0x21800000, 0x00000000, "1,2,3", 1,0 },
{ "ld.cu",    0xff800000, 0x00000000, 0x31800000, 0x00000000, "1,I,4", 2,0 },
{ "ld.cu",    0xff800000, 0x00000000, 0x31800000, 0x00000000, "K,4",   2,1 },
{ "ld.s",     0xff800000, 0x00000000, 0x22000000, 0x00000000, "1,2,3", 1,0 },
{ "ld.s",     0xff800000, 0x00000000, 0x32000000, 0x00000000, "1,I,4", 2,0 },
{ "ld.s",     0xff800000, 0x00000000, 0x32000000, 0x00000000, "K,4",   2,1 },
{ "ld.su",    0xff800000, 0x00000000, 0x22800000, 0x00000000, "1,2,3", 1,0 },
{ "ld.su",    0xff800000, 0x00000000, 0x32800000, 0x00000000, "1,I,4", 2,0 },
{ "ld.su",    0xff800000, 0x00000000, 0x32800000, 0x00000000, "K,4",   2,1 },
{ "ld",       0xff800000, 0x00000000, 0x23000000, 0x00000000, "1,2,3", 1,0 },
{ "ld",       0xff800000, 0x00000000, 0x33000000, 0x00000000, "1,I,4", 2,0 },
{ "ld",       0xff800000, 0x00000000, 0x33000000, 0x00000000, "K,4",   2,1 },
{ "ld.l",     0xff800000, 0x00000000, 0x23800000, 0x00000000, "1,2,3", 1,0 },
{ "ld.l",     0xff800000, 0x00000000, 0x33800000, 0x00000000, "1,I,4", 2,0 },
{ "ld.l",     0xff800000, 0x00000000, 0x33800000, 0x00000000, "K,4",   2,1 },
{ "ld.f",     0xff800000, 0x00000000, 0x24000000, 0x00000000, "1,2,g", 1,0 },
{ "ld.f",     0xff800000, 0x00000000, 0x34000000, 0x00000000, "1,I,h", 2,0 },
{ "ld.f",     0xff800000, 0x00000000, 0x34000000, 0x00000000, "K,h",   2,1 },
{ "ld.d",     0xff800000, 0x00000000, 0x24800000, 0x00000000, "1,2,g", 1,0 },
{ "ld.d",     0xff800000, 0x00000000, 0x34800000, 0x00000000, "1,I,h", 2,0 },
{ "ld.d",     0xff800000, 0x00000000, 0x34800000, 0x00000000, "K,h",   2,1 },
{ "ld.pf",    0xff800000, 0x00000000, 0x25000000, 0x00000000, "1,2,g", 1,0 },
{ "ld.pf",    0xff800000, 0x00000000, 0x35000000, 0x00000000, "1,I,h", 2,0 },
{ "ld.pf",    0xff800000, 0x00000000, 0x35000000, 0x00000000, "K,h",   2,1 },
{ "ld.pd",    0xff800000, 0x00000000, 0x25800000, 0x00000000, "1,2,g", 1,0 },
{ "ld.pd",    0xff800000, 0x00000000, 0x35800000, 0x00000000, "1,I,h", 2,0 },
{ "ld.pd",    0xff800000, 0x00000000, 0x35800000, 0x00000000, "K,h",   2,1 },

{ "st.c",     0xff800000, 0x00000000, 0x28800000, 0x00000000, "3,1,2", 1,0 },
{ "st.c",     0xff800000, 0x00000000, 0x38800000, 0x00000000, "4,1,I", 2,0 },
{ "st.c",     0xff800000, 0x00000000, 0x38800000, 0x00000000, "4,K",   2,1 },
{ "st.s",     0xff800000, 0x00000000, 0x29000000, 0x00000000, "3,1,2", 1,0 },
{ "st.s",     0xff800000, 0x00000000, 0x39000000, 0x00000000, "4,1,I", 2,0 },
{ "st.s",     0xff800000, 0x00000000, 0x39000000, 0x00000000, "4,K",   2,1 },
{ "st",       0xff800000, 0x00000000, 0x29800000, 0x00000000, "3,1,2", 1,0 },
{ "st",       0xff800000, 0x00000000, 0x39800000, 0x00000000, "4,1,I", 2,0 },
{ "st",       0xff800000, 0x00000000, 0x39800000, 0x00000000, "4,K",   2,1 },
{ "st.f",     0xff800000, 0x00000000, 0x2e000000, 0x00000000, "g,1,2", 1,0 },
{ "st.f",     0xff800000, 0x00000000, 0x3e000000, 0x00000000, "h,1,I", 2,0 },
{ "st.f",     0xff800000, 0x00000000, 0x3e000000, 0x00000000, "h,K",   2,1 },
{ "st.d",     0xff800000, 0x00000000, 0x2e800000, 0x00000000, "g,1,2", 1,0 },
{ "st.d",     0xff800000, 0x00000000, 0x3e800000, 0x00000000, "h,1,I", 2,0 },
{ "st.d",     0xff800000, 0x00000000, 0x3e800000, 0x00000000, "h,K",   2,1 },

{ "mov.if",   0xff800000, 0x00000000, 0x2a000000, 0x00000000, "1,g",   1,1 },
{ "mov.ld",   0xff800000, 0x00000000, 0x2a800000, 0x00000000, "1,g",   1,1 },
{ "mov.fi",   0xff800000, 0x00000000, 0x2b000000, 0x00000000, "e,3",   1,1 },
{ "mov.dl",   0xff800000, 0x00000000, 0x2b800000, 0x00000000, "e,3",   1,1 }, 

{ "add",      0xff800000, 0x00000000, 0x52000000, 0x00000000, "1,2,3", 1,1 },
{ "addl",     0xff800000, 0x00000000, 0x52800000, 0x00000000, "1,2,3", 1,1 },
{ "sub",      0xff800000, 0x00000000, 0x53000000, 0x00000000, "1,2,3", 1,1 },
{ "subl",     0xff800000, 0x00000000, 0x53800000, 0x00000000, "1,2,3", 1,1 },
{ "addi5",    0xff800000, 0x00000000, 0x5a000000, 0x00000000, "1,i,3", 1,1 },
{ "addi25",   0xff800000, 0x00000000, 0x4a000000, 0x00000000, "1,I,4", 2,1 },
{ "addli5",   0xff800000, 0x00000000, 0x5a800000, 0x00000000, "1,i,3", 1,1 },
{ "addli25",  0xff800000, 0x00000000, 0x4a800000, 0x00000000, "1,I,4", 2,1 },
{ "subi5",    0xff800000, 0x00000000, 0x5b000000, 0x00000000, "1,i,3", 1,1 },
{ "subi25",   0xff800000, 0x00000000, 0x4b000000, 0x00000000, "1,I,4", 2,1 },
{ "subli5",   0xff800000, 0x00000000, 0x5b800000, 0x00000000, "1,i,3", 1,1 },
{ "subli25",  0xff800000, 0x00000000, 0x4b800000, 0x00000000, "1,I,4", 2,1 },

{ "and",      0xff800000, 0x00000000, 0x54000000, 0x00000000, "1,2,3", 1,0 },
{ "and",      0xff800000, 0x00000000, 0x4c000000, 0x00000000, "1,I,4", 2,1 },
{ "or",       0xff800000, 0x00000000, 0x54800000, 0x00000000, "1,2,3", 1,0 },
{ "or",       0xff800000, 0x00000000, 0x4c800000, 0x00000000, "1,I,4", 2,1 },
{ "xor",      0xff800000, 0x00000000, 0x55000000, 0x00000000, "1,2,3", 1,0 },
{ "xor",      0xff800000, 0x00000000, 0x4d000000, 0x00000000, "1,I,4", 2,1 },

{ "sra",      0xff800000, 0x00000000, 0x15800000, 0x00000000, "1,2,3", 1,0 },
{ "sra",      0xff800000, 0x00000000, 0x1d800000, 0x00000000, "1,u,3", 1,1 },
{ "srl",      0xff800000, 0x00000000, 0x16000000, 0x00000000, "1,2,3", 1,0 },
{ "srl",      0xff800000, 0x00000000, 0x1e000000, 0x00000000, "1,u,3", 1,1 },
{ "sll",      0xff800000, 0x00000000, 0x16800000, 0x00000000, "1,2,3", 1,0 },
{ "sll",      0xff800000, 0x00000000, 0x1e800000, 0x00000000, "1,u,3", 1,1 },
{ "mult",     0xff800000, 0x00000000, 0x26000000, 0x00000000, "1,2,3", 1,0 },
{ "mult",     0xff800000, 0x00000000, 0x36000000, 0x00000000, "1,I,4", 2,1 },
{ "mult.l",   0xff800000, 0x00000000, 0x26800000, 0x00000000, "1,2,3", 1,0 },
{ "mult.l",   0xff800000, 0x00000000, 0x36800000, 0x00000000, "1,I,4", 2,1 },
{ "mult.lu",  0xff800000, 0x00000000, 0x27800000, 0x00000000, "1,2,3", 1,0 },
{ "mult.lu",  0xff800000, 0x00000000, 0x37800000, 0x00000000, "1,I,4", 2,1 },

{ "fadd",     0xff800000, 0x00000000, 0x91000000, 0x00000000, "e,f,g", 1,1 },
{ "fsub",     0xff800000, 0x00000000, 0x92000000, 0x00000000, "e,f,g", 1,1 },
{ "fmul",     0xff800000, 0x00000000, 0xa3000000, 0x00000000, "e,f,g", 1,1 },
{ "fdiv",     0xff800000, 0x00000000, 0xa4000000, 0x00000000, "e,f,g", 1,1 },
{ "fcmp",     0xff800000, 0x00000000, 0x9a000000, 0x00000000, "e,f",   1,1 },
{ "fadd.d",   0xff800000, 0x00000000, 0x91800000, 0x00000000, "e,f,g", 1,1 },
{ "fsub.d",   0xff800000, 0x00000000, 0x92800000, 0x00000000, "e,f,g", 1,1 },
{ "fmul.d",   0xff800000, 0x00000000, 0xa3800000, 0x00000000, "e,f,g", 1,1 },
{ "fdiv.d",   0xff800000, 0x00000000, 0xa4800000, 0x00000000, "e,f,g", 1,1 },
{ "fcmp.d",   0xff800000, 0x00000000, 0x9a800000, 0x00000000, "e,f",   1,1 },
{ "fadd.x",   0xff800000, 0x00000000, 0x9d000000, 0x00000000, "e,f,g", 1,1 },
{ "fsub.x",   0xff800000, 0x00000000, 0x9e000000, 0x00000000, "e,f,g", 1,1 },
{ "fcmp.x",   0xff800000, 0x00000000, 0x9b000000, 0x00000000, "e,f",   1,1 },
{ "fcnv.fd",  0xff800000, 0x00000000, 0x95000000, 0x00000000, "e,g",   1,1 },
{ "fcnv.df",  0xff800000, 0x00000000, 0x95800000, 0x00000000, "e,g",   1,1 },
{ "fcnv.dx",  0xff800000, 0x00000000, 0x9f000000, 0x00000000, "e,g",   1,1 },
{ "fcnv.xd",  0xff800000, 0x00000000, 0x9f800000, 0x00000000, "e,g",   1,1 },
{ "fcnv.if",  0xff800000, 0x00000000, 0x96000000, 0x00000000, "e,g",   1,1 },
{ "fcnv.fi",  0xff800000, 0x00000000, 0x96800000, 0x00000000, "e,g",   1,1 },
{ "fcnv.id",  0xff800000, 0x00000000, 0x97000000, 0x00000000, "e,g",   1,1 },
{ "fcnv.di",  0xff800000, 0x00000000, 0x97800000, 0x00000000, "e,g",   1,1 },
{ "fcnz.fi",  0xff800000, 0x00000000, 0x98800000, 0x00000000, "e,g",   1,1 },
{ "fcnz.di",  0xff800000, 0x00000000, 0x99800000, 0x00000000, "e,g",   1,1 },

{ "svc",      0xff800000, 0x00000000, 0xc0800000, 0x00000000, "Y",     3,1 },
{ "ldsr",     0xff800000, 0x00000000, 0xc1000000, 0x00000000, "y,z,U", 3,0 },
{ "ldsr",     0xff800000, 0x00000000, 0xc1000000, 0x00000000, "w,U",   3,1 },
{ "stsr",     0xff800000, 0x00000000, 0xc1800000, 0x00000000, "U,y,z", 3,0 },
{ "stsr",     0xff800000, 0x00000000, 0xc1800000, 0x00000000, "U,w",   3,1 },
{ "ld.mcr",   0xff800000, 0x00000000, 0xc9000000, 0x00000000, "y,z,5", 3,0 },
{ "ld.mcr",   0xff800000, 0x00000000, 0xc9000000, 0x00000000, "w,5",   3,1 },
{ "st.mcr",   0xff800000, 0x00000000, 0xc9800000, 0x00000000, "5,y,z", 3,0 },
{ "st.mcr",   0xff800000, 0x00000000, 0xc9800000, 0x00000000, "5,w",   3,1 },
{ "wbst",     0xff800000, 0x00000000, 0xcd000000, 0x00000000, "z",     3,1 },
{ "movgs",    0xff800000, 0x00000000, 0xc2000000, 0x00000000, "1,V",   3,1 },
{ "movsg",    0xff800000, 0x00000000, 0xc2800000, 0x00000000, "V,1",   3,1 },
{ "movgt",    0xff800000, 0x00000000, 0xce000000, 0x00000000, "1",     3,1 },
{ "movtg",    0xff800000, 0x00000000, 0xce800000, 0x00000000, "1",     3,1 },
{ "ptlbe",    0xff800000, 0x00000000, 0xc4000000, 0x00000000, "y,z",   3,1 },
{ "ptlbp",    0xff800000, 0x00000000, 0xc4800000, 0x00000000, "y,z",   3,1 },
{ "ptlba",    0xff800000, 0x00000000, 0xc5000000, 0x00000000, "y,z",   3,0 },
{ "ptlba",    0xff800000, 0x00000000, 0xc5000000, 0x00000000, "w",     3,1 },
{ "flche",    0xff800000, 0x00000000, 0xc6000000, 0x00000000, "y,z",   3,1 },
{ "flcha",    0xff800000, 0x00000000, 0xc6800000, 0x00000000, "y,z",   3,0 },
{ "flcha",    0xff800000, 0x00000000, 0xc6800000, 0x00000000, "w",     3,1 },
{ "aki",      0xff800000, 0x00000000, 0xc7000000, 0x00000000, "X,5",   3,1 },
{ "rti",      0xff800000, 0x00000000, 0xc7800000, 0x00000000, "",      3,1 },
{ "cs",       0xff800000, 0x00000000, 0xc8800000, 0x00000000, "1,3,y,z,5",3,0},
{ "cs",       0xff800000, 0x00000000, 0xc8800000, 0x00000000, "1,3,w,5",3,1},
{ "fwc",      0xff800000, 0x00000000, 0xcf800000, 0x00000000, "Y",     3,1 },

{ "vplcr",    0xfffc1fc0, 0x300f0000, 0xb00004c0, 0x00000000, "7,Q",     3,1 },
{ "vpscr",    0xfffc1fc0, 0x300f0000, 0xb00001c0, 0x00010000, "P,9",     3,1 },
{ "vlcr",     0xfffc1fc0, 0x300f0000, 0xb00004c0, 0x20000000, "7,Q",     3,1 },
{ "vscr",     0xfffc1fc0, 0x300f0000, 0xb00001c0, 0x20010000, "P,9",     3,1 },
{ "vlvl",     0xfffc1fc0, 0x300f0000, 0xb00004c0, 0x30020000, "7",       3,1 },
{ "vlvlf",    0xfffc1fc0, 0x300f0000, 0xb00004c0, 0x30030000, "7",       3,1 },
{ "vlvt",     0xfffc1fc0, 0x300f0000, 0xb00004c0, 0x00020000, "7,T",     3,1 },
{ "vsvt",     0xfffc1fc0, 0x300f0000, 0xb00001c0, 0x00030000, "S,9",     3,1 },
{ "vltr",     0xfffc1fc0, 0x300f0000, 0xb00008c0, 0x10020000, "6",       3,1 },
{ "vpt",      0xfffc1fc0, 0x300f0000, 0xb00000c0, 0x30000000, "",        3,1 },

{ "vl",	      0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x10000000, "6,7,C,O", 3,1 },
{ "vla",      0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x10000000, "6,7,C,O", 3,1 },
{ "vlb",      0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x10000000, "6,7,C,O", 3,1 },
{ "vld",      0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x20000000, "6,7,C,O", 3,1 },
{ "vlda",     0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x20000000, "6,7,C,O", 3,1 },
{ "vldb",     0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x20000000, "6,7,C,O", 3,1 },
{ "vle",      0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x30000000, "6,7,C,O", 3,1 },
{ "vlea",     0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x30000000, "6,7,C,O", 3,1 },
{ "vleb",     0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x30000000, "6,7,C,O", 3,1 },
{ "vlm",      0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x00000000, "6,7,N",   3,1 },
{ "vlma",     0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x00000000, "6,7,N",   3,1 },
{ "vlmb",     0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x00000000, "6,7,N",   3,1 },
{ "vil",      0xfffc1fc0, 0x300f0000, 0xb0000800, 0x10000000, "6,A,C,O", 3,1 },
{ "vild",     0xfffc1fc0, 0x300f0000, 0xb0000800, 0x20000000, "6,A,C,O", 3,1 },
{ "vile",     0xfffc1fc0, 0x300f0000, 0xb0000800, 0x30000000, "6,A,C,O", 3,1 },

{ "vlcp",     0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x10020000, "6,7,C,O", 3,1 },
{ "vlcpa",    0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x10020000, "6,7,C,O", 3,1 },
{ "vlcpb",    0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x10020000, "6,7,C,O", 3,1 },
{ "vlcpd",    0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x20020000, "6,7,C,O", 3,1 },
{ "vlcpda",   0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x20020000, "6,7,C,O", 3,1 },
{ "vlcpdb",   0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x20020000, "6,7,C,O", 3,1 },
{ "vlcpe",    0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x30020000, "6,7,C,O", 3,1 },
{ "vlcpea",   0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x30020000, "6,7,C,O", 3,1 },
{ "vlcpeb",   0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x30020000, "6,7,C,O", 3,1 },

{ "vst",      0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x10010000, "B,6,7,O", 3,1 },
{ "vsta",     0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x10010000, "B,6,7,O", 3,1 },
{ "vstb",     0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x10010000, "B,6,7,O", 3,1 },
{ "vstd",     0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x20010000, "B,6,7,O", 3,1 },
{ "vstda",    0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x20010000, "B,6,7,O", 3,1 },
{ "vstdb",    0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x20010000, "B,6,7,O", 3,1 },
{ "vste",     0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x30010000, "B,6,7,O", 3,1 },
{ "vstea",    0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x30010000, "B,6,7,O", 3,1 },
{ "vsteb",    0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x30010000, "B,6,7,O", 3,1 },
{ "vstm",     0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x00010000, "M,6,7",   3,1 },
{ "vstma",    0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x00010000, "M,6,7",   3,1 },
{ "vstmb",    0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x00010000, "M,6,7",   3,1 },
{ "vist",     0xfffc1fc0, 0x300f0000, 0xb0000800, 0x10010000, "B,6,A,O", 3,1 },
{ "vistd",    0xfffc1fc0, 0x300f0000, 0xb0000800, 0x20010000, "B,6,A,O", 3,1 },
{ "viste",    0xfffc1fc0, 0x300f0000, 0xb0000800, 0x30010000, "B,6,A,O", 3,1 },

{ "vstx",     0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x10030000, "B,6,7,O", 3,1 },
{ "vstxa",    0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x10030000, "B,6,7,O", 3,1 },
{ "vstxb",    0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x10030000, "B,6,7,O", 3,1 },
{ "vstxd",    0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x20030000, "B,6,7,O", 3,1 },
{ "vstxda",   0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x20030000, "B,6,7,O", 3,1 },
{ "vstxdb",   0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x20030000, "B,6,7,O", 3,1 },
{ "vstxe",    0xfffc1ff8, 0x300f0000, 0xb0000c80, 0x30030000, "B,6,7,O", 3,1 },
{ "vstxea",   0xfffc1ff8, 0x300f0000, 0xb0000c88, 0x30030000, "B,6,7,O", 3,1 },
{ "vstxeb",   0xfffc1ff8, 0x300f0000, 0xb0000c90, 0x30030000, "B,6,7,O", 3,1 },

{ "va",	      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x10080000, "A,B,C,O", 3,1 },
{ "vad",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x20080000, "A,B,C,O", 3,1 },
{ "vae",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x30080000, "A,B,C,O", 3,1 },
{ "vas",      0xfffc1fc0, 0x300f0000, 0xb0000480, 0x10080000, "7,B,C,O", 3,1 },
{ "vasd",     0xfffc1fc0, 0x300f0000, 0xb0800480, 0x20080000, "E,B,C,O", 3,1 },
{ "vase",     0xfffc1fc0, 0x300f0000, 0xb0800480, 0x30080000, "E,B,C,O", 3,1 },
{ "vs",	      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x10090000, "A,B,C,O", 3,1 },
{ "vsd",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x20090000, "A,B,C,O", 3,1 },
{ "vse",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x30090000, "A,B,C,O", 3,1 },
{ "vss",      0xfffc1fc0, 0x300f0000, 0xb0000480, 0x10090000, "7,B,C,O", 3,1 },
{ "vssd",     0xfffc1fc0, 0x300f0000, 0xb0800480, 0x20090000, "E,B,C,O", 3,1 },
{ "vsse",     0xfffc1fc0, 0x300f0000, 0xb0800480, 0x30090000, "E,B,C,O", 3,1 },
{ "vm",	      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x100a0000, "A,B,C,O", 3,1 },
{ "vmd",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x200a0000, "A,B,C,O", 3,1 },
{ "vme",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x300a0000, "A,B,C,O", 3,1 },
{ "vms",      0xfffc1fc0, 0x300f0000, 0xb0000480, 0x100a0000, "7,B,C,O", 3,1 },
{ "vmsd",     0xfffc1fc0, 0x300f0000, 0xb0800480, 0x200a0000, "E,B,C,O", 3,1 },
{ "vmse",     0xfffc1fc0, 0x300f0000, 0xb0800480, 0x300a0000, "E,B,C,O", 3,1 },
{ "vdd",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x200b0000, "A,B,C,O", 3,1 },
{ "vde",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x300b0000, "A,B,C,O", 3,1 },
{ "vdsd",     0xfffc1fc0, 0x300f0000, 0xb0800480, 0x200b0000, "E,B,C,O", 3,1 },
{ "vdse",     0xfffc1fc0, 0x300f0000, 0xb0800480, 0x300b0000, "E,B,C,O", 3,1 },

{ "vcf",      0xfffc1fc0, 0x300f0000, 0xb0000040, 0x10000000, "A,B,N,O", 3,1 },
{ "vch",      0xfffc1fc0, 0x300f0000, 0xb0000040, 0x10010000, "A,B,N,O", 3,1 },
{ "vcl",      0xfffc1fc0, 0x300f0000, 0xb0000040, 0x10020000, "A,B,N,O", 3,1 },
{ "vcne",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x10030000, "A,B,N,O", 3,1 },
{ "vce",      0xfffc1fc0, 0x300f0000, 0xb0000040, 0x10040000, "A,B,N,O", 3,1 },
{ "vcnl",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x10050000, "A,B,N,O", 3,1 },
{ "vcnh",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x10060000, "A,B,N,O", 3,1 },
{ "vct",      0xfffc1fc0, 0x300f0000, 0xb0000040, 0x10070000, "A,B,N,O", 3,1 },

{ "vcfd",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x20000000, "A,B,N,O", 3,1 },
{ "vchd",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x20010000, "A,B,N,O", 3,1 },
{ "vcld",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x20020000, "A,B,N,O", 3,1 },
{ "vcned",    0xfffc1fc0, 0x300f0000, 0xb0000040, 0x20030000, "A,B,N,O", 3,1 },
{ "vced",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x20040000, "A,B,N,O", 3,1 },
{ "vcnld",    0xfffc1fc0, 0x300f0000, 0xb0000040, 0x20050000, "A,B,N,O", 3,1 },
{ "vcnhd",    0xfffc1fc0, 0x300f0000, 0xb0000040, 0x20060000, "A,B,N,O", 3,1 },
{ "vctd",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x20070000, "A,B,N,O", 3,1 },

{ "vcfe",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x30000000, "A,B,N,O", 3,1 },
{ "vche",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x30010000, "A,B,N,O", 3,1 },
{ "vcle",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x30020000, "A,B,N,O", 3,1 },
{ "vcnee",    0xfffc1fc0, 0x300f0000, 0xb0000040, 0x30030000, "A,B,N,O", 3,1 },
{ "vcee",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x30040000, "A,B,N,O", 3,1 },
{ "vcnle",    0xfffc1fc0, 0x300f0000, 0xb0000040, 0x30050000, "A,B,N,O", 3,1 },
{ "vcnhe",    0xfffc1fc0, 0x300f0000, 0xb0000040, 0x30060000, "A,B,N,O", 3,1 },
{ "vcte",     0xfffc1fc0, 0x300f0000, 0xb0000040, 0x30070000, "A,B,N,O", 3,1 },

{ "vcsf",     0xfffc1fc0, 0x300f0000, 0xb0000440, 0x10080000, "7,B,N,O", 3,1 },
{ "vcsh",     0xfffc1fc0, 0x300f0000, 0xb0000440, 0x10090000, "7,B,N,O", 3,1 },
{ "vcsl",     0xfffc1fc0, 0x300f0000, 0xb0000440, 0x100a0000, "7,B,N,O", 3,1 },
{ "vcsne",    0xfffc1fc0, 0x300f0000, 0xb0000440, 0x100b0000, "7,B,N,O", 3,1 },
{ "vcse",     0xfffc1fc0, 0x300f0000, 0xb0000440, 0x100c0000, "7,B,N,O", 3,1 },
{ "vcsnl",    0xfffc1fc0, 0x300f0000, 0xb0000440, 0x100d0000, "7,B,N,O", 3,1 },
{ "vcsnh",    0xfffc1fc0, 0x300f0000, 0xb0000440, 0x100e0000, "7,B,N,O", 3,1 },
{ "vcst",     0xfffc1fc0, 0x300f0000, 0xb0000440, 0x100f0000, "7,B,N,O", 3,1 },

{ "vcsfd",    0xfffc1fc0, 0x300f0000, 0xb0800440, 0x20080000, "E,B,N,O", 3,1 },
{ "vcshd",    0xfffc1fc0, 0x300f0000, 0xb0800440, 0x20090000, "E,B,N,O", 3,1 },
{ "vcsld",    0xfffc1fc0, 0x300f0000, 0xb0800440, 0x200a0000, "E,B,N,O", 3,1 },
{ "vcsned",   0xfffc1fc0, 0x300f0000, 0xb0800440, 0x200b0000, "E,B,N,O", 3,1 },
{ "vcsed",    0xfffc1fc0, 0x300f0000, 0xb0800440, 0x200c0000, "E,B,N,O", 3,1 },
{ "vcsnld",   0xfffc1fc0, 0x300f0000, 0xb0800440, 0x200d0000, "E,B,N,O", 3,1 },
{ "vcsnhd",   0xfffc1fc0, 0x300f0000, 0xb0800440, 0x200e0000, "E,B,N,O", 3,1 },
{ "vcstd",    0xfffc1fc0, 0x300f0000, 0xb0800440, 0x200f0000, "E,B,N,O", 3,1 },

{ "vcsfe",    0xfffc1fc0, 0x300f0000, 0xb0800440, 0x30080000, "E,B,N,O", 3,1 },
{ "vcshe",    0xfffc1fc0, 0x300f0000, 0xb0800440, 0x30090000, "E,B,N,O", 3,1 },
{ "vcsle",    0xfffc1fc0, 0x300f0000, 0xb0800440, 0x300a0000, "E,B,N,O", 3,1 },
{ "vcsnee",   0xfffc1fc0, 0x300f0000, 0xb0800440, 0x300b0000, "E,B,N,O", 3,1 },
{ "vcsee",    0xfffc1fc0, 0x300f0000, 0xb0800440, 0x300c0000, "E,B,N,O", 3,1 },
{ "vcsnle",   0xfffc1fc0, 0x300f0000, 0xb0800440, 0x300d0000, "E,B,N,O", 3,1 },
{ "vcsnhe",   0xfffc1fc0, 0x300f0000, 0xb0800440, 0x300e0000, "E,B,N,O", 3,1 },
{ "vcste",    0xfffc1fc0, 0x300f0000, 0xb0800440, 0x300f0000, "E,B,N,O", 3,1 },

{ "vn",	      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x10040000, "A,B,C,O", 3,1 },
{ "vns",      0xfffc1fc0, 0x300f0000, 0xb0400480, 0x10040000, "7,B,C,O", 3,1 },
{ "vnm",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x00040000, "L,M,N,H", 3,1 },
{ "vo",	      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x10050000, "A,B,C,O", 3,1 },
{ "vos",      0xfffc1fc0, 0x300f0000, 0xb0400480, 0x10050000, "7,B,C,O", 3,1 },
{ "vom",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x00050000, "L,M,N,H", 3,1 },
{ "vx",	      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x10060000, "A,B,C,O", 3,1 },
{ "vxs",      0xfffc1fc0, 0x300f0000, 0xb0400480, 0x10060000, "7,B,C,O", 3,1 },
{ "vxm",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x00060000, "L,M,N,H", 3,1 },
{ "vsll",     0xfffc1fc0, 0x300f0000, 0xb0000480, 0x30070000, "7,B,C,O", 3,1 },
{ "vsrl",     0xfffc1fc0, 0x300f0000, 0xb0000480, 0x20070000, "7,B,C,O", 3,1 },
{ "vesll",    0xfffc1fc0, 0x300f0000, 0xb0000000, 0x10070000, "A,B,C,O", 3,1 },
{ "vtm",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x00070000, "M,N,H",   3,1 },

{ "vfm",      0xfffc1fc0, 0x300f0000, 0xb00002c0, 0x000c0000, "L,8,H",   3,1 },
{ "vfx",      0xfffc1fc0, 0x300f0000, 0xb00003c0, 0x100c0000, "A,8,9,O", 3,1 },
{ "vfxd",     0xfffc1fc0, 0x300f0000, 0xb00803c0, 0x200c0000, "A,8,F,O", 3,1 },
{ "vfxe",     0xfffc1fc0, 0x300f0000, 0xb00803c0, 0x300c0000, "A,8,F,O", 3,1 },
{ "vfn",      0xfffc1fc0, 0x300f0000, 0xb00003c0, 0x100e0000, "A,8,9,O", 3,1 },
{ "vfnd",     0xfffc1fc0, 0x300f0000, 0xb00803c0, 0x200e0000, "A,8,F,O", 3,1 },
{ "vfne",     0xfffc1fc0, 0x300f0000, 0xb00803c0, 0x300e0000, "A,8,F,O", 3,1 },
{ "vfmx",     0xfffc1fc0, 0x300f0000, 0xb00002c0, 0x000d0000, "L,8,H",   3,1 },
{ "vfxx",     0xfffc1fc0, 0x300f0000, 0xb00003c0, 0x100d0000, "A,8,9,O", 3,1 },
{ "vfxdx",    0xfffc1fc0, 0x300f0000, 0xb00803c0, 0x200d0000, "A,8,F,O", 3,1 },
{ "vfxex",    0xfffc1fc0, 0x300f0000, 0xb00803c0, 0x300d0000, "A,8,F,O", 3,1 },
{ "vfnx",     0xfffc1fc0, 0x300f0000, 0xb00003c0, 0x100f0000, "A,8,9,O", 3,1 },
{ "vfndx",    0xfffc1fc0, 0x300f0000, 0xb00803c0, 0x200f0000, "A,8,F,O", 3,1 },
{ "vfnex",    0xfffc1fc0, 0x300f0000, 0xb00803c0, 0x300f0000, "A,8,F,O", 3,1 },
{ "vgs",      0xfffc1fc0, 0x300f0000, 0xb0000480, 0x000a0000, "7,C,O",   3,1 },
{ "vgm",      0xfffc1fc0, 0x300f0000, 0xb0000480, 0x00070000, "7,N,H",   3,1 },
{ "vsm",      0xfffc1fc0, 0x300f0000, 0xb00001c0, 0x100a0000, "A,9,O",   3,1 },
{ "vsmd",     0xfffc1fc0, 0x300f0000, 0xb00801c0, 0x200a0000, "A,F,O",   3,1 },
{ "vsme",     0xfffc1fc0, 0x300f0000, 0xb00801c0, 0x300a0000, "A,F,O",   3,1 },
{ "vsmm",     0xfffc1fc0, 0x300f0000, 0xb00001c0, 0x000a0000, "L,9,H",   3,1 },

{ "vls",      0xfffc1fc0, 0x300f0000, 0xb0000480, 0x100e0000, "7,C,O",   3,1 },
{ "vlsd",     0xfffc1fc0, 0x300f0000, 0xb0800480, 0x200e0000, "E,C,O",   3,1 },
{ "vlse",     0xfffc1fc0, 0x300f0000, 0xb0800480, 0x300e0000, "E,C,O",   3,1 },
{ "vts",      0xfffc1fc0, 0x300f0000, 0xb00005c0, 0x10080000, "7,B,9",   3,1 },
{ "vtsd",     0xfffc1fc0, 0x300f0000, 0xb00805c0, 0x20080000, "7,B,F",   3,1 },
{ "vtse",     0xfffc1fc0, 0x300f0000, 0xb00805c0, 0x30080000, "7,B,F",   3,1 },

{ "vvid",     0xfffc1fc0, 0x300f0000, 0xb0000000, 0x20040000, "B,C,O",   3,1 },
{ "vvie",     0xfffc1fc0, 0x300f0000, 0xb0000000, 0x30040000, "B,C,O",   3,1 },
{ "vvdi",     0xfffc1fc0, 0x300f0000, 0xb0000000, 0x20050000, "B,C,O",   3,1 },
{ "vvei",     0xfffc1fc0, 0x300f0000, 0xb0000000, 0x30050000, "B,C,O",   3,1 },
{ "vvde",     0xfffc1fc0, 0x300f0000, 0xb0000000, 0x30060000, "B,C,O",   3,1 },
{ "vved",     0xfffc1fc0, 0x300f0000, 0xb0000000, 0x20060000, "B,C,O",   3,1 },
{ "vlp",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x100e0000, "B,C,O",   3,1 },
{ "vlpd",     0xfffc1fc0, 0x300f0000, 0xb0000000, 0x200e0000, "B,C,O",   3,1 },
{ "vlpe",     0xfffc1fc0, 0x300f0000, 0xb0000000, 0x300e0000, "B,C,O",   3,1 },
{ "vlr",      0xfffc1fc0, 0x300f0000, 0xb0000000, 0x100c0000, "B,C,O",   3,1 },

};

#define NUMOPCODES ((sizeof tachyon_opcodes)/(sizeof *tachyon_opcodes))

