/* @(#)tachyon.c	1.16    7/4/91 */

/* tachyon.c -- Assemble for the TaChyon
   Copyright (C) 1989 Free Software Foundation, Inc.

This file is part of GAS, the GNU Assembler.

GAS is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GAS; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


#include <stdio.h>
#include <ctype.h>

#include "tachyon-opcode.h"
#include "as.h"
#include "frags.h"
#include "struc-symbol.h"
#include "flonum.h"
#include "expr.h"
#include "hash.h"
#include "md.h"
#include "tachyon.h"
#include "write.h"
#include "read.h"
#include "symbols.h"

void md_begin();
void md_end();
void md_number_to_chars();
void md_assemble();
char *md_atof();
void md_convert_frag();
void md_create_short_jump();
void md_create_long_jump();
int  md_estimate_size_before_relax();
void md_number_to_imm();
void md_number_to_disp();
void md_number_to_field();
void md_ri_to_chars();
void emit_relocations();
static void tachyon_ip();
char *malloc();
unsigned int listing_where();
void listing_insn();
void listing_init();
void listing_new();

relax_typeS md_relax_table[] = { 0 };

/* handle of the OPCODE hash table */
static struct hash_control *op_hash = NULL;

static void s_seg(), s_using(),  s_data1(), s_tachyon_align(), s_lcommon(), s_common();
extern void s_globl(), s_long(), s_short(), big_cons(), cons(), float_cons(), s_skip();
extern void pseudo_set();

pseudo_typeS
md_pseudo_table[] = {
    { "align",      s_tachyon_align, 0 },
    { "char",       cons,            1 },
    { "common",     s_common,        0 },
    { "extended",   float_cons,    'x' },
    { "global",     s_globl,         0 },
    { "half",       cons,            2 },
    { "hexa",       big_cons,       16 },
    { "long",       big_cons,        8 },
    { "octa",       big_cons,        8 },
    { "lcommon",    s_lcommon,       0 },
    { "seg",        s_seg,           0 },
    { "skip",       s_skip,          0 },
    { "using",      s_using,         0 },
    { "word",       cons,            4 },
    { NULL,         0,               0 },
};

int md_short_jump_size = 8;
int md_long_jump_size = 8;
int omagic  =  (200 << 16) | OMAGIC;  /* Magic number for header */

/* This array holds the chars that always start a comment.  If the
    pre-processor is disabled, these aren't very useful */
char comment_chars[] = "|!";

/* This array holds the chars that only start a comment at the beginning of
   a line.  If the line seems to have the form '# 123 filename'
   .line and .file directives will appear in the pre-processed output */
/* Note that input_file.c hand checks for '#' at the beginning of the
   first line of the input file.  This is because the compiler outputs
   #NO_APP at the beginning of its output. */
/* Also note that '/*' will always start a comment */
char line_comment_chars[] = "#";

/* Chars that can be used to separate mant from exp in floating point nums */
char EXP_CHARS[] = "eE";

/* Chars that mean this number is a floating point constant */
/* As in 0f12.456 */
/* or    0d1.2345e12 */
char FLT_CHARS[] = "rRsSfFdDxXpP";

/* Also be aware that MAXIMUM_NUMBER_OF_CHARS_FOR_FLOAT may have to be
   changed in read.c .  Ideally it shouldn't have to know about it at all,
   but nothing is ideal around here.
 */
int size_reloc_info = sizeof(struct reloc_info_tachyon);

static unsigned char octal[256];
#define isoctal(c)  octal[c]
static unsigned char toHex[256];

struct tachyon_it {
    char    *error;
    unsigned opcode_u,opcode_l;
    struct nlist *nlistp;
    expressionS exp;
    int pcrel;
    enum reloc_type reloc;
} the_insn[3];

unsigned int final_type,final_opcode_u,final_opcode_l;

#ifdef __STDC__
static void print_insn(struct tachyon_it *insn);
static int getExpression(char *str);
#else
static void print_insn();
static int getExpression();
#endif
static char *expr_end;
static int opcode_size;

#define SPECIAL_CASE_SET    1

/*
 * sort of like s_lcomm
 *
 */
static void
s_lcommon()
{
    char *name;
    char c;
    char *p;
    int temp;
    symbolS *symbolP;

    name = input_line_pointer;
    c = get_symbol_end();
    p = input_line_pointer;
    *p = c;
    SKIP_WHITESPACE();
    if ( * input_line_pointer != ',' ) {
        as_warn("Expected comma after name");
        ignore_rest_of_line();
        return;
    }
    input_line_pointer ++;
    if ((temp = get_absolute_expression()) < 0) {
        as_warn("BSS length (%d.) <0! Ignored.", temp);
        ignore_rest_of_line();
        return;
    }
    *p = 0;
    symbolP = symbol_find_or_make(name);
    *p = c;
    if (strncmp(input_line_pointer, ",\"bss\"", 6) != 0) {
        as_warn("bad .lcommon segment: `%s'", input_line_pointer);
        return;
    }
    input_line_pointer += 6;
    if (symbolP->sy_other == 0 
        && symbolP->sy_desc  == 0
        && ((symbolP->sy_type  == N_BSS
        && symbolP->sy_value == local_bss_counter)
        || ((symbolP->sy_type & N_TYPE) == N_UNDF
        && symbolP->sy_value == 0))) {
            symbolP->sy_value = local_bss_counter;
            symbolP->sy_type  = N_BSS;
            symbolP->sy_frag  = & bss_address_frag;
            local_bss_counter += temp;
    } else {
        as_warn( "Ignoring attempt to re-define symbol from %d. to %d.",
            symbolP->sy_value, local_bss_counter );
    }
    demand_empty_rest_of_line();
    return;
}

static void
s_common()
{
    register char *name;
    register char c;
    register char *p;
    register int temp;
    register symbolS *  symbolP;

    name = input_line_pointer;
    c = get_symbol_end();
    /* just after name is now '\0' */
    p = input_line_pointer;
    *p = c;
    SKIP_WHITESPACE();
    if ( * input_line_pointer != ',' ) {
        as_warn("Expected comma after symbol-name");
        ignore_rest_of_line();
        return;
    }
    input_line_pointer ++; /* skip ',' */
    if ( (temp = get_absolute_expression ()) < 0 ) {
        as_warn(".COMMon length (%d.) <0! Ignored.", temp);
        ignore_rest_of_line();
        return;
    }
    *p = 0;
    symbolP = symbol_find_or_make (name);
    *p = c;
    if (   (symbolP->sy_type & N_TYPE) != N_UNDF ||
        symbolP->sy_other != 0 || symbolP->sy_desc != 0) {
        as_warn( "Ignoring attempt to re-define symbol");
        ignore_rest_of_line();
        return;
    }
    if (symbolP->sy_value) {
        if (symbolP->sy_value != temp) {
            as_warn( "Length of .comm \"%s\" is already %d. Not changed to %d.",
                symbolP->sy_name, symbolP->sy_value, temp);
        }
    } else {
        symbolP->sy_value = temp;
        symbolP->sy_type |= N_EXT;
    }
    know(symbolP->sy_frag == &zero_address_frag);
    if (strncmp(input_line_pointer, ",\"bss\"", 6) != 0) {
        as_warn("bad .common segment: `%s'", input_line_pointer);
        return;
    }
    input_line_pointer += 6;
    demand_empty_rest_of_line();
    return;
}

struct using_table { int inuse; symbolS *symbolP; } USING_TABLE[32];

static void
s_using()
{
        char *name;
        char delim;
        char *end_name;
        symbolS *symbolP;
	int tmp;
 
        name = input_line_pointer;
        delim = get_symbol_end();
        end_name = input_line_pointer;
        *end_name = delim;
        SKIP_WHITESPACE();
        if ( * input_line_pointer != ',' ) {
                *end_name = 0;
                as_warn("Expected comma after $gr_name \"%s\"", name);
                *end_name = delim;
                ignore_rest_of_line();
                return;
        }
        input_line_pointer ++;
        *end_name = 0;
	if (*name != '$' || *(name+1) != 'g' || *(name+2) != 'r') {
                as_warn("Expected $gr_name \"%s\"", name);
                *end_name = delim;
                ignore_rest_of_line();
                return;
	}
	tmp = -1;
	sscanf(name+3, "%d", &tmp);
	if (tmp < 0 || tmp > 31) {
                as_warn("Illegal $gr_name \"%s\" (must be $gr0-$gr31)", name);
                *end_name = delim;
                ignore_rest_of_line();
                return;
	}
        symbolP = symbol_find_or_make (name);
        *end_name = delim;
        pseudo_set (symbolP);
	if (symbolP->sy_forward)
                as_warn("undefined symbol");
	else {
		USING_TABLE[tmp].inuse   = 1;
		USING_TABLE[tmp].symbolP = symbolP;
	}
        demand_empty_rest_of_line ();
}

static void
s_seg()
{

    if (strncmp(input_line_pointer, "\"text\"", 6) == 0) {
        input_line_pointer += 6;
        s_text();
        return;
    }
    if (strncmp(input_line_pointer, "\"data\"", 6) == 0) {
        input_line_pointer += 6;
        s_data();
        return;
    }
    if (strncmp(input_line_pointer, "\"data1\"", 7) == 0) {
        input_line_pointer += 7;
        s_data1();
        return;
    }
    as_warn("Unknown segment type");
    demand_empty_rest_of_line();
    return;
}

static void
s_data1()
{
    subseg_new(SEG_DATA, 1);
    demand_empty_rest_of_line();
    return;
}

static void
s_tachyon_align()
{
    register unsigned int temp;
    register long int temp_fill;
    unsigned int i;

    temp = get_absolute_expression ();
#define MAX_ALIGNMENT (1 << 16)
    if ( temp > MAX_ALIGNMENT ) {
        as_warn("Alignment too large: %d. assumed.", temp = MAX_ALIGNMENT);
    }

    /*
     * For the tachyon, `.align (1<<n)' actually means `.align n'
     * so we have to convert it.
     */
    if (temp != 0) {
        for (i = 0; (temp & 1) == 0; temp >>= 1, ++i)
            ;
    }
    if (temp != 1) {
        as_warn("Alignment not a power of 2");
    }
    temp = i;
    if (*input_line_pointer == ',') {
        input_line_pointer ++;
        temp_fill = get_absolute_expression ();
    } else {
        temp_fill = 0;
    }
    /* Only make a frag if we HAVE to. . . */
    if (temp && ! need_pass_2) {
        frag_align (temp, (int)temp_fill);
    }
    demand_empty_rest_of_line();
    return;
}

/*
 * This function is called once, at assembler startup time.  This should
 * set up all the tables, etc that the MD part of the assembler needs
 */
void
md_begin()
{
    register char *retval = NULL;
    register int i;
    int j = 0;

    listing_init();

    if ((op_hash = hash_new()) == NULL)
        as_fatal("Virtual memory exhausted");

    for (i = 0; i < NUMOPCODES; ++i) {
        if ((~tachyon_opcodes[i].mask_u & tachyon_opcodes[i].match_u)
         || (~tachyon_opcodes[i].mask_l & tachyon_opcodes[i].match_l)) {
            printf("bad opcode - `%s %s'\n",
                tachyon_opcodes[i].name, tachyon_opcodes[i].args);
            ++j;
        }
    }

    if (j)
        exit(1);

    for (i = 0; i < NUMOPCODES; ++i) {
      retval = hash_insert(op_hash, tachyon_opcodes[i].name, &tachyon_opcodes[i]);
      if(retval && *retval) {
          as_fatal("Internal Error:  Can't hash %s: %s",
            tachyon_opcodes[i].name, retval);
      }
      while (!tachyon_opcodes[i].last)
          ++i;
    }
    for (i = '0'; i < '8'; ++i)
        octal[i] = 1;
    for (i = '0'; i <= '9'; ++i)
        toHex[i] = i - '0';
    for (i = 'a'; i <= 'f'; ++i)
        toHex[i] = i + 10 - 'a';
    for (i = 'A'; i <= 'F'; ++i)
        toHex[i] = i + 10 - 'A';
    return;
}

void
md_end()
{
    return;
}

comp_op_all(secondary, op_typev, type) int secondary; int op_typev, type;
{
    register int i;

    if (secondary) {
      for (i=0; i<16; i++)
        if (op_type_OP0_OP1   [i][op_typev] == type)
          return(1);
      for (i=0; i<16; i++)
        if (op_type_OP2_bimm12[i][op_typev] == type)
          return(1);
      for (i=0; i<16; i++)
        if (op_type_OP2_fmul  [i][op_typev] == type)
          return(1);
      return (0);
    }
    else {
      for (i=0; i<16; i++)
        if (op_type_OP0_OP1   [op_typev][i] == type)
          return(1);
      for (i=0; i<16; i++)
        if (op_type_OP2_bimm12[op_typev][i] == type)
          return(1);
      for (i=0; i<16; i++)
        if (op_type_OP2_fmul  [op_typev][i] == type)
          return(1);
      return (0);
    }
}

comp_op(secondary, op_array, op_typev, type)
    int  secondary; int (*op_array)[16]; int op_typev, type;
{
    register int i;

    if (secondary) {
      for (i=0; i<16; i++)
        if (op_array [i][op_typev] == type)
          return(1);
      return (0);
    }
    else {
      for (i=0; i<16; i++)
        if (op_array [op_typev][i] == type)
          return(1);
      return (0);
    }
}

seek_op_all(secondary, op_typev) int secondary; int op_typev;
{
    register int i, found;

    if (secondary) {
      for (i=0; i<16; i++)
        if ((found = op_type_OP0_OP1   [i][op_typev]) >= TYPE_0)
          return(found);
      for (i=0; i<16; i++)
        if ((found = op_type_OP2_bimm12[i][op_typev]) >= TYPE_0)
          return(found);
      for (i=0; i<16; i++)
        if ((found = op_type_OP2_fmul  [i][op_typev]) >= TYPE_0)
          return(found);
      return(-1);
    }
    else { 
      for (i=0; i<16; i++)
        if ((found = op_type_OP0_OP1   [op_typev][i]) >= TYPE_0)
          return(found);
      for (i=0; i<16; i++)
        if ((found = op_type_OP2_bimm12[op_typev][i]) >= TYPE_0)
          return(found);
      for (i=0; i<16; i++)
        if ((found = op_type_OP2_fmul  [op_typev][i]) >= TYPE_0)
          return(found);
      return(-1);
    }
}

seek_op(secondary, op_array, op_typev)
    int  secondary; int (*op_array)[16]; int op_typev;
{
    register int i, found;

    if (secondary) {
      for (i=0; i<16; i++)
        if ((found = op_array [i][op_typev]) >= TYPE_0)
          return(found);
      return(-1);
    }
    else { 
      for (i=0; i<16; i++)
        if ((found = op_array [op_typev][i]) >= TYPE_0)
          return(found);
      return(-1);
    }
}

void
md_assemble(str,type,unit,stat,loc)
    char *str;
    int type;/* current insn type (0-15:specified  16:any  256:illegal) */
    int unit;/* current insn unit-id (0-15:specified  16:any  256:illegal) */
    int stat;/* 0:current op is first & has successor.
                1:current op is first & has no successor.
                2:current op is not first & has successor.
                3:current op is not first & has no successor. */
    int loc; /* described location of operation(0,1,2) 3:overflow */
{
    static char *toP;
    int i,j,found;
    int op_indr[3];
    int op_type[3];
    int tmp;

/* fprintf(stderr, "type=%d, stat=%d, loc=%d : %s\n",type,stat,loc,str); */
    if (!(stat & 2)) {
      op_indr[0] = 0; op_indr[1] = 1; op_indr[2] = 2;
      op_type[0] =    op_type[1] =    op_type[2] = OPER_NOP;
      final_type = TYPE_0;

      if (flagseen['l'])
        listing_new((unsigned char *) NULL, (unsigned char *) NULL, 8, 0, 1);
    }

    /* create the_insn[loc] */
    if (loc < 3) {
      if (flagseen['l']) {
        if (str == NULL)
	  str = "nop";
        listing_cur->str[loc] = malloc(strlen(str)+1);
        strcpy(listing_cur->str[loc], str);
      }
      tachyon_ip(str,loc);
      op_type[loc] = the_insn[loc].opcode_u >> 28;
    }

    if (!(stat & 1))
      return;

    toP = frag_more(8);
    if (flagseen['l'])
      listing_bufmod(frag_base(), (unsigned char *) toP);

/******* all OP is nop *******/
    if (!op_type[0] && !op_type[1] && !op_type[2]) {
      if (type <= TYPE_F)
        final_type = type;
      else
        final_type = TYPE_7;
      goto comp_com;
    }

/******* only one OP is specified *******/
    /* case of single OP0 in instruction */
    if (op_type[0] && !op_type[1] && !op_type[2]) {
      switch (tmp = op_single_loc[op_type[0]]) {
      case 1:
        op_indr[0] = 1;
        op_indr[1] = 0;
      case 0:
      case -1:
        found = 0;
        if (type <= TYPE_F) {
          type &= TYPE_F;
          found = comp_op_all(tmp == 1, op_type[0], type);
          if (!found)
            as_warn("@type mismatch(single op0): @%X is ignored", type);
          else
            final_type = type;
        }
        if (!found) {
          if (op_type[0] == OPER_JMP) { /* force jmp(type2)->jmp(type8) */
            op_indr[0] = 1;
            op_indr[1] = 0;
	    found = 8;
	  }
	  else
            found = seek_op_all(tmp == 1, op_type[0]);
          if (found < TYPE_0) {
            as_warn("Illegal location of op0(single op0)");
            op_type[0] = OPER_NOP;
            listing_cur->str[0] = NULL;
          }
          else
            final_type = found;
        }
        break;
      case 2:
        op_indr[0] = 2;
        op_indr[2] = 0;
        found = 0;
        if (type <= TYPE_F) {
          type &= TYPE_F;
          found = (type == TYPE_6 || type == TYPE_7);
          if (!found)
            as_warn("@type mismatch(single op0): @%X is ignored", type);
          else
            final_type = type;
        }
        if (!found)
          final_type = TYPE_6;
        break;
      default:
        as_warn("Illegal location of op0(single op0)");
        op_type[0] = OPER_NOP;
        listing_cur->str[0] = NULL;
        break;
      }
      goto comp_com;
    }

    /* case of single OP1 in instruction */
    if (!op_type[0] && op_type[1] && !op_type[2]) {
      switch (op_single_loc[op_type[1]]) {
      case 1:
      case -1:
	if (op_type[1] == OPER_FLO_M) {
          op_indr[1] = 2;
          op_indr[2] = 1;
	  goto op2_active;
        }
        found = 0;
        if (type <= TYPE_F) {
          type &= TYPE_F;
          found = comp_op_all(1, op_type[1], type);
          if (!found)
            as_warn("@type mismatch(single op1): @%X is ignored", type);
          else
            final_type = type;
        }
        if (!found) {
          found = seek_op_all(1, op_type[1]);
          if (found < TYPE_0) {
            as_warn("Illegal location of op1(single op1)");
            op_type[1] = OPER_NOP;
            listing_cur->str[1] = NULL;
          }
          else
            final_type = found;
        }
        break;
      case 2:
        op_indr[1] = 2;
        op_indr[2] = 1;
        found = 0;
        if (type <= TYPE_F) {
          type &= TYPE_F;
          found = (type == TYPE_6 || type == TYPE_7);
          if (!found)
            as_warn("@type mismatch(single op1): @%X is ignored", type);
          else
            final_type = type;
        }
        if (!found)
          final_type = TYPE_6;
        break;
      default:
        as_warn("Illegal location of op1(single op1)");
        op_type[1] = OPER_NOP;
        listing_cur->str[1] = NULL;
        break;
      }
      goto comp_com;
    }

    /* case of single OP2 in instruction */
op2_active:
    if (!op_type[op_indr[0]] && !op_type[op_indr[1]] && op_type[op_indr[2]]) {
      if (!op_avail_OP2[op_type[op_indr[2]]]) {
        as_warn("Illegal op2(single op2)");
        op_type[op_indr[2]] = OPER_NOP;
        listing_cur->str[op_indr[2]] = NULL;
        goto comp_com;
      }
      switch (op_single_loc[op_type[op_indr[2]]]) {
      case -1:
        found = 0;
        if (type <= TYPE_F) {
          type &= TYPE_F;
          found = ((op_type[op_indr[2]] == OPER_FLO_A && (type == TYPE_4 || type == TYPE_5)) || 
                   (op_type[op_indr[2]] == OPER_FLO_M && (type >= TYPE_2 && type <= TYPE_5)));
          if (!found)
            as_warn("@type mismatch(single op2): @%X is ignored", type);
          else
            final_type = type;
        }
        if (!found)
          final_type = TYPE_4;
        break;
      case 2:
        found = 0;
        if (type <= TYPE_F) {
          type &= TYPE_F;
          found = (type == TYPE_6 || type == TYPE_7);
          if (!found)
            as_warn("@type mismatch(single op2): @%X is ignored", type);
          else
            final_type = type;
        }
        if (!found)
          final_type = TYPE_6;
        break;
      default:
        as_warn("Illegal location of op2(single op2)");
        op_type[op_indr[2]] = OPER_NOP;
        listing_cur->str[op_indr[2]] = NULL;
        break;
      }
      goto comp_com;
    }

/******* OP0 & OP1 are specified *******/
    /* case of OP0 & OP1 are active */
    if (op_type[0] && op_type[1]) {
      if (!op_avail_OP2[op_type[2]]) {
        as_warn("Illegal op2(op0/op1)");
        op_type[0] = op_type[1] = op_type[2] = OPER_NOP;
        listing_cur->str[0] = listing_cur->str[1] = listing_cur->str[2] = NULL;
        goto comp_com;
      }
      switch (op_single_loc[op_type[0]]) {
      case 1:
        if (op_avail_OP2[op_type[1]] && op_type[2] == OPER_NOP) {
          op_indr[0] = 2;
          op_indr[1] = 0;
          op_indr[2] = 1;
          goto op1_op2_active;
        }
        else {
          as_warn("Collision op0 & op1 & op2(op0/op1)");
          op_type[0] = op_type[1] = op_type[2] = OPER_NOP;
          listing_cur->str[0] = listing_cur->str[1] = listing_cur->str[2] = NULL;
          break;
        }
	break;
      case 2:
        as_warn("Flying op0 over op1(op0/op1)");
        op_type[0] = op_type[1] = op_type[2] = OPER_NOP;
        listing_cur->str[0] = listing_cur->str[1] = listing_cur->str[2] = NULL;
        break;
      case 0:
      case -1:
        if (op_type[2] == OPER_NOP) {
	  switch (op_type[1]) {
	  case OPER_FLO_M: /* & fmult */
            if (op_type[0] == OPER_FLO_A) { /* fadd & fmult */
              op_indr[0] = 2;
              op_indr[1] = 0;
              op_indr[2] = 1;
              goto op1_op2_active;
	    }
            else { /* & fmult */
              op_indr[1] = 2;
              op_indr[2] = 1;
              goto op0_op2_active;
            }
	    break;
          case OPER_BRCI12: /* & branch imm12 */
            if (op_type[0] == OPER_LSM) { /* l/s/m & brcs */
              op_indr[0] = 2;
              op_indr[1] = 0;
              op_indr[2] = 1;
              goto op1_op2_active;
	    }
	    else {
              op_indr[1] = 2;
              op_indr[2] = 1;
              goto op0_op2_active;
	    }
	    break;
          }
	}

        /* OP0 & OP1 are really active */
        found = 0;
        if (type <= TYPE_F) {
          type &= TYPE_F;
          switch (op_type[2]) {
          case OPER_BRCI12:
            found = (op_type_OP2_bimm12[op_type[0]][op_type[1]] == type);
	    break;
	  case OPER_FLO_A:
            found = (op_type_OP2_fadd  [op_type[0]][op_type[1]] == type);
	    break;
          case OPER_FLO_M:
            found = (op_type_OP2_fmul  [op_type[0]][op_type[1]] == type);
	    break;
          default: /* case 0 */
            if (!(found = (op_type_OP0_OP1   [op_type[0]][op_type[1]] == type)))
            if (!(found = (op_type_OP2_bimm12[op_type[0]][op_type[1]] == type)))
                  found = (op_type_OP2_fmul  [op_type[0]][op_type[1]] == type);
	    break;
          }
          if (!found)
            as_warn("@type mismatch(op0/op1): @%X is ignored", type);
          else
            final_type = type;
        }
        if (!found) {
          switch (op_type[2]) {
          case OPER_BRCI12:
            found = op_type_OP2_bimm12[op_type[0]][op_type[1]];
	    break;
	  case OPER_FLO_A:
            found = op_type_OP2_fadd [op_type[0]][op_type[1]];
	    break;
          case OPER_FLO_M:
            found = op_type_OP2_fmul [op_type[0]][op_type[1]];
	    break;
          default: /* case 0 */
            if ((found = op_type_OP0_OP1   [op_type[0]][op_type[1]]) < TYPE_0)
            if ((found = op_type_OP2_bimm12[op_type[0]][op_type[1]]) < TYPE_0)
                 found = op_type_OP2_fmul  [op_type[0]][op_type[1]];
	    break;
          }
          if (found < TYPE_0) {
            as_warn("Illegal combination of op0 & op1 & op2(op0/op1)");
            op_type[0] = op_type[1] = op_type[2] = OPER_NOP;
            listing_cur->str[0] = listing_cur->str[1] = listing_cur->str[2] = NULL;
          }
          else
            final_type = found;
        }
        break;
      }
      goto comp_com;
    }

    /* case of OP0 & OP2 are active */
op0_op2_active:
    if (op_type[op_indr[0]] && !op_type[op_indr[1]] && op_type[op_indr[2]]) {
      if (op_single_loc[op_type[op_indr[0]]] <= 0 && op_avail_OP2[op_type[op_indr[2]]]) {
        found = 0;
        if (type <= TYPE_F) {
          type &= TYPE_F;
          switch (op_type[op_indr[2]]) {
          case OPER_BRCI12:
            found = comp_op(0, op_type_OP2_bimm12, op_type[op_indr[0]], type);
	    break;
	  case OPER_FLO_A:
            found = comp_op(0, op_type_OP2_fadd,   op_type[op_indr[0]], type);
	    break;
          default:
            found = comp_op(0, op_type_OP2_fmul,   op_type[op_indr[0]], type);
	    break;
	  }
          if (!found)
            as_warn("@type mismatch(op0/op2): @%X is ignored", type);
          else
            final_type = type;
        }
        if (!found) {
          switch (op_type[op_indr[2]]) {
          case OPER_BRCI12:
            found = seek_op(0, op_type_OP2_bimm12, op_type[op_indr[0]]);
	    break;
	  case OPER_FLO_A:
            found = seek_op(0, op_type_OP2_fadd,   op_type[op_indr[0]]);
	    break;
          default:
            found = seek_op(0, op_type_OP2_fmul,   op_type[op_indr[0]]);
	    break;
	  }
          if (found < TYPE_0) {
            as_warn("Illegal combination of op%d & op%d(op0/op2)", op_indr[0], op_indr[2]);
            op_type[op_indr[0]] = op_type[op_indr[2]] = OPER_NOP;
            listing_cur->str[op_indr[0]] = listing_cur->str[op_indr[2]] = NULL;
          }
          else
            final_type = found;
        }
      }
      else {
        as_warn("Illegal combination of op%d & op%d(op0/op2)", op_indr[0], op_indr[2]);
        op_type[op_indr[0]] = op_type[op_indr[2]] = OPER_NOP;
        listing_cur->str[op_indr[0]] = listing_cur->str[op_indr[2]] = NULL;
      }
      goto comp_com;
    }

    /* case of OP1 & OP2 are active */
op1_op2_active:
    if (!op_type[op_indr[0]] && op_type[op_indr[1]] && op_type[op_indr[2]]) {
      if (((tmp = op_single_loc[op_type[op_indr[1]]]) == 1 || tmp < 0) && op_avail_OP2[op_type[op_indr[2]]]) {
        found = 0;
        if (type <= TYPE_F) {
          type &= TYPE_F;
          switch (op_type[op_indr[2]]) {
          case OPER_BRCI12:
            found = comp_op(1, op_type_OP2_bimm12, op_type[op_indr[1]], type);
            break;
          case OPER_FLO_A:
            found = comp_op(1, op_type_OP2_fadd,   op_type[op_indr[1]], type);
            break;
          default:
            found = comp_op(1, op_type_OP2_fmul,   op_type[op_indr[1]], type);
            break;
	  }
          if (!found)
            as_warn("@type mismatch(OP1/OP2): @%X is ignored", type);
          else
            final_type = type;
        }
        if (!found) {
          switch (op_type[op_indr[2]]) {
          case OPER_BRCI12:
            found = seek_op(1, op_type_OP2_bimm12, op_type[op_indr[1]]);
   	    break;
	  case OPER_FLO_A:
            found = seek_op(1, op_type_OP2_fadd,   op_type[op_indr[1]]);
   	    break;
          default:
            found = seek_op(1, op_type_OP2_fmul,   op_type[op_indr[1]]);
   	    break;
	  }
          if (found < TYPE_0) {
            as_warn("Illegal combination of op%d & op%d(op1/op2)", op_indr[1], op_indr[2]);
            op_type[op_indr[1]] = op_type[op_indr[2]] = OPER_NOP;
            listing_cur->str[op_indr[1]] = listing_cur->str[op_indr[2]] = NULL;
          }
          else
            final_type = found;
        }
      }
      else {
        as_warn("Illegal combination of op%d & op%d(op1/op2)", op_indr[1], op_indr[2]);
        op_type[op_indr[1]] = op_type[op_indr[2]] = OPER_NOP;
        listing_cur->str[op_indr[1]] = listing_cur->str[op_indr[2]] = NULL;
      }
      goto comp_com;
    }

/******* assume combination completed *******/
comp_com:
    /* exchange OP0 & OP1 if op_type is 0, 6 or a */
    if (type_exchange_loc[final_type]) {
      tmp = op_indr[0]; op_indr[0] = op_indr[1]; op_indr[1] = tmp;
    }

    /* put out the symbol-dependent stuff */
    for (i=0; i<3; i++) {
      if (op_type[op_indr[i]] && the_insn[op_indr[i]].reloc != NO_RELOC) {
        if (the_insn[op_indr[i]].reloc == RELOC_5) {
          switch (i) {
          case 0:
            the_insn[op_indr[i]].reloc = RELOC0_5;
            break;
          case 1:
            the_insn[op_indr[i]].reloc = RELOC1_5;
            break;
          default:
            as_warn("Illegal reloc");
            break;
          }
        }
        else if (the_insn[op_indr[i]].reloc == RELOC_U5) {
          switch (i) {
          case 0:
            the_insn[op_indr[i]].reloc = RELOC0_U5;
            break;
          case 1:
            the_insn[op_indr[i]].reloc = RELOC1_U5;
            break;
          default:
            as_warn("Illegal reloc");
            break;
          }
        }
        fix_new(
            frag_now,                     /* which frag */
            (toP - frag_now->fr_literal), /* where */
            8,                            /* size */
            the_insn[op_indr[i]].exp.X_add_symbol,
            the_insn[op_indr[i]].exp.X_subtract_symbol,
            the_insn[op_indr[i]].exp.X_add_number,
            the_insn[op_indr[i]].pcrel,
            the_insn[op_indr[i]].reloc
        );
      }
    }

    /* put out the opcode */
    if (op_type[op_indr[0]]) {
      final_opcode_u = final_type << 28 |
                       (the_insn[op_indr[0]].opcode_u & 0x0fffffff);
      final_opcode_l =  the_insn[op_indr[0]].opcode_l;
    }
    else {
      final_opcode_u = final_type << 28;
      final_opcode_l = 0;
    }
    if (op_type[op_indr[1]]) {
      final_opcode_u |=(the_insn[op_indr[1]].opcode_u & 0x0fffffff) >> 20;
      final_opcode_l |=(the_insn[op_indr[1]].opcode_u & 0x0fffffff) << 12
                     | (the_insn[op_indr[1]].opcode_l >> 20);
    }
    if (op_type[op_indr[2]]) {
      final_opcode_l |=(the_insn[op_indr[2]].opcode_u & 0x0fffffff) >> 8;
    }

    if (flagseen['l']) {
      listing_cur->indr[0] = op_indr[0];
      listing_cur->indr[1] = op_indr[1];
      listing_cur->indr[2] = op_indr[2];
    }

    if (unit < 16) {
      if (final_type == TYPE_F)
        final_opcode_u |= unit << 24;
      else
        as_warn("@.unit-id is ignored");
    }

    md_number_to_chars(toP  , final_opcode_u, 4);
    md_number_to_chars(toP+4, final_opcode_l, 4);
}

static void
tachyon_ip(str,loc)
    char *str; int loc;
{
    char *s;
    const char *args;
    char c;
    unsigned long i;
    struct tachyon_opcode *insn;
    char *argsStart;
    unsigned int mask;
    unsigned int opcode_u,opcode_l;
    int match = FALSE;
    int comma = 0;

    for (s = str; islower(*s) || isdigit(*s) || *s == '.'; ++s)
        ;
    switch (*s) {

    case '\0':
        break;

    case ' ':
        *s++ = '\0';
        break;

    default:
            as_warn("Unknown opcode: `%s'", str);
            exit(1);
    }
    if ((insn = (struct tachyon_opcode *) hash_find(op_hash, str)) == NULL) {
        as_warn("`%s' not in hash table.", str);
        exit(1);
    }
    argsStart = s;
    for (;;) {
        opcode_u = insn->match_u;
        opcode_l = insn->match_l;
        bzero(&the_insn[loc], sizeof(the_insn[loc]));
        the_insn[loc].reloc = NO_RELOC;

        /*
         * Build the opcode, checking as we go to make
         * sure that the operands match
         */
        for (args = insn->args; ; ++args) {
            switch (*args) {

            case '\0':  /* end of args */
                if (*s == '\0') {
                    match = TRUE;
                }
                break;

            case ',':
            case ' ':
                if (*s++ == *args)
                    continue;
                break;

            case '#':   /* must be at least one digit */
                if (isdigit(*s++)) {
                    while (isdigit(*s)) {
                        ++s;
                    }
                    continue;
                }
                break;

            case 's':   /* integer/floating state register (20 bit OP)*/
            case 't':   /* integer/floating state register (40 bit OP)*/
 /*
  * +-+---+------------+---+
  * |4|opc|        I12 | s |
  * +-+---+------------+---+
  *                      ^- (mask & 0x1f) << 8 -> opcode_u
  * +-+---+---------------------+---+
  * |4|opc|        I32          | t |
  * +-+---+---------------------+---+
  *                               ^- (mask & 0x1f) << 20 -> opcode_l
  */
                if (strncmp(s, "true", 4) == 0)        { s += 4; mask = 0; }
                else if (strncmp(s, "oz", 2) == 0)     { s += 2; mask = 8; }
                else if (strncmp(s, "iccp.", 5) == 0)  { s += 5;
                    if      (*s == 'z')                { s ++;   mask = 1; }
                    else if (*s == 'n')                { s ++;   mask = 6; }
                    else if (*s == 'v')                { s ++;   mask = 7; }
                    else if (*s == 'c')                { s ++;   mask = 4; }
                    else if (strncmp(s, "le", 2) == 0) { s += 2; mask = 3; }
                    else if (*s == 'l')                { s ++;   mask = 2; }
                    else if (strncmp(s, "ule",3) == 0) { s += 3; mask = 5; }
                    else if (strncmp(s, "ul", 2) == 0) { s += 2; mask = 4; }
                    else
                      break;
                }
                else if (strncmp(s, "iccs.", 5) == 0)  { s += 5;
                    if      (*s == 'z')                { s ++;   mask = 9; }
                    else if (*s == 'n')                { s ++;   mask =14; }
                    else if (*s == 'v')                { s ++;   mask =15; }
                    else if (*s == 'c')                { s ++;   mask =12; }
                    else if (strncmp(s, "le", 2) == 0) { s += 2; mask =11; }
                    else if (*s == 'l')                { s ++;   mask =10; }
                    else if (strncmp(s, "ule",3) == 0) { s += 3; mask =13; }
                    else if (strncmp(s, "ul", 2) == 0) { s += 2; mask =12; }
                    else
                      break;
                }
                else if (strncmp(s, "fcc.", 4) == 0)   { s += 4;
                    if      (*s == 'e')                { s ++;   mask =16; }
                    else if (strncmp(s, "ue", 2) == 0) { s += 2; mask =22; }
                    else if (*s == 'u')                { s ++;   mask =19; }
                    else if (strncmp(s, "le", 2) == 0) { s += 2; mask =20; }
                    else if (*s == 'l')                { s ++;   mask =17; }
                    else if (strncmp(s, "ge", 2) == 0) { s += 2; mask =21; }
                    else if (*s == 'g')                { s ++;   mask =18; }
                    else
                      break;
                }
                else if (isdigit(*s)) {
                    mask = *s++;
                    if (isdigit(*s)) {
                        mask = 10 * (mask - '0') + (*s++ - '0');
                        if (mask >= 32) {
                            break;
                        }
                    } else {
                        mask -= '0';
                    }
                }
                else
                    break;
                switch (*args) {
                case 's':
                    opcode_u |= mask << 8;
                    continue;
                case 't':
                    opcode_l |= mask << 20;
                    continue;
                }
                break;

            case 'c':   /* coprocessor state register */
            case 'd':   /* coprocessor state register */
                if (strncmp(s, "asa", 3) == 0)         { s += 3;
                    if      (*s == 'm')                { s ++;   mask = 1; }
                    else if (*s == 'f')                { s ++;   mask = 2; }
                    else if (*s == 'u')                { s ++;   mask = 3; }
                    else			       {         mask = 0; }
                }
                else if (strncmp(s, "es", 2) == 0)     { s += 2;
                    if      (*s == '0')                { s ++;   mask = 8; }
                    else if (*s == '1')                { s ++;   mask = 9; }
                    else if (*s == '2')                { s ++;   mask =10; }
                    else if (*s == '3')                { s ++;   mask =11; }
                    else
                      break;
                }
                else if (strncmp(s, "cb", 2) == 0)     { s += 2;
                    if      (*s == 'a')                { s ++;   mask =12; }
                    else if (*s == 'u')                { s ++;   mask =13; }
                    else if (*s == 's')                { s ++;   mask =14; }
                    else if (*s == '0')                { s ++;   mask =16; }
                    else if (*s == '1')                { s ++;
		       if   (*s >= '0' && *s <= '5')   { mask=26+*s++-'0'; }
		       else 			       {         mask =17; }}
		    else if (*s == '2')		       { s ++;   mask =18; }
		    else if (*s == '3')		       { s ++;   mask =19; }
		    else if (*s == '4')		       { s ++;   mask =20; }
		    else if (*s == '5')		       { s ++;   mask =21; }
		    else if (*s == '6')		       { s ++;   mask =22; }
		    else if (*s == '7')		       { s ++;   mask =23; }
		    else if (*s == '8')		       { s ++;   mask =24; }
		    else if (*s == '9')		       { s ++;   mask =25; }
                    else
                      break;
                }
                else if (isdigit(*s)) {
                    mask = *s++;
                    if (isdigit(*s)) {
                        mask = 10 * (mask - '0') + (*s++ - '0');
                        if (mask < 32 || mask >= 64) {
                            break;
                        }
                    } else {
                        break;
                    }
                }
                else
                    break;
                switch (*args) {
                case 'c':
                    opcode_u |= mask << 8;
                    continue;
                case 'd':
                    opcode_l |= mask << 20;
                    continue;
                }
                break;

            case '1':   /* next operand must be a GPR */
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
            case 'y':
 /*
  *             v- (mask & 0x1f) << 18 -> opcode_u
  * +-+-----+----+----+----+
  * |4| opc | r1 | r2 | r3 |
  * +-+-----+----+----+----+
  *                ^- (mask & 0x1f) << 13 -> opcode_u
  * +-+---+-----------+----+
  * |4|opc|        I12| r3 |
  * +-+---+-----------+----+
  *                     ^- (mask & 0x1f) << 8 -> opcode_u
  * +-+-----+----+-------------+----+
  * |4| opc | r1 | I25         | r4 |
  * +-+-----+----+-------------+----+
  *                              ^- (mask & 0x1f) << 20 -> opcode_l
  * +-+-----+----------------------------+----+
  * |4| opc |                            | r5 |
  * +-+-----+----------------------------+----+
  *                                        ^- (mask & 0x1f) -> opcode_l
  * +-+-----+----++----++----+-----------+----+
  * |4| opc | r1 || r3 || ry |    I25    | r5 |
  * +-+-----+----++----++----+-----------+----+
  *                         ^- (mask & 0x1f) << 30 -> opcode_l
  *                       ^- (mask & 0x1f) >> 2 -> opcode_u
  * +-+-----+----+----+++----+-+----+----+----+
  * |4| uid |    | r6 ||| r7 | | r8 |    | r9 |
  * +-+-----+----+----+++----+-+----+----+----+
  *                        ^- (mask & 0x1f) << 30 -> opcode_l
  *                       ^- (mask & 0x1f) >> 2 -> opcode_u
  */
                if (*s++ == 'g' && *s++ == 'r' && isdigit(*s)) {
                    mask = *s++;
                    if (isdigit(*s)) {
                        mask = 10 * (mask - '0') + (*s++ - '0');
                        if (mask >= 32) {
                            break;
                        }
                    } else {
                        mask -= '0';
                    }
                    switch (*args) {
                    case '1':
                        opcode_u |= mask << 18;
                        continue;
                    case '2':
                    case '6':
                        opcode_u |= mask << 13;
                        continue;
                    case '3':
                        opcode_u |= mask << 8;
                        continue;
                    case '4':
                    case '8':
                        opcode_l |= mask << 20;
                        continue;
                    case '5':
                    case '9':
                        opcode_l |= mask;
                        continue;
                    case '7':
                    case 'y':
                        opcode_u |= mask >> 2;
                        opcode_l |= mask << 30;
                        continue;
                    }
                }
                break;

            case 'e':    /* next operand is a floating point register */
            case 'f':
            case 'g':
            case 'h':
	    case 'E':
	    case 'F':
                if (*s++ == 'f' && *s++ == 'r' && isdigit(*s)) {
                    mask = *s++;
                    if (isdigit(*s)) {
                        mask = 10 * (mask - '0') + (*s++ - '0');
                        if (mask >= 32) {
                            break;
                        }
                    } else {
                        mask -= '0';
                    }
                    switch (*args) {
                    case 'e':
                        opcode_u |= mask << 18;
                        continue;
                    case 'f':
                        opcode_u |= mask << 13;
                        continue;
                    case 'g':
                        opcode_u |= mask << 8;
                        continue;
                    case 'h':
                        opcode_l |= mask << 20;
                        continue;
                    case 'E':
                        opcode_u |= mask >> 2;
                        opcode_l |= mask << 30;
                        continue;
                    case 'F':
                        opcode_l |= mask;
                        continue;
                    }
                }
                break;

            case 'U':    /* next operand is a special register */
            case 'V':
                if (*s++ == 's' && *s++ == 'r' && isdigit(*s)) {
                    mask = *s++ - '0';
                    if (isdigit(*s))
                        mask = 10 * mask + (*s++ - '0');
                    if (isdigit(*s))
                        mask = 10 * mask + (*s++ - '0');
                    if (isdigit(*s))
                        mask = 10 * mask + (*s++ - '0');
                    if ((*args == 'U' && mask >= 32) || (*args == 'V' && mask >= 1024)) {
                        break;
                    }
                    switch (*args) {
                    case 'U':
                        opcode_l |= mask;
                        continue;
                    case 'V':
                        opcode_u |= mask << 8;
                        continue;
                    }
                }
                break;

            case 'A':    /* next operand is a VCP register */
            case 'B':
            case 'C':
            case 'D':
                if (*s++ == 'v' && *s++ == 'r' && isdigit(*s)) {
                    mask = *s++ - '0';
                    if (isdigit(*s))
                        mask = 10 * mask + (*s++ - '0');
                    if (isdigit(*s))
                        mask = 10 * mask + (*s++ - '0');
                    if (mask >= 256) {
                        break;
                    }
                    switch (*args) {
                    case 'A':
                        opcode_u |= mask >> 2;
                        opcode_l |= mask << 30;
                        continue;
                    case 'B':
                        opcode_l |= mask << 20;
                        continue;
                    case 'C':
                        opcode_l |= mask;
                        continue;
                    case 'D':
                        opcode_l |= mask << 8;
                        continue;
                    }
                }
                break;

            case 'L':    /* next operand is a VCP mask register */
            case 'M':
            case 'N':
            case 'O':
                if (*s++ == 'm' && *s++ == 'r' && isdigit(*s)) {
                    mask = *s++ - '0';
                    if (isdigit(*s))
                        mask = 10 * mask + (*s++ - '0');
                    if (isdigit(*s))
                        mask = 10 * mask + (*s++ - '0');
                    if (mask >= 256) {
                        break;
                    }
                    switch (*args) {
                    case 'L':
                        opcode_u |= mask >> 2;
                        opcode_l |= mask << 30;
                        continue;
                    case 'M':
                        opcode_l |= mask << 20;
                        continue;
                    case 'N':
                        opcode_l |= mask;
                        continue;
                    case 'O':
                        opcode_l |= mask << 8;
                        continue;
                    }
                }
                break;

	    case 'P':   /* operand is a VCP VCR */
	    case 'Q':
                if (*s++ == 'v' && *s++ == 'c' && *s++ == 'r' && isdigit(*s)) {
                    mask = *s++;
                    if (isdigit(*s)) {
                        mask = 10 * (mask - '0') + (*s++ - '0');
                        if (mask >= 32) {
                            break;
                        }
                    } else {
                        mask -= '0';
                    }
                    switch (*args) {
                    case 'P':
                        opcode_u |= mask >> 2;
                        opcode_l |= mask << 30;
                        continue;
                    case 'Q':
                        opcode_l |= mask;
                        continue;
		    }
                }
                break;

	    case 'S':   /* operand is a VCP VUT */
	    case 'T':
                if (*s++ == 'v' && *s++ == 'u' && *s++ == 't' && isdigit(*s)) {
                    mask = *s++;
                    if (isdigit(*s)) {
                        mask = 10 * (mask - '0') + (*s++ - '0');
                        if (mask >= 32) {
                            break;
                        }
                    } else {
                        mask -= '0';
                    }
                    switch (*args) {
                    case 'S':
                        opcode_u |= mask >> 2;
                        opcode_l |= mask << 30;
                        continue;
                    case 'T':
                        opcode_l |= mask;
                        continue;
		    }
                }
                break;

	    case 'H':   /* operand is a VCP M bits */
                if (isdigit(*s)) {
		    mask = *s++ - '0';
                    if (mask >= 4) {
                        break;
                    }
                    opcode_l |= mask << 8;
                    continue;
                }
                break;

            case 'i':   /* 5 bit immediate */
                the_insn[loc].reloc = RELOC_5;
                goto immediate;

            case 'u':   /* 5 bit unsigned immediate */
                the_insn[loc].reloc = RELOC_U5;
                goto immediate;

            case 'j':   /* 12 bit immediate */
                the_insn[loc].reloc = RELOC_12;
                goto immediate;

            case 'I':   /* 25 bit immediate */
                the_insn[loc].reloc = RELOC_25;
                goto immediate;

            case 'J':   /* 32 bit immediate */
                the_insn[loc].reloc = RELOC_32;
                goto immediate;

            case 'K':   /* 25 bit immediate (32 bit -> base_GR + 25 bit) */
                the_insn[loc].reloc = RELOC_BASE25;
                goto immediate;

            case 'r':   /* 12 bit PC relative immediate */
                the_insn[loc].reloc = RELOC_PCREL12;
                the_insn[loc].pcrel = 1;
                goto immediate;

            case 'R':   /* 32 bit PC relative immediate */
                the_insn[loc].reloc = RELOC_PCREL32;
                the_insn[loc].pcrel = 1;
                goto immediate;

            case 'X':   /* 4 bit unsigned immediate(for type-B) */
                the_insn[loc].reloc = RELOC_SYSU4;
                goto immediate;

            case 'Y':   /* 8 bit unsigned immediate(for type-B) */
                the_insn[loc].reloc = RELOC_SYSU8;
                goto immediate;

            case 'z':   /* 25 bit Immediate(for cs) */
                the_insn[loc].reloc = RELOC_SYS25;
                goto immediate;

            case 'w':   /* 25 bit immediate(32 bit -> base_GR + 25 bit for type-B) */
                the_insn[loc].reloc = RELOC_SYSBASE25;
                goto immediate;

            immediate:
                if(*s==' ')
                    s++;
                if (getExpression(s,loc))
                    break;
                s = expr_end;
                continue;
                break;

            default:
                abort();
            }
            break;
        }
        error:
        if (match == FALSE) {
            /* args don't match */
            if (!insn->last) {
                ++insn;
                s = argsStart;
                continue;
            } else {
                as_warn("Illegal operands");
                return;
            }
        }
        break;
    }

    the_insn[loc].opcode_u = opcode_u;
    the_insn[loc].opcode_l = opcode_l;
    return;
}

static int
getExpression(str,loc)
    char *str; int loc;
{
    char *save_in;
    segT seg;

    save_in = input_line_pointer;
    input_line_pointer = str;
    switch (seg = expression(&the_insn[loc].exp)) {

    case SEG_ABSOLUTE:
    case SEG_TEXT:
    case SEG_DATA:
    case SEG_BSS:
    case SEG_UNKNOWN:
    case SEG_DIFFERENCE:
    case SEG_BIG:
    case SEG_NONE:
        break;

    default:
        the_insn[loc].error = "bad segment";
        return 1;
    }
    expr_end = input_line_pointer;
    input_line_pointer = save_in;
    return 0;
}


#define MAX_LITTLENUMS 8

/*
    This is identical to the md_atof in atof-tachyon.c.  I think this is right,
    but I'm not sure.

   Turn a string in input_line_pointer into a floating point constant of type
   type, and store the appropriate bytes in *litP.  The number of LITTLENUMS
   emitted is stored in *sizeP .  An error message is returned, or NULL on OK.
 */
char *
md_atof(type,litP,sizeP)
    char type;
    char *litP;
    int *sizeP;
{
    int prec;
    LITTLENUM_TYPE words[MAX_LITTLENUMS];
    LITTLENUM_TYPE *wordP;
    char        *t;
    char        *atof_tachyon();

    switch(type) {

    case 'f':
    case 'F':
    case 's':
    case 'S':
        prec = 2;
        break;

    case 'd':
    case 'D':
    case 'r':
    case 'R':
        prec = 4;
        break;

    case 'x':
    case 'X':
        prec = 8;
        break;

    case 'p':
    case 'P':
        prec = 6;
        break;

    default:
        *sizeP=0;
        return "Bad call to MD_ATOF()";
    }
    t=atof_tachyon(input_line_pointer,type,words);
    if(t)
        input_line_pointer=t;
    *sizeP=prec * sizeof(LITTLENUM_TYPE);
    for(wordP=words;prec--;) {
        md_number_to_chars(litP,(long)(*wordP++),sizeof(LITTLENUM_TYPE));
        litP+=sizeof(LITTLENUM_TYPE);
    }
    return "";  /* Someone should teach Dean about null pointers */
}

/*
 * Write out big-endian.
 */
void
md_number_to_chars(buf,val,n)
    char *buf;
    long val;
    int n;
{

    switch(n) {

    case 16:
        *buf++ = 0;
        *buf++ = 0;
        *buf++ = 0;
        *buf++ = 0;
        *buf++ = 0;
        *buf++ = 0;
        *buf++ = 0;
        *buf++ = 0;
    case 8:
        *buf++ = 0;
        *buf++ = 0;
        *buf++ = 0;
        *buf++ = 0;
    case 4:
        *buf++ = val >> 24;
        *buf++ = val >> 16;
    case 2:
        *buf++ = val >> 8;
    case 1:
        *buf = val;
        break;

    default:
        as_warn("Expected 2**n value (n=0-4)");
    }
    return;
}

void
md_number_to_imm(buf,val,n, fixP, seg_type, insn_addr, symbol_type)
    char *buf;
    long val;
    int n;
    fixS *fixP;
    int seg_type;
    long insn_addr;
    int symbol_type;
{
    if (seg_type != N_TEXT || (enum reloc_type) fixP->fx_r_type == NO_RELOC) {
        switch (n) {
        case 1:
                *buf = val;
                break;
        case 2:
                *buf++ = (val>>8);
                *buf = val;
                break;
        case 4:
                *buf++ = (val>>24);
                *buf++ = (val>>16);
                *buf++ = (val>>8);
                *buf = val;
                break;
        default:
                as_warn("can't reloc text in data segment");
                break;
        }
        return;
    }

    assert(n == 4);
    assert((enum reloc_type) fixP->fx_r_type < NO_RELOC);

 /*
  * +-+-----+----+----+----+
  * |4| opc | r1 |I0_5| r3 |
  * +-+-----+----+----+----+
  *                ^-buf[1].R2bit + buf[2].L3bit
  * +----------------------+-----+----+----+----+
  * |                      | opc | r1 |I1_5| r3 |
  * +----------------------+-----+----+----+----+
  *                                     ^-buf[4].R6bit & buf[4].L7bit
  * +-+-----------------------------------------+---+-----------+----+
  * |4|                                         |opc|        I12| r3 |
  * +-+-----------------------------------------+---+-----------+----+
  *                     buf[5].R1bit + buf[6] + buf[7].L3bit-^
  * +-+--------------------+-----+----+-------------------------+----+
  * |4|                    | opc | r1 | I25                     | r3 |
  * +-+--------------------+-----+----+-------------------------+----+
  *            buf[4].R6bit + buf[5] + buf[6] + buf[7].L3bit-^
  * +-+--------------------+---+--------------------------------+----+
  * |4|                    |opc|        I32                     | r3 |
  * +-+--------------------+---+--------------------------------+----+
  *   buf[3].R5bit + buf[4] + buf[5] + buf[6] + buf[7].L3bit-^
  *
  * <SYSTEM>
  * +-+-----+----------------------------------------------+----+----+
  * |4| opc |                                              | I4 |    |
  * +-+-----+----------------------------------------------+----+----+
  *                              buf[6].R1bit + buf[7].L3bit-^
  * +-+-----+-------------------------------------------+-------+----+
  * |4| opc |                                           |    I7 |    |
  * +-+-----+-------------------------------------------+-------+----+
  *                              buf[6].R4bit + buf[7].L3bit-^
  * +-+-----+-------------------------+-------------------------+----+
  * |4| opc |                         | I25                     |    |
  * +-+-----+-------------------------+-------------------------+----+
  *            buf[4].R6bit + buf[5] + buf[6] + buf[7].L3bit-^
  */
    switch (fixP->fx_r_type) {
    case RELOC_SYSU4:
        if (!((unsigned) (val) < 0x10))
          as_warn("[obj_address=0x%08.8X]:Imm value exceeds unsigned 4 bit field: %d", insn_addr, val);
        buf[6] |= (val >>  3) & 1;
        buf[7] |= (val <<  5) & 0xff;
        break;
    case RELOC_SYSU8:
        if (!((unsigned) (val) < 0x100))
          as_warn("[obj_address=0x%08.8X]:Imm value exceeds unsigned 8 bit field: %d", insn_addr, val);
        buf[6] |= (val >>  3) & 0x1f;
        buf[7] |= (val <<  5) & 0xff;
        break;
    case RELOC0_5:
        if (!((unsigned) (val + 0x10) < 0x30))
          as_warn("[obj_address=0x%08.8X]:Imm value exceeds 5 bit field: %d", insn_addr, val);
        buf[1] |= (val >>  3) & 3;
        buf[2] |= (val <<  5) & 0xff;
        break;
    case RELOC1_5:
        if (!((unsigned) (val + 0x10) < 0x30))
          as_warn("[obj_address=0x%08.8X]:Imm value exceeds 5 bit field: %d", insn_addr, val);
        buf[4] |= (val <<  1) & 0x3f;
        break;
    case RELOC0_U5:
        if (!((unsigned) (val) < 0x20))
          as_warn("[obj_address=0x%08.8X]:Imm value exceeds unsigned 5 bit field: %u", insn_addr, val);
        buf[1] |= (val >>  3) & 3;
        buf[2] |= (val <<  5) & 0xff;
        break;
    case RELOC1_U5:
        if (!((unsigned) (val) < 0x20))
          as_warn("[obj_address=0x%08.8X]:Imm value exceeds unsigned 5 bit field: %u", insn_addr, val);
        buf[4] |= (val <<  1) & 0x3f;
        break;
    case RELOC_PCREL12:
	if (val & 7)
          as_warn("[obj_address=0x%08.8X]:bad alignment: 0x%08.8X", insn_addr, val);
        val >>= 3;
        if (!((unsigned) (val + 0x800) < 0x1800))
          as_warn("[obj_address=0x%08.8X]:Imm value exceeds 12 bit field: %d", insn_addr, val);
        buf[5] |= (val >> 11) & 1;
        buf[6]  = (val >>  3) & 0xff;
        buf[7] |= (val <<  5) & 0xff;
        break;
    case RELOC_25:
    case RELOC_SYS25:
        if (!((unsigned) (val + 0x1000000) < 0x3000000))
          as_warn("[obj_address=0x%08.8X]:Imm value exceeds 25 bit field: %d", insn_addr, val);
        buf[4] |= (val >> 19) & 0x3f;
        buf[5]  = (val >> 11) & 0xff;
        buf[6]  = (val >>  3) & 0xff;
        buf[7] |= (val <<  5) & 0xff;
        break;
    case RELOC_BASE25:
      {
	register int tmp;
	register long offset;

	for (tmp = 0; tmp < 32; tmp++) {
	  if (USING_TABLE[tmp].inuse) {
	    if ((USING_TABLE[tmp].symbolP->sy_nlist.n_type & N_TYPE)
             == (symbol_type & N_TYPE)) {
	      offset = val - USING_TABLE[tmp].symbolP->sy_nlist.n_value;
	      if (offset >= -0x1000000 && offset < 0x1000000)
	        break;
	    }
	  }
	}
        if (tmp >= 32) {	
          as_warn("[obj_address=0x%08.8X]:Can't find valid base_register (target address: 0x%08.8X)", insn_addr, val);
	  break;
	}
        buf[3] |= (tmp >> 2) & 7;
        buf[4] |= (tmp << 6) & 0xc0;
        buf[4] |= (offset >> 19) & 0x3f;
        buf[5]  = (offset >> 11) & 0xff;
        buf[6]  = (offset >>  3) & 0xff;
        buf[7] |= (offset <<  5) & 0xff;
	break;
      }
    case RELOC_SYSBASE25:
      {
	register int tmp;
	register long offset;

	for (tmp = 0; tmp < 32; tmp++) {
	  if (USING_TABLE[tmp].inuse) {
	    if ((USING_TABLE[tmp].symbolP->sy_nlist.n_type & N_TYPE)
             == (symbol_type & N_TYPE)) {
	      offset = val - USING_TABLE[tmp].symbolP->sy_nlist.n_value;
	      if (offset >= -0x1000000 && offset < 0x1000000)
	        break;
	    }
	  }
	}
        if (tmp >= 32) {	
          as_warn("[obj_address=0x%08.8X]:Can't find valid base_register (target address: 0x%08.8X)", insn_addr, val);
	  break;
	}
        buf[3] |= (tmp >> 2) & 7;
        buf[4] |= (tmp << 6) & 0xc0;
        buf[4] |= (offset >> 19) & 0x3f;
        buf[5]  = (offset >> 11) & 0xff;
        buf[6]  = (offset >>  3) & 0xff;
        buf[7] |= (offset <<  5) & 0xff;
	break;
      }
    case RELOC_PCREL32:
	if (val & 7)
          as_warn("[obj_address=0x%08.8X]:bad alignment: 0x%08.8X", insn_addr, val);
        val >>= 3;
    case RELOC_32:
        buf[3] |= (val >> 27) & 0x1f;
        buf[4]  = (val >> 19) & 0xff;
        buf[5]  = (val >> 11) & 0xff;
        buf[6]  = (val >>  3) & 0xff;
        buf[7] |= (val <<  5) & 0xff;
        break;
    case RELOC_JMP_TBL:
        *buf++ = (val>>24);
        *buf++ = (val>>16);
        *buf++ = (val>>8);
        *buf = val;
        break;
    case NO_RELOC:
    default:
        as_warn("[obj_address=0x%08.8X]:bad relocation type: 0x%02x", insn_addr, fixP->fx_r_type);
        break;
    }
    return;
}

/* should never be called for tachyon */
void
md_create_short_jump(ptr, from_addr, to_addr, frag, to_symbol)
    char *ptr;
    long from_addr, to_addr;
{
    fprintf(stderr, "tachyon_create_short_jmp\n");
    abort();
}

/* should never be called for tachyon */
void
md_number_to_disp(buf,val,n)
    char        *buf;
    long        val;
{
    fprintf(stderr, "md_number_to_disp\n");
    abort();
}

/* should never be called for tachyon */
void
md_number_to_field(buf,val,fix)
    char *buf;
    long val;
    char *fix;
{
    fprintf(stderr, "tachyon_number_to_field\n");
    abort();
}

void
md_ri_to_chars(ri_p, ri)
     struct reloc_info_tachyon *ri_p, ri;
{
    /* 
     * if you want to cross-assemble from
     * a little-endian then you might need to
     * do some byte-swapping here.
     */
    return;
}

/* should never be called for tachyon */
void
md_convert_frag(fragP)
    register fragS *fragP;
{
    fprintf(stderr, "tachyon_convert_frag\n");
    abort();
}

/* should never be called for tachyon */
void
md_create_long_jump(ptr, from_addr, to_addr, frag, to_symbol)
    char        *ptr;
    long        from_addr,
                to_addr;
    fragS       *frag;
    symbolS     *to_symbol;
{
    fprintf(stderr, "tachyon_create_long_jump\n");
    abort();
}

/* should never be called for tachyon */
int
md_estimate_size_before_relax(fragP, segtype)
    register fragS *fragP;
{
    fprintf(stderr, "tachyon_estimate_size_before_relax\n");
    abort();
    return 0;
}

#if 0
/* for debugging only */
static void
print_insn(insn)
    struct tachyon_it *insn;
{
    char *Reloc[] = {
    "RELOC_5",
    "RELOC0_5",
    "RELOC1_5",
    "RELOC_U5",
    "RELOC0_U5",
    "RELOC1_U5",
    "RELOC_12",
    "RELOC_25",
    "RELOC_32",
    "RELOC_BASE25",
    "RELOC_PCREL12",
    "RELOC_PCREL32",
    "RELOC_SYSU4",
    "RELOC_SYSU8",
    "RELOC_SYS25",
    "RELOC_SYSBASE25",
    "RELOC_JMP_TBL",
    "RELOC_RELATIVE",
    "NO_RELOC"
    };

    if (insn->error) {
        fprintf(stderr, "ERROR: %s\n");
    }
    fprintf(stderr, "opcode=0x%08x\n", insn->opcode);
    fprintf(stderr, "reloc = %s\n", Reloc[insn->reloc]);
    fprintf(stderr, "exp =  {\n");
    fprintf(stderr, "\t\tX_add_symbol = %s\n",
        insn->exp.X_add_symbol ?
        (insn->exp.X_add_symbol->sy_name ? 
        insn->exp.X_add_symbol->sy_name : "???") : "0");
    fprintf(stderr, "\t\tX_sub_symbol = %s\n",
        insn->exp.X_subtract_symbol ?
            (insn->exp.X_subtract_symbol->sy_name ? 
                insn->exp.X_subtract_symbol->sy_name : "???") : "0");
    fprintf(stderr, "\t\tX_add_number = %d\n",
        insn->exp.X_add_number);
    fprintf(stderr, "}\n");
    return;
}
#endif

/*
 * TaChyon relocations are completely different, so it needs
 * this machine dependent routine to emit them.
 */
void
emit_relocations(fixP, segment_address_in_file)
    register fixS *fixP;
    relax_addressT segment_address_in_file;
{
    struct reloc_info_tachyon ri;
    register symbolS *symbolP;
    extern char *next_object_file_charP;
    long add_number;

    bzero((char *) &ri, sizeof(ri));
    for (; fixP; fixP = fixP->fx_next) {

        if (fixP->fx_r_type >= (char) NO_RELOC) {
            fprintf(stderr, "fixP->fx_r_type = %d\n", fixP->fx_r_type);
            abort();
        }

        if ((symbolP = fixP->fx_addsy) != NULL) {
            ri.r_address = fixP->fx_frag->fr_address +
                fixP->fx_where - segment_address_in_file;
            if ((symbolP->sy_type & N_TYPE) == N_UNDF) {
                ri.r_extern = 1;
                ri.r_index = symbolP->sy_number;
            } else {
                ri.r_extern = 0;
                ri.r_index = symbolP->sy_type & N_TYPE;
            }
            if (symbolP && symbolP->sy_frag) {
                ri.r_addend = symbolP->sy_frag->fr_address;
            }
            ri.r_type = (enum reloc_type) fixP->fx_r_type;
            if (fixP->fx_pcrel) {
/*              ri.r_addend -= fixP->fx_where;          */
                ri.r_addend -= ri.r_address;            
            } else {
                ri.r_addend = fixP->fx_addnumber;
            }

/*          md_ri_to_chars((char *) &ri, ri);        */
            append(&next_object_file_charP, (char *)& ri, sizeof(ri));
        }
    }
    return;
}

int
md_parse_option(argP,cntP,vecP)
    char **argP;
    int *cntP;
    char ***vecP;
{
    return 1;
}

void
listing_stdout()
{
    listing_cur = listing_top;

    while (listing_cur = listing_cur->nextptr) {
      /* if (!flagseen['s']) */
      listing_insn(listing_cur->lineno, 
		   listing_cur->top,
		   listing_cur->buf,
		   listing_cur->offset, 
		   listing_cur->size,
		   listing_cur->type);
#if 1
      putchar('\n');
#endif
#if 0
      switch (listing_cur->type) {
      case 0: /* insn */
        switch ((* (listing_cur->buf)) >> 4) {
        case TYPE_2: case TYPE_3: case TYPE_4: case TYPE_5: case TYPE_6: case TYPE_7:
          printf("  %-18s & %-18s & %-s\n",
            listing_cur->str[listing_cur->indr[0]]?listing_cur->str[listing_cur->indr[0]]:"nop",
            listing_cur->str[listing_cur->indr[1]]?listing_cur->str[listing_cur->indr[1]]:"nop",
            listing_cur->str[listing_cur->indr[2]]?listing_cur->str[listing_cur->indr[2]]:"nop");
          break;
        case TYPE_8: case TYPE_9: case TYPE_A: case TYPE_B: case TYPE_C: case TYPE_D:
          printf("  %-18s & %-18s",
            listing_cur->str[listing_cur->indr[0]]?listing_cur->str[listing_cur->indr[0]]:"nop",
            listing_cur->str[listing_cur->indr[1]]?listing_cur->str[listing_cur->indr[1]]:"nop");
          if (listing_cur->str[listing_cur->indr[2]])
            printf(" & ???");
          putchar('\n');
          break;
        default:
          printf("  %-18s",
            listing_cur->str[listing_cur->indr[0]]?listing_cur->str[listing_cur->indr[0]]:"nop");
          if (listing_cur->str[listing_cur->indr[1]])
            printf(" & ???");
          if (listing_cur->str[listing_cur->indr[2]])
            printf(" & ???");
          putchar('\n');
          break;
        }
        break;
      case 1: /* data */
      case 2: /* string */
        putchar('\n');
        break;
      default: /* error */
        printf("???\n");
        break;
      }
#endif
    }
}

void
listing_insn(lineno, top, buf, offs, size, type) 
    unsigned lineno;
    unsigned char *top, *buf;
    unsigned offs, size, type; /* if type==2(string) size is meanless */
{
    register unsigned char a, b;
    register unsigned i, j, addr;
    register unsigned LIMIT;

    if (flagseen['s'])
      LIMIT = 8;
    else
      LIMIT = 16;

    addr = buf - top + offs;
    printf("%-5d %08.8X - ", lineno, addr);
    /* printf("%-5d %08.8X %08.8X %08.8X - ", lineno, buf, top, offs); */

    switch (type) {
    case 0: /* insn */
      /* mark if op_type is 8 (op0 - op1 are exchanged) */
      printf("%02.2X%02.2X%02.2X%02.2X ", *(buf  ),*(buf+1),*(buf+2),*(buf+3));
      printf("%02.2X%02.2X%02.2X%02.2X" , *(buf+4),*(buf+5),*(buf+6),*(buf+7));
      printf(" [%c%X",type_exchange_loc[*buf>>4]?'X':' ',*buf>>4);
      if (!flagseen['s']) {
        switch (type_listing[*buf>>4]) {
        case 0: /* 20+20+20 */
          printf(" %02.2X%02.2X%02.2X%02.2X",
            ((*(buf  ) << 1) & 0x1e) | (*(buf+1) >> 7),
            ((*(buf+1) >> 2) & 0x1f),
            ((*(buf+1) << 3) & 0x18) | (*(buf+2) >> 5),
            ((*(buf+2)     ) & 0x1f));
          printf(" %02.2X%02.2X%02.2X%02.2X",
            ((*(buf+3) >> 3)       ),
            ((*(buf+3) << 2) & 0x1c) | (*(buf+4) >> 6),
            ((*(buf+4) >> 1) & 0x1f),
            ((*(buf+4) << 4) & 0x10) | (*(buf+5) >> 4));
          printf(" %02.2X%02.2X%02.2X%02.2X]",
            ((*(buf+5) << 1) & 0x1e) | (*(buf+6) >> 7),
            ((*(buf+6) >> 2) & 0x1f),
            ((*(buf+6) << 3) & 0x18) | (*(buf+7) >> 5),
            ((*(buf+7)     ) & 0x1f));
          break;
        case 1: /* 20+20+20(12bit) */
          printf(" %02.2X%02.2X%02.2X%02.2X",
            ((*(buf  ) << 1) & 0x1e) | (*(buf+1) >> 7),
            ((*(buf+1) >> 2) & 0x1f),
            ((*(buf+1) << 3) & 0x18) | (*(buf+2) >> 5),
            ((*(buf+2)     ) & 0x1f));
          printf(" %02.2X%02.2X%02.2X%02.2X",
            ((*(buf+3) >> 3)       ),
            ((*(buf+3) << 2) & 0x01c) | (*(buf+4) >> 6),
            ((*(buf+4) >> 1) & 0x01f),
            ((*(buf+4) << 4) & 0x010) | (*(buf+5) >> 4));
          printf(" %01.1X_%03.3X_%02.2X]",
            ((*(buf+5) >> 1) & 0x007),
            ((*(buf+5) <<11) & 0x800) | (*(buf+6) << 3) | (*(buf+7) >> 5),
            ((*(buf+7)     ) & 0x01f));
          break;
        case 2: /* 20+40(25bit) */
          printf(" %02.2X%02.2X%02.2X%02.2X",
            ((*(buf  ) << 1) & 0x01e) | (*(buf+1) >> 7),
            ((*(buf+1) >> 2) & 0x01f),
            ((*(buf+1) << 3) & 0x018) | (*(buf+2) >> 5),
            ((*(buf+2)     ) & 0x01f));
          printf(" %02.2X_%02.2X__%07.7X_%02.2X]",
            ((*(buf+3) >> 3)            ),
            ((*(buf+3) << 2) & 0x000001c) | ((*(buf+4) >> 6)),
            ((*(buf+4) <<19) & 0x1f80000) | ((*(buf+5) <<11) & 0x7f800) |
            ((*(buf+6) << 3) & 0x00007f8) | ((*(buf+7) >> 5)),
            ((*(buf+7)     ) & 0x000001f));
          break;
        case 3: /* 20+40(32bit) */
          printf(" %02.2X%02.2X%02.2X%02.2X",
            ((*(buf  ) << 1) & 0x01e) | (*(buf+1) >> 7),
            ((*(buf+1) >> 2) & 0x01f),
            ((*(buf+1) << 3) & 0x018) | (*(buf+2) >> 5),
            ((*(buf+2)     ) & 0x01f));
          printf(" %01.1X_____%08.8X_%02.2X]",
            ((*(buf+3) >> 5)             ),
            ((*(buf+3) <<27) & 0xf8000000) | ((*(buf+4) <<19) & 0x7f80000) |
            ((*(buf+5) <<11) & 0x0007f800) | ((*(buf+6) << 3) & 0x00007f8) |
            ((*(buf+7) >> 5)             ),
            ((*(buf+7)     ) & 0x0000001f));
          break;
        case 4: /* 60(5+5+10+10+25+5) */
          printf(" %02.2X_%02.2X",
            ((*(buf  ) << 1) & 0x0000001e) | ((*(buf+1) >> 7)),
            ((*(buf+1) >> 2) & 0x0000001f));
          printf("_%03.3X_%02.2X_%02.2X_%07.7X_%02.2X]",
            ((*(buf+1) << 8) & 0x00000300) | ((*(buf+2))),
            ((*(buf+3) >> 3)             ),
            ((*(buf+3) << 2) & 0x0000001c) | ((*(buf+4) >> 6)),
            ((*(buf+4) <<19) & 0x01f80000) | ((*(buf+5) <<11) & 0x7f800) |
            ((*(buf+6) << 3) & 0x000007f8) | ((*(buf+7) >> 5)),
            ((*(buf+7)     ) & 0x0000001f));
          break;
        case 5: /* 60(4+4+16+20+20) */
          printf(" %01.1X",
            ((*(buf  )     ) & 0x0f));
          printf(" %02.2X%02.2X%02.2X",
            ((*(buf+1) >> 2) & 0x3f),
            ((*(buf+1) << 3) & 0x18) | (*(buf+2) >> 5),
            ((*(buf+2)     ) & 0x1f));
	  printf(" %01.1X_%02.2X %01.1X_%02.2X %01.1X_%02.2X_%02.2X]",
            ((*(buf+3) >> 6)       ),
            ((*(buf+3) << 2) & 0xff) | ((*(buf+4) >> 6)),
	    ((*(buf+4) >> 4) & 0x03),
            ((*(buf+4) << 4) & 0xff) | ((*(buf+5) >> 4) & 0x0f),
            ((*(buf+5)     ) & 0x0f),
            ((*(buf+6)     )       ),
            ((*(buf+7)     )       ));
          break;
        default:
          break;
        }
      }
      else {
        switch (type_listing[*buf>>4]) {
        case 0: /* 20+20+20 */
          printf(" #________]");
          break;
        case 1: /* 20+20+20(12bit) */
          printf(" #_____%03.3X]",
            ((*(buf+5) <<11) & 0x800) | (*(buf+6) << 3) | (*(buf+7) >> 5));
          break;
        case 2: /* 20+40(25bit) */
          printf(" #_%07.7X]",
            ((*(buf+4) <<19) & 0x1f80000) | ((*(buf+5) <<11) & 0x7f800) |
            ((*(buf+6) << 3) & 0x00007f8) | ((*(buf+7) >> 5)));
          break;
        case 3: /* 20+40(32bit) */
          printf(" #%08.8X]",
            ((*(buf+3) <<27) & 0xf8000000) | ((*(buf+4) <<19) & 0x7f80000) |
            ((*(buf+5) <<11) & 0x0007f800) | ((*(buf+6) << 3) & 0x00007f8) |
            ((*(buf+7) >> 5)));
          break;
        case 4: /* 60(5+5+10+10+25+5) */
          printf(" #_%07.7X]",
            ((*(buf+4) <<19) & 0x1f80000) | ((*(buf+5) <<11) & 0x7f800) |
            ((*(buf+6) << 3) & 0x00007f8) | ((*(buf+7) >> 5)));
          break;
        case 5: /* 60(4+4+16+20+20) */
          printf(" #________]");
          break;
        default:
          break;
        }
      }
      break;
    case 1: /* data */
        i = 0;
data_cont_loop:
        for (j = 0; j < size; j++, i++) {
          if (j && j % LIMIT == 0) {
	    i = 0;
            printf("->\n%-5d %08.8X - ", lineno, addr+j);
	  }
          if ((b = *(listing_cur->buf+j)>>4)>9) putchar(b+'A'-10); else putchar(b+'0');
          if ((b = *(listing_cur->buf+j)&15)>9) putchar(b+'A'-10); else putchar(b+'0');
          if (size > 4 && j < size-1 && j % 4 == 3 && i < LIMIT-1)
            putchar(' ');
        }
        if (listing_cur->term)
          break;
        else if (i >= LIMIT) {
          putchar('-');
          putchar('>');
          break;
        }
        else {
          putchar(' ');
          listing_cur = listing_cur->nextptr;
          goto data_cont_loop;
        }
      break;
    case 2: /* string */
      i = 0;
str_cont_loop:
        if ((b = *(listing_cur->buf)>>4)>9) putchar(b+'A'-10);else putchar(b+'0');
        if ((b = *(listing_cur->buf)&15)>9) putchar(b+'A'-10);else putchar(b+'0');
        if (listing_cur->term) 
          break;
        else if (++i >= LIMIT) {
          putchar('-');
          putchar('>');
          break;
        }
        else {
          if (i % 4 == 0)
            putchar(' ');
          listing_cur = listing_cur->nextptr;
          goto str_cont_loop;
        }
      break;
    }
}

void
listing_init()
{
    listing_top = listing_cur = (struct listing*) malloc(sizeof(listing_buf));
    bzero (listing_cur, sizeof(listing_buf));
}

void
listing_new(top, buf, size, type, term)
register unsigned char *top, *buf; unsigned size, type, term;
{
    listing_cur->nextptr = (struct listing*) malloc(sizeof(listing_buf));
    listing_cur          = listing_cur->nextptr;
    bzero (listing_cur, sizeof(listing_buf));
    listing_cur->top     = top;
    listing_cur->buf     = buf;
    listing_cur->size    = size;
    listing_cur->type    = type;
    listing_cur->term    = term;
    listing_cur->offset  = 0;
    listing_cur->lineno  = listing_where();
}

listing_bufmod(top, buf)
register unsigned char *top, *buf;
{
    listing_cur->top     = top;
    listing_cur->buf     = buf;
}

listing_termmod(term)
register unsigned term;
{
    listing_cur->term    = term;
}

void
listing_string_terminate()
{
    listing_cur->term = 1;
}

void
listing_address_correct(top, offset)
register unsigned char * top; unsigned offset;
{
    register struct listing * i = listing_top;

    while (i = i->nextptr) {
      if (i->top == top)
        i->offset = offset;
    }
}
