/* @(#)ip-decode.h	1.4    1/7/93 */

/* TaChyon/KAI Instruction simulator      */
/*  Copyright (c) 1989 by FUJITSU LIMITED */
/*            Primary writer: Y.Nakashima */

/*
 * ip-decode.h - mnemonic table
 */

enum ip_operation_type
{
	O_NL,
	O_IP,
	O_AL, /* AP_LOAD  */
	O_AS, /* AP_STORE */
	O_FH, /* FP_HIGH  */
	O_FL, /* FP_LOW   */
	O_VP, /* VP       */
};

static struct tachyon_opcode
{
	char *name;
	char *args;
	int  (*func)();
	char  type;
};

int	o_nop();

int	o_call(), o_brc0(), o_brnc0(), o_brc1(), o_brnc1(), o_seti();

int	o_brcs0(), o_brncs0(), o_brcs1(), o_brncs1();

int	o_jmp(), o_ld_c(), o_ld_cu(), o_ld_s(), o_ld_su(), o_ld(),
	o_ld_l(), o_ld_f(), o_ld_d(), o_ld_pf(), o_ld_pd(),
        o_st_c(), o_st_s(), o_st(), o_st_d(), o_st_f(),
	o_mov_if(), o_mov_fi(), o_mov_ld(), o_mov_dl(), 
        o_mult(), o_mult_l(), o_mult_lu();

int	o_ld_c25(), o_ld_cu25(), o_ld_s25(), o_ld_su25(), o_ld25(),
	o_ld_l25(), o_ld_f25(), o_ld_d25(), o_ld_pf25(), o_ld_pd25(),
	o_st_c25(), o_st_s25(), o_st25(), o_addi25(),
	o_addli25(), o_subi25(), o_subli25(), o_and25(), o_or25(), 
	o_xor25(), o_st_f25(), o_st_d25(),
	o_mult25(), o_mult_l25(), o_mult_lu25();

int	o_add(), o_addl(), o_sub(), o_subl(), o_and(), o_or(), o_xor(),
	o_sra(), o_srl(), o_sll(), o_addi5(), o_addli5(), o_subi5(),
	o_subli5(), o_sra5(), o_srl5(), o_sll5();

int	o_fadd(), o_fadd_d(), o_fadd_x(),
	o_fsub(), o_fsub_d(), o_fsub_x(),
	o_fcmp(), o_fcmp_d(), o_fcmp_x(),
	o_fmul(), o_fmul_d(), o_fdiv(), o_fdiv_d(), 
	o_fcnv_fd(), o_fcnv_df(), o_fcnv_if(), o_fcnv_fi(), 
	o_fcnv_id(), o_fcnv_di(), o_fcnv_dx(), o_fcnv_xd(),
	o_fcnz_fi(), o_fcnz_di();

int	o_svc(), o_fwc(), o_ldsr(), o_stsr(), 
	o_ld_mcr(), o_st_mcr(), o_wbst(), 
	o_movgs(), o_movsg(), o_movgt(), o_movtg(),
	o_ptlbe(), o_ptlbp(), o_ptlba(), 
	o_flche(), o_flcha(), o_aki(), o_rti(), o_cs();

int	o_COP();

/*
Kinds of operands:
      =: same as assembler
      <: assembler only
      >: disasm only
      
 = #    Number used by optimizer.  It is ignored.
 = 1    rs1 register.
 = 2    rs2 register.
 = 3    rs3 register(for 20 bit OP).
 = 4    rs3 register(for 40 bit OP).
 = 5    rs3/rs4 register(for 60 bit type-B OP).
 = e    frs1 floating point register.
 = f    frs2 floating point register.
 = g    frs3 floating point register(for 20 bit OP).
 = h    frs3 floating point register(for 40 bit OP).
 = i     5 bit Immediate.
 = u     5 bit Unsigned Immediate.(for shift OP).
 = j    12 bit immediate.
 = I    25 bit Immediate.
 = J    32 bit Immediate.
 < K    25 bit Immediate(32 bit -> base_GR1 + 25 bit).
 = r    12 bit PC relative immediate.
 = R    32 bit PC relative immediate.
 = s    integer/floating state register(for 20 bit OP).
 = t    integer/floating state register(for 40 bit OP).
 = c    Coprocessor state register(for 20 bit OP).
 = d    Coprocessor state register(for 40 bit OP).

 = U    SRrsr( 5 bits wide 0-31 for type-B).
 = V    SRrsr(10 bits wide 0-1023 for type-B).
 = X     4 bit Unsigned Immediate(for type-B).
 = Y     8 bit Unsigned Immediate(for type-B).
 = y    Base register(for type-B).
 = z    25 bit Immediate(for type-B).
 < w    25 bit Immediate(32 bit -> base_GR2 + 25 bit (for type-B)).

 = A    VCP vR1 register (26-33)
 = B    VCP vR2 register (36-43)
 = C    VCP vR3 register (48-55)
 = D    VCP vR4 register (56-63)
 = L    VCP vR1 mask register (26-33)
 = M    VCP vR2 mask register (36-43)
 = N    VCP vR3 mask register (48-55)
 = O    VCP vR4 mask register (56-63)
 = 6    VCP rr1 GR (14-18)
 = 7    VCP rr2 GR (29-33)
 = 8    VCP rw1 GR (39-43)
 = 9    VCP rw2 GR (59-63)
 = E    VCP rr2 FR (29-33)
 = F    VCP rw2 FR (59-63)
 = P    VCP R2 VCR (29-33)
 = Q    VCP R3 VCR (59-63)
 = S    VCP R2 VUT (29-33)
 = T    VCP R3 VUT (59-63)
 = H    VCP M (54-55)
 > G    VCP ID (26-28) (disassembler only)
*/

static struct tachyon_opcode oper_undef[] = {
{ "???",        "",           NULL,	  O_NL },
};


static struct tachyon_opcode oper_cbri32[] = {
{ "nop",        "", 	      o_nop,	  O_NL },
{ "call",       "R,4",        o_call,	  O_IP },
{ "brc",        "t,R",        o_brc0,	  O_IP },
{ "brnc",       "t,R",        o_brnc0,	  O_IP },
{ "brc",        "d,R",        o_brc1,	  O_IP },
{ "brnc",    	"d,R",        o_brnc1,	  O_IP },
{ "seti",       "J,4",        o_seti,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
};


static struct tachyon_opcode oper_brci12[] = {
{ "nop",        "",           o_nop,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "brcs",       "s,r",        o_brcs0,	  O_IP },
{ "brncs",      "s,r",        o_brncs0,	  O_IP },
{ "brcs",       "c,r",        o_brcs1,	  O_IP },
{ "brncs",      "c,r",        o_brncs1,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
};


static struct tachyon_opcode oper_lsmj[] = {
{ "nop",        "",           o_nop,	  O_NL },
{ "jmp",        "1",          o_jmp,	  O_IP },
{ "ld.c",       "1,2,3",      o_ld_c,	  O_AL },
{ "ld.cu",      "1,2,3",      o_ld_cu,	  O_AL },
{ "ld.s",       "1,2,3",      o_ld_s,	  O_AL },
{ "ld.su",      "1,2,3",      o_ld_su,	  O_AL },
{ "ld",         "1,2,3",      o_ld,	  O_AL },
{ "ld.l",       "1,2,3",      o_ld_l,	  O_AL },
{ "ld.f",       "1,2,g",      o_ld_f,	  O_AL },
{ "ld.d",       "1,2,g",      o_ld_d,	  O_AL },
{ "ld.pf",      "1,2,g",      o_ld_pf,	  O_AL },
{ "ld.pd",      "1,2,g",      o_ld_pd,	  O_AL },
{ "mult",       "1,2,3",      o_mult,	  O_IP },
{ "mult.l",     "1,2,3",      o_mult_l,	  O_IP },
{ "???",        "",           NULL,       O_NL },
{ "mult.lu",    "1,2,3",      o_mult_lu,  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "st.c",       "3,1,2",      o_st_c,	  O_AS },
{ "st.s",       "3,1,2",      o_st_s,	  O_AS },
{ "st",         "3,1,2",      o_st,	  O_AS },
{ "mov.if",     "1,g",        o_mov_if,	  O_IP },
{ "mov.ld",     "1,g",        o_mov_ld,	  O_IP },
{ "mov.fi",     "e,3",        o_mov_fi,	  O_IP },
{ "mov.dl",     "e,3",        o_mov_dl,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "st.f",       "g,1,2",      o_st_f,	  O_AS },
{ "st.d",       "g,1,2",      o_st_d,	  O_AS },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
};


static struct tachyon_opcode oper_lsm[] = {
{ "nop",        "",           o_nop,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "ld.c",       "1,2,3",      o_ld_c,	  O_AL },
{ "ld.cu",      "1,2,3",      o_ld_cu,	  O_AL },
{ "ld.s",       "1,2,3",      o_ld_s,	  O_AL },
{ "ld.su",      "1,2,3",      o_ld_su,	  O_AL },
{ "ld",         "1,2,3",      o_ld,	  O_AL },
{ "ld.l",       "1,2,3",      o_ld_l,	  O_AL },
{ "ld.f",       "1,2,g",      o_ld_f,	  O_AL },
{ "ld.d",       "1,2,g",      o_ld_d,	  O_AL },
{ "ld.pf",      "1,2,g",      o_ld_pf,	  O_AL },
{ "ld.pd",      "1,2,g",      o_ld_pd,	  O_AL },
{ "mult",       "1,2,3",      o_mult,	  O_IP },
{ "mult.l",     "1,2,3",      o_mult_l,	  O_IP },
{ "???",        "",           NULL,       O_NL },
{ "mult.lu",    "1,2,3",      o_mult_lu,  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "st.c",       "3,1,2",      o_st_c,	  O_AS },
{ "st.s",       "3,1,2",      o_st_s,	  O_AS },
{ "st",         "3,1,2",      o_st,	  O_AS },
{ "mov.if",     "1,g",        o_mov_if,	  O_IP },
{ "mov.ld",     "1,g",        o_mov_ld,	  O_IP },
{ "mov.fi",     "e,3",        o_mov_fi,	  O_IP },
{ "mov.dl",     "e,3",        o_mov_dl,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "st.f",       "g,1,2",      o_st_f,	  O_AS },
{ "st.d",       "g,1,2",      o_st_d,	  O_AS },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
};


static struct tachyon_opcode oper_alsmi25[] = {
{ "nop",        "",           o_nop,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "ld.c",       "1,I,4",      o_ld_c25,	  O_AL },
{ "ld.cu",      "1,I,4",      o_ld_cu25,  O_AL },
{ "ld.s",       "1,I,4",      o_ld_s25,	  O_AL },
{ "ld.su",      "1,I,4",      o_ld_su25,  O_AL },
{ "ld",         "1,I,4",      o_ld25,	  O_AL },
{ "ld.l",       "1,I,4",      o_ld_l25,	  O_AL },
{ "ld.f",       "1,I,h",      o_ld_f25,	  O_AL },
{ "ld.d",       "1,I,h",      o_ld_d25,	  O_AL },
{ "ld.pf",      "1,I,h",      o_ld_pf25,  O_AL },
{ "ld.pd",      "1,I,h",      o_ld_pd25,  O_AL },
{ "mult",       "1,I,4",      o_mult25,   O_IP },
{ "mult.l",     "1,I,4",      o_mult_l25, O_IP },
{ "???",        "",           NULL,       O_NL },
{ "mult.lu",    "1,I,4",      o_mult_lu25,O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "st.c",       "4,1,I",      o_st_c25,	  O_AS },
{ "st.s",       "4,1,I",      o_st_s25,	  O_AS },
{ "st",         "4,1,I",      o_st25,	  O_AS },
{ "addi25",     "1,I,4",      o_addi25,	  O_IP },
{ "addli25",    "1,I,4",      o_addli25,  O_IP },
{ "subi25",     "1,I,4",      o_subi25,	  O_IP },
{ "subli25",    "1,I,4",      o_subli25,  O_IP },
{ "and",        "1,I,4",      o_and25,	  O_IP },
{ "or",         "1,I,4",      o_or25,	  O_IP },
{ "xor",        "1,I,4",      o_xor25,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "st.f",       "h,1,I",      o_st_f25,	  O_AS },
{ "st.d",       "h,1,I",      o_st_d25,	  O_AS },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
};


static struct tachyon_opcode oper_alri25[] = {
{ "nop",        "",           o_nop,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "addi25",     "1,I,4",      o_addi25,	  O_IP },
{ "addli25",    "1,I,4",      o_addli25,  O_IP },
{ "subi25",     "1,I,4",      o_subi25,	  O_IP },
{ "subli25",    "1,I,4",      o_subli25,  O_IP },
{ "and",        "1,I,4",      o_and25,	  O_IP },
{ "or",         "1,I,4",      o_or25,	  O_IP },
{ "xor",        "1,I,4",      o_xor25,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
};


static struct tachyon_opcode oper_alri5[] = {
{ "nop",        "",           o_nop,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "add",        "1,2,3",      o_add,	  O_IP },
{ "addl",       "1,2,3",      o_addl,	  O_IP },
{ "sub",        "1,2,3",      o_sub,	  O_IP },
{ "subl",       "1,2,3",      o_subl,	  O_IP },
{ "and",        "1,2,3",      o_and,	  O_IP },
{ "or",         "1,2,3",      o_or, 	  O_IP },
{ "xor",        "1,2,3",      o_xor,	  O_IP },
{ "sra",        "1,2,3",      o_sra,	  O_IP },
{ "srl",        "1,2,3",      o_srl,	  O_IP },
{ "sll",        "1,2,3",      o_sll,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "addi5",      "1,i,3",      o_addi5,	  O_IP },
{ "addli5",     "1,i,3",      o_addli5,	  O_IP },
{ "subi5",      "1,i,3",      o_subi5,	  O_IP },
{ "subli5",     "1,i,3",      o_subli5,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "sra",        "1,u,3",      o_sra5,	  O_IP },
{ "srl",        "1,u,3",      o_srl5,	  O_IP },
{ "sll",        "1,u,3",      o_sll5,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
};


static struct tachyon_opcode oper_ali5[] = {
{ "nop",        "",           o_nop,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "add",        "1,2,3",      o_add,	  O_IP },
{ "addl",       "1,2,3",      o_addl,	  O_IP },
{ "sub",        "1,2,3",      o_sub,	  O_IP },
{ "subl",       "1,2,3",      o_subl,	  O_IP },
{ "and",        "1,2,3",      o_and,	  O_IP },
{ "or",         "1,2,3",      o_or, 	  O_IP },
{ "xor",        "1,2,3",      o_xor,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "addi5",      "1,i,3",      o_addi5,	  O_IP },
{ "addli5",     "1,i,3",      o_addli5,	  O_IP },
{ "subi5",      "1,i,3",      o_subi5,	  O_IP },
{ "subli5",     "1,i,3",      o_subli5,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
};


static struct tachyon_opcode oper_flo[] = {
{ "nop",        "",           o_nop,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "fadd",       "e,f,g",      o_fadd,	  O_FH },
{ "fadd.d",     "e,f,g",      o_fadd_d,	  O_FH },
{ "fsub",       "e,f,g",      o_fsub,	  O_FH },
{ "fsub.d",     "e,f,g",      o_fsub_d,	  O_FH },
{ "fmul",       "e,f,g",      o_fmul,	  O_FL },
{ "fmul.d",     "e,f,g",      o_fmul_d,	  O_FL },
{ "fdiv",       "e,f,g",      o_fdiv,	  O_FL },
{ "fdiv.d",     "e,f,g",      o_fdiv_d,	  O_FL },
{ "fcnv.fd",    "e,g",        o_fcnv_fd,  O_FH },
{ "fcnv.df",    "e,g",        o_fcnv_df,  O_FH },
{ "fcnv.if",    "e,g",        o_fcnv_if,  O_FH },
{ "fcnv.fi",    "e,g",        o_fcnv_fi,  O_FH },
{ "fcnv.id",    "e,g",        o_fcnv_id,  O_FH },
{ "fcnv.di",    "e,g",        o_fcnv_di,  O_FH },
{ "???",        "",           NULL,	  O_NL },
{ "fcnz.fi",    "e,g",        o_fcnz_fi,  O_FH },
{ "???",        "",           NULL,	  O_NL },
{ "fcnz.di",    "e,g",        o_fcnz_di,  O_FH },
{ "fcmp",       "e,f",        o_fcmp,	  O_FH },
{ "fcmp.d",     "e,f",        o_fcmp_d,	  O_FH },
{ "fcmp.x",     "e,f",        o_fcmp_x,	  O_FH },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "fadd.x",     "e,f,g",      o_fadd_x,	  O_FH },
{ "???",        "",           NULL,	  O_NL },
{ "fsub.x",     "e,f,g",      o_fsub_x,	  O_FH },
{ "???",        "",           NULL,	  O_NL },
{ "fcnv.dx",    "e,g",        o_fcnv_dx,  O_FH },
{ "fcnv.xd",    "e,g",        o_fcnv_xd,  O_FH },
};


static struct tachyon_opcode oper_floa[] = {
{ "nop",        "",           o_nop,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "fadd",       "e,f,g",      o_fadd,	  O_FH },
{ "fadd.d",     "e,f,g",      o_fadd_d,	  O_FH },
{ "fsub",       "e,f,g",      o_fsub,	  O_FH },
{ "fsub.d",     "e,f,g",      o_fsub_d,	  O_FH },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "fcnv.fd",    "e,g",        o_fcnv_fd,  O_FH },
{ "fcnv.df",    "e,g",        o_fcnv_df,  O_FH },
{ "fcnv.if",    "e,g",        o_fcnv_if,  O_FH },
{ "fcnv.fi",    "e,g",        o_fcnv_fi,  O_FH },
{ "fcnv.id",    "e,g",        o_fcnv_id,  O_FH },
{ "fcnv.di",    "e,g",        o_fcnv_di,  O_FH },
{ "???",        "",           NULL,	  O_NL },
{ "fcnz.fi",    "e,g",        o_fcnz_fi,  O_FH },
{ "???",        "",           NULL,	  O_NL },
{ "fcnz.di",    "e,g",        o_fcnz_di,  O_FH },
{ "fcmp",       "e,f",        o_fcmp,	  O_FH },
{ "fcmp.d",     "e,f",        o_fcmp_d,	  O_FH },
{ "fcmp.x",     "e,f",        o_fcmp_x,	  O_FH },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "fadd.x",     "e,f,g",      o_fadd_x,	  O_FH },
{ "???",        "",           NULL,	  O_NL },
{ "fsub.x",     "e,f,g",      o_fsub_x,	  O_FH },
{ "???",        "",           NULL,	  O_NL },
{ "fcnv.dx",    "e,g",        o_fcnv_dx,  O_FH },
{ "fcnv.xd",    "e,g",        o_fcnv_xd,  O_FH },
};


static struct tachyon_opcode oper_flom[] = {
{ "nop",        "",           o_nop,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "fmul",       "e,f,g",      o_fmul,	  O_FL },
{ "fmul.d",     "e,f,g",      o_fmul_d,	  O_FL },
{ "fdiv",       "e,f,g",      o_fdiv,	  O_FL },
{ "fdiv.d",     "e,f,g",      o_fdiv_d,	  O_FL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
};


static struct tachyon_opcode oper_sys[] = {
{ "nop",        "",           o_nop,	  O_NL },
{ "svc",        "Y",          o_svc,	  O_IP },
{ "ldsr",       "y,z,U",      o_ldsr,	  O_AL },
{ "stsr",       "U,y,z",      o_stsr,	  O_AS },
{ "movgs",      "1,V",        o_movgs,	  O_IP },
{ "movsg",      "V,1",        o_movsg,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "ptlbe",      "y,z",        o_ptlbe,	  O_IP },
{ "ptlbp",      "y,z",        o_ptlbp,	  O_IP },
{ "ptlba",      "y,z",        o_ptlba,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "flche",      "y,z",        o_flche,	  O_IP },
{ "flcha",      "y,z",        o_flcha,	  O_IP },
{ "aki",        "X,5",        o_aki,	  O_IP },
{ "rti",        "",           o_rti,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "cs",         "1,3,y,z,5",  o_cs,	  O_AS },
{ "ld.mcr",     "y,z,5",      o_ld_mcr,	  O_AL },
{ "st.mcr",     "5,y,z",      o_st_mcr,	  O_AS },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "???",        "",           NULL,	  O_NL },
{ "wbst",       "z",          o_wbst,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "movgt",      "1",          o_movgt,	  O_IP },
{ "movtg",      "1",          o_movtg,	  O_IP },
{ "???",        "",           NULL,	  O_NL },
{ "fwc",        "Y",          o_fwc,	  O_IP },
};


static struct tachyon_opcode oper_cop[] = {
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vnm",        "L,M,N,H",    o_COP,      O_VP },
{ "vom",        "L,M,N,H",    o_COP,      O_VP },
{ "vxm",        "L,M,N,H",    o_COP,      O_VP },
{ "vtm",        "M,N,H",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },

{ "vil",        "6,A,C,O",    o_COP,      O_VP },
{ "vist",       "B,6,A,O",    o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vn",	        "A,B,C,O",    o_COP,      O_VP },
{ "vo",	        "A,B,C,O",    o_COP,      O_VP },
{ "vx",	        "A,B,C,O",    o_COP,      O_VP },
{ "vesll",      "A,B,C,O",    o_COP,      O_VP },
{ "va",	        "A,B,C,O",    o_COP,      O_VP },
{ "vs",	        "A,B,C,O",    o_COP,      O_VP },
{ "vm",	        "A,B,C,O",    o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vlr",        "B,C,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vlp",        "B,C,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },

{ "vild",       "6,A,C,O",    o_COP,      O_VP },
{ "vistd",      "B,6,A,O",    o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vvid",       "B,C,O",      o_COP,      O_VP },
{ "vvdi",       "B,C,O",      o_COP,      O_VP },
{ "vved",       "B,C,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vad",        "A,B,C,O",    o_COP,      O_VP },
{ "vsd",        "A,B,C,O",    o_COP,      O_VP },
{ "vmd",        "A,B,C,O",    o_COP,      O_VP },
{ "vdd",        "A,B,C,O",    o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vlpd",       "B,C,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },

{ "vile",       "6,A,C,O",    o_COP,      O_VP },
{ "viste",      "B,6,A,O",    o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vvie",       "B,C,O",      o_COP,      O_VP },
{ "vvei",       "B,C,O",      o_COP,      O_VP },
{ "vvde",       "B,C,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vae",        "A,B,C,O",    o_COP,      O_VP },
{ "vse",        "A,B,C,O",    o_COP,      O_VP },
{ "vme",        "A,B,C,O",    o_COP,      O_VP },
{ "vde",        "A,B,C,O",    o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vlpe",       "B,C,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },


{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },

{ "vcf",        "A,B,N,O",    o_COP,      O_VP },
{ "vch",        "A,B,N,O",    o_COP,      O_VP },
{ "vcl",        "A,B,N,O",    o_COP,      O_VP },
{ "vcne",       "A,B,N,O",    o_COP,      O_VP },
{ "vce",        "A,B,N,O",    o_COP,      O_VP },
{ "vcnl",       "A,B,N,O",    o_COP,      O_VP },
{ "vcnh",       "A,B,N,O",    o_COP,      O_VP },
{ "vct",        "A,B,N,O",    o_COP,      O_VP },
{ "vcsf",       "7,B,N,O",    o_COP,      O_VP },
{ "vcsh",       "7,B,N,O",    o_COP,      O_VP },
{ "vcsl",       "7,B,N,O",    o_COP,      O_VP },
{ "vcsne",      "7,B,N,O",    o_COP,      O_VP },
{ "vcse",       "7,B,N,O",    o_COP,      O_VP },
{ "vcsnl",      "7,B,N,O",    o_COP,      O_VP },
{ "vcsnh",      "7,B,N,O",    o_COP,      O_VP },
{ "vcst",       "7,B,N,O",    o_COP,      O_VP },

{ "vcfd",       "A,B,N,O",    o_COP,      O_VP },
{ "vchd",       "A,B,N,O",    o_COP,      O_VP },
{ "vcld",       "A,B,N,O",    o_COP,      O_VP },
{ "vcned",      "A,B,N,O",    o_COP,      O_VP },
{ "vced",       "A,B,N,O",    o_COP,      O_VP },
{ "vcnld",      "A,B,N,O",    o_COP,      O_VP },
{ "vcnhd",      "A,B,N,O",    o_COP,      O_VP },
{ "vctd",       "A,B,N,O",    o_COP,      O_VP },
{ "vcsfd",      "E,B,N,O",    o_COP,      O_VP },
{ "vcshd",      "E,B,N,O",    o_COP,      O_VP },
{ "vcsld",      "E,B,N,O",    o_COP,      O_VP },
{ "vcsned",     "E,B,N,O",    o_COP,      O_VP },
{ "vcsed",      "E,B,N,O",    o_COP,      O_VP },
{ "vcsnld",     "E,B,N,O",    o_COP,      O_VP },
{ "vcsnhd",     "E,B,N,O",    o_COP,      O_VP },
{ "vcstd",      "E,B,N,O",    o_COP,      O_VP },

{ "vcfe",       "A,B,N,O",    o_COP,      O_VP },
{ "vche",       "A,B,N,O",    o_COP,      O_VP },
{ "vcle",       "A,B,N,O",    o_COP,      O_VP },
{ "vcnee",      "A,B,N,O",    o_COP,      O_VP },
{ "vcee",       "A,B,N,O",    o_COP,      O_VP },
{ "vcnle",      "A,B,N,O",    o_COP,      O_VP },
{ "vcnhe",      "A,B,N,O",    o_COP,      O_VP },
{ "vcte",       "A,B,N,O",    o_COP,      O_VP },
{ "vcsfe",      "E,B,N,O",    o_COP,      O_VP },
{ "vcshe",      "E,B,N,O",    o_COP,      O_VP },
{ "vcsle",      "E,B,N,O",    o_COP,      O_VP },
{ "vcsnee",     "E,B,N,O",    o_COP,      O_VP },
{ "vcsee",      "E,B,N,O",    o_COP,      O_VP },
{ "vcsnle",     "E,B,N,O",    o_COP,      O_VP },
{ "vcsnhe",     "E,B,N,O",    o_COP,      O_VP },
{ "vcste",      "E,B,N,O",    o_COP,      O_VP },


{ "vlm",        "G,6,7,N",    o_COP,      O_VP },
{ "vstm",       "G,M,6,7",    o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vgm",        "7,N,H",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vgs",        "7,C,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },

{ "vl",	        "G,6,7,C,O",  o_COP,      O_VP },
{ "vst",        "G,B,6,7,O",  o_COP,      O_VP },
{ "vlcp",       "G,6,7,C,O",  o_COP,      O_VP },
{ "vstx",       "G,B,6,7,O",  o_COP,      O_VP },
{ "vns",        "7,B,C,O",    o_COP,      O_VP },
{ "vos",        "7,B,C,O",    o_COP,      O_VP },
{ "vxs",        "7,B,C,O",    o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vas",        "7,B,C,O",    o_COP,      O_VP },
{ "vss",        "7,B,C,O",    o_COP,      O_VP },
{ "vms",        "7,B,C,O",    o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vls",        "7,C,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },

{ "vld",        "G,6,7,C,O",  o_COP,      O_VP },
{ "vstd",       "G,B,6,7,O",  o_COP,      O_VP },
{ "vlcpd",      "G,6,7,C,O",  o_COP,      O_VP },
{ "vstxd",      "G,B,6,7,O",  o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vsrl",       "7,B,C,O",    o_COP,      O_VP },
{ "vasd",       "E,B,C,O",    o_COP,      O_VP },
{ "vssd",       "E,B,C,O",    o_COP,      O_VP },
{ "vmsd",       "E,B,C,O",    o_COP,      O_VP },
{ "vdsd",       "E,B,C,O",    o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vlsd",       "E,C,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },

{ "vle",        "G,6,7,C,O",  o_COP,      O_VP },
{ "vste",       "G,B,6,7,O",  o_COP,      O_VP },
{ "vlcpe",      "G,6,7,C,O",  o_COP,      O_VP },
{ "vstxe",      "G,B,6,7,O",  o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vsll",       "7,B,C,O",    o_COP,      O_VP },
{ "vase",       "E,B,C,O",    o_COP,      O_VP },
{ "vsse",       "E,B,C,O",    o_COP,      O_VP },
{ "vmse",       "E,B,C,O",    o_COP,      O_VP },
{ "vdse",       "E,B,C,O",    o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vlse",       "E,C,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },


{ "vplcr",      "7,Q",        o_COP,      O_VP },
{ "vpscr",      "P,9",        o_COP,      O_VP },
{ "vlvt",       "7,T",        o_COP,      O_VP },
{ "vsvt",       "S,9",        o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vsmm",       "L,9,H",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vfm",        "L,8,H",      o_COP,      O_VP },
{ "vfmx",       "L,8,H",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },

{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vltr",       "6",          o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vts",        "7,B,9",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vsm",        "A,9,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vfx",        "A,8,9,O",    o_COP,      O_VP },
{ "vfxx",       "A,8,9,O",    o_COP,      O_VP },
{ "vfn",        "A,8,9,O",    o_COP,      O_VP },
{ "vfnx",       "A,8,9,O",    o_COP,      O_VP },

{ "vlcr",       "7,Q",        o_COP,      O_VP },
{ "vscr",       "P,9",        o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vtsd",       "7,B,F",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vsmd",       "A,F,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vfxd",       "A,8,F,O",    o_COP,      O_VP },
{ "vfxdx",      "A,8,F,O",    o_COP,      O_VP },
{ "vfnd",       "A,8,F,O",    o_COP,      O_VP },
{ "vfndx",      "A,8,F,O",    o_COP,      O_VP },

{ "vpt",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vlvl",       "7",          o_COP,      O_VP },
{ "vlvlf",      "7",          o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vtse",       "7,B,F",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vsme",       "A,F,O",      o_COP,      O_VP },
{ "???",        "",           o_COP,      O_VP },
{ "vfxe",       "A,8,F,O",    o_COP,      O_VP },
{ "vfxex",      "A,8,F,O",    o_COP,      O_VP },
{ "vfne",       "A,8,F,O",    o_COP,      O_VP },
{ "vfnex",      "A,8,F,O",    o_COP,      O_VP },
};

/*
 * true if ICCP & ICCS don't match with OP0 & OP1
 */

static char type_exchange_loc[16] = {
/* Never exchange location (op0/op1)                 */
/* 0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f    */
/* 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }; */
/* Exchange location (op0/op1) ... Don't use         */
/* 0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f    */
   0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0 };

/*
 * true if MULT/MOV is in OP0
 */

static char type_multi_in_op0[16] = {
/* 0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f    */
   0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0 };

/*
 * instruction - opcode table
 */

static struct tachyon_insn
{
    struct tachyon_opcode  *op0;
    struct tachyon_opcode  *op1;
    struct tachyon_opcode  *op2;
    unsigned char          mask0;
    unsigned char          mask1;
    unsigned char          mask2;
    unsigned char          sft0;
    unsigned char          sft1;
    unsigned char          sft2;
};


static struct tachyon_insn insn_type[] = {
{ oper_undef, 	NULL,		NULL,		0x00, 0x00, 0x00,  0,  0,  0 },
{ oper_undef, 	NULL,		NULL,		0x00, 0x00, 0x00,  0,  0,  0 },
{ oper_lsmj, 	oper_floa,	oper_flom,	0x1f, 0x1f, 0x1f, 23, 23, 23 },
{ oper_alri5, 	oper_floa,	oper_flom,	0x1f, 0x1f, 0x1f, 23, 23, 23 },
{ oper_alri5, 	oper_lsmj,	oper_flo,	0x1f, 0x1f, 0x1f, 23, 23, 23 },
{ oper_alri5, 	oper_ali5,	oper_flo,	0x1f, 0x1f, 0x1f, 23, 23, 23 },
{ oper_alri5, 	oper_lsm,	oper_brci12,	0x1f, 0x1f, 0x07, 23, 23, 25 },
{ oper_alri5, 	oper_ali5,	oper_brci12,	0x1f, 0x1f, 0x07, 23, 23, 25 },
{ oper_lsmj, 	oper_alri25,	NULL,		0x1f, 0x1f, 0x00, 23, 23,  0 },
{ oper_alri5, 	oper_alsmi25,	NULL,		0x1f, 0x1f, 0x00, 23, 23,  0 },
{ oper_flo, 	oper_alsmi25,	NULL,		0x1f, 0x1f, 0x00, 23, 23,  0 },
{ oper_sys, 	NULL,		NULL,		0x1f, 0x00, 0x00, 23,  0,  0 },
{ oper_lsm, 	oper_cbri32,	NULL,		0x1f, 0x07, 0x00, 23, 25,  0 },
{ oper_alri5, 	oper_cbri32,	NULL,		0x1f, 0x07, 0x00, 23, 25,  0 },
{ oper_flo, 	oper_cbri32,	NULL,		0x1f, 0x07, 0x00, 23, 25,  0 },
{ oper_cop, 	NULL,		NULL,	        0xc0, 0x30, 0x0f,  0, 24, 16 },
};


/*
 * instruction format
 */

static struct iform0
{
  unsigned insn_l;
  unsigned insn_u;
};

static struct iform1
{
  unsigned dm_0 : 32;
  unsigned dm_1 : 8;
  unsigned r3   : 5;
  unsigned r2   : 5;
  unsigned r1   : 5;
  unsigned op   : 5;
  unsigned ty   : 4;
};

static struct iform2
{
  unsigned dm_0 : 32;
  unsigned dm_1 : 8;
  unsigned r3   : 5;
  unsigned im   : 12;
  unsigned op   : 3;
  unsigned ty   : 4;
};

static struct iform3
{
  unsigned dm_0 : 20;
  unsigned r3   : 5;
  unsigned im_l : 7;
  unsigned im_u : 18;
  unsigned r1   : 5;
  unsigned op   : 5;
  unsigned ty   : 4;
};

static struct iform4
{
  unsigned dm_0 : 20;
  unsigned r3   : 5;
  unsigned im_l : 7;
  unsigned im_u : 25;
  unsigned op   : 3;
  unsigned ty   : 4;
};

static struct iform5
{
  unsigned dm_l : 5;
  unsigned im   : 8;
  unsigned dm_m : 19;
  unsigned dm_u : 23;
  unsigned op   : 5;
  unsigned ty   : 4;
};

static struct iform6
{
  unsigned r3   : 5;
  unsigned im   : 25;
  unsigned r1_l : 2;
  unsigned r1_u : 3;
  unsigned dm   : 20;
  unsigned op   : 5;
  unsigned ty   : 4;
};

static struct iform7
{
  unsigned dm_l : 32;
  unsigned dm_u : 8;
  unsigned r2   : 10;
  unsigned r1   : 5;
  unsigned op   : 5;
  unsigned ty   : 4;
};

static struct iform8
{
  unsigned r3   : 5;
  unsigned im   : 4;
  unsigned dm_l : 23;
  unsigned dm_u : 23;
  unsigned op   : 5;
  unsigned ty   : 4;
};

static struct iform9
{
  unsigned dm_l : 32;
  unsigned dm_u : 23;
  unsigned op   : 5;
  unsigned ty   : 4;
};

static struct iforma
{
  unsigned r4   : 5;
  unsigned im   : 25;
  unsigned r2_l : 2;
  unsigned r2_u : 3;
  unsigned dm_2 : 5;
  unsigned r3   : 5;
  unsigned dm_1 : 5;
  unsigned r1   : 5;
  unsigned op   : 5;
  unsigned ty   : 4;
};

static struct iformb
{
  unsigned R3     : 8;
  unsigned R4     : 8;
  unsigned op2    : 4;
  unsigned R2     : 8;
  unsigned op1    : 2;
  unsigned R1_l   : 2;
  unsigned R1_u   : 6;
  unsigned op0    : 2;
  unsigned r_use  : 5;
  unsigned r1     : 5;
  unsigned r_type : 6;
  unsigned u_id   : 4;
  unsigned ty     : 4;
};

static union insn_form
{
  struct iform0 ful;
  struct iform1 i5_5_5_5;
  struct iform2 i3_12_5;
  struct iform3 i5_5_25_5;
  struct iform4 i3_32_5;
  struct iform5 sys_svc;
  struct iform6 sys_25;
  struct iform7 sys_mov;
  struct iform8 sys_aki;
  struct iform9 sys_rti;
  struct iforma cs;
  struct iformb cop;
};

/*
 * star0/1 alias
 */

static char * star0[] = {
 "true",	"iccp.z",	"iccp.l",	"iccp.le",
 "iccp.ul",	"iccp.ule",	"iccp.n",	"iccp.v",
 "oz",		"iccs.z",	"iccs.l",	"iccs.le",
 "iccs.ul",	"iccs.ule",	"iccs.n",	"iccs.v",
 "fcc.e",	"fcc.l",	"fcc.g",	"fcc.u",
 "fcc.le",	"fcc.ge",	"fcc.ue",	"23",
 "24",		"25",		"26",		"27",
 "28",		"29",		"30",		"31",
};

static char * star1[] = {
 "asa",		"asam",		"asaf",		"asau",
 "36",		"37",		"38",		"39",
 "es0",		"es1",		"es2",		"es3",
 "cba",		"cbu",		"cbs",		"47",
 "cb0",		"cb1",		"cb2",  	"cb3",  
 "cb4",   	"cb5",   	"cb6",   	"cb7",  
 "cb8",   	"cb9",   	"cb10",   	"cb11",  
 "cb12",   	"cb13",   	"cb14",   	"cb15",  
};
