static char SccsId[] = "@(#)atof.c	1.1    6/12/91";

/* TaChyon/KAI Instruction simulator      */
/*  Copyright (c) 1989 by FUJITSU LIMITED */
/*            Primary writer: Y.Nakashima */

/*    atof.c   90/02/06   */

#define    Tachyon_float_bias     0x7f
#define    Tachyon_double_bias    0x3ff
#define    Tachyon_extended_bias  0x3fff
#define    Tachyon_float_emax     0xff
#define    Tachyon_double_emax    0x7ff
#define    Tachyon_extended_emax  0x7fff

static struct Tachyon_float_format
{
    unsigned s         :1;
    unsigned e         :8;
    unsigned f         :23;
};
 
static struct Tachyon_double_format
{
    unsigned s         :1;
    unsigned e         :11;
    unsigned f1        :20;
    unsigned f2        :32;
};
 
static struct Tachyon_extended_format
{
    unsigned s         :1;
    unsigned e         :15;
    unsigned rev0      :16;
    unsigned rev1      :32;
    unsigned l         :1;
    unsigned f1        :31;
    unsigned f2        :32;
};

#ifdef     UTS

#define    F_float_bias           0x40
#define    F_float_emax           0x7f
#define    F_cardinal_number      4 /* 16 */
#define    F_sign_size            1
#define    F_exponent_size        7
#define    F_mantissa_size        24

#else
/* sun assumed */

#define    F_float_bias           0x7f
#define    F_float_emax           0xff
#define    F_cardinal_number      1 /* 2 */
#define    F_sign_size            1
#define    F_exponent_size        8
#define    F_mantissa_size        23

#endif

static struct F_format
{
    unsigned s:F_sign_size;
    unsigned e:F_exponent_size;
    unsigned n:F_mantissa_size;
};

static union
{
    struct   F_format fd;
    float    sf;
} ud;

static union
{
    struct Tachyon_float_format      f_data;
    struct Tachyon_double_format     d_data;
    struct Tachyon_extended_format   e_data;
    unsigned long                    fr[4];
} un_fr;

f_to_tachyon( f, mode, d )
    float    *f;
    int      mode; /* 'f' or 'd' or 'x' */
    unsigned *d;
{
    unsigned s, n;
    int      e;
    int      tmp;

    for( tmp=0 ; tmp<3 ; tmp++ ){
        un_fr.fr[tmp] = 0;
    }
    ud.sf = *f;
    e = ( ud.fd.e - F_float_bias ) * F_cardinal_number;
    s = ud.fd.s;
    n = ud.fd.n << ( 32 - F_mantissa_size );

#ifdef     UTS

    while( n && !( n >> 31 ) ){
        n = n << 1;
        e--;
    }
    n = n << 1;
    e--;

#endif

    switch( mode ){
    case 'f' :
            un_fr.f_data.s = s;
            if( ( e + Tachyon_float_bias >= Tachyon_float_emax ) ||
                ( e + Tachyon_float_bias <  1 ) ){
                if( !ud.fd.e ){
                    e = 0;
                }
                else{
                    return( 1 );
                }
            }
            else{
                un_fr.f_data.e = e + Tachyon_float_bias;
                un_fr.f_data.f = n >> 9;
            }
            *d = un_fr.fr[0];
            break;
    case 'd' :
            un_fr.d_data.s = s;
            if( ( e + Tachyon_double_bias >= Tachyon_double_emax ) ||
                ( e + Tachyon_double_bias <  1 ) ){
                if( !ud.fd.e ){
                    e = 0;
                }
                else{
                    return( 1 );
                }
            }
            else{
                un_fr.d_data.e = e + Tachyon_double_bias;
                un_fr.d_data.f1= n >> 12;
                un_fr.d_data.f2= n << 20;
            }
            *d++ = un_fr.fr[0];
            *d = un_fr.fr[1];
            break;
    case 'x' :
            un_fr.e_data.s = s;
            if( ( e + Tachyon_extended_bias >= Tachyon_extended_emax ) ||
                ( e + Tachyon_extended_bias <  1 ) ){
                if( !ud.fd.e ){
                    e = 0;
                    un_fr.e_data.l = 0;
                }
                else{
                    return( 1 );
                }
            }
            else{
                un_fr.e_data.e = e + Tachyon_extended_bias;
                un_fr.e_data.l = 1;
                un_fr.e_data.f1= n >> 1;
                un_fr.e_data.f2= n << 31;
            }
            *d++ = un_fr.fr[0];
            *d++ = un_fr.fr[1];
            *d++ = un_fr.fr[2];
            *d   = un_fr.fr[3];
            break;
    defualt :
            break;
    }
    return( 0 );
}

tachyon_to_f( d, mode, f )
    unsigned *d;
    int      mode; /* 'f' or 'd' or 'x' */
    float    *f;
{
    unsigned s, n;
    int      e;
    int      tmp;
    int      z;

    switch( mode ){
    case 'f' :
            un_fr.fr[0] = *d;
            s = un_fr.f_data.s;
            z = un_fr.f_data.e;
            e = un_fr.f_data.e - Tachyon_float_bias;
            n = un_fr.f_data.f << 9;
            break;
    case 'd' :
            un_fr.fr[0] = *d++;
            un_fr.fr[1] = *d;
            s = un_fr.d_data.s;
            z = un_fr.d_data.e;
            e = un_fr.d_data.e - Tachyon_double_bias;
            n = (un_fr.d_data.f1 << 12) + (un_fr.d_data.f2 >> 20);
            break;
    case 'x' :
            un_fr.fr[0] = *d++;
            un_fr.fr[1] = *d++;
            un_fr.fr[2] = *d++;
            un_fr.fr[3] = *d;
            s = un_fr.e_data.s;
            z = un_fr.e_data.e;
            e = un_fr.e_data.e - Tachyon_extended_bias;
            if( !un_fr.e_data.l ){
                return( 1 );
            }
            n = (un_fr.e_data.f1 << 1) + (un_fr.e_data.f2 >> 31);
            break;
    defualt :
            return( 0 );
            break;
    }
    ud.fd.s = s;

#ifdef     UTS

    e++;
    n = ( n >> 1 ) | 0x80000000;

#endif

    if( z ){
        for( tmp=0 ; e % F_cardinal_number ; tmp++ ){
            e++;
            n = n >> 1;
        }
        e = e / F_cardinal_number;
        if( ( e + F_float_bias >= F_float_emax ) ||
            ( e + F_float_bias < 1 ) ){
            return( 1 );
        }
        ud.fd.e = e + F_float_bias;
        ud.fd.n = n >> ( 32 - F_mantissa_size );
    }
    else{
        ud.fd.e = ud.fd.n = 0;
    }
    *f = ud.sf;
    return( 0 );
}

