static char SccsId[] = "@(#)command.c	1.29    12/16/93";

/* TaChyon/KAI Instruction simulator      */
/*  Copyright (c) 1989 by FUJITSU LIMITED */
/*            Primary writer: Y.Nakashima */

/* command.c    89/10/1 */

#include "tim.h"

#define         COMMAND_LINE     (256)
static char     COMMAND_BUF[COMMAND_LINE*2];
static char     PREV_COMMAND_BUF[COMMAND_LINE*2];
static char     tmp_buf[COMMAND_LINE*2];

int	        interrupt;
static jmp_buf  sjbuf;
static unsigned c_pe;		/* current processor # */
static unsigned c_mode;		/* interp. of scalar address {real|virtual} */
static unsigned c_node;		/* interp. of vector address {real|virtual} */
static unsigned c_echo;		/* echo back */
static unsigned c_logview;	/* log viewer */
static unsigned c_forcew;	/* ignore write protect */
static unsigned c_symbol;	/* display address by symbol or hex */
static unsigned c_coint;	/* interrupt level of logical console */

#define    skipxdigit() { for (; isxdigit (*cmdP); cmdP++); }
#define    skipdigit()  { for (; isdigit (*cmdP); cmdP++); }
#define    skipspc()    { for (; isspace (*cmdP); cmdP++); }
#define    isspc()      (!*cmdP || isspace (*cmdP))
#define    optbit(c)    ((unsigned)0x80000000 >> ((c)-0x60))

/* for nested scripts */

#define         SCRIPT_NUM   (4)
static struct	script { FILE *script_fp; char *param[10]; };
static struct	script SCRIPT[SCRIPT_NUM];
static int      script_ptr = 0;

/* for viewer */

#define         VIEW_NUM       (40)
#define         LINE_CHAR_SIZE (82)

static struct   view_line {
                  char     compress;    /* 'x': hold 4 contents */
                                        /* 'c': hold 8 contents */
	          unsigned pe[4];       /* pe # */
                  char     mode[4];     /* 'r'eal / 'v'irtual */
                  char     facility[4]; /* type of facility */
                  unsigned select[4];   /* selection of vr/mr number */
                  unsigned num[4];      /* reg # */
	          char     prv[LINE_CHAR_SIZE]; /* previous contents */
                  char     new[LINE_CHAR_SIZE]; /* current contents */
		};

static struct   view_line VIEW_LINE[VIEW_NUM];
static int      view_end = -1;

command(source, fp, sp, log)
  char source; /* 'f' for file || 's' for string */
  FILE *fp;
  char *sp;
  int log;   /* 0:no log, 1:log_command() */
{
  int         onintr_cmd();
  int         onintr_sys();
#ifdef CONSOLE
  int         onintr_con1();
  int         onintr_con2();
#endif
  char        *get_srname();
#ifdef MCR_AREA
  char        *get_mcrname();
#endif
  long        gettme();
  static char *save_cmdP, *cmdP;

  if (c_mode        == 0) c_mode        = 'r';
  if (c_node        == 0) c_node        = 'r';
  if (c_echo        == 0) c_echo        = 'n';
  if (c_logview     == 0) c_logview     = 'n';
  if (c_forcew      == 0) c_forcew      = 'n';
  if (c_symbol      == 0) c_symbol      = 'n';
  if (c_coint       == 0) c_coint       = 0;

  for (;;) {

    if (source == 'f') {
      signal(SIGINT, SIG_IGN);
      setjmp(sjbuf);
      signal(SIGINT, onintr_cmd);
#ifdef CONSOLE
      signal(SIGUSR1, SIG_IGN);
      signal(SIGUSR1, onintr_con1);
      signal(SIGUSR2, SIG_IGN);
      signal(SIGUSR2, onintr_con2);
#endif
    }

    view(c_logview);

    if (source == 'f') { /* from file */
      prompt(c_pe, fp);
      if (!fgets(cmdP = COMMAND_BUF, COMMAND_LINE, fp)) {
        message("EOF\n");
        return;
      }
      if (!*cmdP) /* illegal file (? binary ?) */
        return;
    }
    else if (source == 's') { /* from string PTR */
      cmdP = sp;
      source = ' ';
    }
    else /* nothing */
      return;

  /* GDB commands */
    if (*cmdP == '$') {
#ifdef GDB
      if (kiodebug)
        message("command(): Pkt received: %s", cmdP);
      gdb_command(fp, cmdP+1, c_pe);
#else
      message("gdb command interface not available. (compiled with no -DGDB)\n");
#endif
      continue;
    }


  /* INTERACTIVE commands */
    expand_command(cmdP, PREV_COMMAND_BUF, COMMAND_BUF, fp);
    strncpy(PREV_COMMAND_BUF, COMMAND_BUF, COMMAND_LINE);
    if (log) log_command(COMMAND_BUF);
    delete_comment(COMMAND_BUF);
    cmdP = COMMAND_BUF;

    if (isdigit(*cmdP)) {
      unsigned i;
      
      i = *cmdP++ - '0';
      if (isdigit(*cmdP))
        i = 10 * i + (*cmdP++ - '0');
      if (isdigit(*cmdP))
        i = 10 * i + (*cmdP++ - '0');
      if (*cmdP == ':') {
        if (i < PE_NUM) {
          c_pe = i;
          cmdP++;
          skipspc();
        }
        else {
          message("PE#%u not exist\n", i);
          continue;
        }
      }
      else {
	message("':' expected\n");
        continue;
      }
    }

    save_cmdP = cmdP;
    switch (*cmdP++) {
    case 0: /* do nothing */
      continue;

    case '\\': /* execute UNIX command */
      sprintf(tmp_buf, "csh -c \"%s\"", cmdP);
      req_system(tmp_buf);
      continue;

    case 'e':
      if (!strncmp("nv", cmdP, 2)) { /* display environment variables */
        cmdP += 2;
        if (!isspc()) break;
        ignore_rest_of_line(cmdP);
        message("MODULE       =\"%s\"\n", ENV_MODULE);
        message("INITRC       =\"%s\"\n", ENV_INITRC);
        message("COMMAND      =\"%s\"\n", ENV_COMMAND);
        message("CMDDIR       =\"%s\"\n", ENV_CMDDIR);
        message("HISTORY      =\"%s\"\n", ENV_HISTORY);
        message("DISPLOG      =\"%s\"\n", ENV_DISPLOG);
        message("GDBREPL      =\"%s\"\n", ENV_GDBREPL);
        message("ALLFAC       =\"%s\"\n", ENV_ALLFAC);
        message("DASD         =\"%s\"\n", ENV_DASD);
#ifdef SSUMOVER
        message("SSU          =\"%s\"\n", ENV_SSU);
#endif
#ifdef CONSOLE
        message("CONSOLEIN    =\"%s\"\n", ENV_COIN);
        message("CONSOLEOUT   =\"%s\"\n", ENV_COUT);
#endif
        message("fast_mv      =%s\n", fast_mv?"yes":"no");
        message("old_mode     =%s\n", old_mode?"yes":"no");
        message("kiodebug     =%s\n", kiodebug?"yes":"no");
        message("current pe#  =%u\n", c_pe);
        message("mode         ='%c'\n", c_mode);
#ifdef VCP
        message("node         ='%c'\n", c_node);
#endif
#ifdef MCR_AREA
        message("msmap        =%u(MB)\n", (pe[c_pe].mcr[MSMAP].ful >> 20) + 1);
#endif
        message("echo         ='%c'\n", c_echo);
        message("logview      ='%c'\n", c_logview);
        message("forcew       ='%c'\n", c_forcew);
        message("symbol       ='%c'\n", c_symbol);
#ifdef CONSOLE
        message("coint        =%u\n", c_coint);
#endif
        continue;
      }
      else if (!strncmp("cho", cmdP, 3)) { /* echo back string */
        cmdP += 3;
        if (!isspc()) break;
        if (*cmdP)
          message("%s\n", cmdP+1);
        else 
          message("%s\n", cmdP);
        continue;
      }
      break;

    case 'h':
      if (!strncmp("elp", cmdP, 3)) { /* display menu */
        cmdP += 3;
        if (!isspc()) break;
        ignore_rest_of_line(cmdP);
        sprintf(tmp_buf, "jless '-P%%f(%%i of %%m) [line %%lb/%%L %%pB\\%%]$' %s", HELP_PATH);
        req_system(tmp_buf);
        continue;
      }
      else if (!strncmp("is", cmdP, 2)) { /* jless +G timhis */
        cmdP += 2;
        if (!isspc()) break;
        ignore_rest_of_line(cmdP);
	if (history_fp) {
          sprintf(tmp_buf, "cat -n %s | jless '-P%%f(%%i of %%m) [line %%lb/%%L %%pB\\%%]$' +G", ENV_HISTORY);
          req_system(tmp_buf);
	}
	else {
	  message("history file inactive\n");
	}
        continue;
      }
      break;

    case 'q':
      if (!strncmp("uit", cmdP, 3)) { /* exit tim */
        cmdP += 3;
        if (!isspc()) break;
        ignore_rest_of_line(cmdP);
        return;
      }
      break;

    case 'a':
      if (isspc()) { /* assemble one line */
        FILE *open_file();
        FILE *write_fp;
        static struct stat   stbuf;
        static time_t        cur_time;
        static unsigned      start, dummy;
        char                 *read_f;

        if (extract_memrange(c_pe, &cmdP, 8, &start, &dummy))
          continue;
	skipspc();
	start &= ~7;
        if (!*cmdP) read_f = "";
        else if (*cmdP == '<') {
          cmdP++;
          skipspc();
          read_f = cmdP;
        }
        else {
	  message("'<' expected\n");
	  continue;
	}

        time(&cur_time);
        if (!*read_f) {
          read_f = ENV_TMPAS;
          message("start from 0x%08.8X, '.' for end...\n",start);
          if (!(write_fp = open_file(read_f, "w")))
            continue;
          for (;;) {
            arrow_prompt(fp);
            if (!fgets(tmp_buf, COMMAND_LINE, fp)) {
              message("EOF\n");
              break;
            }
            expand_command(tmp_buf, PREV_COMMAND_BUF, tmp_buf, fp);
            strncpy(PREV_COMMAND_BUF, tmp_buf, COMMAND_LINE);
            if (log) log_command(tmp_buf);
            delete_comment(tmp_buf);
            if (*tmp_buf == '.' && !*(tmp_buf+1))
              break;
            fprintf(write_fp, "   %s\n", tmp_buf);
          }
          fclose(write_fp);
        }
	else
	  message("start from 0x%08.8X\n",start);
        if (displog_fp) fclose(displog_fp);
        sprintf(tmp_buf, "((tas -w -ls -L %s -o %s | asl - %s) && tld -N -M %s -Ttext %X -o %s) 2>&1 | sed \"s/^##FORMFEED##//\" | tee -a %s",
          read_f, ENV_TMPOB, read_f, ENV_TMPOB, start, ENV_TMPLD, ENV_DISPLOG);
        if (system(tmp_buf) == -1)
          message("can't fork \"%s\"\n", tmp_buf);
        displog_fp = open_file(ENV_DISPLOG, "a");
        if (stat(ENV_TMPLD, &stbuf))
          message("can't get status of %s\n", ENV_TMPLD);
        else if (cur_time <= stbuf.st_mtime)
          start += read_module(c_pe, c_mode, 0, ENV_TMPLD, 0, 0, c_forcew);
        unlink(ENV_TMPAS);
        unlink(ENV_TMPOB);
        unlink(ENV_TMPLD);
        continue;
      }
      break;

    case 'l':
      if (!strncmp("og", cmdP, 2)) { /* jless +G timlog */
        cmdP += 2;
        if (!isspc()) break;
        ignore_rest_of_line(cmdP);
	if (displog_fp) {
          sprintf(tmp_buf, "jless '-P%%f(%%i of %%m) [line %%lb/%%L %%pB\\%%]$' +G %s", ENV_DISPLOG);
          req_system(tmp_buf);
	}
	else {
	  message("display_log inactive\n");
	}
        continue;
      }
      else if (!strncmp("oadm", cmdP, 4)) { /* read load_module */
        unsigned start, dummy;

        cmdP += 4;
        if (*cmdP == '/') {
          cmdP++;
          if (extract_memrange(c_pe, &cmdP, 8, &start, &dummy))
            continue;
          skipspc();
          if (!*cmdP) {
	    message("no {filename}\n");
	    continue;
	  }
          read_module(c_pe, c_mode, start, cmdP, 0, 1, c_forcew);
        }
        else if (isspc()) {
          skipspc();
          if (!*cmdP) {
	    message("no {filename}\n");
	    continue;
	  }
          read_module(c_pe, c_mode, 0, cmdP, 1, 0, c_forcew);
        }
        else
          break;
        continue;
      }
      else if (!strncmp("oada", cmdP, 4)) { /* read all_facility */
        cmdP += 4;
        if (!isspc()) break;
        skipspc();
        if (!*cmdP) {
	  message("no {filename}\n");
	  continue;
	}
        read_all(cmdP);
        continue;
      }
      else if (isspc()) { /* disassemble */
        static unsigned start, end;
        unsigned        count;

        if (extract_memrange(c_pe, &cmdP, 8, &start, &end))
          continue;
        ignore_rest_of_line(cmdP);
        start &= ~7;
        end   |=  7;
        count = (end - start) / 8 + 1;
        disasm(c_pe, c_mode, start, count, 0);
        start += count * 8;
        end   += count * 8;
        continue;
      }
      break;

    case 'v':
      if (!command_v(cmdP, fp, log)) continue;
      break;

    case 'd':
      if (!command_d(cmdP, fp, log)) continue;
      break;

    case 'r':
      if (!command_r(cmdP, fp, log)) continue;
      break;
    
    case 'w':
      if (!command_w(cmdP, fp, log)) continue;
      break;
    
    case 'f':
      if (!command_f(cmdP, fp, log)) continue;
      break;
    
    case 's':
      if (!strncmp("et", cmdP, 2)) { /* set variables */
        FILE *open_file();
	unsigned itmp;
	char ctmp;

        cmdP += 2;
        if (!isspc()) break;
        skipspc();
	if (!strncmp("DISPLOG", cmdP, 7)) { /* DISPLOG */
          cmdP += 7;
          if (!isspc()) break;
          skipspc();
	  if (displog_fp) {
	    fclose(displog_fp);
	    free(ENV_DISPLOG);
          }
          if (displog_fp = open_file(cmdP, "w"))
            strcpy(ENV_DISPLOG = malloc(strlen(cmdP) + 1), cmdP);
          else
            ENV_DISPLOG = "";
          message("DISPLOG = \"%s\"\n", ENV_DISPLOG);
	}
	else if (!strncmp("mode", cmdP, 4)) { /* mode */
          cmdP += 4;
          if (!isspc()) break;
          skipspc();
          if (sscanf(cmdP, "%c", &ctmp) != 1) {
	    message("{r|v} expected\n");
	    continue;
	  }
          if (ctmp != 'r' && ctmp != 'v') {
            message("{r|v} expected\n");
            continue;
          }
          message("mode = '%c'\n", c_mode = ctmp);
	}
#ifdef VCP
	else if (!strncmp("node", cmdP, 4)) { /* node */
          cmdP += 4;
          if (!isspc()) break;
          skipspc();
          if (sscanf(cmdP, "%c", &ctmp) != 1) {
	    message("{r|v} expected\n");
	    continue;
	  }
          if (ctmp != 'r' && ctmp != 'v') {
            message("{r|v} expected\n");
            continue;
          }
          message("node = '%c'\n", c_node = ctmp);
	}
#endif
#ifdef MCR_AREA
	else if (!strncmp("msmap", cmdP, 5)) { /* msmap */
          cmdP += 5;
          if (!isspc()) break;
          skipspc();
          if (sscanf(cmdP, "%u", &itmp) != 1) {
	    message("{num} expected\n");
	    continue;
	  }
          if (itmp !=   1 && itmp !=   2 && itmp !=   4 && itmp !=   8
           && itmp !=  16 && itmp !=  32 && itmp !=  64 && itmp !=  96
           && itmp != 128 && itmp != 192 && itmp != 256 && itmp != 384
           && itmp != 512 && itmp != 768 && itmp != 1024) {
            message("{1|2|4|8|16} expected (PPM not applied) *or*\n");
            message("{32|64|96|128|192|256|384|512|768|1024} expected (PPM applied)\n");
            continue;
          }
          pe[c_pe].mcr[MSMAP].ful = ((itmp << 20) - 1) & 0xffff0000;
          message("msmap = %u(MB)\n", (pe[c_pe].mcr[MSMAP].ful >> 20) + 1);
	}
#endif
	else if (!strncmp("echo", cmdP, 4)) { /* echo */
          cmdP += 4;
          if (!isspc()) break;
          skipspc();
          if (sscanf(cmdP, "%c", &ctmp) != 1) {
            message("{y|n} expected\n");
            continue;
          }
          if (ctmp != 'y' && ctmp != 'n') {
            message("{y|n} expected\n");
            continue;
          }
          message("echo = '%c'\n", c_echo = ctmp);
	}
	else if (!strncmp("logview", cmdP, 7)) { /* logview */
          cmdP += 7;
          if (!isspc()) break;
          skipspc();
          if (sscanf(cmdP, "%c", &ctmp) != 1) {
            message("{y|n} expected\n");
            continue;
          }
          if (ctmp != 'y' && ctmp != 'n') {
            message("{y|n} expected\n"); 
            continue; 
          }
          message("logview = '%c'\n", c_logview = ctmp);
	}
	else if (!strncmp("forcew", cmdP, 6)) { /* forcew */
          cmdP += 6;
          if (!isspc()) break;
          skipspc();
          if (sscanf(cmdP, "%c", &ctmp) != 1) {
            message("{y|n} expected\n");
            continue;
          }
          if (ctmp != 'y' && ctmp != 'n') {
            message("{y|n} expected\n"); 
            continue; 
          }
          message("forcew = '%c'\n", c_forcew = ctmp);
	}
	else if (!strncmp("symbol", cmdP, 6)) { /* symbol */
          cmdP += 6;
          if (!isspc()) break;
          skipspc();
          if (sscanf(cmdP, "%c", &ctmp) != 1) {
            message("{y|n} expected\n");
            continue;
          }
          if (ctmp != 'y' && ctmp != 'n') {
            message("{y|n} expected\n"); 
            continue; 
          }
          message("symbol = '%c'\n", c_symbol = ctmp);
	}
#ifdef CONSOLE
	else if (!strncmp("coint", cmdP, 5)) { /* coint */
          cmdP += 5;
          if (!isspc()) break;
          skipspc();
          if (sscanf(cmdP, "%u", &itmp) != 1) {
	    message("{num} expected\n");
	    continue;
	  }
          if (itmp >= IR_NUM) {
            message("{0-15} expected\n");
            continue;
          }
          message("coint = %u\n", c_coint = itmp);
	}
#endif
	else
	  break;
	continue;
      }
      else if (!strncmp("yms", cmdP, 3)) { /* display symbols */
        unsigned memaddr;

        cmdP += 3;
        if (!isspc()) break;
        skipspc();
        if (!*cmdP) { /* print all symbols */
          disp_symbols(c_pe);
        }
        else {
          if (get_memnumber(c_pe, cmdP, &memaddr)) {
	    if (sscanf(cmdP, "%x", &memaddr) != 1) {
              message("*unknown*\n");
              continue;
            }
          }
          disp_symbol(c_pe, memaddr);
        }
	continue;
      }
      else if (*cmdP == 'b' || *cmdP == 'c') { /* set break/checkpoints */
        static unsigned interval, type, dummy;
        static unsigned start;
        unsigned bp_number;
        unsigned real_addr, abs_addr;
        unsigned b_or_c; /* 'b', 'c' */
        unsigned access, tmp;

        b_or_c = *cmdP++;
        if (!(*cmdP == '/' || isspc())) break;
	if (extract_option(&cmdP, &interval, &type, 1))
          continue;
        skipspc();
        if (interval == 0) interval = 1;
        if (type == 0) type = optbit('x');
        access = _N_;
	if (type & optbit('r')) access |= _R_;
	if (type & optbit('w')) access |= _W_;
	if (type & optbit('x')) access |= _X_;
	if (type & ~(0x80000000 | optbit('r') | optbit('w') | optbit('x'))) {
          message("illegal type\n");
	  continue;
	}
	if (access == _N_) {
          message("must specify r|w|x\n");
	  continue;
	}
        if (!*cmdP) { /* print points */
          for (tmp = 0; tmp < BP_NUM; tmp++) {
            if (pe[c_pe].bp_table[tmp].access) {
              message("%c", pe[c_pe].bp_table[tmp].access == access ? '*' : ' ');
              message("No.%2u: ", tmp);
              message("type=%s access=%s%s%s ",
                pe[c_pe].bp_table[tmp].b_or_c == 'b' ? "break" : "check",
                pe[c_pe].bp_table[tmp].access & _R_ ? "R" : "-",
                pe[c_pe].bp_table[tmp].access & _W_ ? "W" : "-",
                pe[c_pe].bp_table[tmp].access & _X_ ? "X" : "-");
              message("interval=%u count_up=%u\n",
		pe[c_pe].bp_table[tmp].interval,
		pe[c_pe].bp_table[tmp].count);
              message("        mode='%c' ",
		pe[c_pe].bp_table[tmp].mode);
              message("addr=0x%08.8X ",
                pe[c_pe].bp_table[tmp].address);
              message("(real_addr=0x%08.8X)\n",
                pe[c_pe].bp_table[tmp].real_addr,
                pe[c_pe].bp_table[tmp].abs_addr);
              message("        command=%s\n",
		pe[c_pe].bp_table[tmp].command);
            }
          }
        }
        else { /* set points */
	  if (extract_memrange(c_pe, &cmdP, 8, &start, &dummy))
            continue;
          skipspc();
          if (*cmdP == ',') {
	    cmdP++;
	    skipspc();
	  }
	  else {
            message(",{pos} expected\n");
            continue;
          }
          if (sscanf(cmdP, "%u", &bp_number) != 1) {
            message("{pos} expected\n");
            continue;
          }
          if (bp_number >= BP_NUM) {
            message("{pos} exceeds %u\n", BP_NUM-1);
            continue;
          }
          if (pe[c_pe].bp_table[bp_number].access) {
	    message("%u: already used\n", bp_number);
	    continue;
	  }
	  if (get_mem_real_addr(c_pe, c_mode, _R_, start, &real_addr, 'n'))
	    continue;
	  if (get_abs_addr(c_pe, _R_, real_addr, &abs_addr, 'n'))
	    continue;
          pe[c_pe].bp_table[bp_number].interval  = interval;
          pe[c_pe].bp_table[bp_number].count     = 0;
          pe[c_pe].bp_table[bp_number].access    = access;
          pe[c_pe].bp_table[bp_number].b_or_c    = b_or_c;
          pe[c_pe].bp_table[bp_number].address   = start;
          pe[c_pe].bp_table[bp_number].real_addr = real_addr;
          pe[c_pe].bp_table[bp_number].abs_addr  = abs_addr;
          pe[c_pe].bp_table[bp_number].mode      = c_mode;
          skipdigit();
          skipspc();
          strcpy(pe[c_pe].bp_table[bp_number].command = malloc(strlen(cmdP) + 1), cmdP);
          pe[c_pe].mbp[abs_addr/8] |= (0x0080 >> (abs_addr & 7));
          message("*No.%2u: ", bp_number);
          message("type=%s access=%s%s%s ",
            b_or_c == 'b' ? "break" : "check",
	    access & 4 ? "R" : "-",
            access & 2 ? "W" : "-",
            access & 1 ? "X" : "-");
          message("interval=%u count_up=%u\n",
	    interval, 0);
          message("        mode='%c' ", c_mode);
          message("addr=0x%08.8X ",
	    start);
          message("(real_addr=0x%08.8X)\n",
            real_addr, abs_addr);
          message("        command=%s\n", cmdP);
          start++;
        }
        continue;
      }
      else if (*cmdP == 'm') { /* set memory map */
#ifdef MANUAL_MAP
        static unsigned type, dummy;
        static unsigned start, end;
        unsigned map_number, abs_addr, bytes;
        unsigned prot, tmp;

        cmdP++;
        if (!(*cmdP == '/' || isspc())) break;
	if (extract_option(&cmdP, &dummy, &type, 1))
          continue;
        skipspc();
        if (type == 0) type = optbit('r') | optbit('w') | optbit('x');
        prot = _N_;
	if (type & optbit('r')) prot |= _R_;
	if (type & optbit('w')) prot |= _W_;
	if (type & optbit('x')) prot |= _X_;
	if (type & ~(0x80000000 | optbit('r') | optbit('w') | optbit('x'))) {
          message("illegal type\n");
	  continue;
	}
	if (prot == _N_) {
          message("must specify r|w|x\n");
	  continue;
	}
        if (!*cmdP) { /* print memory map */
          for (tmp = 0; tmp < MAP_NUM; tmp++) {
            if (pe[c_pe].map[tmp].prot) {
              message("%c", pe[c_pe].map[tmp].prot == prot ? '*' : ' ');
              message("No.%2u: prot=%s%s%s ",
                tmp,
	        pe[c_pe].map[tmp].prot & _R_ ? "R" : "-",
	        pe[c_pe].map[tmp].prot & _W_ ? "W" : "-",
	        pe[c_pe].map[tmp].prot & _X_ ? "X" : "-");
              message("real_addr=0x%08.8X-0x%08.8X ",
                pe[c_pe].map[tmp].start,
                pe[c_pe].map[tmp].end);
              message("abs_addr=0x%08.8X\n", pe[c_pe].map[tmp].abs_addr);
            }
          }
	  cmd_mem_free_status(c_pe);
        }
        else { /* set memory map */
	  if (extract_memrange(c_pe, &cmdP, 1, &start, &end))
            continue;
          skipspc();
          start &= ~63;
          end   |=  63;
	  bytes = end - start + 1;
          if (*cmdP == ',') {
	    cmdP++;
	    skipspc();
	  }
	  else {
	    message(",{pos} expected\n");
	    continue;
	  }
          if (sscanf(cmdP, "%u", &map_number) != 1) {
	    message("{pos} expected\n");
	    continue;
	  }
          skipdigit();
          ignore_rest_of_line(cmdP);

          if (map_number >= MAP_NUM) {
	    message("{pos} exceeds %u\n", MAP_NUM-1);
	    continue;
	  }
          if (pe[c_pe].map[map_number].prot) {
	    message("%u: already used\n", map_number);
	    continue;
	  }
	  if (cmd_mem_map(c_pe, map_number, prot, start, end, &abs_addr))
	    continue;
          message("*No.%2u: prot=%s%s%s ",
            map_number,
	    prot & _R_ ? "R" : "-",
	    prot & _W_ ? "W" : "-",
	    prot & _X_ ? "X" : "-");
          message("real_addr=0x%08.8X-0x%08.8X ",
            start,
            end);
          message("abs_addr=0x%08.8X\n", abs_addr);
          start += bytes;
          end   += bytes;
        }
#else
        message("manual mapping not available. (compiled with no -DMANUAL_MAP)\n");
        message("fixed map: prot=RWX real_addr=0x00000000-0x%08.8X\n", memsize-1);
#endif
        continue;
      }
      else if (*cmdP == 'i') { /* set interval interrupt */
#ifdef MANUAL_INTR
        static unsigned type; /* 'c':continuous 's':single */
        static unsigned interval, dummy;
        unsigned intno;
        unsigned tmp;

	cmdP++;
        if (!(*cmdP == '/' || isspc())) break;
        if (extract_option(&cmdP, &dummy, &type, 0))
          continue;
	skipspc();
        if (type == 0) type = 'c';
        if (type!='c'&&type!='s') {
          message("illegal: type='%c'\n", type);
          continue;
        }
        if (!*cmdP) { /* print intervals */
          for (tmp = 1; tmp < IR_NUM; tmp++) {
            if (pe[c_pe].ir_table[tmp].interval) {
              message("%c", pe[c_pe].ir_table[tmp].type == type ? '*' : ' ');
	      message("No.%2u: ", tmp);
              message("type=%s interval=%u count_up=%u\n",
                pe[c_pe].ir_table[tmp].type == 'c' ? "continuous" : "single    ",
                pe[c_pe].ir_table[tmp].interval,
                pe[c_pe].ir_table[tmp].count);
            }
          }
        }
        else { /* set interval */
          if (extract_regrange(0, &cmdP, &dummy, &interval, &dummy))
            continue;
          skipspc();
          if (interval == 0) interval = 1;
          if (*cmdP == ',') {
            cmdP++;
            skipspc();
          }
          else {
            message(",{pos} expected\n");
            continue;
          }
          if (sscanf(cmdP, "%u", &intno) != 1) {
            message("{pos} expected\n");
            continue;
          }
          skipdigit();
          ignore_rest_of_line(cmdP);
          if (intno == 0 || intno >= IR_NUM) {
            message("{pos} out of 1-%u\n", IR_NUM-1);
            continue;
          }
          if (pe[c_pe].ir_table[intno].interval) {
            message("%u: already set\n", intno);
            continue;
          }
          pe[c_pe].ir_table[intno].interval = interval;
          pe[c_pe].ir_table[intno].count    = 0;
          pe[c_pe].ir_table[intno].type     = type;
	  message("*No.%2u: ", intno);
          message("type=%s interval=%u count_up=%u\n",
            type == 'c' ? "continuous" : "single    ", interval, 0);
        }
#else
        message("manual interruption not available. (compiled with no -DMANUAL_INTR)\n");
#endif
	continue;
      }
      break;

    case 'c':
      if (*cmdP == 'p') { /* clear break/checkpoint */
        static unsigned start, end, dummy;
        unsigned count, abs_addr;
        unsigned i, tmp;

        cmdP++;
        if (!isspc()) break;
        if (extract_regrange(0, &cmdP, &dummy, &start, &end))
          continue;
        ignore_rest_of_line(cmdP);
	count = end - start + 1;
	for (i=start; i<=end; i++) {
          if (i >= BP_NUM) break;
          if (pe[c_pe].bp_table[i].access == 0)
	    continue;
          pe[c_pe].bp_table[i].access = 0;
          free(pe[c_pe].bp_table[i].command);
          abs_addr = pe[c_pe].bp_table[i].abs_addr;
          pe[c_pe].mbp[abs_addr/8] &= 0xff7f >> (abs_addr & 7);
          for (tmp = 0; tmp < BP_NUM; tmp++) {
            if (pe[c_pe].bp_table[tmp].access && pe[c_pe].bp_table[tmp].abs_addr == abs_addr)
              pe[c_pe].mbp[abs_addr/8] |= 0x0080 >> (abs_addr & 7);
          }
        }
	start += count;
	end   += count;
        continue;
      }
      else if (*cmdP == 'm') { /* clear memory map */
#ifdef MANUAL_MAP
        static unsigned start, end, dummy;
        unsigned count;
        unsigned i;

        cmdP++;
        if (!isspc()) break;
        if (extract_regrange(0, &cmdP, &dummy, &start, &end))
          continue;
        ignore_rest_of_line(cmdP);
	count = end - start + 1;
	for (i=start; i<=end && i<MAP_NUM; i++)
	  cmd_mem_free(c_pe, i, pe[c_pe].map[i].start, pe[c_pe].map[i].end, pe[c_pe].map[i].abs_addr);
	start += count;
	end   += count;
#else
        message("manual mapping not available. (compiled with no -DMANUAL_MAP)\n");
        message("fixed map: prot=RWX real_addr=0x00000000-0x%08.8X\n", memsize-1);
#endif
        continue;
      }
      else if (*cmdP == 'i') { /* clear intervals */
#ifdef MANUAL_INTR
        static unsigned start, end, dummy;
        unsigned i, count;

        cmdP++;
        if (!isspc()) break;
        if (extract_regrange(0, &cmdP, &dummy, &start, &end))
          continue;
        ignore_rest_of_line(cmdP);
	count = end - start + 1;
	for (i=start; i<=end; i++) {
          if (i >= IR_NUM) break;
          pe[c_pe].ir_table[i].interval = 0;
        }
	start += count;
	end   += count;
#else
        message("manual interruption not available. (compiled with no -DMANUAL_INTR)\n");
#endif
        continue;
      }
      else if (*cmdP == 'v') { /* clear view */
        static int start, end, dummy;
        int i, tmp, tmp2;

        cmdP++;
        if (!isspc()) break;
        if (extract_regrange(0, &cmdP, &dummy, &start, &end))
          continue;
        ignore_rest_of_line(cmdP);
        for (i=start; i<=end && start<=view_end; i++) {
          if (start < 0 || start >= VIEW_NUM) break;
          for (tmp=start; tmp<view_end; tmp++) {
            VIEW_LINE[tmp].compress = VIEW_LINE[tmp+1].compress;
            for (tmp2=0; tmp2<4; tmp2++) {
              VIEW_LINE[tmp].pe[tmp2]       = VIEW_LINE[tmp+1].pe[tmp2];
              VIEW_LINE[tmp].mode[tmp2]     = VIEW_LINE[tmp+1].mode[tmp2];
              VIEW_LINE[tmp].facility[tmp2] = VIEW_LINE[tmp+1].facility[tmp2];
              VIEW_LINE[tmp].select[tmp2]   = VIEW_LINE[tmp+1].select[tmp2];
              VIEW_LINE[tmp].num[tmp2]      = VIEW_LINE[tmp+1].num[tmp2];
	    }
          }
          VIEW_LINE[tmp].compress = 0;
          for (tmp2=0; tmp2<4; tmp2++)
            VIEW_LINE[tmp].facility[tmp2] = 0;
          if (view_end >= 0)
	    view_end--;
          scr_setr(view_end + 2);
	}
        continue;
      }
      else if (*cmdP == 'x') { /* clear statistics */
        int i, j, k;

        cmdP++;
        if (!isspc()) break;
        ignore_rest_of_line(cmdP);
#ifdef VCP
        for (i=0; i<VR_NUM; i++) {
          pe[c_pe].vr_rw_conflict[i] = 0;
          pe[c_pe].vr_wr_conflict[i] = 0;
          pe[c_pe].vr_ww_conflict[i] = 0;
	}
        for (i=0; i<MR_NUM; i++) {
          pe[c_pe].mr_rw_conflict[i] = 0;
          pe[c_pe].mr_wr_conflict[i] = 0;
          pe[c_pe].mr_ww_conflict[i] = 0;
	}
        for (i=0; i<256; i++) {
          pe[c_pe].vinsn_stat[i] = 0;
        }
#endif
        for (i=0; i<16; i++) {
          for (j=0; j<32; j++) {
            for (k=0; k<3; k++) {
              pe[c_pe].sinsn_stat[i][j][k] = 0;
            }
          }
        }
        pe[c_pe].brcs_b_t = 0;
        pe[c_pe].brcs_b_n = 0;
        pe[c_pe].brcs_f_n = 0;
        pe[c_pe].brcs_f_t = 0;
        pe[c_pe].brc_b_t  = 0;
        pe[c_pe].brc_b_n  = 0;
        pe[c_pe].brc_f_n  = 0;
        pe[c_pe].brc_f_t  = 0;
        pe[c_pe].text_access = 0;
        pe[c_pe].data_access = 0;
        pe[c_pe].text_tlb_access = 0;
        pe[c_pe].text_tlb_hit    = 0;
        pe[c_pe].data_tlb_access = 0;
        pe[c_pe].data_tlb_hit    = 0;
        pe[c_pe].text_cache_access = 0;
        pe[c_pe].text_cache_hit    = 0;
        pe[c_pe].data_cache_access = 0;
        pe[c_pe].data_cache_hit    = 0;

        continue;
      }
      break;

    case 'p':
      if (*cmdP == 't') { /* purge tlb */
        unsigned i;

        cmdP++;
        if (!isspc()) break;
        ignore_rest_of_line(cmdP);
        for (i = 0; i < TLB_NUM; i++) {
          pe[c_pe].text_tlb[i].vmc &= ~4;
          pe[c_pe].data_tlb[i].vmc &= ~4;
        }
        continue;
      }
      else if (*cmdP == 'c') { /* purge cache */
        unsigned i;

        cmdP++;
        if (!isspc()) break;
        ignore_rest_of_line(cmdP);
        for (i = 0; i < CACHE_NUM; i++) {
          pe[c_pe].text_cache[i].v = 0;
          pe[c_pe].data_cache[i].v = 0;
        }
	continue;
      }
      break;

    case 'g':
    case 't':
      {
      unsigned t_pe;
      unsigned save_clock;

      if (!strncmp("gres", cmdP-1, 4)) { /* gres */
        cmdP+=3;
        if (!isspc()) break;
        ignore_rest_of_line(cmdP);
        general_reset(c_pe);
        continue;
      }
      else if (!strncmp("go", cmdP-1, 2)) { /* go */
        cmdP++;
        if (!isspc()) break;
        ignore_rest_of_line(cmdP);
        message("executing...\n");
        restme();
        save_clock = tim_clock;
	scan_on = 0;
	bp_match = 0;
	cp_match = 0;
	interrupt = 0;
        for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
	  pe[T_PE].cntl_block = -1;
	  pe[T_PE].bp_entry   = -1;
	  pe[T_PE].cp_entry   = -1;
	}
        signal(SIGINT,  SIG_IGN);
        signal(SIGINT,  onintr_sys);
        for (;;) {
          for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
	    if (pe[T_PE].ip_status != STATUS_HALT) {
              sim_pe(t_pe);
	      if (pe[T_PE].cntl_block != -1)
                serialize(-1, 0);
            }
          }
          tim_clock++;
	  if (all_halts())
	    goto run_halt;
          if (cp_match || bp_match || interrupt)
            serialize(-1, 0);
          if (cp_match) {
            for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
	      if (pe[T_PE].cp_entry >= 0) {
                message("PE#%u check at 0x%08.8X (real_addr=0x%08.8X) by cp%u.\n",
	          t_pe,
	          pe[T_PE].bp_table[pe[T_PE].cp_entry].address,
	          pe[T_PE].bp_table[pe[T_PE].cp_entry].real_addr,
	          pe[T_PE].cp_entry);
                message("exec_clock=%u (exec_time=%umsec)\n", tim_clock-save_clock, gettme());
		if (dasd_fd) close(dasd_fd);
#ifdef SSUMOVER
    		if (ssu_fd) close(ssu_fd);
#endif
                command('s', fp, pe[T_PE].bp_table[pe[T_PE].cp_entry].command, 1);
    		dasd_fd = open_dasd(ENV_DASD);
#ifdef SSUMOVER
    		ssu_fd = open_ssu(ENV_SSU);
#endif
	        pe[T_PE].cp_entry = -1;
              }
	    }
	    cp_match = 0;
	  }
	  if (bp_match || interrupt)
            goto run_break;
        }
      }
      else if (!strncmp("t", cmdP-1, 1)) { /* trace */
        static unsigned type, count, dummy;
        unsigned scantype, scount;

        if (!(*cmdP == '/' || isspc())) break;
	if (extract_option(&cmdP, &dummy, &type, 1))
          continue;
        skipspc();
        if (type == 0) type = optbit('l');
        scantype = SCAN_NULL;
	if (type & optbit('a')) scantype |= SCAN_ALL;
	if (type & optbit('l')) scantype |= SCAN_LIST;
	if (type & optbit('b')) scantype |= SCAN_BRANCH;
	if (type & optbit('i')) scantype |= SCAN_INTR;
	if (type & optbit('m')) scantype |= SCAN_MEM;
	if (type & optbit('s')) scantype |= SCAN_STATUS;
	if (type & ~(0x80000000 | optbit('a') | optbit('l') | optbit('b') | optbit('i') | optbit('m') | optbit('s'))) {
          message("illegal type\n");
	  continue;
	}

        if (extract_regrange(0, &cmdP, &dummy, &count, &dummy))
          continue;
        ignore_rest_of_line(cmdP);
        if (count == 0) count = 1;
        restme();
        save_clock = tim_clock;
	bp_match = 0;
	cp_match = 0;
	interrupt = 0;
        for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
          pe[T_PE].cntl_block = -1;
	  pe[T_PE].bp_entry   = -1;
	  pe[T_PE].cp_entry   = -1;
	}
        signal(SIGINT,  SIG_IGN);
        signal(SIGINT,  onintr_sys);
        for (scount = 0; scount < count; scount++) {
          for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
	    if (pe[T_PE].ip_status != STATUS_HALT) {
	      if ((scantype & SCAN_ALL) || t_pe == c_pe)
		scan_on = scantype;
	      else
		scan_on = 0;
              sim_pe(t_pe);
	      if (pe[T_PE].cntl_block != -1)
                serialize(c_pe, scantype);
	    }
          }
          tim_clock++;
	  if (all_halts())
	    goto run_halt;
          if (cp_match || bp_match || interrupt)
	    serialize(c_pe, scantype);
          if (cp_match) {
            for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
	      if (pe[T_PE].cp_entry >= 0) {
                message("PE#%u check at 0x%08.8X (real_addr=0x%08.8X) by cp%u.\n",
	          t_pe,
	          pe[T_PE].bp_table[pe[T_PE].cp_entry].address,
	          pe[T_PE].bp_table[pe[T_PE].cp_entry].real_addr,
	          pe[T_PE].cp_entry);
                message("exec_clock=%u (exec_time=%umsec)\n", tim_clock-save_clock, gettme());
		if (dasd_fd) close(dasd_fd);
#ifdef SSUMOVER
    		if (ssu_fd) close(ssu_fd);
#endif
                command('s', fp, pe[T_PE].bp_table[pe[T_PE].cp_entry].command, 1);
    		dasd_fd = open_dasd(ENV_DASD);
#ifdef SSUMOVER
    		ssu_fd = open_ssu(ENV_SSU);
#endif
	        pe[T_PE].cp_entry = -1;
              }
	    }
	    cp_match = 0;
	  }
	  if (bp_match || interrupt)
            goto run_break;
          if (scantype != SCAN_NULL && scount < count-1) 
	    view(c_logview);
        }
	serialize(c_pe, scantype);
        continue;
      }
      break;
run_break:
      if (bp_match) {
        for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
	  if (pe[T_PE].bp_entry >= 0 && pe[T_PE].bp_entry < BP_NUM) {
            message("PE#%u break at 0x%08.8X (real_addr=0x%08.8X) by bp%u.\n",
	      t_pe,
	      pe[T_PE].bp_table[pe[T_PE].bp_entry].address,
	      pe[T_PE].bp_table[pe[T_PE].bp_entry].real_addr,
	      pe[T_PE].bp_entry);
            message("exec_clock=%u (exec_time=%umsec)\n", tim_clock-save_clock, gettme());
  	    if (dasd_fd) close(dasd_fd);
#ifdef SSUMOVER
    	    if (ssu_fd) close(ssu_fd);
#endif
            command('s', fp, pe[T_PE].bp_table[pe[T_PE].bp_entry].command, 1);
    	    dasd_fd = open_dasd(ENV_DASD);
#ifdef SSUMOVER
    	    ssu_fd = open_ssu(ENV_SSU);
#endif
          }
	  else if (pe[T_PE].bp_entry == EXT_FUNC_BREAK) {
            message("exec_clock=%u (exec_time=%umsec)\n", tim_clock-save_clock, gettme());
          }
        }
      }
      else {
        message("manual break. exec_clock=%u (exec_time=%umsec)\n", tim_clock-save_clock, gettme());
      }
      continue;
run_halt:
      message("all PE(s) halted. exec_clock=%u (exec_time=%umsec)\n", tim_clock-save_clock, gettme());
      continue;
      }
      break;

    default: /* regard as script */
      break;
    }

    /* regard as script */
    {
      FILE *open_rc();
      int i;

      if (script_ptr >= SCRIPT_NUM) {
        message("too nested script\n");
        continue;
      }
      for (cmdP = save_cmdP; *cmdP; cmdP++) { /* extract filename */
        if (isspace(*cmdP)) { *cmdP++ = NULL; break; }
      }
      if (*save_cmdP == '/')
        sprintf(tmp_buf, "%s", save_cmdP);
      else if (!*ENV_CMDDIR)
        sprintf(tmp_buf, "%s", save_cmdP);
      else
        sprintf(tmp_buf, "%s/%s", ENV_CMDDIR, save_cmdP);
      if (SCRIPT[script_ptr].script_fp = open_rc(tmp_buf)) {
        for (i=1; i<=9; i++) { /* extract parameters */
          skipspc();
          save_cmdP = cmdP;
          for (; *cmdP; cmdP++) { /* extract parameter */
            if (isspace(*cmdP)) { *cmdP++ = NULL; break; }
          }
          strcpy(SCRIPT[script_ptr].param[i] = malloc(strlen(save_cmdP)+1), save_cmdP);
        }
        command('f', SCRIPT[script_ptr++].script_fp, NULL, 1);
        fclose(SCRIPT[--script_ptr].script_fp);
        for (i=1; i<=9; i++)
          free(SCRIPT[script_ptr].param[i]);
      }
    }
  }
}

command_v(cmdP, fp, log) char * cmdP; FILE *fp; int log;
    /* return 1 if failed, 0 if succeeded */
{
  unsigned once;

  for (once = 0; !once; once++) {
    switch (once) {
    case 0:
      switch (*cmdP++) {
#ifdef MCR_AREA
      case 'l': /* mcr */
#endif
      case 'm': /* scalar */
#ifdef VCP
      case 'n': /* vector */
#endif
        if (*cmdP == '/' || isspc()) { /* view mem */
          static unsigned type, dummy;
          static unsigned start, end;
          unsigned count, addr;
          int lineno;
          unsigned facility, tmp;

          facility = *(cmdP-1);
          if (extract_option(&cmdP, &dummy, &type, 0))
            continue;
          if (type == 0) type = 'x';
          if (type!='x'&&type!='c') {
            message("illegal: type='%c'\n", type);
            continue;
          }
#ifdef MCR_AREA
          if (facility == 'l') {
            if (extract_regrange(facility, &cmdP, &dummy, &start, &end))
              continue;
          }
          else {
#endif
            if (extract_memrange(c_pe, &cmdP, 4, &start, &end))
              continue;
#ifdef MCR_AREA
          }
#endif
          skipspc();
          if (*cmdP == ',') {
	    cmdP++;
	    skipspc();
	  }
	  else {
	    message(",{pos} expected\n");
	    continue;
	  }
          if (sscanf(cmdP, "%d", &lineno) != 1) {
	    message("{pos} expected\n");
	    continue;
	  }
          skipdigit();
          ignore_rest_of_line(cmdP);
          if (lineno < 0 || lineno >= VIEW_NUM) {
	    message("{pos} negative or exceeds %d\n", VIEW_NUM-1);
	    continue;
	  }

	  switch (type) {
	  case 'x': /* expanded format */
            start &= ~3;
            end   |= 3;
            count = (end - start) / 4 + 1;
            for (addr=start; addr<=end; addr+=4) {
              for (tmp = 0; tmp < 4; tmp++) {
                if (VIEW_LINE[lineno].compress != 'c'
                 &&!VIEW_LINE[lineno].facility[tmp]) break;
              }
              if (tmp < 4) {
                VIEW_LINE[lineno].compress      = 'x';
                VIEW_LINE[lineno].pe[tmp]       = c_pe;
                VIEW_LINE[lineno].mode[tmp]     = facility=='m' ? c_mode : c_node;
                VIEW_LINE[lineno].facility[tmp] = facility;
                VIEW_LINE[lineno].num[tmp]      = addr;
                for (; view_end < lineno; view_end++)
                  *VIEW_LINE[view_end+1].prv    = '*';
              }
              else if (++lineno < VIEW_NUM) {
                addr-=4;
                continue;
              }
              else {
                message("too many elements\n");
                break;
              }
            }
            scr_setr(view_end + 2);
            start += count * 4;
            end += count * 4;
            continue;
            break;

          case 'c': /* compressed format */
            start &= ~31;
            end   |= 31;
            count = (end - start) / 32 + 1;
            for (addr=start; addr<=end; addr+=32) {
              if (!VIEW_LINE[lineno].facility[0]) {
                VIEW_LINE[lineno].compress    = 'c';
                VIEW_LINE[lineno].pe[0]       = c_pe;
                VIEW_LINE[lineno].mode[0]     = facility=='m' ? c_mode : c_node;
                VIEW_LINE[lineno].facility[0] = facility;
                VIEW_LINE[lineno].num[0]      = addr;
                for (; view_end < lineno; view_end++)
                  *VIEW_LINE[view_end+1].prv  = '*';
              }
              else if (++lineno < VIEW_NUM) {
                addr-=32;
                continue;
              }
              else {
                message("too many elements\n");
                break;
              }
            }
            scr_setr(view_end + 2);
            start += count * 32;
            end += count * 32;
            continue;
            break;
          }
        }
        break;

      case 'g': /* gr */
      case 'f': /* fr */
      case 's': /* sr */
      case 'o': /* tod */
#ifdef VCP
      case 'v': /* vr */
      case 'k': /* mr */
      case 'c': /* vcr */
      case 'u': /* vut */
      case 't': /* tr */
#endif
        if (*cmdP == '/' || isspc()) { /* view regs */
          static unsigned type, dummy;
          static unsigned select, start, end;
          unsigned count, addr, step;
          int lineno;
          unsigned facility, tmp;

          facility = *(cmdP-1);
          if (extract_option(&cmdP, &dummy, &type, 0))
            continue;
          if (type == 0) type = 'x';
          if (type!='x'&&type!='c') {
            message("illegal: type='%c'\n", type);
            continue;
          }
#ifdef VCP
          if ((facility == 'v' || facility == 'k') && start == 0) start = end = 1;
#endif
          if (extract_regrange(facility, &cmdP, &select, &start, &end))
            continue;
          skipspc();
#ifdef VCP
          if ((facility == 'v' || facility == 'k') && start == 0) {
	    message("{index} begins from 1\n");
	    continue;
          }
#endif
          if (*cmdP == ',') {
	    cmdP++;
	    skipspc();
	  }
	  else {
	    message(",{pos} expected\n");
	    continue;
	  }
          if (sscanf(cmdP, "%d", &lineno) != 1) {
	    message("{pos} expected\n");
	    continue;
	  }
          skipdigit();
          ignore_rest_of_line(cmdP);
          if (lineno < 0 || lineno >= VIEW_NUM) {
	    message("{pos} negative or exceeds %d\n", VIEW_NUM-1);
	    continue;
	  }

          switch (type) {
          case 'x': /* expanded format */
	    switch (facility) {
            case 'o':
              start = 0;
              end   = 0;
              count = 1;
	      step  = 1;
	      break;

#ifdef VCP
            case 'k':
              start = ((start-1) & ~63) + 1;
              end   = ((end-1) | 63) + 1;
              count = (end - start) / 64 + 1;
	      step  = 64;
	      break;
#endif

	    default:
              count = end - start + 1;
	      step  = 1;
	      break;
	    }

            for (addr=start; addr<=end; addr+=step) {
              for (tmp = 0; tmp < 4; tmp++) {
                if (VIEW_LINE[lineno].compress != 'c'
                 &&!VIEW_LINE[lineno].facility[tmp]) break;
              }
              switch (facility) {
	      case 'o':
	        if (tmp > 0) break;
                VIEW_LINE[lineno].facility[tmp+1] = ' ';

	      case 'f':
#ifdef VCP
	      case 'v':
	      case 'k':
#endif
	        if (tmp > 2) break;
                VIEW_LINE[lineno].facility[tmp+1] = ' ';

	      default:
	        if (tmp > 3) break;
                VIEW_LINE[lineno].compress          = 'x';
                VIEW_LINE[lineno].pe[tmp]           = c_pe;
                VIEW_LINE[lineno].mode[tmp]         = c_mode;
                VIEW_LINE[lineno].facility[tmp]     = facility;
                VIEW_LINE[lineno].select[tmp]       = select;
                VIEW_LINE[lineno].num[tmp]          = addr;
                for (; view_end < lineno; view_end++)
                  *VIEW_LINE[view_end+1].prv        = '*';
	        continue;
              }
              if (++lineno < VIEW_NUM) {
                addr-=step;
                continue;
              }
              else {
                message("too many elements\n");
                break;
              }
            }
            scr_setr(view_end + 2);
            start += count * step;
            end += count * step;
            continue;
            break;

          case 'c': /* compressed format */
	    switch (facility) {
	    case 'f':
              start &= ~3;
              end   |= 3;
              count = (end - start) / 4 + 1;
	      step  = 4;
	      break;

	    case 'o':
              start = 0;
              end   = 0;
              count = 1;
	      step  = 1;
	      break;

#ifdef VCP
	    case 'v':
            case 'k':
              start = ((start-1) & ~3) + 1;
              end   = ((end-1) | 3) + 1;
              count = (end - start) / 4 + 1;
	      step  = 4;
	      break;
#endif

	    default:
              start &= ~7;
              end   |= 7;
              count = (end - start) / 8 + 1;
	      step  = 8;
	      break;
	    }

            for (addr=start; addr<=end; addr+=step) {
              if (!VIEW_LINE[lineno].facility[0]) {
                VIEW_LINE[lineno].compress        = 'c';
                VIEW_LINE[lineno].pe[0]           = c_pe;
                VIEW_LINE[lineno].mode[0]         = c_mode;
                VIEW_LINE[lineno].facility[0]     = facility;
                VIEW_LINE[lineno].select[0]       = select;
                VIEW_LINE[lineno].num[0]          = addr;
                for (; view_end < lineno; view_end++)
                  *VIEW_LINE[view_end+1].prv      = '*';
	        continue;
              }
              if (++lineno < VIEW_NUM) {
                addr-=step;
                continue;
              }
              else {
                message("too many elements\n");
                break;
              }
            }
            scr_setr(view_end + 2);
            start += count * step;
            end += count * step;
            continue;
            break;
          }
        }
        break;

      default:
        break;
      }
      break;
    }
    return(1); /* failed */
  }
  return(0); /* succeeded */
}

command_d(cmdP, fp, log) char * cmdP; FILE *fp; int log;
      /* return 1 if failed, 0 if succeeded */
{
  unsigned once;

  for (once = 0; !once; once++) {
    switch (once) {
    case 0:
      if (!strncmp("umpm", cmdP, 4)) { /* dump memory to file */
	unsigned a_text, a_data, a_bss, a_entry;

        cmdP += 4;
        if (!isspc()) break;
        skipspc();
        if (!*cmdP) {
	  message("no {filename}\n");
	  continue;
	}
        access_header(c_pe, 'p', 0, 0, 0, 0);
        access_header(c_pe, 'r', &a_text, &a_data, &a_bss, &a_entry);
update_a_text:
	message("a_text "); arrow_prompt(fp);
        if (!fgets(tmp_buf, COMMAND_LINE, fp)) { message("EOF\n"); return; }
        expand_command(tmp_buf, PREV_COMMAND_BUF, tmp_buf, fp);
        strncpy(PREV_COMMAND_BUF, tmp_buf, COMMAND_LINE);
        if (log) log_command(tmp_buf);
        delete_comment(tmp_buf);
        if (sscanf(tmp_buf, "%x", &a_text) != 1)
          if (*tmp_buf) { message("a_text?\n"); goto update_a_text; }
        a_text &= ~7;
update_a_data:
	message("a_data "); arrow_prompt(fp);
        if (!fgets(tmp_buf, COMMAND_LINE, fp)) { message("EOF\n"); return; }
        expand_command(tmp_buf, PREV_COMMAND_BUF, tmp_buf, fp);
        strncpy(PREV_COMMAND_BUF, tmp_buf, COMMAND_LINE);
        if (log) log_command(tmp_buf);
        delete_comment(tmp_buf);
        if (sscanf(tmp_buf, "%x", &a_data) != 1)
          if (*tmp_buf) { message("a_data?\n"); goto update_a_data; }
        a_data &= ~7;
update_a_bss: 
	message("a_bss  "); arrow_prompt(fp);
        if (!fgets(tmp_buf, COMMAND_LINE, fp)) { message("EOF\n"); return; }
        expand_command(tmp_buf, PREV_COMMAND_BUF, tmp_buf, fp);
        strncpy(PREV_COMMAND_BUF, tmp_buf, COMMAND_LINE);
        if (log) log_command(tmp_buf);
        delete_comment(tmp_buf);
        if (sscanf(tmp_buf, "%x", &a_bss ) != 1)
          if (*tmp_buf) { message("a_bss?\n"); goto update_a_bss; }
update_a_entry:
	message("a_entry"); arrow_prompt(fp);
        if (!fgets(tmp_buf, COMMAND_LINE, fp)) { message("EOF\n"); return; }
        expand_command(tmp_buf, PREV_COMMAND_BUF, tmp_buf, fp);
        strncpy(PREV_COMMAND_BUF, tmp_buf, COMMAND_LINE);
        if (log) log_command(tmp_buf);
        delete_comment(tmp_buf);
        if (sscanf(tmp_buf, "%x", &a_entry) != 1)
          if (*tmp_buf) { message("a_entry?\n"); goto update_a_entry; }
        access_header(c_pe, 'w', &a_text, &a_data, &a_bss, &a_entry);
        write_module(c_pe, c_mode, cmdP);
        continue;
      }
      else if (!strncmp("umpa", cmdP, 4)) { /* dump all to file */
        cmdP += 4;
        if (!isspc()) break;
        skipspc();
        if (!*cmdP) {
	  message("no {filename}\n");
	  continue;
	}
	write_all(cmdP);
	continue;
      }
      switch (*cmdP++) {
      case 'a':
        if (isspc()) { /* display address translation process */
          static unsigned start, dummy;

          if (extract_memrange(c_pe, &cmdP, PAGE_SIZE, &start, &dummy))
            continue;
          ignore_rest_of_line(cmdP);
#ifdef VCP
	  message("---- scalar ----\n");
#endif
          cmd_mem_dat(c_pe, _R_, start, &dummy, 'n', 1);
#ifdef VCP
	  message("---- vector ----\n");
          cmd_vop_dat(c_pe, _R_, start, &dummy, 'n', 1);
#endif
	  start += PAGE_SIZE;
          continue;
        }
        break;

#ifdef MCR_AREA
      case 'l': /* mcr */
        if (isspc()) { /* dump mcr */
          static unsigned start, dummy;
          unsigned i, j;
          unsigned d, c;

          if (extract_regrange('l', &cmdP, &dummy, &start, &dummy))
            continue;
          ignore_rest_of_line(cmdP);
          start &= ~3;
          message("    addr +0 +1 +2 +3 +4 +5 +6 +7  ");
          message("+8 +9 +A +B +C +D +E +F  ");
          message("0123456789ABCDEF\n");
          for (i = 0; i < 16; i++) {
            message("%08.8X", start);
            for (j = 0; j < 16; j+=4) {
              if (j == 8) message(" ");
              if (cmd_mcr_access(c_pe,start+j,4,_R_,&d))
                goto exit_dumpmcr;
              message(" %02.2X %02.2X %02.2X %02.2X",
                (unsigned char) ((d >> 24) & 0xff),
                (unsigned char) ((d >> 16) & 0xff),
                (unsigned char) ((d >>  8) & 0xff),
                (unsigned char) ( d        & 0xff));
            }
            message("  ");
            for (j = 0; j < 16; j+=4) {
              if (cmd_mcr_access(c_pe,start+j,4,_R_,&d))
                goto exit_dumpmcr;
              c = (d >> 24) & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
              c = (d >> 16) & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
              c = (d >>  8) & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
              c =  d        & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
            }
            message("\n");
            start += 16;
          }
exit_dumpmcr:
          continue;
        }
        break;
#endif

      case 'm': /* scalar */
        if (isspc()) { /* dump mem */
          static unsigned start, dummy;
          unsigned i, j;
          unsigned d, c;

          if (extract_memrange(c_pe, &cmdP, 256, &start, &dummy))
            continue;
          ignore_rest_of_line(cmdP);
          start &= ~3;
          message("    addr +0 +1 +2 +3 +4 +5 +6 +7  ");
          message("+8 +9 +A +B +C +D +E +F  ");
          message("0123456789ABCDEF\n");
          for (i = 0; i < 16; i++) {
            message("%08.8X", start);
            for (j = 0; j < 16; j+=4) {
              if (j == 8) message(" ");
              if (cmd_mem_access(c_pe,c_mode,start+j,4,_R_,&d,'n'))
                goto exit_dumpmem;
              message(" %02.2X %02.2X %02.2X %02.2X",
	        (unsigned char) ((d >> 24) & 0xff),
	        (unsigned char) ((d >> 16) & 0xff),
	        (unsigned char) ((d >>  8) & 0xff),
	        (unsigned char) ( d        & 0xff));
            }
            message("  ");
            for (j = 0; j < 16; j+=4) {
              if (cmd_mem_access(c_pe,c_mode,start+j,4,_R_,&d,'n'))
                goto exit_dumpmem;
              c = (d >> 24) & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
              c = (d >> 16) & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
              c = (d >>  8) & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
              c =  d        & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
            }
            message("\n");
	    start += 16;
          }
exit_dumpmem:
          continue;
        }
        break;

#ifdef VCP
      case 'n': /* vector */
        if (isspc()) { /* dump mem */
          static unsigned start, dummy;
          unsigned i, j;
          unsigned d, c;

          if (extract_memrange(c_pe, &cmdP, 256, &start, &dummy))
            continue;
          ignore_rest_of_line(cmdP);
          start &= ~3;
          message("    addr +0 +1 +2 +3 +4 +5 +6 +7  ");
          message("+8 +9 +A +B +C +D +E +F  ");
          message("0123456789ABCDEF\n");
          for (i = 0; i < 16; i++) {
            message("%08.8X", start);
            for (j = 0; j < 16; j+=4) {
              if (j == 8) message(" ");
              if (cmd_vop_access(c_pe,c_node,start+j,4,_R_,&d,'n'))
                goto exit_dumpnem;
              message(" %02.2X %02.2X %02.2X %02.2X",
                (unsigned char) ((d >> 24) & 0xff),
                (unsigned char) ((d >> 16) & 0xff),
                (unsigned char) ((d >>  8) & 0xff),
                (unsigned char) ( d        & 0xff));
            }
            message("  ");
            for (j = 0; j < 16; j+=4) {
              if (cmd_vop_access(c_pe,c_node,start+j,4,_R_,&d,'n'))
                goto exit_dumpnem;
              c = (d >> 24) & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
              c = (d >> 16) & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
              c = (d >>  8) & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
              c =  d        & 0xff;
              if (isascii(c) && isprint(c)) message("%c", c);
              else message(".");
            }
            message("\n");
	    start += 16;
          }
exit_dumpnem:
          continue;
        }
        break;
#endif

      case 'i':
        switch (*cmdP++) {
        case 't':
          if (*cmdP == '/' || isspc()) { /* dump text TLB */
            static unsigned type; /* 'd':detail 's':summary */
            static unsigned start, dummy;
            unsigned i, j;

            if (extract_option(&cmdP, &dummy, &type, 0))
              continue;
            if (type == 0) type = 's';
            if (type!='d'&&type!='s') {
              message("illegal: type='%c'\n", type);
              continue;
            }
	    switch (type) {
	    case 'd':
              if (extract_regrange(0, &cmdP, &dummy, &start, &dummy))
                continue;
              ignore_rest_of_line(cmdP);
              message("           pid  la  paddr vmc wp prot  ");
              message("  pid  la  paddr vmc wp prot\n");
              for (i = 0; i < 16; i++) {
                message("I_TLB[%3u] ", start);
	        if (start >= TLB_NUM) {
                  message("not exist\n");
                  break;
                }
                for (j = 0; j < 2; j++) {
                  message("%04.4X %03.3X %05.5X ",
		    pe[c_pe].text_tlb[start].pid,
		    pe[c_pe].text_tlb[start].la,
		    pe[c_pe].text_tlb[start].pa);
                  message("%c%c%c ",
		    pe[c_pe].text_tlb[start].vmc&4?'V':'-',
		    pe[c_pe].text_tlb[start].vmc&2?'M':'-',
		    pe[c_pe].text_tlb[start].vmc&1?'C':'-');
                  message(" %1X ",
		    pe[c_pe].text_tlb[start].wp);
                  message("%1X%1X%1X%1X    ",
		    pe[c_pe].text_tlb[start].prot>>3&1,
		    pe[c_pe].text_tlb[start].prot>>2&1,
		    pe[c_pe].text_tlb[start].prot>>1&1,
		    pe[c_pe].text_tlb[start].prot   &1);
	          if (++start >= TLB_NUM)
		    break;
                }
                message("\n");
              }
              continue;

	    case 's':
              ignore_rest_of_line(cmdP);
              message("           000 001 002 003 004 005 006 007  ");
              message("008 009 010 011 012 013 014 015\n");
              for (i = 0; i < TLB_NUM; i+=16) {
                message("I_TLB[%3u]", i);
                for (j = 0; j < 16; j++) {
	          if (!(j % 8)) message(" ");
                  message("%c%c%c ",
		    pe[c_pe].text_tlb[i+j].vmc&4?'V':'-',
		    pe[c_pe].text_tlb[i+j].vmc&2?'M':'-',
		    pe[c_pe].text_tlb[i+j].vmc&1?'C':'-');
                }
                message("\n");
              }
	      continue;
	    }
          }
          break;

	case 'c':
          if (*cmdP == '/' || isspc()) { /* dump text CACHE */
            static unsigned type; /* 'd':detail 's':summary */
            static unsigned start, dummy;
            unsigned i, j;

            if (extract_option(&cmdP, &dummy, &type, 0))
              continue;
            if (type == 0) type = 's';
            if (type!='d'&&type!='s') {
              message("illegal: type='%c'\n", type);
              continue;
            }
            switch (type) {
            case 'd':
              if (extract_regrange(0, &cmdP, &dummy, &start, &dummy))
                continue;
              ignore_rest_of_line(cmdP);
              message("I_CHE[%3u] ", start);
	      if (start >= CACHE_NUM) {
                message("not exist\n");
                continue;
              }
              message("la=0x%08.8X pa=0x%08.8X v=%u\n",
	        start<<6,
	        pe[c_pe].text_cache[start].pa<<15|start<<6,
	        pe[c_pe].text_cache[start].v);
	      for (i=0; i<16; i++) {
	        if (i == 8) message("\n");
	        if (!(i & 3)) message(" ");
                message(" %08.8X", *((unsigned *) (pe[c_pe].text_cache[start].body)+i));
	      }
	      message("\n");
	      start++;
              continue;

	    case 's':
              ignore_rest_of_line(cmdP);
              message("           0-------+-----15 16------+-----31 ");
              message("32------+-----47 48------+-----63\n");
              for (i = 0; i < CACHE_NUM; i+=64) {
                message("I_CHE[%3u]", i);
                for (j = 0; j < 64; j++) {
                  if (!(j & 15)) message(" ");
                  message("%c", pe[c_pe].text_cache[i+j].v?'V':'-');
                }
                message("\n");
              }  
              continue;
	    }
          }
          break;
	}
	break;

      case 'd':
        switch (*cmdP++) {
        case 't':
          if (*cmdP == '/' || isspc()) { /* dump data TLB */
            static unsigned type; /* 'd':detail 's':summary */
            static unsigned start, dummy;
            unsigned i, j;

            if (extract_option(&cmdP, &dummy, &type, 0))
              continue;
            if (type == 0) type = 's';
            if (type!='d'&&type!='s') {
              message("illegal: type='%c'\n", type);
              continue;
            }
            switch (type) {
            case 'd':
              if (extract_regrange(0, &cmdP, &dummy, &start, &dummy))
                continue;
              ignore_rest_of_line(cmdP);
              message("           pid  la  paddr vmc wp prot  ");
              message("  pid  la  paddr vmc wp prot\n");
              for (i = 0; i < 16; i++) {
                message("D_TLB[%3u] ", start);
	        if (start >= TLB_NUM) {
                  message("not exist\n");
                  break;
                }
                for (j = 0; j < 2; j++) {
                  message("%04.4X %03.3X %05.5X ",
		    pe[c_pe].data_tlb[start].pid,
		    pe[c_pe].data_tlb[start].la,
		    pe[c_pe].data_tlb[start].pa);
                  message("%c%c%c ",
		    pe[c_pe].data_tlb[start].vmc&4?'V':'-',
		    pe[c_pe].data_tlb[start].vmc&2?'M':'-',
		    pe[c_pe].data_tlb[start].vmc&1?'C':'-');
                  message(" %1X ",
		    pe[c_pe].data_tlb[start].wp);
                  message("%1X%1X%1X%1X    ",
		    pe[c_pe].data_tlb[start].prot>>3&1,
		    pe[c_pe].data_tlb[start].prot>>2&1,
		    pe[c_pe].data_tlb[start].prot>>1&1,
		    pe[c_pe].data_tlb[start].prot   &1);
	          if (++start >= TLB_NUM)
		    break;
                }
                message("\n");
              }
              continue;

	    case 's':
              ignore_rest_of_line(cmdP);
              message("           000 001 002 003 004 005 006 007  ");
              message("008 009 010 011 012 013 014 015\n");
              for (i = 0; i < TLB_NUM; i+=16) {
                message("D_TLB[%3u]", i);
                for (j = 0; j < 16; j++) {
                  if (!(j & 7)) message(" ");
                  message("%c%c%c ",
                    pe[c_pe].data_tlb[i+j].vmc&4?'V':'-',
                    pe[c_pe].data_tlb[i+j].vmc&2?'M':'-',
                    pe[c_pe].data_tlb[i+j].vmc&1?'C':'-');
                }
                message("\n");
              }            
              continue;    
            }
	  }
	  break;

	case 'c':
          if (*cmdP == '/' || isspc()) { /* dump data CACHE */
            static unsigned type; /* 'd':detail 's':summary */
            static unsigned start, dummy;
            unsigned i, j;

            if (extract_option(&cmdP, &dummy, &type, 0))
              continue;
            if (type == 0) type = 's';
            if (type!='d'&&type!='s') {
              message("illegal: type='%c'\n", type);
              continue;
            }
            switch (type) {
            case 'd':
              if (extract_regrange(0, &cmdP, &dummy, &start, &dummy))
                continue;
              ignore_rest_of_line(cmdP);
              message("D_CHE[%3u] ", start);
	      if (start >= CACHE_NUM) {
	        message("not exist\n");
                continue;
              }
              message("la=0x%08.8X pa=0x%08.8X v=%u\n",
	        start<<6,
	        pe[c_pe].data_cache[start].pa<<15|start<<6,
	        pe[c_pe].data_cache[start].v);
	      for (i=0; i<16; i++) {
	        if (i == 8) message("\n");
	        if (!(i & 3)) message(" ");
                message(" %08.8X", *((unsigned *) (pe[c_pe].data_cache[start].body)+i));
	      }
	      message("\n");
	      start++;
              continue;

            case 's':
              ignore_rest_of_line(cmdP);
              message("           0-------+-----15 16------+-----31 ");
              message("32------+-----47 48------+-----63\n");
              for (i = 0; i < CACHE_NUM; i+=64) {
                message("D_CHE[%3u]", i);
                for (j = 0; j < 64; j++) {
                  if (!(j & 15)) message(" ");
                  message("%c", pe[c_pe].data_cache[i+j].v?'V':'-');
                }
                message("\n");
              }
              continue;
            }
          }
          break;
        }
        break;

      case 'g':
        if (isspc()) { /* dump gr */
          unsigned i, j, d;

          ignore_rest_of_line(cmdP);
          for (i = 0; i < GR_NUM; i+=8) {
            message("gr%-3u", i);
            for (j = 0; j < 8; j++) {
              if (cmd_gr_access(c_pe,i+j,_R_,&d))
                goto exit_dumpgr;
              message(" %08.8X", d);
	    }
            message("\n");
          }
exit_dumpgr:
          continue;
        }
        break;

      case 'f':
        if (isspc()) { /* dump fr */
          unsigned i, j, d1, d2;

          ignore_rest_of_line(cmdP);
          for (i = 0; i < FR_NUM; i+=4) {
            message("fr%-3u", i);
            for (j = 0; j < 4; j++) {
              if (cmd_fr_access(c_pe,i+j,_R_,&d1, &d2))
                goto exit_dumpfr;
              message(" %08.8X_%08.8X", d1, d2);
            }
            message("\n");
          }
exit_dumpfr:
          continue;
        }
        break;

      case 's':
        if (isspc()) { /* dump sr */
          unsigned i, j, d;

          ignore_rest_of_line(cmdP);
          for (i = 0; i < SR_REALNUM; i+=8) {
            message("sr%-3u", i);
            for (j = 0; j < 8; j++) {
              if (cmd_sr_access(c_pe,i+j,_R_,&d))
                goto exit_dumpsr;
              message(" %08.8X", d);
            }
            message("\n");
          }
exit_dumpsr:
          continue;
        }
        break;

      case 'o':
        if (isspc()) { /* dump tod */
          unsigned d1, d2;

          ignore_rest_of_line(cmdP);
	  if (cmd_tod_access(c_pe,_R_,&d1,&d2))
            goto exit_dumptod;
          message("tod   %08.8X_%08.8X\n", d1, d2);
exit_dumptod:
          continue;
        }
        break;

#ifdef VCP
      case 'v':
        if (isspc()) { /* dump vr */
          static unsigned select, start, dummy;
	  unsigned i, j, d1, d2;

          if (extract_regrange('v', &cmdP, &select, &start, &dummy))
            continue;
          ignore_rest_of_line(cmdP);
          if (start  == 0) start = 1;
          for (i = 0; i < VRL; i+=4) {
            message("vr%-3u.%-4u", select, start);
            for (j = 0; j < 4; j++) {
              if (cmd_vr_access(c_pe,select,start+j,_R_,&d1,&d2))
                goto exit_dumpvr;
              message(" %08.8X%08.8X", d1, d2);
            }
            message("\n");
	    start += 4;
          }
exit_dumpvr:
          continue;
        }
        break;

      case 'k':
        if (isspc()) { /* dump mr */
          static unsigned select, start, dummy;
          unsigned i, j, d;

          if (extract_regrange('k', &cmdP, &select, &start, &dummy))
            continue;
          ignore_rest_of_line(cmdP);
          if (start  == 0) start = 1;
          for (i = 0; i < VRL; i+=4) {
            message("mr%-3u.%-4u", select, start);
            for (j = 0; j < 4; j++) {
              if (cmd_mr_access(c_pe,select,start+j,1,_R_,&d))
                goto exit_dumpmr;
              message(" %01.1X               ", d);
            }
            message("\n");
	    start += 4;
          }
exit_dumpmr:
          continue;
        } 
        break;

      case 'c':
        if (isspc()) { /* dump vcr */
          unsigned i, j, d;

          ignore_rest_of_line(cmdP);
          for (i = 0; i < VCR_NUM; i+=8) {
            message("vcr%-2u", i);
            for (j = 0; j < 8; j++) {
              if (cmd_vcr_access(c_pe,i+j,_R_,&d))
                goto exit_dumpvcr;
              message(" %08.8X", d);
            }
            message("\n");
          }
exit_dumpvcr:
          continue;
        }
        break;

      case 'u':
        if (isspc()) { /* dump vut */
          unsigned i, j, d;

          ignore_rest_of_line(cmdP);
          for (i = 0; i < VUT_NUM; i+=8) {
            message("vut%-2u", i);
            for (j = 0; j < 8; j++) {
              if (cmd_vut_access(c_pe,i+j,_R_,&d))
                goto exit_dumpvut;
              message(" %08.8X", d);
            }
            message("\n");
          }
exit_dumpvut: 
          continue; 
        } 
        break;

      case 't':
        if (isspc()) { /* dump tr */
          unsigned i, j, d;

          ignore_rest_of_line(cmdP);
          for (i = 0; i < TR_NUM; i+=8) {
            message("tr%-3u", i);
            for (j = 0; j < 8; j++) {
              if (cmd_tr_access(c_pe,i+j,_R_,&d))
                goto exit_dumptr;
              message(" %08.8X", d);
            }
            message("\n");
          }
exit_dumptr:
          continue;
        }  
        break;
#endif

      case 'x': /* display statistics */
        if (isspc()) {
          unsigned i, j;

          ignore_rest_of_line(cmdP);
#ifdef VCP
	  message("### VR confliction statistics ###:\n");
	  message("         RW(clock)  WR(clock)  WW(clock)         RW(clock)  WR(clock)  WW(clock)\n");
          for (i=0; i<VR_NUM/2; i++) {
            message(" vr%-3u: %10u %10u %10u",
              i,
              pe[c_pe].vr_rw_conflict[i],
              pe[c_pe].vr_wr_conflict[i],
              pe[c_pe].vr_ww_conflict[i]);
            message(" vr%-3u: %10u %10u %10u\n",
              i+VR_NUM/2,
              pe[c_pe].vr_rw_conflict[i+VR_NUM/2],
              pe[c_pe].vr_wr_conflict[i+VR_NUM/2],
              pe[c_pe].vr_ww_conflict[i+VR_NUM/2]);
          }
	  message("### MR confliction statistics ###:\n");
	  message("         RW(clock)  WR(clock)  WW(clock)         RW(clock)  WR(clock)  WW(clock)\n");
          for (i=0; i<MR_NUM/2; i++) {
            message(" mr%-3u: %10u %10u %10u",
              i,
              pe[c_pe].mr_rw_conflict[i],
              pe[c_pe].mr_wr_conflict[i],
              pe[c_pe].mr_ww_conflict[i]);
            message(" mr%-3u: %10u %10u %10u\n",
              i+MR_NUM/2,
              pe[c_pe].mr_rw_conflict[i+MR_NUM/2],
              pe[c_pe].mr_wr_conflict[i+MR_NUM/2],
              pe[c_pe].mr_ww_conflict[i+MR_NUM/2]);
          }
	  message("### VECTOR INSTRUCTION statistics ###:\n");
          for (i=0; i<64; i++) {
            message(" opc%-3u: %10u ",  i,     pe[c_pe].vinsn_stat[i]);
            message(" opc%-3u: %10u ",  i+64,  pe[c_pe].vinsn_stat[i+64]);
            message(" opc%-3u: %10u ",  i+128, pe[c_pe].vinsn_stat[i+128]);
            message(" opc%-3u: %10u\n", i+192, pe[c_pe].vinsn_stat[i+192]);
          }
#endif
	  message("### SCALAR INSTRUCTION statistics ###:\n");
          for (i=2; i<15; i++) {
            message("*** type%01.1X      OP0        OP1        OP2               OP0        OP1        OP2\n", i);
            switch (i) {
            case 2: case 3: case 4: case 5: case 6: case 7:
              for (j=0; j<16; j++) {
                message(" opc%-2u: %10u %10u %10u",
                  j,
                  pe[c_pe].sinsn_stat[i][j][0],
                  pe[c_pe].sinsn_stat[i][j][1],
                  pe[c_pe].sinsn_stat[i][j][2]);
                message(" opc%-2u: %10u %10u %10u\n",
                  j+16,
                  pe[c_pe].sinsn_stat[i][j+16][0],
                  pe[c_pe].sinsn_stat[i][j+16][1],
                  pe[c_pe].sinsn_stat[i][j+16][2]);
              }
              break;

            case 8: case 9: case 10: case 12: case 13: case 14:
              for (j=0; j<16; j++) {
                message(" opc%-2u: %10u %10u ----------",
                  j,
                  pe[c_pe].sinsn_stat[i][j][0],
                  pe[c_pe].sinsn_stat[i][j][1]);
                message(" opc%-2u: %10u %10u ----------\n",
                  j+16,
                  pe[c_pe].sinsn_stat[i][j+16][0],
                  pe[c_pe].sinsn_stat[i][j+16][1]);
              }
              break;

            case 11:
              for (j=0; j<16; j++) {
                message(" opc%-2u: %10u ---------- ----------",
                  j,
                  pe[c_pe].sinsn_stat[i][j][0]);
                message(" opc%-2u: %10u ---------- ----------\n",
                  j+16,
                  pe[c_pe].sinsn_stat[i][j+16][0]);
              }
              break;
            }
          }
	  message("### BRANCH statistics ###:\n");
          message("  cond-brcs back  taken: %10u\n", pe[c_pe].brcs_b_t);
          message("  cond-brcs back !taken: %10u\n", pe[c_pe].brcs_b_n);
          message("  cond-brcs forw !taken: %10u\n", pe[c_pe].brcs_f_n);
          message("  cond-brcs forw  taken: %10u\n", pe[c_pe].brcs_f_t);
          message("  cond-brc  back  taken: %10u\n", pe[c_pe].brc_b_t);
          message("  cond-brc  back !taken: %10u\n", pe[c_pe].brc_b_n);
          message("  cond-brc  forw !taken: %10u\n", pe[c_pe].brc_f_n);
          message("  cond-brc  forw  taken: %10u\n", pe[c_pe].brc_f_t);
	  message("### ACCESS REQUEST statistics ###:\n");
	  message("           request\n");
          message("  text: %10u\n", pe[c_pe].text_access);
          message("  data: %10u\n", pe[c_pe].data_access);
	  message("### TLB statistics ###:\n");
	  message("            access        hit    ratio\n");
          message("  text: %10u %10u   %5.1f%%\n", 
            pe[c_pe].text_tlb_access,
            pe[c_pe].text_tlb_hit,
            (float) pe[c_pe].text_tlb_hit * 100 / (float) pe[c_pe].text_tlb_access);
          message("  data: %10u %10u   %5.1f%%\n", 
            pe[c_pe].data_tlb_access,
            pe[c_pe].data_tlb_hit,
            (float) pe[c_pe].data_tlb_hit * 100 / (float) pe[c_pe].data_tlb_access);
	  message("### CACHE statistics ###:\n");
	  message("            access        hit    ratio\n");
          message("  text: %10u %10u   %5.1f%%\n", 
            pe[c_pe].text_cache_access,
            pe[c_pe].text_cache_hit,
            (float) pe[c_pe].text_cache_hit * 100 / (float) pe[c_pe].text_cache_access);
          message("  data: %10u %10u   %5.1f%%\n", 
            pe[c_pe].data_cache_access,
            pe[c_pe].data_cache_hit,
            (float) pe[c_pe].data_cache_hit * 100 / (float) pe[c_pe].data_cache_access);
          continue;
        }
        break;

      case 'y': /* hidden command (display internal status) */
        if (isspc()) {
          static unsigned start, dummy;
          unsigned mcra, mopa, i, j, d;

          if (extract_regrange('l', &cmdP, &dummy, &start, &dummy))
            continue;
          ignore_rest_of_line(cmdP);
          start &= ~3;
	  message("TOD_5NANO:       %08.8X\n", pe[c_pe].tod_5nano);
	  message("FIRMSTATE:       %08.8X\n", pe[c_pe].firmstate);
	  message("NULLIFY:         %08.8X\n", pe[c_pe].nullify);
	  message("LPC:             %08.8X\n", pe[c_pe].lpc);
	  message("IP_STATUS:       %08.8X\n", pe[c_pe].ip_status);
	  message("HANGUP_TIMER:    %08.8X\n", pe[c_pe].hangup_timer);
	  message("TIMEOUT_REQ:     %08.8X\n", pe[c_pe].timeout_req);
	  message("CNTL_BLOCK:      %08.8X\n", pe[c_pe].cntl_block);
	  message("SVC_INTERRUPT:   %08.8X\n", pe[c_pe].svc_interrupt);
	  message("FWC_INTERRUPT:   %08.8X\n", pe[c_pe].fwc_interrupt);
	  message("EXT_INTERRUPT:   %08.8X\n", pe[c_pe].ext_interrupt);
	  message("EXT_INTR_LEVEL:  %08.8X\n", pe[c_pe].ext_intr_level);
	  message("SLRP_INTR_LEVEL: %08.8X\n", pe[c_pe].slrp_intr_level);
	  message("EXT_INTR_DELAY:  %08.8X\n", pe[c_pe].ext_intr_delay);
	  message("SLRP_INTR_DELAY: %08.8X\n", pe[c_pe].slrp_intr_delay);
#ifdef FXT_INTR
	  message("FXT_INTR_LEVEL:  %08.8X\n", pe[c_pe].fxt_intr_level);
	  message("FXT_INTR_DELAY:  %08.8X\n", pe[c_pe].fxt_intr_delay);
#endif
	  message("WORK_IRD0:       %08.8X\n", *(unsigned*)&pe[c_pe].work_ird0);
	  message("WORK_IRD1:       %08.8X\n", *(unsigned*)&pe[c_pe].work_ird1);
	  message("         +0+1+2+3+4+5+6+7+8+9+A+B+C+D+E+F+0+1+2+3+4+5+6+7+8+9+A+B+C+D+E+F\n");
	  message("GR_SCORE:");
	  for (i=0; i<GR_NUM; i++) message(" %u", pe[c_pe].gr_score[i]);
	  message("\nFR_SCORE:");
	  for (i=0; i<FR_NUM; i++) message(" %u", pe[c_pe].fr_score[i]);
	  message("\nSR_SCORE: %u\n", pe[c_pe].sr_score);
	  message("IF_MULTIPLE_CYCLE: %u\n", pe[c_pe].if_multiple_cycle);
	  message("IP_MULTIPLE_CYCLE: %u\n", pe[c_pe].ip_multiple_cycle);
#ifdef MCR_AREA
          mcra = start;
	  message("SMCR:    +0 +1 +2 +3 +4 +5 +6 +7  ");
          message("+8 +9 +A +B +C +D +E +F\n");
          for (i = 0; i < 16; i++) {
            message("%08.8X", mcra);
            for (j = 0; j < 16; j+=4) {
              if (j == 8) message(" ");
              if (cmd_smcr_access(c_pe,mcra+j,4,_R_,&d))
                goto exit_dumpsmcr;
              message(" %02.2X %02.2X %02.2X %02.2X",
                (unsigned char) ((d >> 24) & 0xff),
                (unsigned char) ((d >> 16) & 0xff),
                (unsigned char) ((d >>  8) & 0xff),
                (unsigned char) ( d        & 0xff));
            }
            message("\n");
            mcra += 16;
          }
exit_dumpsmcr:
#endif
	  message("AP_DATQ: av=%u delay=%u addr=0x%08.8X\n", 
	    pe[c_pe].ap_datq[0].av,
	    pe[c_pe].ap_datq[0].delay,
	    pe[c_pe].ap_datq[0].addr);
	  message("AP_CHEQ: av=%u delay=%u addr=0x%08.8X\n", 
	    pe[c_pe].ap_cheq[0].av,
	    pe[c_pe].ap_cheq[0].delay,
	    pe[c_pe].ap_cheq[0].addr);
	  message("FQ_TOP:     %08.8X\n", pe[c_pe].fq_top);
	  message("AVN_FPWD:   %08.8X\n", pe[c_pe].avn_fpwd);
#ifdef VCP
	  message("VP_SCORE:   %u", pe[c_pe].vp_score);
	  message("\nVR_RSCORE: ");
	  for (i=0; i<VR_NUM; i++) message(" %u", pe[c_pe].vr_rscore[i]);
	  message("\nVR_WSCORE: ");
	  for (i=0; i<VR_NUM; i++) message(" %u", pe[c_pe].vr_wscore[i]);
	  message("\nMR_RSCORE: ");
	  for (i=0; i<MR_NUM; i++) message(" %u", pe[c_pe].mr_rscore[i]);
	  message("\nMR_WSCORE: ");
	  for (i=0; i<MR_NUM; i++) message(" %u", pe[c_pe].mr_wscore[i]);
	  message("\nVOP_ISCORE: %u", pe[c_pe].vop_iscore);
	  message("\nVOP_RSCORE:");
	  for (i=0; i<ID_NUM; i++) message(" %u", pe[c_pe].vop_rscore[i]);
	  message("\nVOP_WSCORE:");
	  for (i=0; i<ID_NUM; i++) message(" %u", pe[c_pe].vop_wscore[i]);
	  message("\nVP_OPIQ:   ");
          for (i=0; i<VP_OPIQNUM; i++) message(" %u", pe[c_pe].vp_opiq[i].av);
	  message("\nVP_OPMQ:   ");
          for (i=0; i<VP_OPMQNUM; i++) message(" %u", pe[c_pe].vp_opmq[i].av);
	  message("\nVP_OPCQ:   ");
          for (i=0; i<VP_OPCQNUM; i++) message(" %u", pe[c_pe].vp_opcq[i].av);
	  message("\nVP_LDQ:    ");
          for (i=0; i<VP_LDQNUM; i++) message(" %u", pe[c_pe].vp_ldq[i].av);
	  message("\nVP_STQ:    ");
          for (i=0; i<VP_STQNUM; i++) message(" %u", pe[c_pe].vp_stq[i].av);
	  message("\nVP_ADDQ:   ");
          for (i=0; i<VP_ADDQNUM; i++) message(" %u", pe[c_pe].vp_addq[i].av);
	  message("\nVP_MULQ:   ");
          for (i=0; i<VP_MULQNUM; i++) message(" %u", pe[c_pe].vp_mulq[i].av);
	  message("\nVP_DIVQ:   ");
          for (i=0; i<VP_DIVQNUM; i++) message(" %u", pe[c_pe].vp_divq[i].av);
	  message("\nVP_MSKQ:   ");
          for (i=0; i<VP_MSKQNUM; i++) message(" %u", pe[c_pe].vp_mskq[i].av);
	  message("\nVP_LDP:    ");
          for (i=0; i<VP_LDPLEN; i++) message(" %u", pe[c_pe].vp_ldp[i].av);
	  message("\nVP_STP:    ");
          for (i=0; i<VP_STPLEN; i++) message(" %u", pe[c_pe].vp_stp[i].av);
	  message("\nVP_ADDP:   ");
          for (i=0; i<VP_ADDPLEN; i++) message(" %u", pe[c_pe].vp_addp[i].av);
	  message("\nVP_MULP:   ");
          for (i=0; i<VP_MULPLEN; i++) message(" %u", pe[c_pe].vp_mulp[i].av);
	  message("\nVP_DIVP:   ");
          for (i=0; i<VP_DIVPLEN; i++) message(" %u", pe[c_pe].vp_divp[i].av);
	  message("\nVP_MSKP:   ");
          for (i=0; i<VP_MSKPLEN; i++) message(" %u", pe[c_pe].vp_mskp[i].av);
	  message("\n");
#endif
#ifdef MOVER
	  message("WORK_MRPRMI:         %08.8X\n", *(unsigned*)&pe[c_pe].work_mrprmi);
	  message("WORK_SNDMRI:         %08.8X\n", *(unsigned*)&pe[c_pe].work_sndmri);
	  message("MV_MRPRMI:           %08.8X\n", *(unsigned*)&pe[c_pe].mv_mrprmi);
	  message("MV_RCV_SRCPE:        %08.8X\n", pe[c_pe].mv_rcv_srcpe);
	  message("MV_RCV_STATUS:       %08.8X\n", pe[c_pe].mv_rcv_status);
	  message("MV_RCV_STOPRQ:       %08.8X\n", pe[c_pe].mv_rcv_stoprq);
	  message("MV_SND_STOPRQ:       %08.8X\n", pe[c_pe].mv_snd_stoprq);
	  message("MV_HROUND:           %08.8X\n", pe[c_pe].mv_hround);
	  message("MV_LROUND:           %08.8X\n", pe[c_pe].mv_lround);
	  message("MV_HQO:              %08.8X\n", pe[c_pe].mv_hqo);
	  message("MV_QID:              %08.8X\n", pe[c_pe].mv_qid);
	  message("MV_RBSEL:            %08.8X\n", pe[c_pe].mv_rbsel);
	  message("MV_HEDR.AV:          %08.8X\n", pe[c_pe].mv_hedr.av);
	  message("MV_HEDR.HCOUNT:      %08.8X\n", pe[c_pe].mv_hedr.hcount);
	  message("MV_HEDR.BCOUNT:      %08.8X\n", pe[c_pe].mv_hedr.bcount);
	  message("MV_HEDR.HEDR.DSTNPA: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstnpa);
	  message("MV_HEDR.HEDR.ELNUM:  %08.8X\n", pe[c_pe].mv_hedr.hedr.elnum);
	  message("MV_HEDR.HEDR.SRCECC: %08.8X\n", pe[c_pe].mv_hedr.hedr.srcecc);
	  message("MV_HEDR.HEDR.SRCNPA: %08.8X\n", pe[c_pe].mv_hedr.hedr.srcnpa);
	  message("MV_HEDR.HEDR.SRCQID1:%08.8X\n", pe[c_pe].mv_hedr.hedr.srcqid1);
	  message("MV_HEDR.HEDR.SRCQID2:%08.8X\n", pe[c_pe].mv_hedr.hedr.srcqid2);
	  message("MV_HEDR.HEDR.SRCQID3:%08.8X\n", pe[c_pe].mv_hedr.hedr.srcqid3);
	  message("MV_HEDR.HEDR.SRCQID4:%08.8X\n", pe[c_pe].mv_hedr.hedr.srcqid4);
	  message("MV_HEDR.HEDR.SRCDAM: %08.8X\n", pe[c_pe].mv_hedr.hedr.srcdam);
	  message("MV_HEDR.HEDR.DSTDAM: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstdam);
	  message("MV_HEDR.HEDR.TYPE:   %08.8X\n", pe[c_pe].mv_hedr.hedr.type);
	  message("MV_HEDR.HEDR.SRCPCI: %08.8X\n", pe[c_pe].mv_hedr.hedr.srcpci);
	  message("MV_HEDR.HEDR.DSTPCI: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstpci);
	  message("MV_HEDR.HEDR.SRCCTI: %08.8X\n", pe[c_pe].mv_hedr.hedr.srccti);
	  message("MV_HEDR.HEDR.DSTCTI: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstcti);
	  message("MV_HEDR.HEDR.SRCCEN: %08.8X\n", pe[c_pe].mv_hedr.hedr.srccen);
	  message("MV_HEDR.HEDR.DSTCEN: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstcen);
	  message("MV_HEDR.HEDR.DSTNPF: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstnpf);
	  message("MV_HEDR.HEDR.SRCVAF: %08.8X\n", pe[c_pe].mv_hedr.hedr.srcvaf);
	  message("MV_HEDR.HEDR.DSTVAF: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstvaf);
	  message("MV_HEDR.HEDR.SRCVAS: %08.8X\n", pe[c_pe].mv_hedr.hedr.srcvas);
	  message("MV_HEDR.HEDR.DSTVAS: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstvas);
	  message("MV_HEDR.HEDR.LSYNC:  %08.8X\n", pe[c_pe].mv_hedr.hedr.lsync);
	  message("MV_HEDR.HEDR.HSYNC:  %08.8X\n", pe[c_pe].mv_hedr.hedr.hsync);
	  message("MV_HEDR.HEDR.SVSPID: %08.8X\n", pe[c_pe].mv_hedr.hedr.svspid);
	  message("MV_HEDR.HEDR.IPL:    %08.8X\n", pe[c_pe].mv_hedr.hedr.ipl);
	  message("MV_HEDR.HEDR.DSTPID: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstpid);
	  message("MV_HEDR.HEDR.SRCBAD: %08.8X\n", pe[c_pe].mv_hedr.hedr.srcbad);
	  message("MV_HEDR.HEDR.DSTBAD: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstbad);
	  message("MV_HEDR.HEDR.SRCAD1: %08.8X\n", pe[c_pe].mv_hedr.hedr.srcad1);
	  message("MV_HEDR.HEDR.DSTAD1: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstad1);
	  message("MV_HEDR.HEDR.SRCAD2: %08.8X\n", pe[c_pe].mv_hedr.hedr.srcad2);
	  message("MV_HEDR.HEDR.DSTAD2: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstad2);
	  message("MV_HEDR.HEDR.SRCELN: %08.8X\n", pe[c_pe].mv_hedr.hedr.srceln);
	  message("MV_HEDR.HEDR.DSTELN: %08.8X\n", pe[c_pe].mv_hedr.hedr.dsteln);
	  message("MV_HEDR.HEDR.SRCCTA: %08.8X\n", pe[c_pe].mv_hedr.hedr.srccta);
	  message("MV_HEDR.HEDR.DSTCTA: %08.8X\n", pe[c_pe].mv_hedr.hedr.dstcta);
          mopa = start;
          message("MV_ADDR_TRS: virtual=%08.8X\n", mopa);
          switch (pe[c_pe].mcr[MTC].mtc.ps) {
          default:
            message("   illegal page size: mtc.ps=%u\n", pe[c_pe].mcr[MTC].mtc.ps);
            break;
          case 11:
	  case 13:
	  case 15:
            switch (pe[c_pe].mcr[MTC].mtc.dnupro) {
            case 0:
            case 3:
              message("   illegal dnupro: dnupro=%u\n", pe[c_pe].mcr[MTC].mtc.dnupro);
              break;
            case 1:
              message("   upid=0,mtr=0:real=");
              cmd_mop_dat1(c_pe, 0, 0, mopa);
              message("   upid=0,mtr=1:real=");
              cmd_mop_dat1(c_pe, 0, 1, mopa);
              message("   upid=1,mtr=0:real=");
              cmd_mop_dat1(c_pe, 1, 0, mopa);
              message("   upid=1,mtr=1:real=");
              cmd_mop_dat1(c_pe, 1, 1, mopa);
              break;
            case 2:
              message("   upid=0,mtr=0:real=");
              cmd_mop_dat2(c_pe, 0, 0, mopa);
              message("   upid=0,mtr=1:real=");
              cmd_mop_dat2(c_pe, 0, 1, mopa);
              message("   upid=1,mtr=0:real=");
              cmd_mop_dat2(c_pe, 1, 0, mopa);
              message("   upid=1,mtr=1:real=");
              cmd_mop_dat2(c_pe, 1, 1, mopa);
              break;
            }
            break;
          }
#endif
          continue;
        }
        break;

#ifdef FSBA
      case 'z': /* hidden command (display fsba/fst/dasd status) */
        if (*cmdP == '/' || isspc()) { /* dump FSBA/FST/DASD */
          static unsigned c_fsba, start, dummy;
          unsigned i, j;
          unsigned d, c;

          if (extract_option(&cmdP, &c_fsba, &dummy, 0))
              continue;
          if (c_fsba >= FSBA_NUM) {
	    message("FSBA#%u not exist\n", c_fsba);
            continue;
          }
          if (extract_memrange(c_pe, &cmdP, 256, &start, &dummy))
            continue;
          skipspc();
          ignore_rest_of_line(cmdP);

          message("FSBA#%u\n", c_fsba);
          message("FSCR:    %08.8X\n", *(unsigned *)&pe[c_pe].fscr[c_fsba]);
	  message("FSIR:   ");
          for (i=0; i<FSIR_NUM/4; i++)
            message(" %08.8X", *((unsigned *)&pe[c_pe].fsir[c_fsba]+i));
	  message("\n");
	  message("FSRQ:   ");
          for (i=0; i<FSBA_SLOT_NUM; i++) {
	      if (i && (i & 7) == 0) message("        ");
              message(" %08.8X", *((unsigned *)pe[c_pe].fsrq[c_fsba]+i));
	      if ((i & 7) == 7) message("\n");
          }

#ifdef FST
	  message("FSTCR:  ");
          for (i=0; i<FSTCR_NUM/2; i++) message(" %08.8X", *((unsigned *)&pe[c_pe].fstcr[c_fsba]+i));
	  message("\n");

          start &= ~7;
	  message("FSTMEM:  +0 +1 +2 +3 +4 +5 +6 +7  ");
          message("+8 +9 +A +B +C +D +E +F\n");
          for (i = 0; i < 16; i++) {
            message("%08.8X", start &= FSTMEM_SIZE-1);
            for (j = 0; j < 16; j+=4) {
              if (j == 8) message(" ");
              d = *(unsigned *)&pe[c_pe].fstmem[c_fsba][start+j];
              message(" %02.2X %02.2X %02.2X %02.2X",
	        (unsigned char) ((d >> 24) & 0xff),
	        (unsigned char) ((d >> 16) & 0xff),
	        (unsigned char) ((d >>  8) & 0xff),
	        (unsigned char) ( d        & 0xff));
            }
            message("\n");
	    start += 16;
          }
#endif
#ifdef DASD
          message("STATUS:  %08.8X\n", *(unsigned *)&pe[c_pe].dasd_status);
          message("ACF:     %04.4X\n", *(unsigned short *)&pe[c_pe].acf);
	  message("ACR:    ");
          for (i=0; i<ACR_NUM/4; i++) message(" %08.8X", *((unsigned *)&pe[c_pe].acr+i));
	  message("\nDACR:   ");
          for (i=0; i<ACR_NUM/4; i++) message(" %08.8X", *((unsigned *)&pe[c_pe].dacr+i));
	  message("\n");
          message("DCBA:    %08.8X\n", *(unsigned *)&pe[c_pe].dcba);
          message("DCWA:    %08.8X\n", *(unsigned *)&pe[c_pe].dcwa);
#endif
          continue;
        }
        break;
#endif

      default:
        break;
      }
      break;
    }
    return(1); /* failed */
  }
  return(0); /* succeeded */
}

command_r(cmdP, fp, log) char * cmdP; FILE *fp; int log;
    /* return 1 if failed, 0 if succeeded */
{
  char        *get_srname();
  unsigned once;

  for (once = 0; !once; once++) {
    switch (once) {
    case 0:
      if (!strncmp("andomize", cmdP, 8)) { /* randomize facility */
        unsigned hex;

        cmdP += 8;
        if (!isspc()) break;
        skipspc();
	if (sscanf(cmdP, "%x", &hex) != 1) {
          message("{hex} expected\n");
          continue;
        }
        skipxdigit();
        ignore_rest_of_line(cmdP);
	randomize_all(hex);
	continue;
      }
      switch (*cmdP++) {
#ifdef MCR_AREA
      case 'l': /* mcr */
#endif
      case 'm': /* scalar */
#ifdef VCP
      case 'n': /* vector */
#endif
        if (*cmdP == '/' || isspc()) { /* read mem */
          unsigned facility;
          static unsigned len;
          static unsigned type, dummy;
          static unsigned start, end;
          unsigned count, addr;
          FILE *write_fp; FILE *open_file();
          unsigned d[4];

          facility = *(cmdP-1);
          if (extract_option(&cmdP, &len, &type, 0))
            continue;
          if (len  == 0) len = 4;
          if (type == 0) type = 'x';
          if (type!='d'&&type!='x'&&type!='f'&&type!='z') {
            message("illegal: type='%c'\n", type);
            continue;
          }
#ifdef MCR_AREA
          if (facility == 'l') {
            if (extract_regrange(facility, &cmdP, &dummy, &start, &end))
              continue;
          }
          else {
#endif
            if (extract_memrange(c_pe, &cmdP, len, &start, &end))
              continue;
#ifdef MCR_AREA
          }
#endif
          skipspc();
	  count = (end - start) / len + 1;
          if (!*cmdP) write_fp = NULL;
          else if (*cmdP == '>') {
            cmdP++;
            if (*cmdP == '>') {
              cmdP++;
              skipspc();
              delete_space(cmdP);
              if (!(write_fp = open_file(cmdP, "a")))
                continue;
            }
            else {
              skipspc();
              delete_space(cmdP);
              if (!(write_fp = open_file(cmdP, "w")))
                continue;
            }
          }
          else {
	    message("'>' expected\n");
	    continue;
	  }

          for (addr=start; addr<=end; addr+=len) {
            switch (facility) {
#ifdef MCR_AREA
            case 'l':
              if (cmd_mcr_access(c_pe,addr,len,_R_,d))
                goto exit_readmem;
              break;
#endif

            case 'm':
              if (cmd_mem_access(c_pe,c_mode,addr,len,_R_,d,'n'))
                goto exit_readmem;
              break;

#ifdef VCP
            case 'n':
              if (cmd_vop_access(c_pe,c_node,addr,len,_R_,d,'n'))
                goto exit_readmem;
              break;
#endif
            }
	    if (!write_fp || c_echo == 'y')
              message("%08.8X  = ",addr);
	    if (put_format_d(addr, len, d, type, c_echo, write_fp))
	      goto exit_readmem;
          }

	  start += count * len;
	  end += count * len;
exit_readmem:
          fclose(write_fp);
          continue;
	}
        break;
      case 'g': /* read gr */
      case 'f': /* read fr */
      case 's': /* read sr */
      case 'o': /* read tod */
#ifdef VCP
      case 'v': /* read vr  */
      case 'k': /* read mr  */
      case 'c': /* read vcr */
      case 'u': /* read vut */
      case 't': /* read tr  */
#endif
        if (isspc()) { /* read reg */
          unsigned facility;
          static unsigned select, start, end;
          unsigned count, num;
          FILE *write_fp; FILE *open_file();
	  unsigned d1, d2;

          facility = *(cmdP-1);
#ifdef VCP
          if ((facility == 'v' || facility == 'k') && start == 0) start = end = 1;
#endif
          if (extract_regrange(facility, &cmdP, &select, &start, &end))
            continue;
          skipspc();
#ifdef VCP
          if ((facility == 'v' || facility == 'k') && start == 0) {
	    message("{index} begins from 1\n");
	    continue;
          }
#endif
          if (!*cmdP) write_fp = NULL;
          else if (*cmdP == '>') {
            cmdP++;
            if (*cmdP == '>') {
              cmdP++;
              skipspc();
              delete_space(cmdP);
              if (!(write_fp = open_file(cmdP, "a")))
                continue;
            }
            else {
              skipspc();
              delete_space(cmdP);
              if (!(write_fp = open_file(cmdP, "w")))
                continue;
            }
          }
          else {
	    message("'>' expected\n");
	    continue;
	  }

	  switch (facility) {
          case 'o':
            start = 0;
            end   = 0;
            count = 1;
	    break;

          default:
	    count = end - start + 1;
	    break;
          }

          for (num=start; num<=end; num++) {
            switch (facility) {
	    case 'g':
              if (cmd_gr_access(c_pe,num,_R_,&d1))
	        goto exit_readreg;
              break;

	    case 'f':
              if (cmd_fr_access(c_pe,num,_R_,&d1,&d2))
	        goto exit_readreg;
              break;

	    case 's':
              if (cmd_sr_access(c_pe,num,_R_,&d1))
	        goto exit_readreg;
              break;

	    case 'o':
              if (cmd_tod_access(c_pe,_R_,&d1,&d2))
	        goto exit_readreg;
              break;

#ifdef VCP
            case 'v':
              if (cmd_vr_access(c_pe,select,num,_R_,&d1,&d2))
                goto exit_readreg;
              break;

            case 'k':
              if (cmd_mr_access(c_pe,select,num,1,_R_,&d1))
                goto exit_readreg;
              break;

            case 'c':
              if (cmd_vcr_access(c_pe,num,_R_,&d1))
                goto exit_readreg;
              break;

            case 'u':
              if (cmd_vut_access(c_pe,num,_R_,&d1))
                goto exit_readreg;
              break;

            case 't':
              if (cmd_tr_access(c_pe,num,_R_,&d1))
                goto exit_readreg;
              break;
#endif
            }
            switch (facility) {
            case 'g':
	      if (!write_fp || c_echo == 'y')
                message("gr%-3u     = %08.8X\n",num,d1);
              if (write_fp)
                fprintf(write_fp,"%08.8X\t# gr%-3u\n",d1,num);
              break;

            case 'f':
	      if (!write_fp || c_echo == 'y')
                message("fr%-3u     = %08.8X_%08.8X\n",num,d1,d2);
              if (write_fp)
                fprintf(write_fp,"%08.8X_%08.8X\t# fr%-3u\n",d1,d2,num);
              break;

            case 's':
	      if (!write_fp || c_echo == 'y')
                message("%-7s   = %08.8X\n", get_srname(num),d1);
              if (write_fp)
                fprintf(write_fp,"%08.8X\t# %-7s\n",d1,get_srname(num));
              break;

            case 'o':
	      if (!write_fp || c_echo == 'y')
                message("tod       = %08.8X_%08.8X\n",d1,d2);
              if (write_fp)
                fprintf(write_fp,"%08.8X_%08.8X\t# tod\n",d1,d2);
              break;

#ifdef VCP
            case 'v':
	      if (!write_fp || c_echo == 'y')
                message("vr%-3u.%-4u= %08.8X_%08.8X\n",select,num,d1,d2);
              if (write_fp)
                fprintf(write_fp,"%08.8X_%08.8X\t# vr%-3u.%-4u\n",d1,d2,select,num);
              break;

            case 'k':
	      if (!write_fp || c_echo == 'y')
                message("mr%-3u.%-4u= %01.1X\n",select,num,d1);
              if (write_fp)
                fprintf(write_fp,"%01.1X\t# mr%-3u.%-4u\n",d1,select,num);
              break;

            case 'c':
	      if (!write_fp || c_echo == 'y')
                message("vcr%-3u    = %08.8X\n",num,d1);
              if (write_fp)
                fprintf(write_fp,"%08.8X\t# vcr%-3u\n",d1,num);
              break;

            case 'u':
	      if (!write_fp || c_echo == 'y')
                message("vut%-3u    = %08.8X\n",num,d1);
              if (write_fp)
                fprintf(write_fp,"%08.8X\t# vut%-3u\n",d1,num);
              break;

            case 't':
	      if (!write_fp || c_echo == 'y')
                message("tr%-3u     = %08.8X\n",num,d1);
              if (write_fp)
                fprintf(write_fp,"%08.8X\t# tr%-3u\n",d1,num);
              break;
#endif
            }
          }

	  start += count;
	  end += count;
exit_readreg:
          fclose(write_fp);
          continue;
        }
        break;

      default:
        break;
      }
      break;
    }
    return(1); /* failed */
  }
  return(0); /* succeeded */
}

command_w(cmdP, fp, log) char * cmdP; FILE *fp; int log;
      /* return 1 if failed, 0 if succeeded */
{
  char        *get_srname();
  unsigned once;

  for (once = 0; !once; once++) {
    switch (once) {
    case 0:
      switch (*cmdP++) {
#ifdef MCR_AREA
      case 'l': /* mcr */
#endif
      case 'm': /* scalar */
#ifdef VCP
      case 'n': /* vector */
#endif
        if (*cmdP == '/' || isspc()) { /* write mem */
	  unsigned facility;
          static unsigned len;
          static unsigned type, dummy;
          static unsigned start, end;
          unsigned count, addr;
          FILE *read_fp; FILE *open_file();
          unsigned d[4];

          facility = *(cmdP-1);
          if (extract_option(&cmdP, &len, &type, 0))
            continue;
          if (len  == 0) len = 4;
          if (type == 0) type = 'x';
          if (type!='d'&&type!='x'&&type!='f'&&type!='z') {
            message("illegal: type='%c'\n", type);
            continue;
          }
#ifdef MCR_AREA
          if (facility == 'l') {
            if (extract_regrange(facility, &cmdP, &dummy, &start, &end))
              continue;
          }
          else {
#endif
            if (extract_memrange(c_pe, &cmdP, len, &start, &end))
              continue;
#ifdef MCR_AREA
          }
#endif
          skipspc();
	  count = (end - start) / len + 1;
          if (!*cmdP) read_fp = NULL;
          else if (*cmdP == '<') {
            cmdP++;
            skipspc();
            delete_space(cmdP);
            if (!(read_fp = open_file(cmdP, "r")))
              continue;
          }
          else {
	    message("'<' expected\n");
	    continue;
	  }

          for (addr=start; addr<=end; addr+=len) {
            switch (facility) {
#ifdef MCR_AREA
            case 'l':
              if (cmd_mcr_access(c_pe,addr,len,_R_,d))
	        goto exit_writemem;
              break;
#endif

	    case 'm':
              if (cmd_mem_access(c_pe,c_mode,addr,len,_R_,d,'n'))
	        goto exit_writemem;
              break;

#ifdef VCP
	    case 'n':
              if (cmd_vop_access(c_pe,c_node,addr,len,_R_,d,'n'))
	        goto exit_writemem;
              break;
#endif
            }
	    if (!read_fp || c_echo == 'y') {
              message("%08.8X  = ",addr);
	      if (put_format_d(addr, len, d, type, 0, 0))
	        goto exit_writemem;
            }

            if (read_fp) {
              if (!fgets(cmdP = tmp_buf, COMMAND_LINE, read_fp)) {
                message("EOF");
                goto exit_writemem;
              }
              delete_comment(tmp_buf);
              delete_nl(tmp_buf);
            }
            else {
              arrow_prompt(fp);
              if (!fgets(cmdP = tmp_buf, COMMAND_LINE, fp)) {
                message("EOF\n");
                return;
              }
              expand_command(tmp_buf, PREV_COMMAND_BUF, tmp_buf, fp);
              strncpy(PREV_COMMAND_BUF, tmp_buf, COMMAND_LINE);
              if (log) log_command(tmp_buf);
              delete_comment(tmp_buf);
            }

	    switch (get_format_d(cmdP, len, d, type)) {
            case 0:
	      break;

            case 1: 
	      continue;
	      break;

            case 2:
	      goto retry_writemem;
	      break;

	    case 3:
              goto exit_writemem;
	      break;
            }

	    switch (facility) {
#ifdef MCR_AREA
            case 'l':
              cmd_mcr_access(c_pe,addr,len,_W_,d);
              break;
#endif

            case 'm':
              cmd_mem_access(c_pe,c_mode,addr,len,_W_,d,c_forcew);
	      break;

#ifdef VCP
	    case 'n':
              cmd_vop_access(c_pe,c_node,addr,len,_W_,d,c_forcew);
	      break;
#endif
            }

	    if (!read_fp || c_echo == 'y') {
	      switch (facility) {
#ifdef MCR_AREA
              case 'l':
                cmd_mcr_access(c_pe,addr,len,_R_,d);
		break;
#endif

              case 'm':
                cmd_mem_access(c_pe,c_mode,addr,len,_R_,d,'n');
		break;

#ifdef VCP
              case 'n':
                cmd_vop_access(c_pe,c_node,addr,len,_R_,d,'n');
		break;
#endif
	      }
              message("          = ");
	      if (put_format_d(addr, len, d, type, 0, 0))
		goto exit_writemem;
            }
            continue;
retry_writemem:
	    message("write mem? %s\n", cmdP);
	    addr -= len;
            continue;
	  }

	  start += count * len;
	  end += count * len;
          fclose(read_fp);
          continue;
exit_writemem:
	  message("write mem?\n");
          fclose(read_fp);
          continue;
        }
        break;
      case 'g': /* write gr */
      case 'f': /* write fr */
      case 's': /* write sr */
      case 'o': /* write tod */
#ifdef VCP
      case 'v': /* write vr  */
      case 'k': /* write mr  */
      case 'c': /* write vcr */
      case 'u': /* write vut */
      case 't': /* write tr  */
#endif
        if (isspc()) { /* write reg */
          unsigned facility;
          static unsigned select, start, end;
	  unsigned count, num;
          FILE *read_fp; FILE *open_file();
	  unsigned d1, d2;

          facility = *(cmdP-1);
#ifdef VCP
          if ((facility == 'v' || facility == 'k') && start == 0) start = end = 1;
#endif
          if (extract_regrange(facility, &cmdP, &select, &start, &end))
            continue;
          skipspc();
#ifdef VCP
          if ((facility == 'v' || facility == 'k') && start == 0) {
	    message("{index} begins from 1\n");
	    continue;
          }
#endif
          if (!*cmdP) read_fp = NULL;
          else if (*cmdP == '<') {
            cmdP++;
            skipspc();
            delete_space(cmdP);
            if (!(read_fp = open_file(cmdP, "r")))
              continue;
          }
          else break;

	  switch (facility) {
          case 'o':
            start = 0;
            end   = 0;
            count = 1;
            break;

          default:
            count = end - start + 1;
            break;
          }

          for (num=start; num<=end; num++) {
            switch (facility) {
            case 'g':
              if (cmd_gr_access(c_pe,num,_R_,&d1))
                goto exit_writereg;
              break;

            case 'f':
              if (cmd_fr_access(c_pe,num,_R_,&d1,&d2))
                goto exit_writereg;
              break;

            case 's':
              if (cmd_sr_access(c_pe,num,_R_,&d1))
                goto exit_writereg;
              break;

            case 'o':
              if (cmd_tod_access(c_pe,_R_,&d1,&d2))
                goto exit_writereg;
              break;

#ifdef VCP
            case 'v':
              if (cmd_vr_access(c_pe,select,num,_R_,&d1,&d2))
                goto exit_writereg;
              break;

            case 'k':
              if (cmd_mr_access(c_pe,select,num,1,_R_,&d1))
                goto exit_writereg;
              break;

            case 'c':
              if (cmd_vcr_access(c_pe,num,_R_,&d1))
                goto exit_writereg;
              break;

            case 'u':
              if (cmd_vut_access(c_pe,num,_R_,&d1))
                goto exit_writereg;
              break;

            case 't':
              if (cmd_tr_access(c_pe,num,_R_,&d1))
                goto exit_writereg;
              break;
#endif
            }
	    if (!read_fp || c_echo == 'y') {
              switch (facility) {
              case 'g':
                message("gr%-3u     = %08.8X\n",num,d1);
	        break;

              case 'f':
                message("fr%-3u     = %08.8X_%08.8X\n",num,d1,d2);
	        break;

              case 's':
                message("%-7s   = %08.8X\n",get_srname(num),d1);
	        break;

              case 'o':
                message("tod       = %08.8X_%08.8X\n",d1,d2);
	        break;

#ifdef VCP
              case 'v':
                message("vr%-3u.%-4u= %08.8X_%08.8X\n",select,num,d1,d2);
	        break;

              case 'k':
                message("mr%-3u.%-4u= %01.1X\n",select,num,d1);
	        break;

              case 'c':
                message("vcr%-3u    = %08.8X\n",num,d1);
	        break;

              case 'u':
                message("vut%-3u    = %08.8X\n",num,d1);
	        break;

              case 't':
                message("tr%-3u     = %08.8X\n",num,d1);
	        break;
#endif
	      }
	    }
  
            if (read_fp) {
              if (!fgets(cmdP = tmp_buf, COMMAND_LINE, read_fp)) {
                message("EOF");
                goto exit_writereg;
              }
              delete_comment(tmp_buf);
              delete_nl(tmp_buf);
            }
            else {
              arrow_prompt(fp);
              if (!fgets(cmdP = tmp_buf, COMMAND_LINE, fp)) {
                message("EOF\n");
                return;
              }
              expand_command(tmp_buf, PREV_COMMAND_BUF, tmp_buf, fp);
              strncpy(PREV_COMMAND_BUF, tmp_buf, COMMAND_LINE);
              if (log) log_command(tmp_buf);
              delete_comment(tmp_buf);
            }

            delete__(cmdP);

	    switch (facility) {
	    case 'g':
	    case 's':
#ifdef VCP
	    case 'k':
	    case 'c':
	    case 'u':
	    case 't':
#endif
              if (sscanf(cmdP, "%x", &d1) != 1) {
                skipspc();
	        if (!*cmdP) continue;
	        else goto retry_writereg;
	      }
	      switch (facility) {
	      case 'g':
	        cmd_gr_access(c_pe,num,_W_,&d1);
	        if (!read_fp || c_echo == 'y') {
	          cmd_gr_access(c_pe,num,_R_,&d1);
                  message("          = %08.8X\n",d1);
	        }
	        break;

	      case 's':
	        cmd_sr_access(c_pe,num,_W_,&d1);
	        if (!read_fp || c_echo == 'y') {
	          cmd_sr_access(c_pe,num,_R_,&d1);
                  message("          = %08.8X\n",d1);
	        }
	        break;

#ifdef VCP
	      case 'k':
	        cmd_mr_access(c_pe,select,num,1,_W_,&d1);
	        if (!read_fp || c_echo == 'y') {
	          cmd_mr_access(c_pe,select,num,1,_R_,&d1);
                  message("          = %01.1X\n",d1);
	        }
	        break;

	      case 'c':
	        cmd_vcr_access(c_pe,num,_W_,&d1);
	        if (!read_fp || c_echo == 'y') {
	          cmd_vcr_access(c_pe,num,_R_,&d1);
                  message("          = %08.8X\n",d1);
	        }
	        break;

	      case 'u':
	        cmd_vut_access(c_pe,num,_W_,&d1);
	        if (!read_fp || c_echo == 'y') {
	          cmd_vut_access(c_pe,num,_R_,&d1);
                  message("          = %08.8X\n",d1);
	        }
	        break;

	      case 't':
	        cmd_tr_access(c_pe,num,_W_,&d1);
	        if (!read_fp || c_echo == 'y') {
	          cmd_tr_access(c_pe,num,_R_,&d1);
                  message("          = %08.8X\n",d1);
	        }
	        break;
#endif
              }
	      break;

	    case 'f':
	    case 'o':
#ifdef VCP
	    case 'v':
#endif
              if (sscanf(cmdP, "%x %x", &d1, &d2) != 2) {
                skipspc();
	        if (!*cmdP) continue;
		else goto retry_writereg;
	      }
	      switch (facility) {
	      case 'f':
	        cmd_fr_access(c_pe,num,_W_,&d1,&d2);
	        if (!read_fp || c_echo == 'y') {
	          cmd_fr_access(c_pe,num,_R_,&d1,&d2);
                  message("          = %08.8X_%08.8X\n",d1,d2);
	        }
	        break;

	      case 'o':
	        cmd_tod_access(c_pe,_W_,&d1,&d2);
	        if (!read_fp || c_echo == 'y') {
	          cmd_tod_access(c_pe,_R_,&d1,&d2);
                  message("          = %08.8X_%08.8X\n",d1,d2);
	        }
	        break;

#ifdef VCP
	      case 'v':
	        cmd_vr_access(c_pe,select,num,_W_,&d1,&d2);
	        if (!read_fp || c_echo == 'y') {
	          cmd_vr_access(c_pe,select,num,_R_,&d1,&d2);
                  message("          = %08.8X_%08.8X\n",d1,d2);
	        }
	        break;
#endif
	      }
	      break;
	    }
            continue;
retry_writereg:
	    message("write reg? %s\n", cmdP);
	    num--;
            continue;
          }

	  start += count;
	  end += count;
          fclose(read_fp);
          continue;
exit_writereg:
	  message("write reg?\n");
          fclose(read_fp);
          continue;
        }
        break;

      default:
        break;
      }
      break;
    }
    return(1); /* failed */
  }
  return(0); /* succeeded */
}

command_f(cmdP, fp, log) char * cmdP; FILE *fp; int log;
      /* return 1 if failed, 0 if succeeded */
{
  unsigned once;

  for (once = 0; !once; once++) {
    switch (once) {
    case 0:
      switch (*cmdP++) {
#ifdef MCR_AREA
      case 'l': /* mcr */
#endif
      case 'm': /* scalar */
#ifdef VCP
      case 'n': /* vector */
#endif
        if (*cmdP == '/' || isspc()) { /* fill mem */
	  unsigned facility;
          static unsigned len;
          static unsigned type, dummy;
          static unsigned start, end;
          unsigned count, addr;
          unsigned d[4];

          facility = *(cmdP-1);
          if (extract_option(&cmdP, &len, &type, 0))
            continue;
          if (len  == 0) len = 4;
          if (type == 0) type = 'x';
          if (type!='d'&&type!='x'&&type!='f'&&type!='z') {
            message("illegal: type='%c'\n", type);
            continue;
          }
#ifdef MCR_AREA
          if (facility == 'l') {
            if (extract_regrange(facility, &cmdP, &dummy, &start, &end))
              continue;
          }
          else {
#endif
            if (extract_memrange(c_pe, &cmdP, len, &start, &end))
              continue;
#ifdef MCR_AREA
          }
#endif
          skipspc();
	  count = (end - start) / len + 1;
          if (*cmdP == ',') {
	    cmdP++;
	    skipspc();
	  }
	  else {
	    message(",{val} expected\n");
	    continue;
	  }

	  switch (get_format_d(cmdP, len, d, type)) {
          case 0:
	    break;

          case 1: 
          case 2:
	  case 3:
	    goto exit_fillmem;
	    break;
          }

          for (addr=start; addr<=end; addr+=len) {
	    switch (facility) {
#ifdef MCR_AREA
            case 'l':
              if (cmd_mcr_access(c_pe,addr,len,_W_,d))
                goto exit_fillmem;
              break;
#endif

	    case 'm':
              if (cmd_mem_access(c_pe,c_mode,addr,len,_W_,d,c_forcew))
	        goto exit_fillmem;
              break;

#ifdef VCP
	    case 'n':
              if (cmd_vop_access(c_pe,c_node,addr,len,_W_,d,c_forcew))
	        goto exit_fillmem;
              break;
#endif
            }
	  }

	  start += count * len;
	  end += count * len;
          continue;
exit_fillmem:
	  message("fill mem? %s\n", cmdP);
          continue;
        }
        break;

      case 'g': /* fill gr */
      case 'f': /* fill fr */
      case 's': /* fill sr */
      case 'o': /* fill tod */
#ifdef VCP
      case 'v': /* fill vr  */
      case 'k': /* fill mr  */
      case 'c': /* fill vcr */
      case 'u': /* fill vcu */
      case 't': /* fill tr  */
#endif
        if (isspc()) { /* fill reg */
          unsigned facility;
          static unsigned select, start, end;
	  unsigned count, num;
	  unsigned d1, d2;

          facility = *(cmdP-1);
#ifdef VCP
          if ((facility == 'v' || facility == 'k') && start == 0) start = end = 1;
#endif
          if (extract_regrange(facility, &cmdP, &select, &start, &end))
            continue;
          skipspc();
#ifdef VCP
          if ((facility == 'v' || facility == 'k') && start == 0) {
	    message("{index} begins from 1\n");
	    continue;
          }
#endif
          if (*cmdP == ',') {
	    cmdP++;
	    skipspc();
	  }
	  else {
	    message(",{val} expected\n");
	    continue;
	  }

          delete__(cmdP);

	  switch (facility) {
	  case 'g':
	  case 's':
#ifdef VCP
	  case 'k':
	  case 'c':
	  case 'u':
	  case 't':
#endif
            if (sscanf(cmdP, "%x", &d1) != 1)
	      goto exit_fillreg;
	    break;

	  case 'f':
	  case 'o':
#ifdef VCP
	  case 'v':
#endif
            if (sscanf(cmdP, "%x %x", &d1, &d2) != 2)
	      goto exit_fillreg;
	    break;
	  }

          switch (facility) {
          case 'o':
            start = 0;
            end   = 0;
            count = 1;
            break;

          default:
            count = end - start + 1;
            break;
          }

          for (num=start; num<=end; num++) {
            switch (facility) {
            case 'g':
              if (cmd_gr_access(c_pe,num,_W_,&d1))
                goto exit_fillreg;
              break;

            case 'f':
              if (cmd_fr_access(c_pe,num,_W_,&d1,&d2))
                goto exit_fillreg;
              break;

            case 's':
              if (cmd_sr_access(c_pe,num,_W_,&d1))
                goto exit_fillreg;
              break;

            case 'o':
              if (cmd_tod_access(c_pe,_W_,&d1,&d2))
                goto exit_fillreg;
              break;

#ifdef VCP
            case 'v':
	      if (cmd_vr_access(c_pe,select,num,_W_,&d1,&d2))
                goto exit_fillreg;
              break;

            case 'k':
              if (cmd_mr_access(c_pe,select,num,1,_W_,&d1))
                goto exit_fillreg;
              break;

            case 'c':
              if (cmd_vcr_access(c_pe,num,_W_,&d1))
                goto exit_fillreg;
              break;

            case 'u':
              if (cmd_vut_access(c_pe,num,_W_,&d1))
                goto exit_fillreg;
              break;

            case 't':
              if (cmd_tr_access(c_pe,num,_W_,&d1))
                goto exit_fillreg;
              break;
#endif
            }
          }

	  start += count;
	  end += count;
          continue;
exit_fillreg:
	  message("fill reg? %s\n", cmdP);
          continue;
        }
        break;

      default:
        break;
      }
      break;
    }
    return(1); /* failed */
  }
  return(0); /* succeeded */
}

req_system(s) char *s;
{
      scr_res();
      if (system(s) == -1)
        message("can't fork \"%s\"\n", s);
      scr_up();
      scr_setr(view_end + 2);
      view_touch();
}

sprint_symbol(t_pe, addr, buf) unsigned t_pe, addr; char *buf;
{
  char *sym;

  if (get_memname(t_pe, addr, &sym))
    sprintf(buf, "0x%08.8X", addr);
  else
    sprintf(buf, "0x%08.8X(%s)", addr, sym);
}

append_symbol(t_pe, addr) unsigned t_pe, addr;
{
  char *sym;

  if (get_memname(t_pe, addr, &sym))
    message("\n");
  else
    message("(%s)\n", sym);
}

delete_space(cmdP) char *cmdP;
{
    for (; *cmdP; cmdP++) {
      if (isspace(*cmdP)) {
        *cmdP = NULL;
        break;
      }
    }
}

delete_comment(cmdP) char *cmdP;
{
    for (; *cmdP; cmdP++) {
      if (*cmdP == '#') {
        *cmdP = NULL;
        break;
      }
    }
}

delete_nl(cmdP) char *cmdP;
{
    for (; *cmdP; cmdP++) {
      if (*cmdP == '\n') {
        *cmdP = NULL;
        break;
      }
    }
}

delete__(cmdP) char *cmdP;
{
    for (; *cmdP; cmdP++) {
      if (*cmdP == '_')
        *cmdP = ' ';
    }
}

ignore_rest_of_line(cmdP) char *cmdP;
{
    skipspc();
    if (*cmdP)
      message("ignored: %s\n", cmdP);
}

expand_command(cmd, prev, new, fp)
    char *cmd, *prev, *new;
    FILE *fp;
{
    FILE *open_file();
    static char SRC_BUF[COMMAND_LINE*2];
    static char sav_buf[COMMAND_LINE*2];
    static char tmp_buf[COMMAND_LINE*2];
    char *cmdP;

    delete_nl(cmd);
    strncpy(cmdP = SRC_BUF, cmd, COMMAND_LINE);
    skipspc();
    if (*cmdP == '!') { /* history */
      if (*(++cmdP) == '!') { /* repeat previous(-1) command */
        if (history_num) {
          sprintf(new, "%s%s", prev, cmdP+1);
          message("%s\n", new);
	}
        else
          *new = NULL;
      }
      else { /* repeat previous command */
        FILE *his_fp;
	int lineno, len, tmp;

        if (!history_fp) {
	  message("history file inactive\n");
          *new = NULL;
	  return;
        }
	if (!(his_fp = open_file(ENV_HISTORY, "r"))) {
          *new = NULL;
	  return;
        }
        *sav_buf = NULL;
	if (isalpha(*cmdP)) { /* ![a-zA-Z] */
          len = strlen(cmdP);
	  while (fgets(tmp_buf, COMMAND_LINE, his_fp)) {
	    if (!strncmp(cmdP, tmp_buf, len))
              strncpy(sav_buf, tmp_buf, COMMAND_LINE);
          }
	  if (!*sav_buf) {
	    message("can't find \"%s...\" command\n", cmdP);
            *new = NULL;
	  }
          else {
            delete_nl(sav_buf);
            sprintf(new, "%s", sav_buf);
            message("%s\n", new);
          }
	}
	else if (sscanf(cmdP, "%u", &lineno) == 1) { /* ![0-9] */
	  skipdigit();
	  for (tmp = 1; tmp <= lineno; tmp++) {
            if (!fgets(sav_buf, COMMAND_LINE, his_fp))
	      break;
          }
	  if (tmp < lineno) {
	    message("can't find %uth command\n", lineno);
            *new = NULL;
	  }
          else {
            delete_nl(sav_buf);
            sprintf(new, "%s%s", sav_buf, cmdP); 
            message("%s\n", new); 
          }
	}
	fclose (his_fp);
      }
    }
    else if (*cmdP == '^') { /* modify */
      char *sav = sav_buf;
      char *tmp = tmp_buf;
      char *dst = new;
      int len;

      *sav = *tmp = NULL;
      for (cmdP++; *cmdP;) {
        if (*cmdP == '^')
          break;
        *sav++ = *cmdP++;
        *sav   = NULL;
      }
      for (cmdP++; *cmdP;) {
        if (*cmdP == '^')
          break;
        *tmp++ = *cmdP++;
        *tmp   = NULL;
      }
      sav = sav_buf;
      tmp = tmp_buf;
      len = strlen(sav);
      for (cmdP = prev; *cmdP;) {
        if (!strncmp(sav, cmdP, len)) {
          for (; *sav; sav++)
            cmdP++;
          for (; *tmp; tmp++)
            *dst++ = *tmp;
        }
        else
          *dst++ = *cmdP++;
      }
      *dst = NULL;
      message("%s\n", new);
    }
    else { /* expand parameters (%1 - %9) */
      char *prm;
      char *dst = new;
      int i;

      for (i = 0; i < script_ptr; i++) {
        if (SCRIPT[i].script_fp == fp)
          break;
      }
      if (i < script_ptr) {
        for (; *cmdP;) {
          if (*cmdP == '%' && *(cmdP+1) >='1' && *(cmdP+1) <= '9') {
            prm = SCRIPT[i].param[*(cmdP+1)-'0'];
            cmdP += 2;
            for (; *prm; prm++)
              *dst++ = *prm;
          }
          else
            *dst++ = *cmdP++;
        }
        *dst = NULL;
      }
      else
        strncpy(new, cmdP, COMMAND_LINE);
    }
}

put_format_d(addr, len, d, type, c_echo, write_fp)
    unsigned addr, len, *d, type, c_echo;
    FILE *write_fp;
    /* return 1 if failed, 0 if succeeded */
{
    unsigned k1, k2, k3, k5, k6;
    float kf;

    switch (type) {
    case 'd':
      switch (len) {
      case 1: k1 = (char)  *d; break;
      case 2: k1 = (short) *d; break;
      case 4: k1 = (int)   *d; break;

      default:
	goto illegal_combination;
	break;
      }
      if (!write_fp || c_echo == 'y')
        message("%d\n",k1);
      if (write_fp)
        fprintf(write_fp,"%d\t# %08.8X\n",k1,addr);
      break;

    case 'x':
      switch (len) {
      case 1:
        k1 = (unsigned char) *d;
        if (!write_fp || c_echo == 'y')
          message("%02.2X\n",k1);
        if (write_fp)
          fprintf(write_fp,"%02.2X\t# %08.8X\n",k1,addr);
        break;

      case 2:
        k1 = (unsigned short) *d;
        if (!write_fp || c_echo == 'y')
          message("%04.4X\n",k1);
        if (write_fp)
          fprintf(write_fp,"%04.4X\t# %08.8X\n",k1,addr);
        break;

      case 4:
        if (!write_fp || c_echo == 'y')
          message("%08.8X\n",*d);
        if (write_fp)
          fprintf(write_fp,"%08.8X\t# %08.8X\n",*d,addr);
        break;

      case 8:
        if (!write_fp || c_echo == 'y')
          message("%08.8X_%08.8X\n",*d,*(d+1));
        if (write_fp)
          fprintf(write_fp,
            "%08.8X_%08.8X\t/* %08.8X */\n",*d,*(d+1),addr);
        break;

      case 16:
        if (!write_fp || c_echo == 'y')
          message("%08.8X_%08.8X_%08.8X_%08.8X\n",*d,*(d+1),*(d+2),*(d+3));
        if (write_fp)
          fprintf(write_fp,
            "%08.8X_%08.8X_%08.8X_%08.8X\t/* %08.8X */\n",*d,*(d+1),*(d+2),*(d+3),addr);
        break;

      default:
	goto illegal_combination;
	break;
      }
      break;

    case 'f':
      if (len != 4 && len != 8 && len != 16)
	goto illegal_combination;
      else if (tachyon_to_f(d, (len==4)?'f':(len==8)?'d':'x', &kf)) {
        if (!write_fp || c_echo == 'y')
          message("(non float)\n");
        if (write_fp)
          fprintf(write_fp,"(non float)\t# %08.8X\n",addr);
      }
      else {
        if (!write_fp || c_echo == 'y')
          message("%.20g\n",kf);
        if (write_fp)
          fprintf(write_fp,"%.20g\t# %08.8X\n",kf,addr);
      }
      break;

    case 'z':
      switch (len) {
      case 4:
        k1 = *d>>31; k2 =(*d>>23)&0xff; k3 = *d&0x7fffff;
        if (!write_fp || c_echo == 'y')
          message("%1X %02.2X %06.6X\n",k1,k2,k3);
        if (write_fp)
          fprintf(write_fp, "%1X %02.2X %06.6X\t# %08.8X\n",k1,k2,k3,addr);
        break;

      case 8:
        k1 = *d>>31; k2 =(*d>>20)&0x7ff; k3 = *d&0xfffff;
        if (!write_fp || c_echo == 'y') {
          message("%1X %03.3X %05.5X_%08.8X\n",k1,k2,k3,*(d+1));
        }
        if (write_fp)
          fprintf(write_fp, "%1X %03.3X %05.5X_%08.8X\t# %08.8X\n",k1,k2,k3,*(d+1),addr);
        break;

      case  16:
        k1 = *d>>31; k2 =(*d>>16)&0x7fff; k3 = *d&0xffff;
        k5 = *(d+2)>>31; k6 = *(d+2)&0x7fffffff; 
        if (!write_fp || c_echo == 'y') {
          message("%1X %04.4X %04.4X_%08.8X ",k1,k2,k3,*(d+1));
          message("%1X %08.8X_%08.8X\n",k5,k6,*(d+3));
        }
        if (write_fp) {
          fprintf(write_fp,"%1X %04.4X %04.4X_%08.8X ",k1,k2,k3,*(d+1));
          fprintf(write_fp,"%1X %08.8X_%08.8X\t# %08.8X\n",k5,k6,*(d+3),addr);
        }
        break;

      default:
	goto illegal_combination;
	break;
      }
      break;
    }
    return(0);

illegal_combination:
    message("illegal combination: len=%u, type='%c'\n", len, type);
    return(1);
}

get_format_d(cmdP, len, d, type)
    char *cmdP;
    unsigned len, *d, type;
    /* return 3 if failed (exit),        2 if failed (retry) */
    /*        1 if continue (no change), 0 if succeeded      */
{
    unsigned k1, k2, k3, k5, k6;
    float kf;

    delete__(cmdP);
    skipspc();
    if (!*cmdP) return(1);

    switch (type) {
    case 'd':
      if (sscanf(cmdP, "%d", &k1) != 1) return(2);
      switch (len) {
      case 1: *d = (char)  k1; break;
      case 2: *d = (short) k1; break;
      case 4: *d = (int)   k1; break;

      default:
	goto illegal_combination;
	break;
      }
      break;

    case 'x':
      switch (len) {
      case 1:
        if (sscanf(cmdP, "%x", &k1) != 1) return(2);
        *d = (unsigned char) k1;
        break;

      case 2:
        if (sscanf(cmdP, "%x", &k1) != 1) return(2);
        *d = (unsigned short) k1;
        break;

      case 4:
        if (sscanf(cmdP, "%x", d) != 1) return(2);
        break;

      case 8:
        if (sscanf(cmdP, "%x %x", d, d+1) != 2) return(2);
        break;

      case 16:
        if (sscanf(cmdP, "%x %x %x %x", d, d+1, d+2, d+3) != 4) return(2);
        break;

      default:
	goto illegal_combination;
	break;
      }
      break;

    case 'f':
      if (len != 4 && len != 8 && len != 16)
	goto illegal_combination;
      if (sscanf(cmdP, "%e", &kf) != 1) return(2);
      if (f_to_tachyon(&kf, (len==4)?'f':(len==8)?'d':'x', d)) return(2);
      break;

    case 'z':
      switch (len) {
      case 4:
        if (sscanf(cmdP, "%x %x %x", &k1, &k2, &k3) != 3) return(2);
        *d = k1<<31 | (k2<<23&0x7f800000) | k3&0x7fffff;
        break;

      case 8:
        if (sscanf(cmdP, "%x %x %x %x", &k1, &k2, &k3, d+1) != 4) return(2);
        *d = k1<<31 | (k2<<20&0x7ff00000) | k3&0xfffff;
        break;

      case 16:
        if (sscanf(cmdP, "%x %x %x %x %x %x %x", &k1,&k2,&k3,d+1,&k5,&k6,d+3) != 7) return(2);
        *d     = k1<<31 | (k2<<16&0x7fff0000) | k3&0xffff;
        *(d+2) = k5<<31 | k6&0x7fffffff;
        break;

      default:
	goto illegal_combination;
	break;
      }
      break;
    }
    return(0);

illegal_combination:
    message("illegal combination: len=%u, type='%c'\n", len, type);
    return(3);
}

extract_option(cmdP, len, type, multiple)
  char     **cmdP;
  unsigned *len, *type;
  unsigned multiple; /* 0: return type with single char */
                     /* 1: return type with bit position (multiple type) */
  /* return 1 if failed, 0 if succeeded */
{
  int i;

  if (**cmdP != '/') 
    return(0);
  (*cmdP)++;
  if (isdigit(**cmdP))
    *len =             *(*cmdP)++ - '0';
  for (i=0; i<10; i++) {
    if (isdigit(**cmdP))
      *len = *len * 10 + *(*cmdP)++ - '0';
  }

  if (multiple) {
    *type = 0x80000000;
    for (i=0; i<10; i++) {
      if (isalpha(**cmdP))
        *type |= optbit(*(*cmdP)++);
    }
  } 
  else {
    if (isalpha(**cmdP))
      *type  = *(*cmdP)++;
  }

  for (; **cmdP && isspace (**cmdP); (*cmdP)++);

  return(0);
}

extract_regrange(facility, cmdP, select, start, end)
  unsigned facility; /* 'g': GR,  'f': FR,  's': SR,  'o': TOD */
#ifdef MCR_AREA
                     /* 'l': MCR */
#endif
#ifdef VCP
                     /* 'v': VR,  'k': MR,  'c': VCR */
                     /* 'u': VUT, 't': TR            */
#endif
  char     **cmdP;
  unsigned *select, *start, *end;
  /* return 1 if failed, 0 if succeeded */
{
  unsigned val;
  char *i, c, save, delim;
 
  for (; **cmdP && isspace (**cmdP); (*cmdP)++);

  if (**cmdP != '-' && **cmdP != '@') {
    i = *cmdP;
    if (!isalnum(*i))
      return(0);
    for (;isalnum(*i); i++);
#ifdef VCP
    if (*i == '.') { /* found select */
      save = *i;
      *i = NULL;
      if      (facility == 'v' && (val = get_vrnumber(*cmdP)) < VR_NUM)
        *select = val;
      else if (facility == 'k' && (val = get_mrnumber(*cmdP)) < MR_NUM)
        *select = val;
      else if (sscanf(*cmdP, "%u", &val) == 1)
        *select = val;
      else {
        message("%s not number\n", *cmdP);
        *i = save;
        return(1);
      }
      *i = save;
      *cmdP = ++i;

      for (; **cmdP && isspace (**cmdP); (*cmdP)++);

      if (!isalnum(*i)) {
        if (facility == 'v' || facility == 'k')
          *start = *end = 1;
        else
          *start = *end = 0;
        goto skip_regrange;
      }
      for (;isalnum(*i); i++);
    }
#endif
    save = *i;
    *i = NULL;
    if      (facility == 'g' && (val = get_grnumber(*cmdP)) < GR_NUM)
      *start = val;
    else if (facility == 'f' && (val = get_frnumber(*cmdP)) < FR_NUM)
      *start = val;
    else if (facility == 's' && (val = get_srnumber(*cmdP)) < SR_NUM)
      *start = val;
    else if (facility == 'o')
      *start = val;
#ifdef MCR_AREA
    else if (facility == 'l' && (val = get_mcrnumber(*cmdP)) < MCR_NUM)
      *start = val * 8;
#endif
#ifdef VCP
    else if (facility == 'c' && (val = get_vcrnumber(*cmdP)) < VCR_NUM)
      *start = val;
    else if (facility == 'u' && (val = get_vutnumber(*cmdP)) < VUT_NUM)
      *start = val;
    else if (facility == 't' && (val = get_trnumber(*cmdP)) < TR_NUM)
      *start = val;
#endif
#ifdef MCR_AREA
    else if (facility == 'l' && sscanf(*cmdP, "%x%c", &val, &c) == 1)
      *start = val;
#endif
    else if (facility != 'l' && sscanf(*cmdP, "%u", &val) == 1)
      *start = val;
    else {
      message("%s not number\n", *cmdP);
      *i = save;
      return(1);
    }
    *i = save;
    *cmdP = i;

    for (; **cmdP && isspace (**cmdP); (*cmdP)++);

skip_regrange:

    if (**cmdP != '-' && **cmdP != '@') {
      *end = *start;
      return(0);
    }
  }

  delim = *((*cmdP)++);

  for (; **cmdP && isspace (**cmdP); (*cmdP)++);

  i = *cmdP;
  if (!isalnum(*i)) {
    message("illegal end number\n");
    *end = *start;
    return(1);
  }
  for (;isalnum(*i); i++);
  save = *i;
  *i = NULL;
  if (delim == '-') {
    if      (facility == 'g' && (val = get_grnumber(*cmdP)) < GR_NUM)
      *end = val;
    else if (facility == 'f' && (val = get_frnumber(*cmdP)) < FR_NUM)
      *end = val;
    else if (facility == 's' && (val = get_srnumber(*cmdP)) < SR_NUM)
      *end = val;
    else if (facility == 'o')
      *end = val;
#ifdef MCR_AREA
    else if (facility == 'l' && (val = get_mcrnumber(*cmdP)) < MCR_NUM)
      *end = val * 8;
#endif
#ifdef VCP
    else if (facility == 'c' && (val = get_vcrnumber(*cmdP)) < VCR_NUM)
      *end = val;
    else if (facility == 'u' && (val = get_vutnumber(*cmdP)) < VUT_NUM)
      *end = val;
    else if (facility == 't' && (val = get_trnumber(*cmdP)) < TR_NUM)
      *end = val;
#endif
#ifdef MCR_AREA
    else if (facility == 'l' && sscanf(*cmdP, "%x%c", &val, &c) == 1)
      *end = val;
#endif
    else if (facility != 'l' && sscanf(*cmdP, "%u", &val) == 1)
      *end = val;
    else {
      message("%s not number\n", *cmdP);
      *end = *start;
      *i = save;
      return(1);
    }
  }
  else { /* case '@' */
    if (sscanf(*cmdP, "%u", &val) == 1 && val > 0) {
#ifdef MCR_AREA
      if (facility == 'l')
        *end = *start + (val - 1) * 4;
      else
#endif
        *end = *start + val - 1;
    }
    else {
      message("%s not count\n", *cmdP);
      *end = *start;
      *i = save;
      return(1);
    }
  }

  *i = save;
  *cmdP = i;

  for (; **cmdP && isspace (**cmdP); (*cmdP)++);

  return(0);
}

extract_memrange(t_pe, cmdP, len, start, end)
  unsigned t_pe;
  char     **cmdP;
  unsigned len, *start, *end;
  /* return 1 if failed, 0 if succeeded */
{
  unsigned val;
  char *i, c, save, delim;

  for (; **cmdP && isspace (**cmdP); (*cmdP)++);

  if (**cmdP != '-' && **cmdP != '@') {
    i = *cmdP;
    if (*i!='$'&&*i!='_'&&*i!='.'&&!isalnum(*i))
      return(0);
    for (;*i=='$'||*i=='_'||*i=='.'||isalnum(*i); i++);
    save = *i;
    *i = NULL;
    if ((val = get_grnumber(*cmdP)) < GR_NUM)
      cmd_gr_access(t_pe,val,_R_,start);
    else if ((val = get_srnumber(*cmdP)) < SR_NUM)
      cmd_sr_access(t_pe,val,_R_,start);
    else if (!get_memnumber(t_pe, *cmdP, &val))
      *start = val;
    else if (sscanf(*cmdP, "%x%c", &val, &c) == 1)
      *start = val;
    else {
      message("%s not address\n", *cmdP);
      *i = save;
      return(1);
    }
    *i = save;
    *cmdP = i;

    for (; **cmdP && isspace (**cmdP); (*cmdP)++);

    if (**cmdP != '-' && **cmdP != '@') {
      *end = *start;
      return(0);
    }
  }

  delim = *((*cmdP)++);

  for (; **cmdP && isspace (**cmdP); (*cmdP)++);

  i = *cmdP;
  if (*i!='$'&&*i!='_'&&*i!='.'&&!isalnum(*i)) {
    message("illegal end address\n");
    *end = *start;
    return(1);
  }
  for (;*i=='$'||*i=='_'||*i=='.'||isalnum(*i); i++);
  save = *i;
  *i = NULL;
  if (delim == '-') {
    if ((val = get_grnumber(*cmdP)) < GR_NUM)
      cmd_gr_access(t_pe,val,_R_,end);
    else if ((val = get_srnumber(*cmdP)) < SR_NUM)
      cmd_gr_access(t_pe,val,_R_,end);
    else if (!get_memnumber(t_pe, *cmdP, &val))
      *end = val;
    else if (sscanf(*cmdP, "%x%c", &val, &c) == 1)
      *end = val;
    else {
      message("%s not address\n", *cmdP);
      *end = *start;
      *i = save;
      return(1);
    }
  }
  else { /* case '@' */
    if (sscanf(*cmdP, "%u", &val) == 1 && val > 0)
      *end = *start + (val - 1) * len;
    else {
      message("%s not byte count\n", *cmdP);
      *end = *start;
      *i = save;
      return(1);
    }    
  }

  *i = save;
  *cmdP = i;

  for (; **cmdP && isspace (**cmdP); (*cmdP)++);

  return(0);
}

onintr_cmd()
{
  message("!\n");
  longjmp(sjbuf,1);
}

onintr_sys()
{
  /* if (!batchmode) fflush(stdout); */
  signal(SIGINT, SIG_IGN);
  interrupt++;
}

poll_interruption()
{
  return(interrupt);
}

#ifdef CONSOLE

static unsigned	svpmesp;
static unsigned char svpmes[COMMAND_LINE];

onintr_con1()
{
  unsigned t_pe;

  signal(SIGUSR1, SIG_IGN);

  for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
    if (c_coint > 0 && c_coint < IR_NUM)
      pe[T_PE].ext_interrupt |= ((unsigned)0x80000000 >> c_coint);
  }

  signal(SIGUSR1, onintr_con1);
}

onintr_con2()
{
  unsigned      i, t_pe, hwat, svfw, d;
  unsigned      usvfwbuf[SVFWBYTE*4];
  unsigned char *svfwbuf = (unsigned char*) usvfwbuf;

  signal(SIGUSR2, SIG_IGN);

  scr_bot();

  if (!coin_fd) {
    lmessage("lost connection for console->simulator\n");
    goto exit_onintr_con2;
  }

  *svfwbuf = NULL;

  while (read(coin_fd, svfwbuf, 1) == 1) {
    if (*svfwbuf == '%')
      break;
  }

  if (*svfwbuf != '%' || read(coin_fd, svfwbuf, 1) != 1
   || *svfwbuf != 'R' || read(coin_fd, svfwbuf, 1) != 1) {
    lmessage("lost %%R preamble for console->simulator\n");
    goto exit_onintr_con2;
  }

  switch (*svfwbuf) {
  case '2': /* svpfirm-monitor */
    if (read(coin_fd, svfwbuf, SVFWBYTE) != SVFWBYTE) {
      lmessage("insufficient SVP-FIRM connection\n");
      goto exit_onintr_con2;
    }

    clearsvpmes();

    for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
      if (svp_mem_access(t_pe,HSATBASE,4,_R_,&d)
       || d != ('H'<<24)+('S'<<16)+('A'<<8)+'T')
        addsvpmes("PE#%u: can't find HSAT (in HSA)\n", t_pe);
      else if (svp_mem_access(t_pe,HSATBASE+0x10,4,_R_,&d)
       || d != ('H'<<24)+('W'<<16)+('A'<<8)+'T'
       || svp_mem_access(t_pe,HSATBASE+0x18,4,_R_,&hwat)
       || svp_mem_access(t_pe,hwat,4,_R_,&d)
       || d != ('H'<<24)+('W'<<16)+('A'<<8)+'T')
        addsvpmes("PE#%u: can't find HWAT (in HSA)\n", t_pe);
      else if (svp_mem_access(t_pe,hwat+0x20,4,_R_,&d)
       || d != ('S'<<24)+('V'<<16)+('F'<<8)+'W')
        addsvpmes("PE#%u: can't find SVFW (in HSA)\n", t_pe);
      else if (svp_mem_access(t_pe,hwat+0x2c,4,_R_,&d)
       || d != SVFWBYTE)
        addsvpmes("PE#%u: SVFW length must be %d (in HSA)\n", t_pe, SVFWBYTE);
      else if (svp_mem_access(t_pe,hwat+0x28,4,_R_,&svfw))
        addsvpmes("PE#%u: can't find SVP-FIRM request block (in HSA)\n", t_pe);
      else {
        for (i=0; i<SVFWBYTE; i+=4) {
          if (svp_mem_access(t_pe,hwat+svfw+i,4,_W_,&usvfwbuf[i/4])) {
            addsvpmes("PE#%u: can't write SVP-FIRM request block (in HSA)\n", t_pe);
            break;
          }
        }
      }
      pe[T_PE].mcr[SVFWRW].svfwrw.svfwrp = 1;
    }

    prtsvpmes();
    break;

#ifdef DAPF
  case '4': /* tmpsapc-monitor(attention) */
    for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
      pe[T_PE].dapf_sapc_pending = 4;
    }
    break;

  case '6': /* tmpsapc-monitor(read buffer) */
    pe[0].dapf_sapc_pending = 6;
    if (read(coin_fd, pe[0].dapf_sapc, SAPCFRAME) != SAPCFRAME) {
      lmessage("insufficient SAPC connection\n");
      goto exit_onintr_con2;
    }
    for (t_pe = 1; t_pe < PE_NUM; t_pe++) {
      pe[T_PE].dapf_sapc_pending = 6;
      for (i=0; i<SAPCFRAME; i++)
        pe[T_PE].dapf_sapc[i] = pe[0].dapf_sapc[i];
    }
    break;
#endif

  }

exit_onintr_con2:
  signal(SIGUSR2, onintr_con2);
}

clearsvpmes()
{
  svpmesp = 0;
  svpmes[0] = 0;
}

addsvpmes(s, i, j, k, l, m, n, o)
  char *s; int i, j, k, l, m, n, o;
{
  int a;

  sprintf(tmp_buf, s, i, j, k, l, m, n, o);
  lmessage("%s", tmp_buf);
  for (a=0; tmp_buf[a] && svpmesp<COMMAND_LINE; a++, svpmesp++) {
    if (tmp_buf[a] == '\n') {
      svpmes[svpmesp] = '\r';
      if (svpmesp+1<COMMAND_LINE)
        svpmes[++svpmesp] = '\n';
    }
    else
      svpmes[svpmesp] = tmp_buf[a];
  }
}

prtsvpmes()
{
  if (svpmes[0] && cout_fd) {
    write(cout_fd, "%stat-leading-prtmess", 21);
    write(cout_fd, svpmes, svpmesp);
    write(cout_fd, "&", 1);
  }
}

#endif

extern int has_no_fixed_region;

view_touch()
{
  int i;

  for (i = 0; i <= view_end; i++)
    *VIEW_LINE[i].prv = '*';
}

view(t_logview)
  unsigned t_logview;
{
  char *get_srname();
#ifdef MCR_AREA
  char *get_mcrname();
#endif
  int i, j, touched;
  unsigned pe, select, num, mode;
  unsigned d[2];
  char *bufP;
  char *sym;

  for (i = 0; i <= view_end; i++) {
    *(bufP = VIEW_LINE[i].new) = NULL;
    switch (VIEW_LINE[i].compress) {
    case 'x': /* expanded format */
      for (j = 0; j < 4; j++) {
        pe     = VIEW_LINE[i].pe[j];
        mode   = VIEW_LINE[i].mode[j];
        select = VIEW_LINE[i].select[j];
        num    = VIEW_LINE[i].num[j];
        switch (VIEW_LINE[i].facility[j]) {
#ifdef MCR_AREA
        case 'l':
          if (cmd_mcr_access(pe, num, 4, _R_, d))
            sprintf(bufP, " %1u:%-7s =--------", pe, get_mcrname(num));
          else
            sprintf(bufP, " %1u:%-7s =%08.8X", pe, get_mcrname(num), *d);
          for (; *bufP; bufP++);
          break;
#endif

        case 'm':
          if (c_symbol == 'n' || get_memname(pe, num, &sym))
            sprintf(bufP, " %1u:%08.8X", pe, num);
          else
            sprintf(bufP, " %1u:%-8s", pe, sym);
          for (; *bufP; bufP++);
          if (cmd_mem_access(pe, mode, num, 4, _R_, d, 'n'))
            sprintf(bufP, "=--------");
          else
            sprintf(bufP, "=%08.8X", *d);
          for (; *bufP; bufP++);
          break;

#ifdef VCP
        case 'n':
          if (c_symbol == 'n' || get_memname(pe, num, &sym))
            sprintf(bufP, " %1u:%08.8X", pe, num);
          else
            sprintf(bufP, " %1u:%-8s", pe, sym);
          for (; *bufP; bufP++);
          if (cmd_vop_access(pe, mode, num, 4, _R_, d, 'n'))
            sprintf(bufP, "=--------");
          else
            sprintf(bufP, "=%08.8X", *d);
          for (; *bufP; bufP++);
          break;
#endif

        case 'g':
          if (cmd_gr_access(pe, num, _R_, d))
            sprintf(bufP, " %1u:gr%-3u   =--------", pe, num);
          else 
            sprintf(bufP, " %1u:gr%-3u   =%08.8X", pe, num, *d);
          for (; *bufP; bufP++);
          break;

        case 'f':
          if (cmd_fr_access(pe, num, _R_, d, d+1))
            sprintf(bufP, " %1u:fr%-3u              =--------_--------", pe, num);
          else
            sprintf(bufP, " %1u:fr%-3u              =%08.8X_%08.8X", pe, num, *d, *(d+1));
          for (; *bufP; bufP++);
          break;

        case 's':
          if (cmd_sr_access(pe, num, _R_, d))
            sprintf(bufP, " %1u:%-7s =--------", pe, get_srname(num));
          else
            sprintf(bufP, " %1u:%-7s =%08.8X", pe, get_srname(num), *d);
          for (; *bufP; bufP++);
          break;

        case 'o':
          if (cmd_tod_access(pe, _R_, d, d+1))
            sprintf(bufP, " %1u:tod                =--------_--------", pe);
          else
            sprintf(bufP, " %1u:tod                =%08.8X_%08.8X", pe, *d, *(d+1));
          for (; *bufP; bufP++);
          break;

#ifdef VCP
        case 'v':
          if (cmd_vr_access(pe, select, num, _R_, d, d+1))
            sprintf(bufP, " %1u:vr%-3u.%-4u         =--------_--------", pe, select, num);
          else
            sprintf(bufP, " %1u:vr%-3u.%-4u         =%08.8X_%08.8X", pe, select, num, *d, *(d+1));
          for (; *bufP; bufP++);
          break;

        case 'k':
          if (cmd_mr_access(pe, select, num   , 32, _R_, d  )
           || cmd_mr_access(pe, select, num+32, 32, _R_, d+1))
            sprintf(bufP, " %1u:mr%-3u.%-4u         =--------_--------", pe, select, num);
          else
            sprintf(bufP, " %1u:mr%-3u.%-4u         =%08.8X_%08.8X", pe, select, num, *d, *(d+1));
          for (; *bufP; bufP++);
          break;

        case 'c':
          if (cmd_vcr_access(pe, num, _R_, d))
            sprintf(bufP, " %1u:vcr%-2u   =--------", pe, num);
          else
            sprintf(bufP, " %1u:vcr%-2u   =%08.8X", pe, num, *d);
          for (; *bufP; bufP++);
          break;

        case 'u':
          if (cmd_vut_access(pe, num, _R_, d))
            sprintf(bufP, " %1u:vut%-2u   =--------", pe, num);
	  else
            sprintf(bufP, " %1u:vut%-2u   =%08.8X", pe, num, *d);
          for (; *bufP; bufP++);
          break;

        case 't':
          if (cmd_tr_access(pe, num, _R_, d))
            sprintf(bufP, " %1u:tr%-3u   =--------", pe, num);
	  else
            sprintf(bufP, " %1u:tr%-3u   =%08.8X", pe, num, *d);
          for (; *bufP; bufP++);
          break;
#endif

        default:
          break;
        }
      }
      break;

    case 'c': /* compressed format */
      pe     = VIEW_LINE[i].pe[0];
      mode   = VIEW_LINE[i].mode[0];
      select = VIEW_LINE[i].select[0];
      num    = VIEW_LINE[i].num[0];
      switch (VIEW_LINE[i].facility[0]) {
#ifdef MCR_AREA
      case 'l':
        sprintf(bufP, " %1u:%-7s =", pe, get_mcrname(num));
        for (; *bufP; bufP++);
        for (j=0; j<8; j+=2) {
          if (cmd_mcr_access(pe,num+j,4,_R_,d) || cmd_mcr_access(pe,num+j+1,4,_R_,d+1))
            sprintf(bufP, " ----------------");
          else
            sprintf(bufP, " %08.8X%08.8X", *d, *(d+1));
          for (; *bufP; bufP++);
        }
        break;
#endif

      case 'm':
        if (c_symbol == 'n' || get_memname(pe, num, &sym))
          sprintf(bufP, " %1u:%08.8X=", pe, num);
        else
          sprintf(bufP, " %1u:%-8s=", pe, sym);
        for (; *bufP; bufP++);
        for (j=0; j<32; j+=8) {
          if (cmd_mem_access(pe,mode,num+j,4,_R_,d,'n') || cmd_mem_access(pe,mode,num+j+4,4,_R_,d+1,'n'))
            sprintf(bufP, " ----------------");
          else
            sprintf(bufP, " %08.8X%08.8X", *d, *(d+1));
          for (; *bufP; bufP++);
        }
        break;

#ifdef VCP
      case 'n':
        if (c_symbol == 'n' || get_memname(pe, num, &sym))
          sprintf(bufP, " %1u:%08.8X=", pe, num);
        else
          sprintf(bufP, " %1u:%-8s=", pe, sym);
        for (; *bufP; bufP++);
        for (j=0; j<32; j+=8) {
          if (cmd_vop_access(pe,mode,num+j,4,_R_,d,'n') || cmd_vop_access(pe,mode,num+j+4,4,_R_,d+1,'n'))
            sprintf(bufP, " ----------------");
          else
            sprintf(bufP, " %08.8X%08.8X", *d, *(d+1));
          for (; *bufP; bufP++);
        }
        break;
#endif

      case 'g':
        sprintf(bufP, " %1u:gr%-3u   =", pe, num);
        for (; *bufP; bufP++);
        for (j=0; j<8; j+=2) {
          if (cmd_gr_access(pe,num+j,_R_,d) || cmd_gr_access(pe,num+j+1,_R_,d+1))
            sprintf(bufP, " ----------------");
          else 
            sprintf(bufP, " %08.8X%08.8X", *d, *(d+1));
          for (; *bufP; bufP++);
        }
        break;

      case 'f':
        sprintf(bufP, " %1u:fr%-3u   =", pe, num);
        for (; *bufP; bufP++);
        for (j=0; j<4; j++) {
          if (cmd_fr_access(pe,num+j,_R_,d,d+1))
            sprintf(bufP, " ----------------");
          else
            sprintf(bufP, " %08.8X%08.8X", *d, *(d+1));
          for (; *bufP; bufP++);
        }
        break;

      case 's':
        sprintf(bufP, " %1u:%-7s =", pe, get_srname(num));
        for (; *bufP; bufP++);
        for (j=0; j<8; j+=2) {
          if (cmd_sr_access(pe,num+j,_R_,d) || cmd_sr_access(pe,num+j+1,_R_,d+1))
            sprintf(bufP, " ----------------");
          else
            sprintf(bufP, " %08.8X%08.8X", *d, *(d+1));
          for (; *bufP; bufP++);
        }
        break;

      case 'o':
        sprintf(bufP, " %1u:tod     =", pe);
        for (; *bufP; bufP++);
        if (cmd_tod_access(pe,_R_,d,d+1))
          sprintf(bufP, " ----------------");
        else
          sprintf(bufP, " %08.8X%08.8X", *d, *(d+1));
        for (; *bufP; bufP++);
        break;

#ifdef VCP
      case 'v':
        sprintf(bufP, "%1u:vr%-3u.%-4u", pe, select, num);
        for (; *bufP; bufP++);
        for (j=0; j<4; j++) {
          if (cmd_vr_access(pe,select,num+j,_R_,d,d+1))
            sprintf(bufP, " ----------------");
          else
            sprintf(bufP, " %08.8X%08.8X", *d, *(d+1));
          for (; *bufP; bufP++);
        }
        break;

      case 'k':
        sprintf(bufP, "%1u:mr%-3u.%-4u", pe, select, num);
        for (; *bufP; bufP++);
        for (j=0; j<4; j++) {
          if (cmd_mr_access(pe,select,num+j,1,_R_,d))
            sprintf(bufP, "                -");
          else
            sprintf(bufP, "                %1X", *d, *(d+1));
          for (; *bufP; bufP++);
        }
        break;

      case 'c':
        sprintf(bufP, " %1u:vcr%-2u   =", pe, num);
        for (; *bufP; bufP++);
        for (j=0; j<8; j+=2) {
          if (cmd_vcr_access(pe,num+j,_R_,d) || cmd_vcr_access(pe,num+j+1,_R_,d+1))
            sprintf(bufP, " ----------------");
          else
            sprintf(bufP, " %08.8X%08.8X", *d, *(d+1));
          for (; *bufP; bufP++);
        }
        break;

      case 'u':
        sprintf(bufP, " %1u:vut%-2u   =", pe, num);
        for (; *bufP; bufP++);
        for (j=0; j<8; j+=2) {
          if (cmd_vut_access(pe,num+j,_R_,d) || cmd_vut_access(pe,num+j+1,_R_,d+1))
            sprintf(bufP, " ----------------");
          else
            sprintf(bufP, " %08.8X%08.8X", *d, *(d+1));
          for (; *bufP; bufP++);
        }
        break;

      case 't':
        sprintf(bufP, " %1u:tr%-3u   =", pe, num);
        for (; *bufP; bufP++);
        for (j=0; j<8; j+=2) {
          if (cmd_tr_access(pe,num+j,_R_,d) || cmd_tr_access(pe,num+j+1,_R_,d+1))
            sprintf(bufP, " ----------------");
          else
            sprintf(bufP, " %08.8X%08.8X", *d, *(d+1));
          for (; *bufP; bufP++);
        }
        break;
#endif

      default:
        break;
      }
      break;
    }
  }

  touched = 0;
  for (i = 0; i <= view_end; i++) {
    if (t_logview == 'y' && strcmp(VIEW_LINE[i].prv, VIEW_LINE[i].new)) {
      scr_move(i+1);
      message("%-80s\n", VIEW_LINE[i].new);
      strcpy(VIEW_LINE[i].prv, VIEW_LINE[i].new);
      touched = 1;
    }
    else if (has_no_fixed_region || strcmp(VIEW_LINE[i].prv, VIEW_LINE[i].new)) {
      scr_move(i+1);
      stdoutprintf("%-80s\n", VIEW_LINE[i].new);
      strcpy(VIEW_LINE[i].prv, VIEW_LINE[i].new);
      touched = 1;
    }
  }
  if (touched)
    scr_bot();
}

struct tms      utms;
long            tmssave;

restme()
{
        times(&utms);
        tmssave = utms.tms_utime;
}

long gettme()
{
        times(&utms);

#ifdef HP9000
        return((utms.tms_utime-tmssave)*10);
#else
        return((utms.tms_utime-tmssave)*50/3);
#endif
}

