static char SccsId[] = "@(#)extended.c	1.8    4/22/93";

/* TaChyon/KAI Instruction simulator      */
/*  Copyright (c) 1989 by FUJITSU LIMITED */
/*            Primary writer: Y.Nakashima */

/* extended.c    89/12/16 */

#include "tim.h"

extended_function(t_pe)
    unsigned t_pe;
{
#define OCTA_NUM 8
    unsigned cntl_block = pe[T_PE].cntl_block;
    unsigned mode, i, d[OCTA_NUM*2];

    mode = pe[T_PE].sr[PEV].pev.va?'v':'r';
    if (cmd_mem_access(t_pe,mode,cntl_block,8,_R_,d,'n')) {
      lmessage("EXT_FUNC: bad cntl_block address 0x%08.8X\n", cntl_block);
      return;
    }

    if (d[0] >> 24) /* check insn_type */
      return;

    if (d[1] >= OCTA_NUM) {
      lmessage("EXT_FUNC: cntl_block=0x%08.8X: bad octa size %u\n", cntl_block, d[1]);
      return;
    }

    for (i=d[1]; i>0; i--) {
      if (cmd_mem_access(t_pe,mode,cntl_block+i*8,8,_R_,d+i*2,'n')) {
        lmessage("EXT_FUNC: bad cntl_block address 0x%08.8X\n", cntl_block);
        return;
      }
    }

    for (i=d[1]+1; i<OCTA_NUM; i++)
      d[i*2] = d[i*2+1] = 0;

    switch (d[0]) {

    case 1: /* HALT */
      /* +----------+----------+ 
         |0x00000001|0x00000000|
         +----------+----------+ */
      pe[T_PE].ip_status = STATUS_HALT;
      message("PE#%u halts. (EXT_FUNC1) PC=0x%08.8X\n", t_pe, pe[T_PE].sr[PC].ful);
      return;

    case 2: /* PRINTF */
      /* +----------+----------+ 
         |0x00000002|  1 - 5   |
         +----------+----------+ 
         | *format  | -------- |
         +----------+----------+ 
         | param_1  | param_2  |
         +----------+----------+ 
         |   ...    |   ...    |
         +----------+----------+ 
         | param_7  | -------- |
         +----------+----------+ */
      {
        char *format;
        unsigned real_addr;
        unsigned abs_addr;
        unsigned arg_ent, status;

        if (get_mem_real_addr(t_pe, mode, _R_, d[2], &real_addr, 'n')) {
          lmessage("EXT_FUNC2(PRINTF): cntl_block=0x%08.8X:\nillegal address of format 0x%08.8X\n", cntl_block, d[2]);
          return;
        }
        if (get_abs_addr(t_pe, _R_, real_addr, &abs_addr, 'n')) {
          lmessage("EXT_FUNC2(PRINTF): cntl_block=0x%08.8X:\nillegal address of format 0x%08.8X\n", cntl_block, d[2]);
          return;
        }
        arg_ent = 3;
        status = 0; /* 0:expect %, 1:expect number, 2:expect alpha, 0x10:escape */
        for (format=(char*)&pe[T_PE].mem[abs_addr]; *format; format++) {
	  if (status & 0x10)
	    status &= ~0x10;
	  else if (*format == '\\')
	    status |= 0x10;
	  else if (!status && *format == '%') {
	    status = 1;
	    if (++arg_ent > 10) {
              lmessage("EXT_FUNC2(PRINTF): cntl_block=0x%08.8X: too many arguments\n", cntl_block);
	      return;
	    }
	  }
	  else if (status == 1 && *format == '+')
	    status = 1;
	  else if (status == 1 && *format == '-')
	    status = 1;
	  else if (status == 1 && isdigit(*format))
	    status = 2;
	  else if ((status == 1 || status == 2) && isalpha(*format)) {
	    status = 0;
            if (*format == 'f' || *format == 'e' ||*format == 'g' 
              ||*format == 'E' || *format == 'G') {
	      unsigned tmp;
              if (++arg_ent > 10) {
                lmessage("EXT_FUNC2(PRINTF): cntl_block=0x%08.8X: too many arguments\n", cntl_block);
                return;
              }
	      tmp = d[arg_ent];
	      d[arg_ent] = d[arg_ent-1];
	      d[arg_ent-1] = tmp;
            }
	    else if (*format == 's') {
              if (get_mem_real_addr(t_pe, mode, _R_, d[arg_ent], &d[arg_ent], 'n')) {
                lmessage("EXT_FUNC2(PRINTF): cntl_block=0x%08.8X:\nillegal address of string 0x%08.8X\n", cntl_block, d[arg_ent]);
                return;
              }
              if (get_abs_addr(t_pe, _R_, d[arg_ent], &d[arg_ent], 'n')) {
                lmessage("EXT_FUNC2(PRINTF): cntl_block=0x%08.8X:\nillegal address of string 0x%08.8X\n", cntl_block, d[arg_ent]);
                return;
              }
	      d[arg_ent] = (unsigned)&pe[T_PE].mem[d[arg_ent]];
	    }
	  }
        }

	format = (char*)&pe[T_PE].mem[abs_addr];

        smessage(&pe[T_PE].mem[abs_addr], d[4],d[5],d[6],d[7],d[8],d[9],d[10]);
        return;
      }

    case 3: /* GET CLOCK */
      /* +----------+----------+ 
         |0x00000003|0x00000001|
         +----------+----------+ 
         | -> clock | -------- |
         +----------+----------+ */
      sys_mem_access(t_pe,mode,cntl_block+0x08,4,_W_,&tim_clock,0,0);
      /* message("EXT_FUNC3(CLOCK) clock=%u\n", tim_clock); */
      return;

    case 4: /* INTERVAL INTERRUPTION */
      /* +----------+----------+ 
         |0x00000004|0x00000001|
         +-+----+---+----------+ 
         |s| -- |num| interval |
         +-+----+---+----------+ */
#ifdef MANUAL_INTR
      if ((d[2] & 0xffff) == 0 || (d[2] & 0xffff) >= IR_NUM) {
        lmessage("EXT_FUNC4(INTERVAL): cntl_block=0x%08.8X: illegal number %u\n", cntl_block, d[2] & 0xffff);
        return;
      }
      pe[T_PE].ir_table[d[2] & 0xffff].interval = d[3];
      pe[T_PE].ir_table[d[2] & 0xffff].count    = 0;
      pe[T_PE].ir_table[d[2] & 0xffff].type     = (d[2] & 0x80000000)?'s':'c';
      /* message("EXT_FUNC4(INTERVAL) number=%u interval=%u\n", d[2] & 0xffff, d[3]); */
#else
      lmessage("EXT_FUNC4(INTERVAL): cntl_block=0x%08.8X:\nnot available. (compiled with no -DMANUAL_INTR)\n", cntl_block);
#endif
      return;

    case 5: /* SAVE ALL FACILITY */
      /* +----------+----------+ 
         |0x00000005|0x00000001|
         +----------+----------+ 
         |*filename | -------- |
         +----------+----------+ */
      {
        unsigned real_addr;
        unsigned abs_addr;

        if (get_mem_real_addr(t_pe, mode, _R_, d[2], &real_addr, 'n')) {
          lmessage("EXT_FUNC5(SAVE): cntl_block=0x%08.8X:\nillegal address of filename 0x%08.8X\n", cntl_block, d[2]);
          return;
        }
        if (get_abs_addr(t_pe, _R_, real_addr, &abs_addr, 'n')) {
          lmessage("EXT_FUNC5(SAVE): cntl_block=0x%08.8X:\nillegal address of filename 0x%08.8X\n", cntl_block, d[2]);
          return;
        }
        if (!pe[T_PE].mem[abs_addr]) {
          lmessage("EXT_FUNC5(SAVE): cntl_block=0x%08.8X:\nillegal filename \"%s\"\n", cntl_block, &pe[T_PE].mem[abs_addr]);
          return;
        }
        if (write_all(&pe[T_PE].mem[abs_addr]))
          lmessage("EXT_FUNC5(SAVE): cntl_block=0x%08.8X:\nPE#%u facilities not saved. PC=0x%08.8X filename=%s\n", cntl_block, t_pe, pe[T_PE].sr[PC].ful, &pe[T_PE].mem[abs_addr]);
        else
          message("PE#%u facilities saved. (EXT_FUNC5) PC=0x%08.8X filename=%s\n", t_pe, pe[T_PE].sr[PC].ful, &pe[T_PE].mem[abs_addr]);
        return;
      }

    case 6: /* BREAK */
      /* +----------+----------+ 
         |0x00000006|0x00000000|
         +----------+----------+ */
      bp_match = 1;
      pe[T_PE].bp_entry = EXT_FUNC_BREAK;
      message("PE#%u breaks. (EXT_FUNC6) PC=0x%08.8X\n", t_pe, pe[T_PE].sr[PC].ful);
      return;

    case 7: /* USRFUNC */
      /* +----------+----------+ 
         |0x00000007|0x00000005|
         +----------+----------+ 
         |   PE #   | -------- |
         +----------+----------+ 
         | param_1  | param_2  |
         +----------+----------+ 
         |   ...    |   ...    |
         +----------+----------+ 
         | param_7  | -------- |
         +----------+----------+ */
      {
        unsigned code;

        if (d[2] < PE_NUM) {
          if (code = user_function(d[2],d[4],d[5],d[6],d[7],d[8],d[9],d[10]))
	    lmessage("EXT_FUNC7(USRFUNC): cntl_block=0x%08.8X: failed (code=%u)\n", cntl_block, code);
        }
        else
          lmessage("EXT_FUNC7(USRFUNC): cntl_block=0x%08.8X: illegal PE# %u\n", cntl_block, d[2]);
        return;
      }

    case 8: /* CONSOLE */
      /* +----------+----------+ 
         |0x00000008|0x00000002|
         +----------+--+-+-----+ 
         |  retval  |rw|v| --- |
         +----------+--+-+-----+ 
         | *buffer  |  bytes   |
         +----------+----------+ */
      {
#ifdef CONSOLE
        unsigned real_addr;
        unsigned abs_addr;
        unsigned len;
        unsigned che_ent, i;
        unsigned retval;

        if (get_mem_real_addr(t_pe, (d[3]&0x40000000)?mode:'r', _R_, d[4], &real_addr, 'n')) {
          lmessage("EXT_FUNC8(CONSOLE): cntl_block=0x%08.8X:\nillegal address of buffer 0x%08.8X\n", cntl_block, d[4]);
          return;
        }

        len = d[5];

        if (get_abs_addr(t_pe, _R_, real_addr, &abs_addr, 'n')) {
          lmessage("EXT_FUNC8(CONSOLE): cntl_block=0x%08.8X:\nillegal start address of buffer 0x%08.8X\n", cntl_block, d[4]);
          return;
        }
        if (abs_addr+len-1 >= memsize) {
          lmessage("EXT_FUNC8(CONSOLE): cntl_block=0x%08.8X:\nillegal end address of buffer 0x%08.8X\n", cntl_block, d[4]+len-1);
          return;
        }

        if (d[3] & 0x80000000) {
          /* write */
          if (!cout_fd) {
            lmessage("EXT_FUNC8(CONSOLE): cntl_block=0x%08.8X: lost write connection\n", cntl_block);
            return;
          }
          retval = write(cout_fd, &pe[T_PE].mem[abs_addr], len);
        }
        else {
          /* read */
          if (!coin_fd) {
            lmessage("EXT_FUNC8(CONSOLE): cntl_block=0x%08.8X: lost read connection\n", cntl_block);
            return;
          }
          retval = read(coin_fd, &pe[T_PE].mem[abs_addr], len);

          /* flush scalar_data_cache */
          for(i=real_addr; i<=real_addr+len-1; i+=CACHE_BLOCK_SIZE) {
            che_ent = (i>>6) & (CACHE_NUM-1);
            if (pe[T_PE].data_cache[che_ent].pa == (i>>15))
              pe[T_PE].data_cache[che_ent].v = 0;
          }
          che_ent = ((real_addr+len-1)>>6) & (CACHE_NUM-1);
          if (pe[T_PE].data_cache[che_ent].pa == ((real_addr+len-1)>>15))
            pe[T_PE].data_cache[che_ent].v = 0;
        }

        sys_mem_access(t_pe,mode,cntl_block+8,4,_W_,&retval,0,0);
#else
        lmessage("EXT_FUNC8(CONSOLE): cntl_block=0x%08.8X:\nnot available. (compiled with no -DCONSOLE)\n", cntl_block);
#endif
        return;
      }

    case 9: /* DASD */
      /* +----------+----------+ 
         |0x00000009|0x00000002|
         +----------+--+-+-+---+ 
         |  retval  |rw|v| |blk|
         +----------+--+-+-+---+ 
         | *buffer  |  blocks  |
         +----------+----------+ */
      {
        unsigned real_addr;
        unsigned abs_addr;
	unsigned block_addr;
        unsigned len;
        unsigned che_ent, i;
        unsigned retval;

        if (get_mem_real_addr(t_pe, (d[3]&0x40000000)?mode:'r', _R_, d[4], &real_addr, 'n')) {
          lmessage("EXT_FUNC9(DASD): cntl_block=0x%08.8X:\nillegal address of buffer 0x%08.8X\n", cntl_block, d[4]);
          return;
        }

        block_addr = d[3] & 0xffffff;
        len = d[5]*DASD_BLOCK_SIZE;

        if (block_addr >= dasd_blocks) {
          lmessage("EXT_FUNC9(DASD): cntl_block=0x%08.8X:\nillegal start block address 0x%08.8X\n", cntl_block, block_addr);
          return;
        }
        if (block_addr+d[5]-1 >= dasd_blocks) {
          lmessage("EXT_FUNC9(DASD): cntl_block=0x%08.8X:\nillegal end block address 0x%08.8X\n", cntl_block, block_addr+d[5]-1);
          return;
        }

        if (get_abs_addr(t_pe, _R_, real_addr, &abs_addr, 'n')) {
          lmessage("EXT_FUNC9(DASD): cntl_block=0x%08.8X:\nillegal start address of buffer 0x%08.8X\n", cntl_block, d[4]);
          return;
        }
        if (abs_addr+len-1 >= memsize) {
          lmessage("EXT_FUNC9(DASD): cntl_block=0x%08.8X:\nillegal end address of buffer 0x%08.8X\n", cntl_block, d[4]+len-1);
          return;
        }

        if (d[3] & 0x80000000) {
          /* write */
          if (!dasd_fd) {
            lmessage("EXT_FUNC9(DASD): cntl_block=0x%08.8X: lost write connection\n", cntl_block);
            return;
          }
          if (lseek(dasd_fd, block_addr*DASD_BLOCK_SIZE, L_SET) == -1) {
            lmessage("EXT_FUNC9(DASD): cntl_block=0x%08.8X: seek error\n", cntl_block);
            return;
          }
          retval = write(dasd_fd, &pe[T_PE].mem[abs_addr], len);
        }
        else {
          /* read */
          if (!dasd_fd) {
            lmessage("EXT_FUNC9(DASD): cntl_block=0x%08.8X: lost read connection\n", cntl_block);
            return;
          }
          if (lseek(dasd_fd, block_addr*DASD_BLOCK_SIZE, L_SET) == -1) {
            lmessage("EXT_FUNC9(DASD): cntl_block=0x%08.8X: seek error\n", cntl_block);
            return;
          }
          retval = read(dasd_fd, &pe[T_PE].mem[abs_addr], len);

          /* flush scalar_data_cache */
          for(i=real_addr; i<=real_addr+len-1; i+=CACHE_BLOCK_SIZE) {
            che_ent = (i>>6) & (CACHE_NUM-1);
            if (pe[T_PE].data_cache[che_ent].pa == (i>>15))
              pe[T_PE].data_cache[che_ent].v = 0;
          }
          che_ent = ((real_addr+len-1)>>6) & (CACHE_NUM-1);
          if (pe[T_PE].data_cache[che_ent].pa == ((real_addr+len-1)>>15))
            pe[T_PE].data_cache[che_ent].v = 0;
        }

        sys_mem_access(t_pe,mode,cntl_block+8,4,_W_,&retval,0,0);
        return;
      }

    case 10: /* SVPFIRM */
      /* +----------+----------+ 
         |0x0000000A|0x00000002|
         +----------+----------+ 
         |  retval  | -------- |
         +----------+----------+ 
         | *buffer  | -------- |
         +----------+----------+ */
      {
#ifdef CONSOLE
        unsigned abs_addr;
        unsigned len;
        unsigned che_ent, i;
        unsigned retval;

        if (get_abs_addr(t_pe, _R_, d[4], &abs_addr, 's')) {
          lmessage("EXT_FUNC10(SVPFIRM): cntl_block=0x%08.8X:\nillegal start address of buffer 0x%08.8X\n", cntl_block, d[4]);
          return;
        }
        if (abs_addr+SVFWBYTE-1 >= memsize) {
          lmessage("EXT_FUNC10(SVPFIRM): cntl_block=0x%08.8X:\nillegal end address of buffer 0x%08.8X\n", cntl_block, d[4]+SVFWBYTE-1);
          return;
        }

        /* write */
        if (!cout_fd) {
          lmessage("EXT_FUNC10(SVPFIRM): cntl_block=0x%08.8X: lost write connection\n", cntl_block);
          return;
        }
        write(cout_fd, "%stat-leading-prtsvfw", 21);
        retval = write(cout_fd, &pe[T_PE].mem[abs_addr], SVFWBYTE);

        sys_mem_access(t_pe,mode,cntl_block+8,4,_W_,&retval,0,0);
#else
        lmessage("EXT_FUNC10(SVPFIRM): cntl_block=0x%08.8X:\nnot available. (compiled with no -DCONSOLE)\n", cntl_block);
#endif
        return;
      }

    case 11: /* SCANF */
      /* +----------+----------+ 
         |0x0000000B|  1 - 5   |
         +----------+----------+ 
         | *format  | -------- |
         +----------+----------+ 
         | param_1  | param_2  |
         +----------+----------+ 
         |   ...    |   ...    |
         +----------+----------+ 
         | param_7  | -------- |
         +----------+----------+ */
      {
        char *format;
        unsigned real_addr;
        unsigned abs_addr;
        unsigned arg_ent, status;
        unsigned che_ent;

        if (get_mem_real_addr(t_pe, mode, _R_, d[2], &real_addr, 'n')) {
          lmessage("EXT_FUNC11(SCANF): cntl_block=0x%08.8X:\nillegal address of format 0x%08.8X\n", cntl_block, d[2]);
          return;
        }
        if (get_abs_addr(t_pe, _R_, real_addr, &abs_addr, 'n')) {
          lmessage("EXT_FUNC11(SCANF): cntl_block=0x%08.8X:\nillegal address of format 0x%08.8X\n", cntl_block, d[2]);
          return;
        }
        arg_ent = 3;
        status = 0; /* 0:expect %, 1:expect number, 2:expect alpha, 0x10:escape */
        for (format=(char*)&pe[T_PE].mem[abs_addr]; *format; format++) {
	  if (status & 0x10)
	    status &= ~0x10;
	  else if (*format == '\\')
	    status |= 0x10;
	  else if (!status && *format == '%') {
	    status = 1;
	    if (++arg_ent > 10) {
              lmessage("EXT_FUNC11(SCANF): cntl_block=0x%08.8X: too many arguments\n", cntl_block);
	      return;
	    }
	  }
	  else if (status == 1 && *format == '+')
	    status = 1;
	  else if (status == 1 && *format == '-')
	    status = 1;
	  else if (status == 1 && isdigit(*format))
	    status = 2;
	  else if ((status == 1 || status == 2) && isalpha(*format)) {
	    status = 0;
            if (get_mem_real_addr(t_pe, mode, _R_, d[arg_ent], &d[arg_ent], 'n')) {
              lmessage("EXT_FUNC11(SCANF): cntl_block=0x%08.8X:\nillegal address of string 0x%08.8X\n", cntl_block, d[arg_ent]);
              return;
            }
            if (get_abs_addr(t_pe, _R_, d[arg_ent], &d[arg_ent], 'n')) {
              lmessage("EXT_FUNC11(SCANF): cntl_block=0x%08.8X:\nillegal address of string 0x%08.8X\n", cntl_block, d[arg_ent]);
              return;
            }
	    d[arg_ent] = (unsigned)&pe[T_PE].mem[d[arg_ent]];
	  }
        }
        fscanf(command_fp, &pe[T_PE].mem[abs_addr], d[4],d[5],d[6],d[7],d[8],d[9],d[10]);

	/* flush all scalar_data_cache */
        for(che_ent=0; che_ent<CACHE_NUM; che_ent++)
          pe[T_PE].data_cache[che_ent].v = 0;

        return;
      }

    case 12: /* SPRINTF */
      /* +----------+----------+ 
         |0x0000000C|  1 - 5   |
         +----------+----------+ 
         | *format  | *buffer  |
         +----------+----------+ 
         | param_1  | param_2  |
         +----------+----------+ 
         |   ...    |   ...    |
         +----------+----------+ 
         | param_7  | -------- |
         +----------+----------+ */
      {
        char *format;
        char *buffer;
        unsigned real_addr;
        unsigned abs_addr;
        unsigned arg_ent, status;
        unsigned che_ent;

        if (get_mem_real_addr(t_pe, mode, _R_, d[2], &real_addr, 'n')) {
          lmessage("EXT_FUNC12(SPRINTF): cntl_block=0x%08.8X:\nillegal address of format 0x%08.8X\n", cntl_block, d[2]);
          return;
        }
        if (get_abs_addr(t_pe, _R_, real_addr, &abs_addr, 'n')) {
          lmessage("EXT_FUNC12(SPRINTF): cntl_block=0x%08.8X:\nillegal address of format 0x%08.8X\n", cntl_block, d[2]);
          return;
        }
        arg_ent = 3;
        status = 0; /* 0:expect %, 1:expect number, 2:expect alpha, 0x10:escape */
        for (format=(char*)&pe[T_PE].mem[abs_addr]; *format; format++) {
	  if (status & 0x10)
	    status &= ~0x10;
	  else if (*format == '\\')
	    status |= 0x10;
	  else if (!status && *format == '%') {
	    status = 1;
	    if (++arg_ent > 10) {
              lmessage("EXT_FUNC12(SPRINTF): cntl_block=0x%08.8X: too many arguments\n", cntl_block);
	      return;
	    }
	  }
	  else if (status == 1 && *format == '+')
	    status = 1;
	  else if (status == 1 && *format == '-')
	    status = 1;
	  else if (status == 1 && isdigit(*format))
	    status = 2;
	  else if ((status == 1 || status == 2) && isalpha(*format)) {
	    status = 0;
            if (*format == 'f' || *format == 'e' ||*format == 'g' 
              ||*format == 'E' || *format == 'G') {
	      unsigned tmp;
              if (++arg_ent > 10) {
                lmessage("EXT_FUNC12(SPRINTF): cntl_block=0x%08.8X: too many arguments\n", cntl_block);
                return;
              }
	      tmp = d[arg_ent];
	      d[arg_ent] = d[arg_ent-1];
	      d[arg_ent-1] = tmp;
            }
	    else if (*format == 's') {
              if (get_mem_real_addr(t_pe, mode, _R_, d[arg_ent], &d[arg_ent], 'n')) {
                lmessage("EXT_FUNC12(SPRINTF): cntl_block=0x%08.8X:\nillegal address of string 0x%08.8X\n", cntl_block, d[arg_ent]);
                return;
              }
              if (get_abs_addr(t_pe, _R_, d[arg_ent], &d[arg_ent], 'n')) {
                lmessage("EXT_FUNC12(SPRINTF): cntl_block=0x%08.8X:\nillegal address of string 0x%08.8X\n", cntl_block, d[arg_ent]);
                return;
              }
	      d[arg_ent] = (unsigned)&pe[T_PE].mem[d[arg_ent]];
	    }
	  }
        }

	format = (char*)&pe[T_PE].mem[abs_addr];

        if (get_mem_real_addr(t_pe, mode, _R_, d[3], &real_addr, 'n')) {
          lmessage("EXT_FUNC12(SPRINTF): cntl_block=0x%08.8X:\nillegal address of buffer 0x%08.8X\n", cntl_block, d[3]);
          return;
        }
        if (get_abs_addr(t_pe, _R_, real_addr, &abs_addr, 'n')) {
          lmessage("EXT_FUNC12(SPRINTF): cntl_block=0x%08.8X:\nillegal address of buffer 0x%08.8X\n", cntl_block, d[3]);
          return;
        }

	buffer = (char*)&pe[T_PE].mem[abs_addr];

        sprintf(buffer, format, d[4],d[5],d[6],d[7],d[8],d[9],d[10]);

	/* flush all scalar_data_cache */
        for(che_ent=0; che_ent<CACHE_NUM; che_ent++)
          pe[T_PE].data_cache[che_ent].v = 0;

        return;
      }

    default:
      lmessage("EXT_FUNC: cntl_block=0x%08.8X: EXT_FUNC%u not supported\n", cntl_block, d[0]);
      return;
    }
}

