static char SccsId[] = "@(#)gdb-command.c	1.4    5/13/92";

/* TaChyon/KAI Instruction simulator      */
/*  Copyright (c) 1989 by FUJITSU LIMITED */
/*            Primary writer: Y.Nakashima */

/* gdb-command.c    90/3/12 */

#ifdef GDB

#include "tim.h"

#define tohex(c) ((c) < 10 ? '0'+(c) : 'a'+(c)-10)
#define              GDB_PARAM_LINE     (256)
static unsigned      dummy; /* dummy for alignment */
static char          GDB_PARAM_BUF[GDB_PARAM_LINE];

/* This must be big enough to hold ALL registers */
static unsigned char GDB_BUFFER[PAGE_SIZE];

gdb_command(fp, cmdP, c_pe) 
    FILE *fp;
    char *cmdP;
    unsigned c_pe;
{
    unsigned char *pch;

    switch (*cmdP++) {
    case 'g': /* read registers */
    {
      unsigned addr, len, i, d1, d2;
      
      if (sscanf(cmdP, "%x,%x", &addr, &len) != 2)
        gdbrepl("E01\n"); /* format error occurred */
      else if (addr       >= GR_NUM+FR_NUM+SR_NUM
            || addr+len-1 >= GR_NUM+FR_NUM+SR_NUM)
        gdbrepl("E02\n"); /* address range error occurred */
      else {
	pch = GDB_BUFFER;
        for (i = addr; i < addr+len; i++) {
          if (i < GR_NUM) {
            cmd_gr_access(c_pe, i, _R_, &d1);
	    *pch++ = tohex((d1 >> 28) & 0xf);
	    *pch++ = tohex((d1 >> 24) & 0xf);
	    *pch++ = tohex((d1 >> 20) & 0xf);
	    *pch++ = tohex((d1 >> 16) & 0xf);
	    *pch++ = tohex((d1 >> 12) & 0xf);
	    *pch++ = tohex((d1 >> 8) & 0xf);
	    *pch++ = tohex((d1 >> 4) & 0xf);
	    *pch++ = tohex((d1 >> 0) & 0xf);
          }
          else if (i < GR_NUM+FR_NUM) {
            cmd_fr_access(c_pe, i-GR_NUM, _R_, &d1, &d2);
	    *pch++ = tohex((d1 >> 28) & 0xf);
	    *pch++ = tohex((d1 >> 24) & 0xf);
	    *pch++ = tohex((d1 >> 20) & 0xf);
	    *pch++ = tohex((d1 >> 16) & 0xf);
	    *pch++ = tohex((d1 >> 12) & 0xf);
	    *pch++ = tohex((d1 >> 8) & 0xf);
	    *pch++ = tohex((d1 >> 4) & 0xf);
	    *pch++ = tohex((d1 >> 0) & 0xf);
	    *pch++ = tohex((d2 >> 28) & 0xf);
	    *pch++ = tohex((d2 >> 24) & 0xf);
	    *pch++ = tohex((d2 >> 20) & 0xf);
	    *pch++ = tohex((d2 >> 16) & 0xf);
	    *pch++ = tohex((d2 >> 12) & 0xf);
	    *pch++ = tohex((d2 >> 8) & 0xf);
	    *pch++ = tohex((d2 >> 4) & 0xf);
	    *pch++ = tohex((d2 >> 0) & 0xf);
          }
          else {
            cmd_sr_access(c_pe, i-GR_NUM-FR_NUM, _R_, &d1);
	    *pch++ = tohex((d1 >> 28) & 0xf);
	    *pch++ = tohex((d1 >> 24) & 0xf);
	    *pch++ = tohex((d1 >> 20) & 0xf);
	    *pch++ = tohex((d1 >> 16) & 0xf);
	    *pch++ = tohex((d1 >> 12) & 0xf);
	    *pch++ = tohex((d1 >> 8) & 0xf);
	    *pch++ = tohex((d1 >> 4) & 0xf);
	    *pch++ = tohex((d1 >> 0) & 0xf);
          }
        }
        *pch++ = '\n';
        *pch++ = '\0';
	gdbrepl(GDB_BUFFER);
      }
      break;
    }

    case 'G': /* write registers */
    {
      unsigned bufp, addr, len, i;

      if (sscanf(cmdP, "%x,%x:", &addr, &len) != 2)
        gdbrepl("E03\n"); /* format error occurred */
      else if (addr       >= GR_NUM+FR_NUM+SR_NUM
            || addr+len-1 >= GR_NUM+FR_NUM+SR_NUM)
        gdbrepl("E04\n"); /* address range error occurred */
      else {
        for (; *cmdP && *cmdP != ':'; cmdP++);
        cmdP++;
        for (bufp=0, i=addr; i<addr+len; i++) {
          if (i < GR_NUM) {
            if (gdb_extract_bytes(fp, &cmdP, 4, (unsigned*)&GDB_BUFFER[bufp])) {
              gdbrepl("E05\n");
              return;
            }
            bufp += 4;
          }
          else if (i < GR_NUM+FR_NUM) {
            if (gdb_extract_bytes(fp, &cmdP, 4, (unsigned*)&GDB_BUFFER[bufp  ])
              ||gdb_extract_bytes(fp, &cmdP, 4, (unsigned*)&GDB_BUFFER[bufp+4])) {
              gdbrepl("E06\n");
              return;
            }
            bufp += 8;
          }
          else {
            if (gdb_extract_bytes(fp, &cmdP, 4, (unsigned*)&GDB_BUFFER[bufp])) {
              gdbrepl("E07\n");
              return;
            }
            bufp += 4;
          }
        }
        for (bufp=0, i=addr; i<addr+len; i++) {
          if (i < GR_NUM) {
            cmd_gr_access(c_pe, i, _W_, (unsigned*)&GDB_BUFFER[bufp]);
            bufp += 4;
          }
          else if (i < GR_NUM+FR_NUM) {
            cmd_fr_access(c_pe, i-GR_NUM, _W_, (unsigned*)&GDB_BUFFER[bufp], (unsigned*)&GDB_BUFFER[bufp+4]);
            bufp += 8;
          }
          else {
            cmd_sr_access(c_pe, i-GR_NUM-FR_NUM, _W_, (unsigned*)&GDB_BUFFER[bufp]);
            bufp += 4;
          }
        }
        gdbrepl("OK\n");
      }
      break;
    }

    case 'm': /* read memory */
    {
      unsigned addr, len, i, d, mode;

      pch = GDB_BUFFER;
      if (sscanf(cmdP, "%x,%x", &addr, &len) != 2)
        gdbrepl("E08\n"); /* format error occurred */
      else if (len > PAGE_SIZE/2)
        gdbrepl("E09\n"); /* address range error occurred */
      else {
        mode = pe[c_pe].sr[PEV].pev.va?'v':'r';
        for (i=addr; i<addr+len; i++) {
          if (cmd_mem_access(c_pe, mode, i, 1, _R_, &d, 'n')) {
            gdbrepl("E10\n");
            return;
          }
          *pch++ = tohex ((d >> 4) & 0xf);
          *pch++ = tohex (d & 0xf);

        }
        *pch++ = '\n';
        *pch++ = '\0';
        gdbrepl(GDB_BUFFER);
      }
      break;
    }

    case 'M': /* write memory */
    {
      unsigned bufp, addr, len, i, d, mode;
      
      if (sscanf(cmdP, "%x,%x:", &addr, &len) != 2)
        gdbrepl("E11\n"); /* format error occurred */
      else if (len > PAGE_SIZE)
        gdbrepl("E12\n"); /* address range error occurred */
      else {
        for (; *cmdP && *cmdP != ':'; cmdP++);
        cmdP++;
        for (bufp=0, i=addr; i<addr+len; i++) {
          if (gdb_extract_bytes(fp, &cmdP, 1, &d)) {
            gdbrepl("E13\n");
            return;
          }
          GDB_BUFFER[bufp++] = d;
        }
        mode = pe[c_pe].sr[PEV].pev.va?'v':'r';
        for (bufp=0, i=addr; i<addr+len; i++) {
          d = GDB_BUFFER[bufp++];
          if (cmd_mem_access(c_pe, mode, i, 1, _W_, &d, 'n')) {
            gdbrepl("E14\n");
            return;
          }
        }
        gdbrepl("OK\n");
      }
      break;
    }

    case 'c': /* cont */
    case 's': /* step */
    {
      unsigned cs, addr, signal, count, t_pe, tmp;
      
      cs = *(cmdP-1);
      if      (cs == 'c' && sscanf(cmdP, "%x,%x", &addr, &signal) != 2)
        gdbrepl("E15\n"); /* format error occurred */
      else if (cs == 's' && sscanf(cmdP, "%x,%x:%x", &addr, &signal, &count) != 3)
        gdbrepl("E16\n"); /* format error occurred */
      else {
        gdbrepl("OK\n");
        if (addr != 0xffffffff)
          pe[c_pe].sr[PC].ful = addr;
        scan_on = 0;
        bp_match = 0;
        cp_match = 0;
        for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
          pe[T_PE].cntl_block = -1;
          pe[T_PE].bp_entry   = -1;
          pe[T_PE].cp_entry   = -1;
        } 
        for (tmp = 0; cs == 'c' || tmp < count; tmp++) {
          for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
            if (pe[T_PE].ip_status != STATUS_HALT) {
              sim_pe(t_pe);
              if (pe[T_PE].cntl_block != -1)
                serialize(-1, 0);
            }
          }   
          tim_clock++;
          if (all_halts())
            goto run_halt;
          if (bp_match) {
            serialize(-1, 0);
            goto run_break;
          }
        }   
        if (cs == 's') {
          serialize(-1, 0);
          if (all_halts())
            goto run_halt;
          sprintf(GDB_BUFFER, "T%x\n", pe[c_pe].sr[PC].ful);
	  gdbrepl(GDB_BUFFER);
        }
      }   
      break;

run_break:
      if (bp_match) {
        for (t_pe = 0; t_pe < PE_NUM; t_pe++) {
          if (pe[T_PE].bp_entry >= 0) {
            sprintf(GDB_BUFFER, "B%x\n", pe[T_PE].sr[PC].ful);
	    gdbrepl(GDB_BUFFER);
          }
        }  
      }  
      break;
run_halt:
      /* Is this right ?????? the expression for the pc ?? */
      sprintf(GDB_BUFFER, "S00,%x\n", pe[c_pe].sr[PC].ful);
      gdbrepl(GDB_BUFFER);
      break;
    }  

    case 'B': /* breakpoint set */
    {
      unsigned addr, id, mode, real_addr, abs_addr;

      mode = pe[c_pe].sr[PEV].pev.va?'v':'r';
      if (sscanf(cmdP, "%x,%x", &addr, &id) != 2)
        gdbrepl("E17\n"); /* format error occurred */
      else if (id >= BP_NUM)
        gdbrepl("E18\n"); /* too large id */
      else if (pe[c_pe].bp_table[id].access)
        gdbrepl("E19\n"); /* already used id */
      else if (get_mem_real_addr(c_pe, mode, _R_, addr, &real_addr, 'n'))
        gdbrepl("E20\n"); /* real_addr not exist */
      else if (get_abs_addr(c_pe, _R_, real_addr, &abs_addr, 'n'))
        gdbrepl("E21\n"); /* abs_addr not exist */
      else {
        pe[c_pe].bp_table[id].interval  = 1;
        pe[c_pe].bp_table[id].count     = 0;
        pe[c_pe].bp_table[id].access    = _X_;
        pe[c_pe].bp_table[id].b_or_c    = 'b';
        pe[c_pe].bp_table[id].address   = addr;
        pe[c_pe].bp_table[id].real_addr = real_addr;
        pe[c_pe].bp_table[id].abs_addr  = abs_addr;
        pe[c_pe].bp_table[id].mode      = mode;
        pe[c_pe].bp_table[id].command   = "";
        pe[c_pe].mbp[abs_addr/8] |= (0x0080 >> (abs_addr & 7));
        gdbrepl("OK\n");
      }  
      break;
    }

    case 'b': /* breakpoint clear */
    {
      unsigned addr, id, abs_addr, tmp;

      if (sscanf(cmdP, "%x,%x", &addr, &id) != 2)
        gdbrepl("E22\n"); /* format error occurred */
      else if (id >= BP_NUM)
        gdbrepl("E23\n"); /* too large id */
      else if (!pe[c_pe].bp_table[id].access)
        gdbrepl("E24\n"); /* not set id */
      else {
        pe[c_pe].bp_table[id].access = 0;
        abs_addr = pe[c_pe].bp_table[id].abs_addr;
        pe[c_pe].mbp[abs_addr/8] &= 0xff7f >> (abs_addr & 7);
        for (tmp = 0; tmp < BP_NUM; tmp++) {
          if (pe[c_pe].bp_table[tmp].access && pe[c_pe].bp_table[tmp].abs_addr == abs_addr)
            pe[c_pe].mbp[abs_addr/8] |= 0x0080 >> (abs_addr & 7);
        }
        gdbrepl("OK\n");
      }
      break;
    }

    case 'P': /* breakpoint display */
    {
      unsigned id, tmp, found;

      if (sscanf(cmdP, "%x", &id) != 1)
        gdbrepl("E25\n"); /* format error occurred */
      else if (id != 0xffffffff && id >= BP_NUM)
        gdbrepl("E26\n"); /* too large id */
      else {
        if (id != 0xffffffff) {
          if (pe[c_pe].bp_table[id].access) {
            sprintf(GDB_BUFFER, "%02.2x:%08.8x\n",
                                 id, pe[c_pe].bp_table[id].address);
	    gdbrepl(GDB_BUFFER);
	  }
          else
            gdbrepl("-1:0\n");
        }
        else {
          for (pch = GDB_BUFFER, found = 0, tmp = 0; tmp < BP_NUM; tmp++) {
            if (pe[c_pe].bp_table[tmp].access) {
              if (found++) {
		*pch++ = ',';
		*pch++ = '\0';
              }
              sprintf(pch, "%x:x\0", tmp, pe[c_pe].bp_table[tmp].address);
              pch = (unsigned char *)index(pch, '\0');
            }
          }
          if (found) {
            gdbrepl(GDB_BUFFER);
            gdbrepl("\n");
	  }
          else
            gdbrepl("-1:0\n");
        }  
      }
      break;
    }

    case 'a': /* attach */
    {
      unsigned code, mode, tmp, abs_addr, a_text, a_data, a_bss, a_entry;
      char *tmp_index;

      mode = pe[c_pe].sr[PEV].pev.va?'v':'r';
      if (tmp_index = rindex(cmdP, '#'))
        *tmp_index = '\0';
      if (sscanf(cmdP, "%x,%s", &code, GDB_BUFFER) != 2)
        gdbrepl("E27\n"); /* format error occurred */
      else if (code)
        gdbrepl("E28\n"); /* only 0 is supported */
      else {
        cmd_mem_all_free(c_pe);
        if (cmd_mem_map(c_pe, 0, _R_|_W_|_X_, 0, memsize-1, &abs_addr))
          gdbrepl("E29\n"); /* memory mapping failed */
        else {
          if (!read_module(c_pe, mode, 0, GDB_BUFFER, 0, 0, 'n'))
            gdbrepl("E30\n"); /* read error occurred */
          else {
            access_header(c_pe, 'r', &a_text, &a_data, &a_bss, &a_entry);
	    pe[c_pe].sr[PEV].ful = 0;
	    pe[c_pe].sr[PC].ful = a_entry;
            gdbrepl("OK\n");
            sprintf(GDB_BUFFER, "S05,%x\n", pe[c_pe].sr[PC].ful);
	    gdbrepl(GDB_BUFFER);
          }
        }
      }
      break;
    }

    default: /* do nothing */
      gdbrepl("E99\n"); /* unknown command error */
      break;
    }
}

gdb_extract_bytes(fp, cmdP, len, d)
    FILE *fp;
    char **cmdP;
    unsigned len, *d;
    /* return 1 if failed, 0 if succeeded */
{
    unsigned i, c, retval;

    retval = 0;
    for (i = 0; i < len*2; i++) {
      if (!(*(*cmdP))) {
        if (!fgets(*cmdP = GDB_PARAM_BUF, GDB_PARAM_LINE, fp)) {
          message("EOF\n");
          return(1);
        }
      }
      c = *(*cmdP)++;
      retval <<= 4;
      if      (c >= '0' && c <= '9') retval += c - '0';
      else if (c >= 'A' && c <= 'F') retval += c - 'A' + 10;
      else if (c >= 'a' && c <= 'f') retval += c - 'a' + 10;
      else return(1);
    }
    *d = retval;
    return(0);
}

#endif
