static char SccsId[] = "@(#)ip.c	1.19    8/18/93";

/* TaChyon/KAI Instruction simulator      */
/*  Copyright (c) 1989 by FUJITSU LIMITED */
/*            Primary writer: Y.Nakashima */

/* ip.c    89/10/1 */

#include "tim.h"
#include "facility.h"
#include "ip-decode.h"

sim_ip(t_pe)
    unsigned t_pe;
{
    unsigned o_score0, o_score1, o_score2;
    unsigned mode, i;

/******** get current access mode ********/

    mode = pe[T_PE].sr[PEV].pev.va?'v':'r';

/******** update hangup timer ********/

    if (++pe[T_PE].hangup_timer > HANGUP_LIMIT) {
      pe[T_PE].ip_status = STATUS_HALT;
      lmessage("PE#%u halts. (HANGUP -- illegal FQ) PC=0x%08.8X\n", t_pe, pe[T_PE].sr[PC].ful);
      return;
    }

/******** update TOD/IVT ********/

    if (++pe[T_PE].tod_5nano == 12) {
      if ((pe[T_PE].tod.l += 0x100) == 0)
        pe[T_PE].tod.u++;
#ifdef MCR_AREA
      if (!pe[T_PE].firmstate) {
        pe[T_PE].mcr[IVT].ful--;
#ifdef IVT_INTR
        if (pe[T_PE].mcr[IVT].ful>>31)
          pe[T_PE].ext_interrupt |= IVT_INTERRUPT;
#endif
      }
#endif
    }
    else if (pe[T_PE].tod_5nano >= 25) {
      pe[T_PE].tod_5nano = 0;
      if ((pe[T_PE].tod.l += 0x100) == 0)
        pe[T_PE].tod.u++;
#ifdef MCR_AREA
      if (!pe[T_PE].firmstate) {
        pe[T_PE].mcr[IVT].ful--;
#ifdef IVT_INTR
        if (pe[T_PE].mcr[IVT].ful>>31)
          pe[T_PE].ext_interrupt |= IVT_INTERRUPT;
#endif
      }
#endif
    }

#ifdef MCR_AREA
/******** update WDT ********/

    if (++pe[T_PE].wdt_5nano >= 819200) {
      pe[T_PE].wdt_5nano = 0;
      if (!pe[T_PE].firmstate) {
#ifdef WDT_INTR
        if (pe[T_PE].mcr[WDTW].wdtw.wdt == 0)
          pe[T_PE].timeout_req = 1;
#endif
        pe[T_PE].mcr[WDTW].wdtw.wdt--;
      }
    }
#endif

#ifdef MANUAL_INTR
/******** scan manual interval interruption ********/

    for (i = 1; i < IR_NUM; i++) {
      if (pe[T_PE].ir_table[i].interval) {
        if (pe[T_PE].ir_table[i].interval == ++pe[T_PE].ir_table[i].count) {
          pe[T_PE].ext_interrupt |= ((unsigned)0x80000000 >> i);
          pe[T_PE].ir_table[i].count = 0;
          if (pe[T_PE].ir_table[i].type == 's')
            pe[T_PE].ir_table[i].interval = 0;
        }
      }
    }
#endif

/******** check asyncronous ********/

    if (pe[T_PE].sr[PEV].pev.sync) { /* check asynchronous action */
      if (pe[T_PE].sr[STAR1].star1.asa) {
        if (scan_on & SCAN_STATUS) {
          scan_mfc(t_pe);
          message("         WAIT(SYNC)\n");
        }
        return;
      }
    }

/******** fetch / decode instruction ********/

    if (pe[T_PE].ip_status == STATUS_FREE) {
      struct tachyon_insn *ti;
      unsigned type, index0, index1, index2;
      unsigned dummy;

      if (!(sys_sr_check_score(t_pe) == SCORE_FREE) /* wait for ldsr */
        || (mode=='v' && sys_mem_tlb(t_pe,_X_,pe[T_PE].sr[PC].ful,&dummy) && pe[T_PE].ap_datq[0].av)) {
        if (scan_on & SCAN_STATUS) {
          scan_mfc(t_pe);
          message("         WAIT(IF)\n");
        }
        return;
      }

      if (!pe[T_PE].if_multiple_cycle) {

/* handle timeout / exception / interruption */

        /* scan EXT / EXT is delayed */
        if (pe[T_PE].ext_interrupt && pe[T_PE].sr[PEV].pev.arm) {
          for (pe[T_PE].ext_intr_level = 15; pe[T_PE].ext_intr_level > 0; pe[T_PE].ext_intr_level--) {
            if (pe[T_PE].ext_interrupt & ((unsigned)0x80000000 >> pe[T_PE].ext_intr_level))
              break;
          }
          if (pe[T_PE].ext_intr_level > 0) {
            if (!old_mode)
              pe[T_PE].ext_intr_delay = 0;
            else if (pe[T_PE].ext_intr_delay != 2) {
              pe[T_PE].ext_intr_delay = 1;
              pe[T_PE].ext_intr_level = 0;
            }
          }
          else
            pe[T_PE].ext_intr_delay = 0;
        }
        else {
          pe[T_PE].ext_intr_delay = 0;
          pe[T_PE].ext_intr_level = 0;
        }

        /* scan SLRP / SLRP is delayed */
        if (pe[T_PE].sr[SLRP].ful && pe[T_PE].sr[PEV].pev.arm) {
          for (pe[T_PE].slrp_intr_level = 15; pe[T_PE].slrp_intr_level > 0; pe[T_PE].slrp_intr_level--) {
            if (pe[T_PE].sr[SLRP].ful & ((unsigned)0x80000000 >> pe[T_PE].slrp_intr_level))
              break;
          }
          if (pe[T_PE].slrp_intr_level > 0) {
            if (!old_mode)
	      pe[T_PE].slrp_intr_delay = 0;
            else if (pe[T_PE].slrp_intr_delay != 2) {
              pe[T_PE].slrp_intr_delay = 1;
              pe[T_PE].slrp_intr_level = 0;
            }
          }
          else
            pe[T_PE].slrp_intr_delay = 0;
        }
        else {
          pe[T_PE].slrp_intr_delay = 0;
          pe[T_PE].slrp_intr_level = 0;
        }

#ifdef FXT_INTR
        /* scan FXT / FXT is delayed */
        if ((pe[T_PE].mcr[SVFWRW].svfwrw.svfwrp || pe[T_PE].mcr[FWFWRW].fwfwrw.fwfwrp)
          && pe[T_PE].mcr[FWRMW].fwrmw.fwrm && pe[T_PE].sr[PEV].pev.arm) {
          if (pe[T_PE].fxt_intr_delay != 2) {
            pe[T_PE].fxt_intr_delay = 1;
            pe[T_PE].fxt_intr_level = 0;
          }
          else
            pe[T_PE].fxt_intr_level = 1;
        }
        else {
          pe[T_PE].fxt_intr_delay = 0;
          pe[T_PE].fxt_intr_level = 0;
        }
#endif

        if (pe[T_PE].timeout_req
         || EXCEPTION_EXIST
         || pe[T_PE].svc_interrupt
         || pe[T_PE].fwc_interrupt
         || (((pe[T_PE].ext_intr_level  > pe[T_PE].sr[PEV].pev.al)
         ||   (pe[T_PE].slrp_intr_level > pe[T_PE].sr[PEV].pev.al)
#ifdef FXT_INTR
         ||   (pe[T_PE].fxt_intr_level && pe[T_PE].mcr[FWRMW].fwrmw.fwrm)
#endif
             ) && pe[T_PE].sr[PEV].pev.arm)) {
          handle_interruption(t_pe);
          return;
        }

/* fetch insn */
        sys_mem_access(t_pe,mode,pe[T_PE].sr[PC].ful,8,_X_,pe[T_PE].insn,0,0);
        /* if EXCEPTION occured, if_multiple_cycle = 0 */
      }

      if (pe[T_PE].if_multiple_cycle) {
        if (--pe[T_PE].if_multiple_cycle) {
          if (scan_on & SCAN_STATUS) {
            scan_mfc(t_pe);
            message("         WAIT(IF)\n");
          }
          return;
        }
      }

      if (scan_on & SCAN_BRANCH) {
        if (pe[T_PE].sr[PC].ful != pe[T_PE].lpc + 8) {
          scan_mfc(t_pe);
          message("         BRANCH 0x%08.8X -> 0x%08.8X ", pe[T_PE].lpc, pe[T_PE].sr[PC].ful);
          append_symbol(t_pe, pe[T_PE].sr[PC].ful);
        }
      }

/* check exception / decode */
      if (EXCEPTION_EXIST) {
        /* do nothing */
      }
      else {
        type     = pe[T_PE].insn[0] >> 28;
        pe[T_PE].mx = type_multi_in_op0[type];
        ti       = &(insn_type[type]);
        if (ti->op0 == oper_cop) {
	  index0 = ((pe[T_PE].insn[0] >> ti->sft0) & ti->mask0)
                 | ((pe[T_PE].insn[1] >> ti->sft1) & ti->mask1)
                 | ((pe[T_PE].insn[1] >> ti->sft2) & ti->mask2);
          pe[T_PE].o_type0 = (ti->op0)[index0].type;
	  pe[T_PE].o_func0 = (ti->op0)[index0].func;
#ifdef VCP
	  pe[T_PE].vinsn_stat[index0]++;
#endif
        }
        else if (ti->op0) {
	  index0 = (pe[T_PE].insn[0]>>ti->sft0)&ti->mask0;
          pe[T_PE].o_type0 = (ti->op0)[index0].type;
	  pe[T_PE].sinsn_stat[type][index0][0]++;
	  if ((pe[T_PE].o_func0 = (ti->op0)[index0].func) == NULL)
            pe[T_PE].work_ird0.opx = 1;
        }
        else {
          pe[T_PE].o_type0 = O_NL;
          pe[T_PE].o_func0 = NULL;
        }
        if (ti->op1) {
          pe[T_PE].insn_u1 = type<<28|pe[T_PE].insn[0]<<20&0x0fffffff|pe[T_PE].insn[1]>>12;
          pe[T_PE].insn_l1 = pe[T_PE].insn[1]<<20;
          index1 = (pe[T_PE].insn_u1>>ti->sft1)&ti->mask1;
          pe[T_PE].o_type1 = (ti->op1)[index1].type;
	  pe[T_PE].sinsn_stat[type][index1][1]++;
	  if ((pe[T_PE].o_func1 = (ti->op1)[index1].func) == NULL)
            pe[T_PE].work_ird0.opx = 1;
        }
        else {
	  pe[T_PE].o_type1 = O_NL;
	  pe[T_PE].o_func1 = NULL;
        }
        if (ti->op2) {
          pe[T_PE].insn_u2 = type<<28|pe[T_PE].insn[1]<<8&0x0fffff00;
          pe[T_PE].insn_l2 = 0;
          index2 = (pe[T_PE].insn_u2>>ti->sft2)&ti->mask2;
          pe[T_PE].o_type2 = (ti->op2)[index2].type;
	  pe[T_PE].sinsn_stat[type][index2][2]++;
	  if ((pe[T_PE].o_func2 = (ti->op2)[index2].func) == NULL)
            pe[T_PE].work_ird0.opx = 1;
        }
        else {
	  pe[T_PE].o_type2 = O_NL;
	  pe[T_PE].o_func2 = NULL;
        }
  
        if (pe[T_PE].o_func1 == o_brnc0 && !(pe[T_PE].insn[1] & 0x1f) && pe[T_PE].o_func0 == o_nop) {
          /* enter extended function */
          if (scan_on & SCAN_LIST) disasm(t_pe, mode, pe[T_PE].sr[PC].ful,1,1);
          pe[T_PE].sr[PC].ful = (pe[T_PE].lpc = pe[T_PE].sr[PC].ful) + 8;
	  pe[T_PE].cntl_block = ((pe[T_PE].insn[0]<<27|pe[T_PE].insn[1]>>5)<<3)+pe[T_PE].sr[PC].ful;
	  return;
        }
	
        pe[T_PE].ip_status = STATUS_R_WAIT;
      }
    }

/******** check specification exceptions / score boards (read/write) ********/

    if (pe[T_PE].ip_status == STATUS_R_WAIT) {

/* get info (read/write) */
      o_score0=(pe[T_PE].o_type0!=O_NL)?(pe[T_PE].o_func0)(t_pe,pe[T_PE].mx  ,pe[T_PE].insn[0],pe[T_PE].insn[1],'c'):SCORE_FREE;
      o_score1=(pe[T_PE].o_type1!=O_NL)?(pe[T_PE].o_func1)(t_pe,1-pe[T_PE].mx,pe[T_PE].insn_u1,pe[T_PE].insn_l1,'c'):SCORE_FREE;
      o_score2=(pe[T_PE].o_type2!=O_NL)?(pe[T_PE].o_func2)(t_pe,2            ,pe[T_PE].insn_u2,pe[T_PE].insn_l2,'c'):SCORE_FREE;

/* check exception / score board (read/write) */
      if (EXCEPTION_EXIST) {
        if (IP_EXCEPTION_EXIST) {
          if (scan_on & SCAN_LIST) disasm(t_pe, mode, pe[T_PE].sr[PC].ful,1,1);
          pe[T_PE].sr[PC].ful = (pe[T_PE].lpc = pe[T_PE].sr[PC].ful) + 8;
        }
        pe[T_PE].ip_status = STATUS_FREE;
      }
      else {
        if (o_score0==SCORE_FREE&&o_score1==SCORE_FREE&&o_score2==SCORE_FREE)
          pe[T_PE].ip_status = STATUS_R_CONT;
      }
    }

/******** READ CONTINUED ********/

    if (pe[T_PE].ip_status == STATUS_R_CONT) {
      unsigned ap_oper_exist;
      unsigned fp_oper_exist;
#ifdef VCP
      unsigned vp_oper_exist;
#endif
      unsigned fqe_ent;

      if (EXCEPTION_EXIST) {
        pe[T_PE].ip_status = STATUS_FREE;
      }
      else {
        ap_oper_exist = (pe[T_PE].o_type0==O_AL||pe[T_PE].o_type0==O_AS
                       ||pe[T_PE].o_type1==O_AL||pe[T_PE].o_type1==O_AS);
        fp_oper_exist = (pe[T_PE].o_type0==O_FH||pe[T_PE].o_type0==O_FL
                       ||pe[T_PE].o_type1==O_FH||pe[T_PE].o_type1==O_FL
                       ||pe[T_PE].o_type2==O_FH||pe[T_PE].o_type2==O_FL);
#ifdef VCP
        vp_oper_exist = (pe[T_PE].o_type0==O_VP);
#endif

        if (fp_oper_exist) {
	  for (fqe_ent = 0; fqe_ent < FQ_NUM; fqe_ent++) {
	    if (!pe[T_PE].sr[FQE0H+fqe_ent*2].fqeh.av)
	      break;
	  }
        }
        if ((!ap_oper_exist || !pe[T_PE].ap_datq[0].av)
         && (!fp_oper_exist || fqe_ent < FQ_NUM)
#ifdef VCP
         && (!vp_oper_exist || !(pe[T_PE].vp_opiq[0].av || sys_vp_check_score(t_pe) != SCORE_FREE))
#endif
         ) {

/******** enqueue ap_operation ********/
 
          if (ap_oper_exist)
            pe[T_PE].ap_datq[0].av = 2;

/******** enqueue fp_operation ********/

	  if (fp_oper_exist) {
	    *(unsigned*)&pe[T_PE].sr[FQE0H+fqe_ent*2].fqeh = 0x80000000;
	    *(unsigned*)&pe[T_PE].sr[FQE0L+fqe_ent*2].fqel = 0x00000000;
	    switch (pe[T_PE].o_type0) {
	    case O_FH:
	      *(unsigned*)&pe[T_PE].sr[FQE0H+fqe_ent*2].fqeh |= 0x40000000 | (pe[T_PE].insn[0] >> 8 & 0xfffff);
	      break;
	    case O_FL:
	      *(unsigned*)&pe[T_PE].sr[FQE0L+fqe_ent*2].fqel |= 0x40000000 | (pe[T_PE].insn[0] >> 8 & 0xfffff);
	      break;
	    }
	    switch (pe[T_PE].o_type1) {
	    case O_FH:
	      *(unsigned*)&pe[T_PE].sr[FQE0H+fqe_ent*2].fqeh |= 0x40000000 | (pe[T_PE].insn_u1 >> 8 & 0xfffff);
	      break;
	    case O_FL:
	      *(unsigned*)&pe[T_PE].sr[FQE0L+fqe_ent*2].fqel |= 0x40000000 | (pe[T_PE].insn_u1 >> 8 & 0xfffff);
	      break;
	    }
	    switch (pe[T_PE].o_type2) {
	    case O_FH:
	      *(unsigned*)&pe[T_PE].sr[FQE0H+fqe_ent*2].fqeh |= 0x40000000 | (pe[T_PE].insn_u2 >> 8 & 0xfffff);
	      break;
	    case O_FL:
	      *(unsigned*)&pe[T_PE].sr[FQE0L+fqe_ent*2].fqel |= 0x40000000 | (pe[T_PE].insn_u2 >> 8 & 0xfffff);
	      break;
	    }
          }

#ifdef VCP
/******** enqueue vp_operation ********/

          if (vp_oper_exist)
            pe[T_PE].vp_opiq[0].av = 1;
#endif
  
          pe[T_PE].ip_status = STATUS_R_READY;

        }
      }
    }

/******** execute ip/ap/fp/vp_operation ********/

    if (pe[T_PE].ip_status == STATUS_R_READY) {

      if (scan_on & SCAN_LIST) disasm(t_pe, mode, pe[T_PE].sr[PC].ful,1,1);

      if (pe[T_PE].o_type0!=O_NL) (pe[T_PE].o_func0)(t_pe,pe[T_PE].mx  ,pe[T_PE].insn[0],pe[T_PE].insn[1],'r');
      if (pe[T_PE].o_type1!=O_NL) (pe[T_PE].o_func1)(t_pe,1-pe[T_PE].mx,pe[T_PE].insn_u1,pe[T_PE].insn_l1,'r');
      if (pe[T_PE].o_type2!=O_NL) (pe[T_PE].o_func2)(t_pe,2            ,pe[T_PE].insn_u2,pe[T_PE].insn_l2,'r');

      if (pe[T_PE].ext_intr_delay == 1)
        pe[T_PE].ext_intr_delay = 2;
      if (pe[T_PE].slrp_intr_delay == 1)
        pe[T_PE].slrp_intr_delay = 2;
#ifdef FXT_INTR
      if (pe[T_PE].fxt_intr_delay == 1)
        pe[T_PE].fxt_intr_delay = 2;
#endif

      if (pe[T_PE].nullify)
        pe[T_PE].nullify = 0;
      else
        pe[T_PE].sr[PC].ful = (pe[T_PE].lpc = pe[T_PE].sr[PC].ful) + 8;

      pe[T_PE].ip_status    = STATUS_FREE;
      pe[T_PE].hangup_timer = 0;

      if (pe[T_PE].mx) {
        if (pe[T_PE].o_type0==O_IP) (pe[T_PE].o_func0)(t_pe,1,pe[T_PE].insn[0],pe[T_PE].insn[1],'e');
        if (pe[T_PE].o_type1==O_IP) (pe[T_PE].o_func1)(t_pe,0,pe[T_PE].insn_u1,pe[T_PE].insn_l1,'x');
        if (pe[T_PE].o_type0==O_IP) (pe[T_PE].o_func0)(t_pe,1,pe[T_PE].insn[0],pe[T_PE].insn[1],'x');
      }
      else {
        if (pe[T_PE].o_type1==O_IP) (pe[T_PE].o_func1)(t_pe,1,pe[T_PE].insn_u1,pe[T_PE].insn_l1,'e');
        if (pe[T_PE].o_type0==O_IP) (pe[T_PE].o_func0)(t_pe,0,pe[T_PE].insn[0],pe[T_PE].insn[1],'x');
        if (pe[T_PE].o_type1==O_IP) (pe[T_PE].o_func1)(t_pe,1,pe[T_PE].insn_u1,pe[T_PE].insn_l1,'x');
      }
      if (pe[T_PE].o_type2==O_IP) (pe[T_PE].o_func2)(t_pe,2,pe[T_PE].insn_u2,pe[T_PE].insn_l2,'x');
    }    

/******** print status *******/

    if (scan_on & SCAN_STATUS) {
      switch (pe[T_PE].ip_status) {
      case STATUS_R_WAIT:
        scan_mfc(t_pe);
        message("         WAIT(RW)\n");
        break;
      case STATUS_R_CONT:
        scan_mfc(t_pe);
        message("         WAIT(QU)\n");
        break;
      }
    }
}

handle_interruption(t_pe)
        unsigned t_pe;
{
        unsigned i, t_fsba;

        do {
	  sim_ap(t_pe);
	  sim_fp(t_pe, 1);
#ifdef VCP
	  sim_vp(t_pe);
#endif
#ifdef MOVER
	  sim_mv(t_pe); /* busy status is not reflected on star1.asa */
#endif
#ifdef BARRIER
	  sim_bp(t_pe); /* busy status is not reflected on star1.asa */
#endif
#ifdef SSUMOVER
	  sim_ssumv(t_pe); /* busy status is not reflected on star1.asa */
#endif
#ifdef GSIGP
	  sim_gsigp(t_pe); /* busy status is not reflected on star1.asa */
#endif
#ifdef FSBA
          for (t_fsba=0; t_fsba<FSBA_NUM; t_fsba++)
	    sim_fsba(t_pe, t_fsba); /* busy status is not reflected on star1.asa */
#endif
#ifdef DAPF
	  sim_dapf(t_pe); /* busy status is not reflected on star1.asa */
#endif
          pe[T_PE].sr[STAR1].star1.asa = pe[T_PE].sr[STAR1].star1.asam
                                       | pe[T_PE].sr[STAR1].star1.asaf
                                       | pe[T_PE].sr[STAR1].star1.cbs;
          if (scan_on & SCAN_STATUS) {
            scan_mfc(t_pe);
            message("         WAIT(SYNC)\n");
          }
        } while (pe[T_PE].sr[STAR1].star1.asa);
  
        pe[T_PE].fq_top = 0; /* reset fq_top */
        for (i = 0; i < FQ_NUM; i++ ){
          if(pe[T_PE].sr[FQE0H + i*2].fqeh.av ){
            pe[T_PE].work_ird0.fqr = 1;
            break;
          }
        }

     /* sys_sr_unlock(t_pe);         */
     /* for (i = 0; i < GR_NUM; i++) */
     /*   sys_gr_unlock(t_pe, i);    */
        for (i = 0; i < FR_NUM; i++) {
	  if (pe[T_PE].fr_score[i] == SCORE_EXCEPTION)
            pe[T_PE].fr_score[i] = SCORE_FREE;
	}

        pe[T_PE].if_multiple_cycle = 0;
        pe[T_PE].ip_multiple_cycle = 0;
  
        if (pe[T_PE].timeout_req) {
          /* timeout request */
          pe[T_PE].ip_status = STATUS_HALT;
          lmessage("PE#%u halts. (TIMEOUT) PC=0x%08.8X\n", t_pe, pe[T_PE].sr[PC].ful);
        }
        else if (EXCEPTION_EXIST) {
          /* program interruption */
	  if (pe[T_PE].svc_interrupt) {
            pe[T_PE].sr[PC].ful -= 8;
	    pe[T_PE].svc_interrupt = 0;
	  }
	  else if (pe[T_PE].fwc_interrupt) {
            pe[T_PE].sr[PC].ful -= 8;
	    pe[T_PE].fwc_interrupt = 0;
	  }
          pe[T_PE].sr[OPEV].ful       = pe[T_PE].sr[PEV].ful;
          pe[T_PE].sr[OPC].ful        = pe[T_PE].sr[PC].ful;
          pe[T_PE].sr[LPC].ful        = pe[T_PE].lpc;
	  pe[T_PE].sr[PEV].pev.arm    = 0;
	  pe[T_PE].sr[PEV].pev.u      = 0;
	  pe[T_PE].sr[PEV].pev.spu    = 0;
          pe[T_PE].sr[PEV].pev.fstop |= pe[T_PE].work_ird0.fqr;
	  if (pe[T_PE].work_ird0.opx + pe[T_PE].work_ird0.pox + pe[T_PE].work_ird0.spx + pe[T_PE].work_ird0.ivx
	    + pe[T_PE].work_ird0.cux + pe[T_PE].work_ird0.cir + pe[T_PE].work_ird0.fax + pe[T_PE].work_ird0.fir
	    + pe[T_PE].work_ird0.nlax+ pe[T_PE].work_ird0.pax > 1)
            pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x100;
	  else if (pe[T_PE].work_ird0.opx)
            pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x110;
	  else if (pe[T_PE].work_ird0.pox)
            pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x120;
	  else if (pe[T_PE].work_ird0.spx)
            pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x130;
	  else if (pe[T_PE].work_ird0.ivx)
            pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x140;
	  else if (pe[T_PE].work_ird0.cux)
            pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x150;
	  else if (pe[T_PE].work_ird0.cir) {
	    if (pe[T_PE].work_ird1.crr0  + pe[T_PE].work_ird1.crr1
              + pe[T_PE].work_ird1.crr2  + pe[T_PE].work_ird1.crr3
              + pe[T_PE].work_ird1.crr4  + pe[T_PE].work_ird1.crr5
	      + pe[T_PE].work_ird1.crr6  + pe[T_PE].work_ird1.crr7
              + pe[T_PE].work_ird1.crr8  + pe[T_PE].work_ird1.crr9
              + pe[T_PE].work_ird1.crr10 + pe[T_PE].work_ird1.crr11
	      + pe[T_PE].work_ird1.crr12 + pe[T_PE].work_ird1.crr13
              + pe[T_PE].work_ird1.crr14 + pe[T_PE].work_ird1.crr15 > 1)
              pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x100;
	    else {
	      for (i = 0; i < 16; i++) {
	        if (*(unsigned*)&pe[T_PE].work_ird1 << i & 0x8000) {
		  pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x300 + 0x10*i;
		  break;
	        }
	      }
	    }
	  }
	  else if (pe[T_PE].work_ird0.fax)
            pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x160;
	  else if (pe[T_PE].work_ird0.fir)
            pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x170;
	  else if (pe[T_PE].work_ird0.nlax)
            pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x200 + 0x10*pe[T_PE].sr[CLAXC0].laxc.lirc;
	  else if (pe[T_PE].work_ird0.pax) {
            if ((pe[T_PE].work_ird1.paxk0 > 0) + (pe[T_PE].work_ird1.paxk1 > 0)
              + (pe[T_PE].work_ird1.paxk2 > 0) + (pe[T_PE].work_ird1.paxk3 > 0)
              + (pe[T_PE].work_ird1.paxk4 > 0) + (pe[T_PE].work_ird1.paxk5 > 0)
              + (pe[T_PE].work_ird1.paxk6 > 0) + (pe[T_PE].work_ird1.paxk7 > 0) > 1)
              pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x100;
            else
              pe[T_PE].sr[PC].ful = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x280;
          }
          pe[T_PE].sr[IRD0].ful = *(unsigned*)&pe[T_PE].work_ird0;
          pe[T_PE].sr[IRD1].ful = *(unsigned*)&pe[T_PE].work_ird1;
          *(unsigned*)&pe[T_PE].work_ird0 = *(unsigned*)&pe[T_PE].work_ird1 = 0;
          if (scan_on & SCAN_INTR) {
            scan_mfc(t_pe);
            message("         INTR(PRG) VEC=0x%04.4X IRD0=0x%08.8X IRD1=0x%08.8X PEV=0x%08.8X PID=0x%04.4X TOD=0x%08.8X_%08.8X\n",
	      pe[T_PE].sr[PC].ful - (pe[T_PE].sr[IREB].ireb.ireb << 12),
              pe[T_PE].sr[IRD0].ful, pe[T_PE].sr[IRD1].ful, pe[T_PE].sr[PEV].ful, pe[T_PE].sr[PID].pid.pid,
	      pe[T_PE].tod.u, pe[T_PE].tod.l);
          }
	  if (!pe[T_PE].sr[OPEV].pev.arm) {
            pe[T_PE].ip_status = STATUS_HALT;
	    lmessage("PE#%u halts. (OPEV.ARM=0) OPC=0x%08.8X PC=0x%08.8X\n", t_pe, pe[T_PE].sr[OPC].ful, pe[T_PE].sr[PC].ful);
	  }
        }
        else if (pe[T_PE].svc_interrupt) {
          /* supervisor call interruption */
          pe[T_PE].sr[OPEV].ful     = pe[T_PE].sr[PEV].ful;
          pe[T_PE].sr[OPC].ful      = pe[T_PE].sr[PC].ful;
          pe[T_PE].sr[LPC].ful      = pe[T_PE].lpc;
          pe[T_PE].sr[PEV].pev.arm  = 0;
	  pe[T_PE].sr[PEV].pev.u    = 0;
	  pe[T_PE].sr[PEV].pev.spu  = 0;
	  pe[T_PE].sr[PC].ful       = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x800
			       + 0x10*(pe[T_PE].svc_interrupt & 0xff);
	  pe[T_PE].svc_interrupt = 0;
          if (scan_on & SCAN_INTR) {
            scan_mfc(t_pe);
            message("         INTR(SVC) VEC=0x%04.4X IRD0=0x%08.8X IRD1=0x%08.8X PEV=0x%08.8X PID=0x%04.4X TOD=0x%08.8X_%08.8X\n",
	      pe[T_PE].sr[PC].ful - (pe[T_PE].sr[IREB].ireb.ireb << 12),
              pe[T_PE].sr[IRD0].ful, pe[T_PE].sr[IRD1].ful, pe[T_PE].sr[PEV].ful, pe[T_PE].sr[PID].pid.pid,
	      pe[T_PE].tod.u, pe[T_PE].tod.l);
          }
	  if (!pe[T_PE].sr[OPEV].pev.arm) {
            pe[T_PE].ip_status = STATUS_HALT;
	    lmessage("PE#%u halts. (OPEV.ARM=0) OPC=0x%08.8X PC=0x%08.8X\n", t_pe, pe[T_PE].sr[OPC].ful, pe[T_PE].sr[PC].ful);
	  }
        }
        else if (pe[T_PE].fwc_interrupt) {
          /* firmware call interruption */
          pe[T_PE].sr[OPEV].ful     = pe[T_PE].sr[PEV].ful;
          pe[T_PE].sr[OPC].ful      = pe[T_PE].sr[PC].ful;
          pe[T_PE].sr[LPC].ful      = pe[T_PE].lpc;
          pe[T_PE].sr[PEV].pev.arm  = 0;
	  pe[T_PE].sr[PEV].pev.u    = 0;
	  pe[T_PE].sr[PEV].pev.va   = 0;
	  pe[T_PE].sr[PEV].pev.spu  = 0;
	  pe[T_PE].sr[PC].ful       = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x800
			       + 0x10*(pe[T_PE].fwc_interrupt & 0xff);
	  pe[T_PE].fwc_interrupt = 0;
	  pe[T_PE].firmstate = 1;
          if (scan_on & SCAN_INTR) {
            scan_mfc(t_pe);
            message("         INTR(FWC) VEC=0x%04.4X IRD0=0x%08.8X IRD1=0x%08.8X PEV=0x%08.8X PID=0x%04.4X TOD=0x%08.8X_%08.8X\n",
	      pe[T_PE].sr[PC].ful - (pe[T_PE].sr[IREB].ireb.ireb << 12),
              pe[T_PE].sr[IRD0].ful, pe[T_PE].sr[IRD1].ful, pe[T_PE].sr[PEV].ful, pe[T_PE].sr[PID].pid.pid,
	      pe[T_PE].tod.u, pe[T_PE].tod.l);
          }
	  if (!pe[T_PE].sr[OPEV].pev.arm) {
            pe[T_PE].ip_status = STATUS_HALT;
	    lmessage("PE#%u halts. (OPEV.ARM=0) OPC=0x%08.8X PC=0x%08.8X\n", t_pe, pe[T_PE].sr[OPC].ful, pe[T_PE].sr[PC].ful);
	  }
        }
	else if (pe[T_PE].ext_intr_level > pe[T_PE].sr[PEV].pev.al || pe[T_PE].slrp_intr_level > pe[T_PE].sr[PEV].pev.al) {
          /* re-calculate new level */
          for (pe[T_PE].ext_intr_level = 15; pe[T_PE].ext_intr_level > 0; pe[T_PE].ext_intr_level--) {
            if (pe[T_PE].ext_interrupt & ((unsigned)0x80000000 >> pe[T_PE].ext_intr_level))
              break;
          }
          for (pe[T_PE].slrp_intr_level = 15; pe[T_PE].slrp_intr_level > 0; pe[T_PE].slrp_intr_level--) {
            if (pe[T_PE].sr[SLRP].ful & ((unsigned)0x80000000 >> pe[T_PE].slrp_intr_level))
              break;
          }

          pe[T_PE].sr[OPEV].ful      = pe[T_PE].sr[PEV].ful;
          pe[T_PE].sr[OPC].ful       = pe[T_PE].sr[PC].ful;
          pe[T_PE].sr[LPC].ful       = pe[T_PE].lpc;
	  pe[T_PE].sr[PEV].pev.arm   = 0;
	  pe[T_PE].sr[PEV].pev.u     = 0;
	  pe[T_PE].sr[PEV].pev.spu   = 0;
	  if (pe[T_PE].ext_intr_level >= pe[T_PE].slrp_intr_level) { /* take ext_interrupt */
	    pe[T_PE].sr[PEV].pev.al  = pe[T_PE].ext_intr_level;
            pe[T_PE].sr[PC].ful      = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x400 + 0x10*pe[T_PE].ext_intr_level;
	    pe[T_PE].ext_interrupt  &= ~((unsigned)0x80000000 >> pe[T_PE].ext_intr_level);
            pe[T_PE].ext_intr_delay  = 0;
	  }
	  else { /* take soft level interrupt */
	    pe[T_PE].sr[PEV].pev.al  = pe[T_PE].slrp_intr_level;
            pe[T_PE].sr[PC].ful      = (pe[T_PE].sr[IREB].ireb.ireb << 12) + 0x500 + 0x10*pe[T_PE].slrp_intr_level;
            pe[T_PE].slrp_intr_delay = 0;
	  }
          if (scan_on & SCAN_INTR) {
            scan_mfc(t_pe);
            message("         INTR(LVL) VEC=0x%04.4X IRD0=0x%08.8X IRD1=0x%08.8X PEV=0x%08.8X PID=0x%04.4X TOD=0x%08.8X_%08.8X\n",
	      pe[T_PE].sr[PC].ful - (pe[T_PE].sr[IREB].ireb.ireb << 12),
              pe[T_PE].sr[IRD0].ful, pe[T_PE].sr[IRD1].ful, pe[T_PE].sr[PEV].ful, pe[T_PE].sr[PID].pid.pid,
	      pe[T_PE].tod.u, pe[T_PE].tod.l);
          }
        }

#ifdef FXT_INTR
        else if (pe[T_PE].fxt_intr_level) {
          /* firmware ext. interruption */
          pe[T_PE].sr[OPEV].ful      = pe[T_PE].sr[PEV].ful;
          pe[T_PE].sr[OPC].ful       = pe[T_PE].sr[PC].ful;
          pe[T_PE].sr[LPC].ful       = pe[T_PE].lpc;
	  pe[T_PE].sr[PEV].pev.arm   = 0;
	  pe[T_PE].sr[PEV].pev.u     = 0;
	  pe[T_PE].sr[PEV].pev.va    = 0;
	  pe[T_PE].sr[PEV].pev.spu   = 0;
          pe[T_PE].sr[PC].ful        = 0x20;
          pe[T_PE].fxt_intr_delay    = 0;
	  pe[T_PE].firmstate         = 1;
          if (scan_on & SCAN_INTR) {
            scan_mfc(t_pe);
            message("         INTR(FXT) VEC=0x%04.4X IRD0=0x%08.8X IRD1=0x%08.8X PEV=0x%08.8X PID=0x%04.4X TOD=0x%08.8X_%08.8X\n",
	      pe[T_PE].sr[PC].ful - (pe[T_PE].sr[IREB].ireb.ireb << 12),
              pe[T_PE].sr[IRD0].ful, pe[T_PE].sr[IRD1].ful, pe[T_PE].sr[PEV].ful, pe[T_PE].sr[PID].pid.pid,
	      pe[T_PE].tod.u, pe[T_PE].tod.l);
          }
        }
#endif
}

#include "ip-oper.c"
#include "disasm.c"

