static char SccsId[] = "@(#)ap.c	1.3    4/23/92";

/* TaChyon/KAI Instruction simulator      */
/*  Copyright (c) 1989 by FUJITSU LIMITED */
/*            Primary writer: Y.Nakashima */

/* ap.c    89/10/1 */

#include "tim.h"
#include "facility.h"

sim_ap(t_pe)
    unsigned t_pe;
{
#define TAD pe[T_PE].ap_datq[0]
#define TAC pe[T_PE].ap_cheq[0]

    unsigned lirc, lsf;
    unsigned mode;

    mode = pe[T_PE].sr[PEV].pev.va?'v':'r';

    /* check CACHE */

    if (TAC.av == 2) {
      switch (TAC.mao) {
#ifdef MCR_AREA
      case 0x42: /* ld.mcr */
      case 0x43: /* st.mcr */
	switch (sys_mcr_access(t_pe,TAC.addr,TAC.rw,TAC.d)) {
        case RW_SUCCESS:
          TAC.av = 0;
          goto mem_exec;
	  break;

        case RW_MCR_P_EXCEPTION:
          TAC.av = 0;
          goto cheq_exception;
	  break;
        }
        break;
#endif

      default:
        if (TAC.real_addr >= MCU_MAX_MEM_SIZE) {
          sys_mem_a_exception(t_pe, TAC.real_addr);
          TAC.av = 0;
          goto cheq_exception;
        }

        switch (sys_mem_che(t_pe,mode,TAC.addr,TAC.real_addr,&TAC.abs_addr,TAC.len,TAC.rw,TAC.d,TAC.mao)) {
        case RW_SUCCESS:
          TAC.av = 0;
          goto mem_exec;
	  break;

        case RW_MEM_P_EXCEPTION:
          TAC.av = 0;
          goto cheq_exception;
	  break;

        case RW_MEM_WAIT:
          TAC.av = 1;
          if (TAC.rw & _W_)
            TAC.delay = 1; /* 1t for write */
          else
            TAC.delay = AP_CHEDELAY;
          goto consume_datq;
	  break;
        }
        break;
      }
    } 
    else if (TAC.av == 1) { /* cache miss (idle wait) */
      if (!(--TAC.delay)) {
        TAC.av = 0;
        sys_mem_mem(t_pe,mode,TAC.addr,TAC.real_addr,TAC.abs_addr,TAC.len,TAC.rw,TAC.d);
        goto mem_exec;
      }
      else /* wait */
        goto consume_datq;
    }   
    else /* empty */
      goto consume_datq;

/**********************************************************************/

cheq_exception:

    switch (TAC.mao) {
      case 2:
      /**********
      * ld.c    *
      **********/
      case 3:
      /**********
      * ld.cu   *
      **********/
      case 4:
      /**********
      * ld.s    *
      **********/
      case 5:
      /**********
      * ld.su   *
      **********/
      case 6:
      /**********
      * ld      *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAC.real_addr, TAC.len);
        }
	sys_gr_unlock(t_pe, TAC.drn);
	break;

      case 7:
      /**********
      * ld.l    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAC.real_addr, TAC.len);
        }
	sys_gr_unlock(t_pe, TAC.drn);
	sys_gr_unlock(t_pe, TAC.drn+1);
	break;

      case 8:
      /**********
      * ld.f    *
      **********/
      case 9:
      /**********
      * ld.d    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAC.real_addr, TAC.len);
        }
	sys_fr_exception(t_pe, TAC.drn);
	break;

      case 0xa:
      /**********
      * ld.pf   *
      **********/
      case 0xb:
      /**********
      * ld.pd   *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAC.real_addr, TAC.len);
        }
	sys_fr_exception(t_pe, TAC.drn  );
	sys_fr_exception(t_pe, TAC.drn+1);
	break;

      case 0x11:
      /**********
      * st.c    *
      **********/
      case 0x12:
      /**********
      * st.s    *
      **********/
      case 0x13:
      /**********
      * st      *
      **********/
      case 0x1c:
      /**********
      * st.f    *
      **********/
      case 0x1d:
      /**********
      * st.d    *
      **********/
	if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(W) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAC.real_addr, TAC.len);
        }
        break;

      case 0x31:
      /**********
      * cs      *
      **********/
	break;

      case 0x22:
      /**********
      * ldsr    *
      **********/
	if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAC.real_addr, TAC.len);
        }
	sys_sr_unlock(t_pe);
        break;

      case 0x23:
      /**********
      * stsr    *
      **********/
	if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(W) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAC.real_addr, TAC.len);
        }
	break;

#ifdef MCR_AREA
      case 0x42:
      /**********
      * ld.mcr  *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MCR(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                len=%d exception\n", TAC.len);
        }
	sys_gr_unlock(t_pe, TAC.drn);
	break;

      case 0x43:
      /**********
      * st.mcr  *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MCR(W) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                len=%d exception\n", TAC.len);
        }
	break;
#endif
    }
    goto consume_datq;

/**********************************************************************/

mem_exec:

    switch (TAC.mao) {
      case 2:
      /**********
      * ld.c    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=1  data=0x%02.2X\n", TAC.real_addr, TAC.d[0]);
        }
	TAC.d[0] = (int) ((char) TAC.d[0]);
	sys_gr_access(t_pe, TAC.drn, _W_, TAC.d[0]);
	sys_gr_unlock(t_pe, TAC.drn);
	break;

      case 3:
      /**********
      * ld.cu   *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=1  data=0x%02.2X\n", TAC.real_addr, TAC.d[0]);
        }
	sys_gr_access(t_pe, TAC.drn, _W_, TAC.d[0]);
	sys_gr_unlock(t_pe, TAC.drn);
	break;

      case 4:
      /**********
      * ld.s    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=2  data=0x%04.4X\n", TAC.real_addr, TAC.d[0]);
        }
	TAC.d[0] = (int) ((short) TAC.d[0]);
	sys_gr_access(t_pe, TAC.drn, _W_, TAC.d[0]);
	sys_gr_unlock(t_pe, TAC.drn);
	break;

      case 5:
      /**********
      * ld.su   *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=2  data=0x%04.4X\n", TAC.real_addr, TAC.d[0]);
        }
	sys_gr_access(t_pe, TAC.drn, _W_, TAC.d[0]);
	sys_gr_unlock(t_pe, TAC.drn);
	break;

      case 6:
      /**********
      * ld      *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=4  data=0x%08.8X\n", TAC.real_addr, TAC.d[0]);
        }
	sys_gr_access(t_pe, TAC.drn, _W_, TAC.d[0]);
	sys_gr_unlock(t_pe, TAC.drn);
	break;

      case 7:
      /**********
      * ld.l    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=8  data=0x%08.8X_%08.8X\n", TAC.real_addr, TAC.d[0], TAC.d[1]);
        }
	sys_gr_access(t_pe, TAC.drn  , _W_, TAC.d[0]);
	sys_gr_access(t_pe, TAC.drn+1, _W_, TAC.d[1]);
	sys_gr_unlock(t_pe, TAC.drn  );
	sys_gr_unlock(t_pe, TAC.drn+1);
	break;

      case 8:
      /**********
      * ld.f    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=4  data=0x%08.8X\n", TAC.real_addr, TAC.d[0]);
        }
	TAC.d[1] = 0;
	sys_fr_access(t_pe, TAC.drn, _W_, TAC.d[0], TAC.d[1]);
	sys_fr_unlock(t_pe, TAC.drn);
	break;

      case 9:
      /**********
      * ld.d    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=8  data=0x%08.8X_%08.8X\n", TAC.real_addr, TAC.d[0], TAC.d[1]);
        }
	sys_fr_access(t_pe, TAC.drn, _W_, TAC.d[0], TAC.d[1]);
	sys_fr_unlock(t_pe, TAC.drn);
	break;

      case 0xa:
      /**********
      * ld.pf   *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=8  data=0x%08.8X_%08.8X\n", TAC.real_addr, TAC.d[0], TAC.d[1]);
        }
	TAC.d[2] = 0;
	sys_fr_access(t_pe, TAC.drn  , _W_, TAC.d[0], TAC.d[2]);
	sys_fr_access(t_pe, TAC.drn+1, _W_, TAC.d[1], TAC.d[2]);
	sys_fr_unlock(t_pe, TAC.drn  );
	sys_fr_unlock(t_pe, TAC.drn+1);
	break;

      case 0xb:
      /**********
      * ld.pd   *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=16 data=0x%08.8X_%08.8X\n", TAC.real_addr, TAC.d[0], TAC.d[1]);
	  scan_mfc(t_pe);
          message("                                            0x%08.8X_%08.8X\n", TAC.d[2], TAC.d[3]);
        }
	sys_fr_access(t_pe, TAC.drn  , _W_, TAC.d[0], TAC.d[1]);
	sys_fr_access(t_pe, TAC.drn+1, _W_, TAC.d[2], TAC.d[3]);
	sys_fr_unlock(t_pe, TAC.drn  );
	sys_fr_unlock(t_pe, TAC.drn+1);
	break;

      case 0x11:
      /**********
      * st.c    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(W) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=1  data=0x%02.2X\n", TAC.real_addr, TAC.d[0]);
        }
	break;

      case 0x12:
      /**********
      * st.s    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(W) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=2  data=0x%04.4X\n", TAC.real_addr, TAC.d[0]);
        }
	break;

      case 0x13:
      /**********
      * st      *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(W) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=4  data=0x%08.8X\n", TAC.real_addr, TAC.d[0]);
        }
	break;

      case 0x1c:
      /**********
      * st.f    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(W) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=4  data=0x%08.8X\n", TAC.real_addr, TAC.d[0]);
        }
	break;

      case 0x1d:
      /**********
      * st.d    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(W) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=8  data=0x%08.8X_%08.8X\n", TAC.real_addr, TAC.d[0], TAC.d[1]);
        }
	break;

      case 0x31:
      /**********
      * cs      *
      **********/
        /* do nothing (synchronous operation)
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(S) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=4  data=0x%08.8X\n", TAC.real_addr, TAC.d[0]);
        }
        */
	break;

      case 0x22:
      /**********
      * ldsr    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=4  data=0x%08.8X\n", TAC.real_addr, TAC.d[0]);
        }
	sys_sr_access(t_pe, TAC.drn, _W_, &TAC.d[0]);
	sys_sr_unlock(t_pe);
	break;

      case 0x23:
      /**********
      * stsr    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(W) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=4  data=0x%08.8X\n", TAC.real_addr, TAC.d[0]);
        }
	break;

#ifdef MCR_AREA
      case 0x42:
      /**********
      * ld.mcr  *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MCR(R) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                len=4  data=0x%08.8X\n", TAC.d[0]);
        }
	sys_gr_access(t_pe, TAC.drn, _W_, TAC.d[0]);
	sys_gr_unlock(t_pe, TAC.drn);
	break;

      case 0x43:
      /**********
      * st.mcr  *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MCR(W) addr=0x%08.8X ", TAC.addr);
          append_symbol(t_pe, TAC.addr);
	  scan_mfc(t_pe);
          message("                len=4  data=0x%08.8X\n", TAC.d[0]);
        }
	break;
#endif
    }

/**********************************************************************/

consume_datq:

    /* check TLB */

    if (TAD.av == 2 && !TAC.av) {
      switch (TAD.mao) {
#ifdef MCR_AREA
      case 0x42: /* ld.mcr */
      case 0x43: /* st.mcr */
	TAD.real_addr = TAD.addr;
        TAC    = TAD;
        TAD.av = 0;
	goto exit_ap;
        break;
#endif

      default:
        if (TAD.addr & (TAD.len-1)) { /* not aligned exception */
          TAD.av = 0;
          lirc   = 1;
          sys_mem_l_exception(t_pe,TAD.addr,TAD.len,TAD.rw,TAD.d,TAD.drn,TAD.mao,lirc,0);
	  goto datq_exception;
        }
        else if (mode == 'r') { /* pass TLB */
	  TAD.real_addr = TAD.addr;
	  TAC    = TAD;
	  TAD.av = 0;
	  goto exit_ap;
        }
        else if (mode == 'v' && sys_mem_tlb(t_pe,TAD.rw,TAD.addr,&TAD.real_addr)) {
          TAD.av    = 1;
          TAD.delay = AP_DATDELAY;
	  goto exit_ap;
        }
        else { /* TLB hit */
	  TAC    = TAD;
          TAD.av = 0;
	  goto exit_ap;
        }
        break;
      }
    }
    else if (TAD.av == 1) { /* tlb miss (idle wait) */
      if (!(--TAD.delay)) {
        switch (sys_mem_dat(t_pe,TAD.rw,TAD.addr,&TAD.real_addr,&lirc,&lsf)) {
        case RW_SUCCESS:
	  TAD.av = 2;
	  TAC    = TAD;
	  TAD.av = 0;
	  goto exit_ap;
	  break;

	case RW_MEM_P_EXCEPTION:
	  TAD.av = 0;
	  goto datq_exception;
          break;

        case RW_MEM_L_EXCEPTION:
	  TAD.av = 0;
          sys_mem_l_exception(t_pe,TAD.addr,TAD.len,TAD.rw,TAD.d,TAD.drn,TAD.mao,lirc,lsf);
	  goto datq_exception;
          break;
        }
      }
      else /* wait */
        goto exit_ap;
    }
    else /* empty */
      goto exit_ap;

/**********************************************************************/

datq_exception:

    switch (TAD.mao) {
      case 2:
      /**********
      * ld.c    *
      **********/
      case 3:
      /**********
      * ld.cu   *
      **********/
      case 4:
      /**********
      * ld.s    *
      **********/
      case 5:
      /**********
      * ld.su   *
      **********/
      case 6:
      /**********
      * ld      *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAD.addr);
          append_symbol(t_pe, TAD.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAD.real_addr, TAD.len);
        }
	sys_gr_unlock(t_pe, TAD.drn);
	break;

      case 7:
      /**********
      * ld.l    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAD.addr);
          append_symbol(t_pe, TAD.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAD.real_addr, TAD.len);
        }
	sys_gr_unlock(t_pe, TAD.drn);
	sys_gr_unlock(t_pe, TAD.drn+1);
	break;

      case 8:
      /**********
      * ld.f    *
      **********/
      case 9:
      /**********
      * ld.d    *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAD.addr);
          append_symbol(t_pe, TAD.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAD.real_addr, TAD.len);
        }
	sys_fr_exception(t_pe, TAD.drn);
	break;

      case 0xa:
      /**********
      * ld.pf   *
      **********/
      case 0xb:
      /**********
      * ld.pd   *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAD.addr);
          append_symbol(t_pe, TAD.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAD.real_addr, TAD.len);
        }
	sys_fr_exception(t_pe, TAD.drn  );
	sys_fr_exception(t_pe, TAD.drn+1);
	break;

      case 0x11:
      /**********
      * st.c    *
      **********/
      case 0x12:
      /**********
      * st.s    *
      **********/
      case 0x13:
      /**********
      * st      *
      **********/
      case 0x1c:
      /**********
      * st.f    *
      **********/
      case 0x1d:
      /**********
      * st.d    *
      **********/
	if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(W) addr=0x%08.8X ", TAD.addr);
          append_symbol(t_pe, TAD.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAD.real_addr, TAD.len);
        }
        break;

      case 0x31:
      /**********
      * cs      *
      **********/
	break;

      case 0x22:
      /**********
      * ldsr    *
      **********/
	if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(R) addr=0x%08.8X ", TAD.addr);
          append_symbol(t_pe, TAD.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAD.real_addr, TAD.len);
        }
        sys_sr_unlock(t_pe);
        break;

      case 0x23:
      /**********
      * stsr    *
      **********/
	if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MEM(W) addr=0x%08.8X ", TAD.addr);
          append_symbol(t_pe, TAD.addr);
	  scan_mfc(t_pe);
          message("                real=0x%08.8X len=%d exception\n", TAD.real_addr, TAD.len);
        }
	break;

#ifdef MCR_AREA
      case 0x42:
      /**********
      * ld.mcr  *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MCR(R) addr=0x%08.8X ", TAD.addr);
          append_symbol(t_pe, TAD.addr);
	  scan_mfc(t_pe);
          message("                len=%d exception\n", TAD.len);
        }
	sys_gr_unlock(t_pe, TAD.drn);
	break;

      case 0x43:
      /**********
      * st.mcr  *
      **********/
        if (scan_on & SCAN_MEM) {
	  scan_mfc(t_pe);
          message("         MCR(W) addr=0x%08.8X ", TAD.addr);
          append_symbol(t_pe, TAD.addr);
	  scan_mfc(t_pe);
          message("                len=%d exception\n", TAD.len);
        }
	break;
#endif
    }

/**********************************************************************/

exit_ap:

    pe[T_PE].sr[STAR1].star1.asam = (TAD.av > 0 || TAC.av > 0);
}

