/* @(#)tim.h	1.29    8/7/93 */

/* TaChyon/KAI Instruction simulator      */
/*  Copyright (c) 1989 by FUJITSU LIMITED */
/*            Primary writer: Y.Nakashima */

#include <stdio.h>
#ifdef USG
#include <string.h>
#else
#include <strings.h>
#endif
#include <time.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#include <termio.h>
#include <setjmp.h>
#include <malloc.h>
#include <sys/ttold.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/timeb.h>
#include <sys/file.h>
#include <sys/stat.h>

/* environment variables & files */

char *			ENV_PROMPT1;		/* prompt */
char *			ENV_PROMPT2;		/* prompt */
char *			ENV_MODULE;		/* name of load_module */
char *			ENV_INITRC;		/* name of rc_file */
char *			ENV_COMMAND;		/* name of command_file */
char *			ENV_CMDDIR;		/* name of command_directory */
char *			ENV_HISTORY;		/* name of history_file */
char *			ENV_DISPLOG;		/* name of displaylog_file */
char *			ENV_GDBREPL;		/* name of GDB-reply_file */
char *			ENV_ALLFAC;		/* name of all_facility */
char *			ENV_DASD;		/* name of dasd_image_file */
char *			ENV_SSU;		/* name of ssu_image_file */
char 			ENV_COIN[32];		/* name of console_input_file */
char 			ENV_COUT[32];		/* name of console_output_file */
char 			ENV_TMPAS[32];		/* name of temporal as_input_file */
char 			ENV_TMPOB[32];		/* name of temporal as_output_file */
char 			ENV_TMPLD[32];		/* name of temporal ld_output_file */
FILE *			module_fp;
FILE *			rc_fp;
FILE *			command_fp;
FILE *			history_fp;
FILE *			displog_fp;
FILE *			gdbrepl_fp;
int 			dasd_fd;
int 			dasd_blocks;		/* number of blocks (parameter) */
int 			ssu_fd;
int 			ssu_elements;		/* number of elements (parameter) */
int 			coin_fd;
int 			cout_fd;
int			kiodebug;		/* for GDB's convenience */
int			old_mode;		/* for backward compatibility */
int			fast_mv;		/* for fast mover */
int                     penum;                  /* number of PEs */
int                     pagenum;                /* number of pages for memory */
int                     memsize;                /* size of memory */

/* constants */

#define         	TACHYONMAGIC    	0x00c80000
#define         	OMAGIC          	0x00000107
#define         	NMAGIC          	0x00000108
#define         	ZMAGIC          	0x0000010b

#define        		SCAN_NULL    	 	0
#define        		SCAN_ALL    	 	1
#define        		SCAN_LIST    	 	2
#define        	 	SCAN_BRANCH  	 	4
#define        	 	SCAN_INTR  	 	8
#define        	 	SCAN_MEM     	 	16
#define        	 	SCAN_STATUS  	 	32

#define			_N_	     	 	0
#define			_R_	     	 	4
#define			_W_	     	 	2
#define			_X_	     	 	1
#define			_A_	     	 	6
#define			_O_	     	 	8

#define			SCORE_FREE		0
#define			SCORE_BUSY		1
#define			SCORE_EXCEPTION		2

enum ip_status { 	STATUS_FREE,
		 	STATUS_R_WAIT,
		 	STATUS_R_CONT, 
			STATUS_R_READY, 
			STATUS_HALT,
		};

#ifdef MOVER
enum mv_status {	STATUS_AVAIL,
			STATUS_CDOWN,
		};

enum mv_pa {		MVPA_TINP,
			MVPA_TNTB,
			MVPA_TBLK,
			MVPA_TSLF,
			MVPA_RINP,
			MVPA_RBLK,
			MVPA_TLEN,
			MVPA_RLEN,
			MVPA_TNUM,
			MVPA_RNUM,
			MVPA_MPAA,
			MVPA_DMY0,
			MVPA_DMY1,
			MVPA_DMY2,
			MVPA_DMY3,
			MVPA_DMY4,
		};
#endif

#ifdef DASD
enum dasd_status { 	STATUS_NORMAL,
			STATUS_PCI,
			STATUS_TIC,
		};
#endif

#ifdef SSUMOVER
enum ssumv_status { 	STATUS_SSLEEP,
			STATUS_SBUSY,
			STATUS_SINTR,
		};
#endif

#ifdef DAPF
enum dapf_status { 	STATUS_DSLEEP,
			STATUS_DBUSY,
		};
#endif

enum sys_rw_facility {	RW_SUCCESS,
			RW_REG_EXCEPTION,
			RW_MCR_P_EXCEPTION,
			RW_MEM_P_EXCEPTION,
			RW_MEM_L_EXCEPTION,
			RW_MEM_WAIT,
			RW_BERR,
			RW_TIMEOUT,
		};

/* configuration parameters */

#ifndef SCREEN_PATH
#define			SCREEN_PATH		"/usr/kai/bin/screen"
#endif

#ifndef VTEM_PATH
#define			VTEM_PATH		"/usr/kai/bin/vtem"
#endif

#ifndef HELP_PATH
#define			HELP_PATH		"/usr/kai/lib/tim.help"
#endif

#ifndef PE_NUM
#define			PE_NUM 		 	1
#endif
#if PE_NUM == 1
#define			T_PE		 	0
#else
#define			T_PE		 	t_pe
#endif

#ifndef DEFAULT_PE_NUM
#define			DEFAULT_PE_NUM 		(4) /* default PEs is 4 */
#endif

#ifndef DEFAULT_PAGE_NUM
#define			DEFAULT_PAGE_NUM 	(32) /* default mem_size is 1MB */
#endif

#ifndef PHSA_NUM
#define			PHSA_NUM	 	(16) /* default PHZ is 512KB */
#endif

#define			IR_NUM		 	(16) /* (1-15) */

#ifdef CONSOLE
#define			HSATBASE		(0x00002000) /* base address of HSAT */
#define			SVFWBYTE		(256) /* size of svp-firm req. block */
#endif

#define			HANGUP_LIMIT		(10000)

#define			GR_NUM		 	(32)
#define			FR_NUM		 	(32)
#define			SR_NUM		 	(1024)
#define			SR_REALNUM	 	(144)

#define			IF_DATDELAY		(76)
#define			IF_CHEDELAY		(32)

#define			IP_MOVGS_CYCLE		(2)
#define			IP_MOVSG_CYCLE		(2)
#define			IP_FLCHA_CYCLE		(8)
#define			IP_FLCHE_CYCLE		(8)
#define			IP_LDSR_CYCLE		(8)
#define			IP_PTLBE_CYCLE		(8)
#define			IP_PTLBP_CYCLE		(8)
#define			IP_PTLBA_CYCLE		(8)
#define			IP_RTI_CYCLE		(9)

#ifdef MCR_AREA
#define			MCR_NUM		 	(8192)
#define			MCR_REALNUM	 	(8192)
#endif

#ifdef IVT_INTR
#define			IVT_INTERRUPT	 	(0x00100000)
#endif

#define			AP_DATQNUM	 	(1)
#define			AP_DATDELAY	 	(76) /* 1-xxx (must be 76) */
#define			AP_CHEQNUM	 	(1)
#define			AP_CHEDELAY	 	(32) /* 1-xxx (must be 32) */
#define			TLB_NUM		 	(256)
#define			CACHE_BLOCK_SIZE 	(64)
#define			CACHE_NUM	 	(512)
#define			PAX_MASK	 	((unsigned) 0xfffffff0)
#define			FQ_NUM		 	(6)
#define			FP_WRITE_DELAY	 	(2)
#define			FP_DIVF_DELAY	 	(10)
#define			FP_DIVD_DELAY	 	(18)
#define			HEADER_SIZE	 	(32)

#ifdef MANUAL_MAP
#define			MAP_NUM		 	(64)
#define			MAP_FREE_NUM	 	(MAP_NUM * 2)
#endif

#define			BP_NUM		 	(32) /* (0-31) */
#define			EXT_FUNC_BREAK	 	(BP_NUM)
#define			PAGE_SIZE	 	(32768)
#define			MCU_MAX_MEM_SIZE 	((unsigned) 0x40000000)
#define			MEMHASH_SIZE		(8192)

#ifdef MCR_AREA
#define			PHSA_SIZE	 	(PAGE_SIZE * PHSA_NUM)
#endif

#ifdef VCP
#define			VP_INTERRUPT		(0x00200000)
#define			VR_NUM		 	(256)
#define			VRL		 	(64) /* minimum length of VR */
#define			MR_NUM		 	(256)
#define			VCR_NUM		 	(32)
#define			VUT_NUM		 	(32)
#define			TR_NUM		 	(32)
#define			ID_NUM		 	(3)
#define			VP_OPIQNUM	 	(1)  /* vp operation queue (from ip) */
#define			VP_OPMQNUM	 	(4)  /* vp operation queue (to mem) */
#define			VP_OPCQNUM	 	(4)  /* vp operation queue (to cal) */
#define			VP_LDQNUM	 	(1)  /* vp load queue */
#define			VP_STQNUM	 	(1)  /* vp store queue */
#define			VP_ADDQNUM	 	(1)  /* vp add queue */
#define			VP_MULQNUM	 	(1)  /* vp mult queue */
#define			VP_DIVQNUM	 	(1)  /* vp divide queue */
#define			VP_MSKQNUM	 	(1)  /* vp mask queue */
#define			VP_LDPLEN	 	(28) /* vp load pipe */
#define			VP_STPLEN	 	(18) /* vp store pipe */
#define			VP_ADDPLEN	 	(9)  /* vp add pipe */
#define			VP_MULPLEN	 	(9)  /* vp mult pipe */
#define			VP_DIVPLEN	 	(24) /* vp divide pipe */
#define			VP_MSKPLEN	 	(6)  /* vp mask pipe */
#endif

#ifdef MOVER
#define			MTR_NUM			(32)
#define			NTR_NUM			(256)
#define			MV_INTERRUPT		(0x00080000)
#define			MV_PANUM	 	(16) /* mover PA */
#endif

#ifdef SSUMOVER
#define			SSU_INTERRUPT		(0x00040000)
#define			SSU_DMA_DELAY		(100) /* ssu dma delay parameter */
#define			SSU_ELEMENT_SIZE	(8)  /* element size of ssu */
#endif

#ifdef GSIGP
#define			GSIGP_INTERRUPT		(0x00020000)
#define			GSIGP_DELAY		(100) /* gsigp delay parameter */
#endif

#ifdef FSBA
#ifndef FSBA_NUM
#define			FSBA_NUM	 	(1)
#endif
#if FSBA_NUM == 1
#define			T_FSBA		 	(0)
#else
#define			T_FSBA		 	(t_fsba)
#endif
#define			FSBA_INTERRUPT		(0x02000000)
#define			FSBA_SLOTID		(2)  /* fsba slot id */
#define			FSBA_SLOT_NUM		(16) /* fsba slot number */
#define			FSIR_NUM		(32) /* fsba internal register */
#endif

#ifdef FST
#define			FST_SLOTID		(4)  /* fst slot id */
#define			FST_DMA_DELAY		(100) /* fst dma delay parameter */
#define			FSTCR_NUM		(8)  /* fst control register */
#define			FSTMEM_SIZE	 	(16384) /* fst memory size */
#endif

#define			DASD_BLOCK_SIZE		(512) /* block size of dasd */
#ifdef DASD
#define			DASD_SLOTID		(6)  /* dasd slot id */
#define			DASD_DVA		(4)  /* dasd device address */
#define			DASD_LDVA		(2)  /* dasd logical device address */
#define			DASD_DMA_DELAY		(100) /* dasd dma delay parameter */
#define			ACR_NUM			(8)  /* adapter control register */
#endif

#ifdef DAPF
#define			SVS_INTERRUPT		(0x00400000)
#define			DAPF_DMA_DELAY		(100) /* dapf dma delay parameter */
#define			SAPCMAXX    		(79)
#define			SAPCMAXY    		(23)
#define			SAPCIMAGE    		((SAPCMAXX+1)*(SAPCMAXY+1))
#define			SAPCFRAME    		(SAPCIMAGE+3)
#endif

/* fr structure */
#include "fr-field.def"

/* sr structure */
#include "sr-field.def"
enum sr_alias {
#define	         	DEF_SR(NAME, CMD_R_MASK, CMD_W_MASK, SYS_R_MASK, SYS_W_MASK) NAME,
#include "sr-name.tmp"
#undef			DEF_SR
};

/* tod structure */
#include "tod-field.def"

#ifdef MCR_AREA
/* mcr structure */
#include "mcr-field.def"
enum mcr_alias {
#define         	DEF_MCR(NAME, REV, R_MASK, W_MASK) NAME,
#include "mcr-name.tmp"
#undef			DEF_MCR
};
#endif

/* dat structure */
#include "dat-field.def"

/* cache structure */
#include "che-field.def"

/* ap queue structure */
#include "ap-field.def"

/* fp queue operation_identifier */
#include "fp-field.def"

#ifdef VCP
/* vr structure */
#include "vr-field.def"

/* mr structure */
#include "mr-field.def"

/* vcr structure */
#include "vcr-field.def"

/* tr structure */
#include "tr-field.def"

/* vp queue structure */
#include "vp-field.def"
#endif

#ifdef MOVER
/* mover queue structure */
#include "mv-field.def"
#endif

#ifdef FSBA
/* fsba cr structure */
#include "fscr-field.def"

/* fsba internal register structure */
#include "fsir-field.def"

/* fsba queue structure */
#include "fsba-field.def"
#endif

#ifdef FST
/* fst control register structure */
#include "fstcr-field.def"
#endif

#ifdef DASD
/* adapter configuration register structure */
#include "acf-field.def"

/* adapter control register structure */
#include "acr-field.def"
#endif

/* controls */

int			batchmode;	/* switch batch or not	   	    */
int 			history_num;	/* history number                   */
int 			scan_on;	/* scan flag (ored SCAN_XXX)        */
int			bp_match;	/* breakpoint found 	 	    */
int			cp_match;	/* checkpoint found 		    */

/* break/check point structure */

struct bp_table {	unsigned interval; /* 1-xxx         */
			unsigned count;    /* working timer */
			unsigned access;   /* 0 _R_,_W_,_X_ */
		   	unsigned b_or_c;   /* 0 'b','c'     */
		   	unsigned address;
		   	unsigned real_addr;
		   	unsigned abs_addr;
		   	unsigned mode;     /* 'r'/'v'       */
		   	char *command;
		};

/* interval interrupt structure */

struct ir_table {	unsigned interval; /* 1-xxx         */
		   	unsigned count;    /* working timer */
			unsigned type;     /* 'c'/'s'       */
		};

/* symbol table structure */

struct memshash {	char	 *name;
			unsigned memaddr;
		};

struct memahash {	unsigned type;
			unsigned memaddr;
			char     *name;
		};

/************************ PE FACILITIES ************************/

unsigned	tim_clock;				/* simulation clock */

struct	pe_form	{
		double		dummy;			/* dummy for alignment */

                unsigned	tod_5nano;		/* 5nano * 25 counter */
		unsigned	firmstate;		/* Firmware State */
		unsigned	nullify;		/* Nullify Insn. */
		unsigned	lpc;			/* Last PC */
		enum ip_status	ip_status;		/* IP Status */
                unsigned	insn[2];
		unsigned	insn_u1;
		unsigned	insn_u2;
		unsigned	insn_l1;
		unsigned	insn_l2;
		int		(*o_func0)();
		int		(*o_func1)();
		int		(*o_func2)();
		unsigned	o_type0;
		unsigned	o_type1;
		unsigned	o_type2;
		unsigned	mx;
		unsigned	hangup_timer;		/* hangup timer */
		unsigned	timeout_req;		/* timeout request */
		unsigned	cntl_block;		/* control_block address */
#ifdef MANUAL_INTR
		struct ir_table	ir_table[IR_NUM];	/* 0-15: interval_int. block */
#endif
		unsigned	svc_interrupt; 		/* 0: on/off, 24-31: code */
		unsigned	fwc_interrupt; 		/* firmware call */
		unsigned	ext_interrupt; 		/* 1-15: external int. signal */
		unsigned	ext_intr_level;		/* recognized ext_int. level */
		unsigned	slrp_intr_level;	/* recognized slrp_int. level */
		unsigned	ext_intr_delay;		/* intr. delayed except for VCP */
		unsigned	slrp_intr_delay;	/* intr. delayed */
#ifdef FXT_INTR
		unsigned	fxt_intr_level; 	/* recognized firm ext_int. */
		unsigned	fxt_intr_delay;		/* intr. delayed */
#endif
		struct srform30	work_ird0;		/* Internal IRD0 */
		struct srform31	work_ird1;		/* Internal IRD1 */

		unsigned	sinsn_stat[16][32][3];	/* Scalar Insn. statistics [type][opc][loc] */

		unsigned char	gr_score[GR_NUM];	/* GR score board */
		unsigned char	fr_score[FR_NUM];	/* FR score board */
		unsigned char	sr_score;		/* SR score board (all) */
		unsigned	if_multiple_cycle;	/* multiple cycle for i-fetch */
		unsigned	ip_multiple_cycle;	/* multiple cycle for ip */
		unsigned	gr[GR_NUM];		/* General REGISTERS */
		struct fr_form	fr[FR_NUM];		/* Floating point REGISTERS */
		union sr_form	sr[SR_REALNUM];		/* Special purpose REGISTERS */
		struct tod_form	tod;			/* Time Of Day */
#ifdef MCR_AREA
		unsigned	wdt_5nano;		/* 5nano * 819200 counter */
		union mcr_form	mcr[MCR_REALNUM];	/* Memory mapped REGISTERS */
		union mcr_form	smcr[MCR_REALNUM];	/* Shadow MCR */
#endif
                unsigned	brcs_b_t;		/* branch statistics */
                unsigned	brcs_b_n;		/* branch statistics */
                unsigned	brcs_f_n;		/* branch statistics */
                unsigned	brcs_f_t;		/* branch statistics */
                unsigned	brc_b_t;		/* branch statistics */
                unsigned	brc_b_n;		/* branch statistics */
                unsigned	brc_f_n;		/* branch statistics */
                unsigned	brc_f_t;		/* branch statistics */
                unsigned	text_access;		/* text access counter */
                unsigned	text_tlb_access;	/* text tlb access counter */
                unsigned	text_tlb_hit;		/* text tlb hit counter */
                unsigned	text_cache_access;	/* text cache access counter */
                unsigned	text_cache_hit;		/* text cache hit counter */
                unsigned	data_access;		/* data access counter */
                unsigned	data_tlb_access;	/* data tlb access counter */
                unsigned	data_tlb_hit;		/* data tlb hit counter */
                unsigned	data_cache_access;	/* data cache access counter */
                unsigned	data_cache_hit;		/* data cache hit counter */
		struct apq_form	ap_datq[AP_DATQNUM];	/* ap dat queue */
		struct apq_form	ap_cheq[AP_CHEQNUM];	/* ap che queue */
		struct tlb_form	text_tlb[TLB_NUM];	/* Instruction TLB */
		struct che_form	text_cache[CACHE_NUM];	/* Instruction CACHE */
		struct tlb_form	data_tlb[TLB_NUM]; 	/* Operand TLB */
		struct che_form	data_cache[CACHE_NUM];	/* Operand CACHE */
  		unsigned	fq_top;     		/* FQ entry to be executed */
  		unsigned	avn_fpwd;     		/* num of effective FPWD */
		struct fqi_form	fqi[FQ_NUM];		/* fp write info */
		struct fpw_form	fpwd[FP_WRITE_DELAY];	/* fp write delay */
		struct fpd_form	fpdb;			/* fp divide buffer */

		unsigned char	header[HEADER_SIZE];	/* a.out header */
		unsigned char	*symtbl;		/* a.out symbols */
		unsigned     	symsize;		/* size of SYMTBL */
#ifdef MANUAL_MAP
		struct map_used	map[MAP_NUM];		/* mem_map info */
		struct map_free map_free[MAP_FREE_NUM];	/* free_map info */
#endif
		int		bp_entry;		/* bp_match_entry */
		int		cp_entry;		/* cp_match_entry */
		struct bp_table	bp_table[BP_NUM];	/* break/check point table */
		unsigned char	*mem;			/* Absolute MEMORY */
		unsigned char	*mbp;			/* break/check point tag */
		unsigned 	memsize;		/* size of memory */
		struct memshash	MEMSHASH[MEMHASH_SIZE];
		struct memahash	MEMAHASH[MEMHASH_SIZE];

#ifdef VCP
		unsigned	vut_5nano;		/* 5nano * 200 counter */
		struct vcrform2	work_vcr2;		/* Internal VCR2 */
		struct vcrform2	work_st_vcr2;		/* Internal VCR2 (st) */

		unsigned	vinsn_stat[256];	/* Vector Insn. statistics [opc] */

                unsigned	vr_rw_conflict[VR_NUM];	/* VR RW confliction counter */
                unsigned	vr_wr_conflict[VR_NUM];	/* VR WR confliction counter */
                unsigned	vr_ww_conflict[VR_NUM];	/* VR WW confliction counter */
                unsigned	mr_rw_conflict[MR_NUM];	/* MR RW confliction counter */
                unsigned	mr_wr_conflict[MR_NUM];	/* MR WR confliction counter */
                unsigned	mr_ww_conflict[MR_NUM];	/* MR WW confliction counter */
		unsigned char	vp_score;		/* VP score */
		unsigned char	vr_rscore[VR_NUM];	/* VR read score */
		unsigned char	vr_wscore[VR_NUM];	/* VR write score */
		unsigned char	mr_rscore[MR_NUM];	/* MR read score */
		unsigned char	mr_wscore[MR_NUM];	/* MR write score */
		unsigned char	vop_iscore;		/* VOP MEM indirect */
                unsigned char   vop_rscore[ID_NUM];	/* VOP MEM load_ID */
                unsigned char   vop_wscore[ID_NUM];	/* VOP MEM store_ID */
		struct vr_form 	vr[VR_NUM][VRL];	/* Vector REGISTER */
		struct mr_form	mr[MR_NUM];		/* Mask REGISTER */
		union vcr_form	vcr[VCR_NUM];		/* Vector Control REGISTER */
		unsigned	vut[VUT_NUM];		/* Vector Unit Timer */
		union tr_form	tr[TR_NUM];		/* Translation Register */

  		unsigned	avn_opiq;     		/* num of .av=1 (OP QUEUE (from IP)) */
  		unsigned	avn_opmq; 		/* num of .av=1 (OP QUEUE (to MEM)) */
  		unsigned	avn_opcq; 		/* num of .av=1 (OP QUEUE (to CAL)) */
  		unsigned	avn_ldq;      		/* num of .av=1 (LOAD QUEUE) */
  		unsigned	avn_stq;      		/* num of .av=1 (STORE QUEUE) */
  		unsigned	avn_addq;     		/* num of .av=1 (ADD QUEUE) */
  		unsigned	avn_mulq;     		/* num of .av=1 (MULT QUEUE) */
  		unsigned	avn_divq;     		/* num of .av=1 (DIVIDE QUEUE) */
  		unsigned	avn_mskq;     		/* num of .av=1 (MASK QUEUE) */
  		unsigned	avn_ldp;      		/* num of .av=1 (LOAD PIPE) */
  		unsigned	avn_stp;      		/* num of .av=1 (STORE PIPE) */
  		unsigned	avn_addp;     		/* num of .av=1 (ADD PIPE) */
  		unsigned	avn_mulp;     		/* num of .av=1 (MULT PIPE) */
  		unsigned	avn_divp;     		/* num of .av=1 (DIVIDE PIPE) */
  		unsigned	avn_mskp;     		/* num of .av=1 (MASK PIPE) */

		struct vpq_form	vp_opiq[VP_OPIQNUM];	/* vp queue (from ip)*/
		struct vpq_form vp_opmq[VP_OPMQNUM];	/* vp queue (to mem)*/
		struct vpq_form vp_opcq[VP_OPCQNUM];	/* vp queue (to cal)*/

		struct vpq_form vp_ldq[VP_LDQNUM];	/* vp load queue */
		struct vpq_form	vp_stq[VP_STQNUM];	/* vp store queue */
		struct vpq_form vp_addq[VP_ADDQNUM];	/* vp add queue */
		struct vpq_form vp_mulq[VP_MULQNUM];	/* vp mult queue */
		struct vpq_form vp_divq[VP_DIVQNUM];	/* vp divide queue */
		struct vpq_form vp_mskq[VP_MSKQNUM];	/* vp mask queue */

		struct vp_ldp 	vp_ldp[VP_LDPLEN];	/* vp load pipe */
		struct vp_stp 	vp_stp[VP_STPLEN];	/* vp store pipe */
		struct vp_addp  vp_addp[VP_ADDPLEN];	/* vp add pipe */
		struct vp_mulp  vp_mulp[VP_MULPLEN];	/* vp mult pipe */
		struct vp_divp  vp_divp[VP_DIVPLEN];	/* vp divide pipe */
		struct vp_mskp  vp_mskp[VP_MSKPLEN];	/* vp mask pipe */
#endif

#ifdef MOVER
		unsigned	rqt_5nano;		/* 5nano * 204800 counter */
		struct mcrform4000 work_mrprmi;		/* Internal MRPRMI */
		struct mcrform4008 work_sndmri;		/* Internal SNDMRI */
		struct mcrform4000 mv_mrprmi;		/* mover rcv INT queue */
		unsigned	mv_rcv_srcpe;		/* connected PE# */
		enum mv_status	mv_rcv_status;		/* rcv status */
		unsigned 	mv_rcv_stoprq;		/* rcv stop request */
		unsigned 	mv_snd_stoprq;		/* snd stop request */
		unsigned 	mv_hround;		/* rotation of user higher queue */
		unsigned 	mv_lround;		/* rotation of user lower queue */
		unsigned 	mv_hqo;			/* current HQO */
		int       	mv_qid;			/* current QID */
		int	 	mv_rbsel;		/* current RBSEL */
		struct mv_hedr	mv_hedr;		/* current HEADER */
		unsigned 	mv_pa[MV_PANUM];	/* mover PA */
#endif

#ifdef BARRIER
		unsigned	bpu_5nano;		/* 5nano * 300 (1.5 micro) counter */
#endif

#ifdef SSUMOVER
		enum ssumv_status ssumv_status;		/* ssu mover status */
		unsigned	ssu_delay_count;	/* ssu startup delay counter */
		unsigned	ssu_stop_request;	/* stop request for ssu */
#endif

#ifdef GSIGP
		unsigned	gsigp_delay_count;	/* gsigp startup delay counter */
#endif

#ifdef FSBA
		struct fscr_form fscr[FSBA_NUM];	/* fsba control register */
		struct fsir_form fsir[FSBA_NUM];	/* fsba internal register */
		struct fsrq_form fsrq[FSBA_NUM][FSBA_SLOT_NUM];	/* fsba interruption request queue */
#endif

#ifdef FST
		struct fstcr_form fstcr[FSBA_NUM];	/* fst control register */
		unsigned char	fstmem[FSBA_NUM][FSTMEM_SIZE];	/* fst MEMORY */
#endif

#ifdef DASD
		enum dasd_status dasd_status;		/* dasd status */
		struct acf_form acf;			/* adapter configuration register */
		struct acr_form acr;			/* adapter control register */
		struct acr_form dacr;			/* acr in dcr (for dasd) */
		unsigned 	dcba;			/* dcb address (for dasd) */
		unsigned 	dcwa;			/* dcw address (for dasd) */
#endif

#ifdef DAPF
		enum dapf_status dapf_status;		/* dapf status */
		unsigned	dapf_delay_count;	/* dapf startup delay counter */
		unsigned	dapf_sapc_pending;	/* 4:attention, 6:read buffer */
		char		dapf_sapc[SAPCFRAME];	/* SAPC console frame */
#endif
	};

#if PE_NUM == penum
struct	pe_form	*pe;
#else
struct	pe_form	pe[PE_NUM];
#endif

